#!/usr/bin/env python3

import os
import subprocess
import re
import markdown

TOC_TEXT_SPACING = '1.02'
BODY_TEXT_SPACING = '1.40'
CHANGES_MD = '../CHANGES.md'

VERSION_HISTORY_APPENDIX_NAME = 'D'


# -- Covert HTML to TeX

html_edited = []
with open('Manual.html') as fi:
    skipping_manual_toc = False
    for l in fi.readlines():
        if '</BODY>' in l or '</HTML>' in l:
            continue
        if 'MANUAL_TOC_BEGIN' in l:
            skipping_manual_toc = True
        elif 'MANUAL_TOC_END' in l:
            skipping_manual_toc = False
        if not skipping_manual_toc:
            html_edited += [l]

with open(CHANGES_MD) as chlog:
    chlog_md = chlog.read()
chlog_md = re.sub('#### Version 0.7.8..*', '', chlog_md, flags=re.S)  # flag => multiline
chlog_html = markdown.markdown(chlog_md)
chlog_html = re.sub('<p>', '', chlog_html)
chlog_html = re.sub('</p>', '', chlog_html)
chlog_html = re.sub('(\[[^\]]+\])', '', chlog_html)
chlog_html = '<h1>Appendix ' + VERSION_HISTORY_APPENDIX_NAME + ':&nbsp; Version history</h1>' + chlog_html + '</BODY></HTML>'
html_edited.append(chlog_html)

with open('Manual-edited.html', 'w') as fo:
    print(''.join(html_edited), file=fo)


pandoc_cmd = 'pandoc -s  -f html  -o Manual.tex  -t latex  Manual-edited.html'
subprocess.call(pandoc_cmd.split())


# -- Edit TeX before compiling

with open('Manual.tex') as fi:
    tex_edited = fi.read()

def replace(text, original, modification):
    text_modified = re.sub(original, modification, text)
    assert text_modified != text
    return text_modified



tex_edited = replace(tex_edited, '''\\\\documentclass\\[
\\]{article}''', '''
\\\\documentclass[english,12pt]{article}
\\\\usepackage{fontspec}
\\\\setsansfont[Mapping=tex-text]{Carlito}
%\\\\setmonofont[Scale=0.84]{Bitstream Vera Sans Mono}
\\\\setmonofont[Scale=0.96]{Inconsolata}
\\\\usepackage[a4paper]{geometry}
\\\\setmainfont[Mapping=tex-text,Numbers=OldStyle]{Carlito}
\\\\geometry{verbose,tmargin=2cm,bmargin=2cm,lmargin=2.45cm,rmargin=2.45cm,
             headsep=0.8cm,footskip=1cm}''')

tex_edited = replace(tex_edited, '\\\\usepackage{lmodern}',
                    '''\\\\usepackage{setspace}
\\\\usepackage{titlesec}
%\\\\titlespacing*{\\\\section}{0pt}{1.8\\\\baselineskip}{0.25\\\\baselineskip}
%\\\\titlespacing*{\\\\subsection}{-20pt}{-2.2\\\\baselineskip}{0.06\\\\baselineskip}
%\\\\titlespacing*{\\\\subsubsection}{-20pt}{-2.15\\\\baselineskip}{0.05\\\\baselineskip}
\\\\titlespacing*{\\\\section}{0pt}{2\\\\baselineskip}{0.5\\\\baselineskip}
\\\\titlespacing*{\\\\subsection}{0pt}{0.8\\\\baselineskip}{0.133\\\\baselineskip}
\\\\titlespacing*{\\\\subsubsection}{0pt}{0.575\\\\baselineskip}{0.100\\\\baselineskip}
\\\\usepackage{tocloft}
\\\\setlength{\\\\cftbeforetoctitleskip}{5mm}
\\\\setlength{\\\\cftbeforesecskip}{2.5mm}''')

with open('../CMakeLists.txt') as fi:
    fc = fi.read()
    version_major = re.search('set\\(.*_VERSION_MAJOR (.*)\\)', fc).group(1)
    version_minor = re.search('set\\(.*_VERSION_MINOR (.*)\\)', fc).group(1)
    version_patch = re.search('set\\(.*_VERSION_PATCH (.*)\\)', fc).group(1)
    version = '.'.join([version_major, version_minor, version_patch])

tex_edited = replace(tex_edited, '\\\\hypersetup{',
                    '''
\\\\usepackage{color}
\\\\definecolor{hypercolor}{rgb}{0.0, 0.11, 0.44}
\\\\hypersetup{colorlinks=true,linkcolor=hypercolor,
               urlcolor=hypercolor,citecolor=hypercolor,
               pdfsubject={Manual features software v. ''' + version + '.},')

tex_edited = replace(tex_edited,
    ''',
  hidelinks,''',
    ',')

tex_edited = replace(tex_edited,
    '\\\\maketitle',
    '''\\\\begin{center}{\\\\Large ShuttleTracker -- User's manual}\\\\\\\\[1ex]
@version ''' + version + '''\\\\\\\\[5ex]\\\\end{center}
\\\\spacing{''' + TOC_TEXT_SPACING + '''}
\\\\renewcommand\\\\contentsname{\\\\empty}
\\\\hrule\\\\vspace{-3ex}\\\\setcounter{tocdepth}{2}
\\\\tableofcontents\\\\vspace{2ex}\\\\hrule\\\\vspace{2ex}
\\\\begin{center} \\\\today\\\\end{center}
\\\\thispagestyle{empty}
\\\\spacing{''' + BODY_TEXT_SPACING + '}\\\\newpage')

tex_edited = replace(tex_edited,
    '''\\\\hypertarget{shuttletracker-users-manual}{%
\\\\section{ShuttleTracker -- User's
manual}\\\\label{shuttletracker-users-manual}}''',
    '')

tex_edited = replace(tex_edited,
    '''\\\\hypertarget{h1_appendix_''',
    '''\\\\newpage\\\\appendix\\\\hypertarget{h1_appendix_''')

tex_edited = replace(tex_edited,
    '''\\\\hypertarget{appendix-''',
    '''\\\\newpage\\\\appendix\\\\hypertarget{appendix-''')


# -- Save and compile TeX

with open('Manual.tex', 'w') as fo:
    print(tex_edited, file=fo)

with open(os.devnull, 'w') as dev_null:
    for _ in range(3):
        subprocess.call('xelatex Manual.tex'.split(), stdout=dev_null)


# -- Clean up

for fn in ['Manual.log', 'Manual.out', 'Manual.aux', 'Manual.toc', 'Manual.tex', 'Manual-edited.html']:
    if os.path.exists(fn):
        os.remove(fn)
