// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "Outline.hpp"
#include <opencv2/imgproc.hpp>


// static data
QPointF const Outline::contour_shift = {+0.5, +0.5};


QPolygonF Outline::from_cvcontour_(st::contour_t const& contour, bool use_shift)
{
  auto const point_shift = static_cast<int>(use_shift)*Outline::contour_shift;
  QVector<QPointF> pts;  pts.reserve(static_cast<int>(contour.size()));
#if 0
  int min_x = INT_MAX, max_x = INT_MIN,
      min_y = INT_MAX, max_y = INT_MIN;
#endif
  for (auto const& pt: contour) {
    pts.push_back(QPointF(pt.x, pt.y) + point_shift);
#if 0
  min_x = std::min(min_x,pt.x());
  max_x = std::max(max_x,pt.x());
  min_y = std::min(min_y,pt.y());
  max_y = std::max(max_y,pt.y());
#endif
  }
  return QPolygonF(pts);
} // from_cvcontour_()


void Outline::derive_hull_and_compute_its_area_(bool use_shift)
{
  assert(not cv_contour_.empty());
  st::contour_t cv_hull;
  cv::convexHull(cv_contour_, cv_hull, false);
  qt_hull_ = from_cvcontour_(cv_hull, use_shift);
  hull_area_ = static_cast<float>(cv::contourArea(cv_hull));
}


Outline::Outline(st::contour_t const& contour, bool use_shift)
: cv_contour_(contour)
{
  qt_contour_ = from_cvcontour_(cv_contour_, use_shift);
  contour_area_ = static_cast<float>(cv::contourArea(cv_contour_));

  derive_hull_and_compute_its_area_(use_shift);
}


//
// @arg poly -- not-shifted polygon
//
Outline::Outline(QPolygon const& poly, bool use_shift)
{
  // contour
  auto const point_shift = static_cast<int>(use_shift)*contour_shift;
  for (auto const& pt: poly) {
    qt_contour_.push_back(QPointF(pt.x(), pt.y()) + point_shift);
    cv_contour_.push_back(       {pt.x(), pt.y()});
  }
  contour_area_ = static_cast<float>(cv::contourArea(cv_contour_));

  // hull
  derive_hull_and_compute_its_area_(use_shift);
}


//
// @arg vp -- QVector not-shifted QPoints
void Outline::update(st::contour_t const& c, bool use_shift
  /*, fp_redraw <-- TODO */)
{
  // clear
  qt_contour_.clear();
  cv_contour_.clear();
  qt_hull_.clear();

  // contour
  cv_contour_ = c;
  qt_contour_ = from_cvcontour_(cv_contour_, use_shift);
  contour_area_ = static_cast<float>(cv::contourArea(cv_contour_));

  // hull
  derive_hull_and_compute_its_area_(use_shift);

  // other properties
  //toggled_ = false;
}


Outline::~Outline() { clearDepictions(); }


void Outline::toggle()
{
  mark_ ^= (1u << MarkBit::Toggledness);

  for (auto pi: qAsConst(gxpolygons_)) {
    QPen pen(pi->pen());
    pen.setStyle((not isToggled()) ? Qt::SolidLine : Qt::DotLine);
    pi->setPen(pen);
  }
}


void Outline::setExternallyHighlighted(bool marked)
{
  QColor const light_green = QColor{ 96, 255, 192, 242},
               light_red   = QColor{255, 128,  64, 242};
  mark_ =  (mark_ & ~(1u << MarkBit::ExternalHighlightedness)) //clear
         | (static_cast<unsigned int>(marked) << MarkBit::ExternalHighlightedness); //set
  for (QGraphicsPolygonItem* pi: qAsConst(gxpolygons_)) {
    QPen pen(pi->pen());
    pen.setWidthF(isExternallyHighlighted() ? 1.5           : 0.6);
    pen.setColor( isExternallyHighlighted() ? light_green   : light_red);
    pen.setStyle( isExternallyHighlighted() ? Qt::SolidLine : Qt::DashLine);
    pi->setPen(pen);
  }
}


st::QuantResults
Outline::quantify(cv::Mat const& orig_img, bool calc_inertia) const
{
  return st::quantifyInContours(orig_img, st::contours_vec_t{cv_contour_}, {},
                                calc_inertia);
}


st::QuantResults PerinuclearOutlines::quantify(cv::Mat const& orig_img, bool) const
{
  bool constexpr calc_inertia = false; // enforced
  if (kind_ == Kind::kCompleteO) {
    return st::quantifyInContours(orig_img, st::contours_vec_t{cv_contours_[1]},
                                            st::contours_vec_t{cv_contours_[0]},
                                            calc_inertia);
  } else if (kind_ == Kind::kEmpty) {
    return st::QuantResults{};
  } else {
    return st::quantifyInContours(orig_img, cv_contours_, {}, calc_inertia);
  }
}


float PerinuclearOutlines::area() const
{
#if 0
  int i=0;
  for(auto const& o:cv_contours_)
  { std::cerr << "a["<< i++ << "]=" << cv::contourArea(o) << "\n"; }
  for(auto const& o:outlines_)
  { std::cerr << "A["<< i++ << "]=" << o.area() << "\n"; }
#endif

  if (kind_ == Kind::kCompleteO) {
    assert(outlines_[1].area() - outlines_[0].area() > 0.f);
    return outlines_[1].area() - outlines_[0].area();
  } else if (kind_ == Kind::kEmpty) {
    return 0.f;
  } else {
    float sum_areas = 0.f;
    for (auto const& o: outlines_) { sum_areas += o.area(); }
    return sum_areas;
  }
}
