// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#ifndef SHUTTLETRACKER_GUI_SUPPORT_OUTLINE_HPP
#define SHUTTLETRACKER_GUI_SUPPORT_OUTLINE_HPP

#include "GUI/GUICommon.hpp"
#include "Image/Types.hpp"
#include "Image/Quantification.hpp"  // st::internal::calcEccentricityAnd...()
#include "Image/SharedUtilities.hpp" // st::centerOfMass

#include <QSet>
#include <QPolygonF>
#include <QGraphicsScene>
#include <QGraphicsPolygonItem>
#include <cassert>


class Outline {
public:
  Outline() = delete;
  Outline(st::contour_t const& contour, bool use_shift = true);
  Outline(QPolygon const& poly, bool use_shift = true);
  ~Outline();

protected:
  void update(st::contour_t const& c, bool use_shift = true);
  QPolygonF from_cvcontour_(st::contour_t const& contour, bool use_shift = true);

public:
  QPolygonF const& contour() const { return qt_contour_; }

  st::contour_t const& cv_contour() const { return cv_contour_; }

  float area() const { assert(contour_area_ >= 0.f);  return contour_area_; }

  float hull_area() const { assert(hull_area_ >= 0.f);  return hull_area_; }

  float solidity() const { assert(hull_area() > 0.f);  return area()/hull_area(); }

  // TODO: rename to calcCenter()
  cv::Point2f center() const { return st::centerOfMass(cv_contour_); }

  float eccentricity() const
  { return st::internal::calcGeometricFeatures_(cv_contour_).first; }

  // TODO: API: unnecessarily exposes internals
  QVector<QGraphicsPolygonItem*>& depictions() { return gxpolygons_; }

  void setDepictedBy(QGraphicsPolygonItem* pgp) { gxpolygons_.push_back(pgp); }

  bool isDepictedBy(QGraphicsPolygonItem const* pgp) const
  { return gxpolygons_.contains(const_cast<QGraphicsPolygonItem*>(pgp)); }

  void derive_hull_and_compute_its_area_(bool use_shift);

  void clearDepictions(QGraphicsPolygonItem* other_than = nullptr)
  {
    for (auto pgp: qAsConst(gxpolygons_)) {
      if (pgp != other_than) {
        delete pgp;
    } }
    gxpolygons_.clear();
  }

  void setDepictionsWidth(double w)
  {
    for (auto pgp: qAsConst(gxpolygons_)) {
      assert(pgp);
      QPen pen = pgp->pen();  pen.setWidthF(w);  pgp->setPen(pen);
    }
  }

  st::QuantResults
  quantify(cv::Mat const& orig_img, bool calc_inertia = true) const;

  int mark() const { return mark_; }
  void toggle();
  bool isToggled() const { return ((mark_ >> MarkBit::Toggledness) & 1u); }
  void setExternallyHighlighted(bool transloc);
  bool isExternallyHighlighted() const { return ((mark_ >> MarkBit::ExternalHighlightedness) & 1u); }

  // shift by (half,half) enables drawing the contour in the middle of the pixel
  QPointF static const contour_shift;

protected:
  enum MarkBit { Toggledness = 0,  // first bit
                 ExternalHighlightedness = 1 }; // second bit
  float contour_area_ = -1.f;
  float hull_area_    = -1.f;
  unsigned int mark_  = 0;  // a bitset

  QPolygonF     qt_contour_;
  st::contour_t cv_contour_;
  QPolygonF     qt_hull_;
  QVector<QGraphicsPolygonItem*> gxpolygons_;
}; // class Outline

// -----------------------------------------------------------------------------

struct NuclearOutline : public Outline {

  enum class Provenience {kAutoSolo, kAutoFromSplit, kAutoImproperGroup,
                          kAutoImproperOrphan, kAutoImproperDebris,
                          kManual, kUnknown};

  NuclearOutline(st::contour_t const& contour, Provenience prov, bool shift=true)
  : Outline::Outline(contour, shift),
    provenience_(prov)
  { }

  NuclearOutline(QPolygon const& polygon, Provenience prov, bool shift = true)
  : Outline::Outline(polygon, shift),
    provenience_(prov)
  { }

  Provenience provenience() const { return provenience_; }

  void update(st::contour_t const& c) { Outline::update(c); }

private:

protected:
  Provenience provenience_ = Provenience::kUnknown;
};

// -----------------------------------------------------------------------------

struct PerinuclearOutlines {

  enum class Kind {kCompleteO, kCompleteC, kMultipartC, kEmpty, kUnknown};

  PerinuclearOutlines(st::contours_vec_t const& contours, Kind kind,
                      bool shift=true)
  : kind_{kind},
    cv_contours_{contours}
  {  for (auto const& c: contours) { outlines_.push_back(Outline{c, shift}); } }

  PerinuclearOutlines(PerinuclearOutlines const& other)
  : kind_{other.kind_},
    cv_contours_{other.cv_contours_},
    outlines_{other.outlines_}
  { }

  QList<Outline> const& outlines() const { return outlines_; }
  QList<Outline> &      outlines()       { return outlines_; }
  Kind kind() const { return kind_; }

  bool isDepictedBy(QGraphicsPolygonItem const* pgp) const
  {
    for (auto const& otl: outlines_) { if (otl.isDepictedBy(pgp)) return true; }
    return false;
  }

  st::contours_vec_t const& cv_contours() const { return cv_contours_; }

  void clearDepictions(QGraphicsPolygonItem* other_than = nullptr)
  {  for (auto& o: outlines_) { o.clearDepictions(other_than); }  }

  float area() const;

  // calc_inertia = false effectively enforced in function body;
  // the argument is here for consistency with Outline::quantify()
  st::QuantResults
  quantify(cv::Mat const& orig_img, bool calc_inertia = false) const;

protected:
  Kind kind_ = Kind::kUnknown;
  st::contours_vec_t cv_contours_; // this data is in outlines_ but
                                   // it's convenient to have it in this
                                   // form for quantification
  QList<Outline> outlines_;
};

#endif // !defined(SHUTTLETRACKER_GUI_SUPPORT_OUTLINE_HPP)
