// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "PixmapCache.hpp"
#include "Image/ImageChannel.hpp"
#include <QImage>
#include <QtConcurrentMap>
#include <opencv2/core.hpp>


namespace st { // for ADL lookup
inline size_t qHash(const ImageChannel &channel) { return toHash(channel); }
}


PixmapCache::PixmapCache(ImageSequence const& source_images,
                         ImageFilter const& display_filter,
                         bool caching_on)
: images_(source_images),
  filter_(display_filter),
  caching_on_(caching_on)
{
  initializeColorTables_();

  // The ImageChannel::UNDETERMINED is associated with an overlay.
  // TODO: QT6: emplace()
  int channel_index = 0;
  for (auto const channel: images_.channels()) {
    channels_indices_.insert(channel, channel_index++);
  }
  channels_indices_.insert(ImageChannel::UNDETERMINED, channel_index);

  rebuild();

  black_pixmap_ = QPixmap{QSize{images_.imageSize().width,
                                images_.imageSize().height}};
  black_pixmap_.fill(Qt::black);
}


void PixmapCache::initializeColorTables_()
{
  QMap<ImageChannel, QRgb> const channel_color_tables_masks = {
    {ImageChannel::R,  QRgb{0xffff0000}},
    {ImageChannel::G,  QRgb{0xff00ff00}},
    {ImageChannel::B,  QRgb{0xff0000ff}},
    {ImageChannel::C,  QRgb{0xff00ffff}},
    {ImageChannel::M,  QRgb{0xffff00ff}},
    {ImageChannel::Y,  QRgb{0xffffff00}},
    {ImageChannel::BF, QRgb{0xffffffff}}
  };

  for (auto const channel: images_.channels()) {
    QVector<QRgb>& table = *color_tables_.insert(channel, {});
    auto const mask = channel_color_tables_masks[channel];
    for (unsigned int i=0u; i < 1u<<8; ++i) { table.append(qRgb(i, i, i) & mask); }
  }
}


void PixmapCache::clear()
{
  if (caching_on_) {
    cached_pixmaps_.clear();
    assert(onthefly_pixmaps_.empty());
  } else {
    onthefly_pixmaps_.clear();
    assert(cached_pixmaps_.empty());
  }
}


void PixmapCache::rebuild(long int frame_i)
{
  clear();

  auto const channels = images_.channels();
  auto const channels_indices_std = channels_indices_.toStdMap();


  auto make_pixmap_ = [this](ImageChannel ch, long ti) -> QPixmap {
    return (ch == ImageChannel::UNDETERMINED) ? makePixmapOfOverlay_(ti)
                                              : makePixmapOfChannel_(ch, ti);
  };

  auto make_all_channels_pixmaps_ = [channels_indices_std, make_pixmap_]
                                    (auto& target, long ti) -> void {
    for (auto [ch, ch_i]: channels_indices_std) {
      target[ch_i] = make_pixmap_(ch, ti);
    }
  }; // make_pixmaps_()


  if (caching_on_) { // -- rebuilding for all time frames (arg frame_i neglected)

    try {

      // prepare an empty container
      assert(cached_pixmaps_.size() == 0);
      cached_pixmaps_.reserve(static_cast<int>(images_.timepointsCount()));
      for (auto ti=0; ti<images_.timepointsCount(); ti++) {
        cached_pixmaps_.append(QVector<QPixmap>(channels_indices_.size()));
      } // for ti


      auto make_cached_pixmaps = [this, make_all_channels_pixmaps_](long ti) {
        make_all_channels_pixmaps_(cached_pixmaps_[ti], ti);
      }; // make_cached_pixmaps()


      switch (images_.storageStrategy()) {

      case ImageSequence::StorageStrategy::Mode::RAM: { // then parallel

        QVector<long> timepoint_indices(images_.timepointsCount());
        std::iota(timepoint_indices.begin(), timepoint_indices.end(), 0);
        QtConcurrent::map(timepoint_indices, make_cached_pixmaps)
                     .waitForFinished();

      } break;

      case ImageSequence::StorageStrategy::Mode::DISK: { // then serial

        for (auto ti=0; ti<images_.timepointsCount(); ti++) {
          make_cached_pixmaps(ti);
        }

      } break;

      case ImageSequence::StorageStrategy::Mode::UNDETERMINED: {}
      } // switchy if

    } catch (...) {
      cached_pixmaps_.clear();
      caching_on_ = false;
      rebuild(frame_i);
      printf("Error: Failed to allocate memory for fast display buffers. ");
      printf("Caching turned off.\n");
    }

  } else { // -- no chaching, rebuilding for the current frame only

    assert(0l <= frame_i);
    long int ti = frame_i;

    onthefly_pixmaps_ = QVector<QPixmap>(channels_indices_.size());

    auto make_onthefly_channel_pixmap = [this, make_pixmap_, ti](auto const& ch_chi) {
      const auto [ch, ch_i] = ch_chi;
      onthefly_pixmaps_[ch_i] = make_pixmap_(ch, ti);
    };


    switch (images_.storageStrategy()) {

    case ImageSequence::StorageStrategy::Mode::RAM: { // then parallel
      QtConcurrent::map(channels_indices_std, make_onthefly_channel_pixmap)
                   .waitForFinished();
    } break;

    case ImageSequence::StorageStrategy::Mode::DISK: { // then serial
      make_all_channels_pixmaps_(onthefly_pixmaps_, ti);
    } break;

    case ImageSequence::StorageStrategy::Mode::UNDETERMINED: {}
    } // switch

  } // -- if caching_on_

  // rebuild() can be called (when displaying enhanced view) and would not
  // remember blackening status of panels -- taking care:
  for (auto channel: qAsConst(blackedout_channels_)) { blackout(channel); }
}


void PixmapCache::rebuild(long int frame_i, bool desaturated)
{
  desaturated_ = desaturated;
  rebuild(frame_i);
}

void PixmapCache::blackout(ImageChannel ch)
{ blackedout_channels_.insert(ch); }


void PixmapCache::deblackout(ImageChannel ch)
{ blackedout_channels_.remove(ch); }


QPixmap const& PixmapCache::at(int frame_i, ImageChannel channel)
{
  if (blackedout_channels_.contains(channel)) {
    return black_pixmap_;
  } else {
    assert(channels_indices_.count(channel));
    int const channel_i = channels_indices_[channel];
    if (caching_on_) {
      return cached_pixmaps_[frame_i][channel_i];
    } else {
      rebuild(frame_i);
      return onthefly_pixmaps_[channel_i];
  } }
 }


QPixmap const& PixmapCache::last(int frame_i) const
{
  if (caching_on_) { return cached_pixmaps_[frame_i].back(); }
  else             { return onthefly_pixmaps_.back();        }
}


int PixmapCache::rowSize() const
{
  if (caching_on_) { return cached_pixmaps_.at(0).size(); }
  else             { return onthefly_pixmaps_    .size(); }
}


bool PixmapCache::empty() const
{
  if (caching_on_) {
    assert(onthefly_pixmaps_.empty());
    return cached_pixmaps_.empty();
  } else {
    assert(cached_pixmaps_.empty());
    return onthefly_pixmaps_.empty();
  }
}
