// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#ifndef SHUTTLETRACKER_GUI_SUPPORT_PIXMAPCACHE_HPP
#define SHUTTLETRACKER_GUI_SUPPORT_PIXMAPCACHE_HPP

#include "GUI/GUICommon.hpp"
#include <QList>
#include <QVector>
#include <QSet>
#include <QPixmap>
#include <QImage>
#include <QRgb>


class PixmapCache {

public:
  PixmapCache(ImageSequence const&, ImageFilter const&, bool caching_on = true);

  void clear();
  void rebuild(long int frame_i = 0);
  void rebuild(long int frame_i, bool desaturate);

  void blackout(ImageChannel);
  void deblackout(ImageChannel);

  QPixmap const& at(int frame_i, ImageChannel);
  QPixmap const& last(int frame_i) const;

  bool hasCache() const { return caching_on_; }
  bool empty() const;
  int rowSize() const;


protected:
  void initializeColorTables_();


  static QImage fromCvMat1_(cv::Mat const& m)
  {
    assert(m.type() == CV_8UC1);
    return QImage{static_cast<unsigned char const*>(m.data),
                  static_cast<int>(m.cols),
                  static_cast<int>(m.rows),
                  static_cast<int>(m.step),
                  QImage::Format_Grayscale8};
  }


  static QImage fromCvMat1_(cv::Mat const& m, QVector<QRgb> const& lut)
  {
    assert(m.type() == CV_8UC1);
    QImage i{static_cast<unsigned char const*>(m.data),
                  static_cast<int>(m.cols),
                  static_cast<int>(m.rows),
                  static_cast<int>(m.step),
                  QImage::Format_Indexed8};
    i.setColorTable(lut);
    return i;
  }


  static QImage fromCvMat3_(cv::Mat const& m)
  {
    assert(m.type() == CV_8UC3);
    return QImage{static_cast<unsigned char const*>(m.data),
                  static_cast<int>(m.cols),
                  static_cast<int>(m.rows),
                  static_cast<int>(m.step),
                  QImage::Format_BGR888};
  }


  QPixmap makePixmapOfChannel_(ImageChannel channel, int timepoint_index) const
  {
    cv::Mat m = images_.image(channel, timepoint_index);
    filter_(m);

    if (desaturated_) {
      QImage const i = fromCvMat1_(m);
      return QPixmap::fromImage(std::move(i));
    } else {
      QVector<QRgb> const& lut = color_tables_[channel];
      //QImage const i = fromCvMat1_(m).convertToFormat(QImage::Format_Indexed8, lut);
      QImage const i = fromCvMat1_(m, lut);
      return QPixmap::fromImage(std::move(i));
    }
  }


  QPixmap makePixmapOfOverlay_(int timepoint_index) const
  {
    cv::Mat const m = images_.overlayImage(timepoint_index, filter_);
    // Note: By design, LUT (color_tables_) are not used on overlay images.
    QImage const i = fromCvMat3_(m);
    auto const p = QPixmap::fromImage(std::move(i));
    return p;
  }


  QMap<ImageChannel, QVector<QRgb>> color_tables_;

  ImageSequence const& images_;
  ImageFilter const& filter_;
  bool caching_on_ = true;

  QMap<ImageChannel, int> channels_indices_;

  typedef QList<QVector<QPixmap>> cached_pixmaps_container_t;
  cached_pixmaps_container_t cached_pixmaps_;

  typedef QVector<QPixmap> onthefly_pixmaps_container_t;
  onthefly_pixmaps_container_t onthefly_pixmaps_;

  bool desaturated_ = false;
  QSet<ImageChannel> blackedout_channels_;
  QPixmap black_pixmap_;
};

#endif // !defined(SHUTTLETRACKER_GUI_SUPPORT_PIXMAPCACHE_HPP)
