// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "ScriptEditor.hpp"
#include "LineNumberArea.hpp"
#include <QtGui>
#include <QAbstractItemView>
#include <QScrollBar>
#include <QCompleter>


ScriptEditor::ScriptEditor(QWidget* parent, bool show_line_numbers)
  : QPlainTextEdit{parent},
    highlighter_{nullptr},
    completer_{nullptr},
    current_line_highlighting_enabled_(true),
    showing_line_numbers_(show_line_numbers)
{
  // setBackgroundVisible(true);
  line_number_area_ = new LineNumberArea{this};

  connect(this, SIGNAL(blockCountChanged(int)),
          this, SLOT(updateLineNumberAreaWidth(int)));
  connect(this, SIGNAL(updateRequest(QRect,int)),
          this, SLOT(updateLineNumberArea(QRect,int)));
  connect(this, SIGNAL(cursorPositionChanged()),
          this, SLOT(highlightCurrentLine()));

  updateLineNumberAreaWidth(0);
  highlightCurrentLine();

  QFont font;
#if defined(Q_OS_MACOS)
  font.setFamily("Monaco");
  font.setPointSize(12);
  lineNumberAreaFont.setFamily("Monaco");
  lineNumberAreaFont.setPointSize(12);
#elif defined(Q_OS_WIN32) || defined(Q_OS_WIN64)
  font.setFamily("Consolas"); // other safe choices: Courier, Lucida Console
  font.setPointSize(9);
  line_number_area_font_.setFamily("Courier");
  line_number_area_font_.setPointSize(9);
#else
  font.setFamily("Monospace");
  font.setPointSize(10);
  line_number_area_font_.setFamily("Monospace");
  line_number_area_font_.setPointSize(9);
#endif
  font.setFixedPitch(true);
  line_number_area_font_.setFixedPitch(true);
  setFont(font);
  setTabStopDistance(4*fontMetrics().maxWidth());

  highlighter_ = new ScriptHighlighter(this->document());
}


ScriptEditor::~ScriptEditor()
{
  delete highlighter_;
  delete line_number_area_;
  delete completer_;
}


int ScriptEditor::lineNumberAreaWidth()
{
  if (not showing_line_numbers_) { return 0; }

  int digits = 1;
  int max = qMax(1, blockCount());
  while (max >= 10) {
    max /= 10;
    ++digits;
  }

  // '2' because of: space in the front + trailing space
  int space = fontMetrics().horizontalAdvance(QLatin1Char('8')) * (digits + 2);

  return space;
}


void ScriptEditor::updateLineNumberAreaWidth(int /* newBlockCount */)
{
  setViewportMargins(lineNumberAreaWidth(), 0, 0, 0);
}


void ScriptEditor::updateLineNumberArea(QRect const& rect, int dy)
{
  if (dy) {
    line_number_area_->scroll(0, dy);
  } else {
    line_number_area_->update(0, rect.y(), line_number_area_->width(), rect.height());
  }
  if (rect.contains(viewport()->rect())) {
    updateLineNumberAreaWidth(0);
  }
}


void ScriptEditor::resizeEvent(QResizeEvent *e)
{
  QPlainTextEdit::resizeEvent(e);

  QRect cr = contentsRect();
  line_number_area_->setGeometry(QRect(cr.left(), cr.top(), lineNumberAreaWidth(),
                                    cr.height()));
}


void ScriptEditor::highlightCurrentLine()
{
  QList<QTextEdit::ExtraSelection> extraSelections;

  if (!isReadOnly() and current_line_highlighting_enabled_) {
    QTextEdit::ExtraSelection selection;
    QColor lineColor = QColor(48, 48, 40);
    selection.format.setBackground(lineColor);
    selection.format.setProperty(QTextFormat::FullWidthSelection, true);
    selection.cursor = textCursor();
    selection.cursor.clearSelection();
    extraSelections.append(selection);
  }

  setExtraSelections(extraSelections);
}


void ScriptEditor::setCurrentLineHighlighting(bool whether)
{
  current_line_highlighting_enabled_ = whether;
}


void ScriptEditor::lineNumberAreaPaintEvent(QPaintEvent* event)
{
  QPainter painter{line_number_area_};
  QColor const background_color = QColor{32,32,32};
  painter.fillRect(event->rect(), background_color);

  QTextBlock block = firstVisibleBlock();
  int block_number = block.blockNumber();
  int top = static_cast<int>(blockBoundingGeometry(block)
                             .translated(contentOffset()).top());
  int bottom = top + static_cast<int>(blockBoundingRect(block).height());

  while (block.isValid() && top <= event->rect().bottom()) {
    if (block.isVisible() && bottom >= event->rect().top()) {
      QString number = QString::number(block_number + 1) + " ";
      painter.setPen(Qt::darkGray);
      painter.setFont(line_number_area_font_);
      painter.drawText(0, top, line_number_area_->width(), fontMetrics().height(),
          Qt::AlignRight, number);
    }
    block = block.next();
    top = bottom;
    bottom = top + static_cast<int>(blockBoundingRect(block).height());
    ++block_number;
  }
}


void ScriptEditor::setCompleterWordlist(QStringList const& words)
{
  completer_ = new QCompleter{this};
  completer_->setModel(
      /*modelFromFile(":/resources/completion/keywords.txt")*/
      new QStringListModel(words, completer_));
  completer_->setModelSorting(QCompleter::CaseInsensitivelySortedModel);
  completer_->setCaseSensitivity(Qt::CaseInsensitive);
  completer_->setWrapAround(false);
  setCompleter(completer_);
}


void ScriptEditor::setCompleter(QCompleter* acompleter)
{
  if (completer_) { disconnect(completer_, nullptr, this, nullptr); }

  completer_ = acompleter;

  if (not completer_) { return; }

  completer_->setWidget(this);
  completer_->setCompletionMode(QCompleter::PopupCompletion);
  completer_->setCaseSensitivity(Qt::CaseInsensitive);
  connect(completer_, SIGNAL(activated(QString)),
          this, SLOT(insertCompletion(QString)));
}


QAbstractItemModel* ScriptEditor::modelFromFile(QString const& fileName)
{
  QFile file(fileName);
  if (!file.open(QFile::ReadOnly)) { return new QStringListModel(completer_); }

  QStringList words;
  while (!file.atEnd()) {
    QByteArray line = file.readLine();
    if (!line.isEmpty()) {
      words << line.trimmed();
  } }

  return new QStringListModel(words, completer_);
}


void ScriptEditor::goToLine(int line_number)
{
  QTextCursor text_cursor(this->document()->findBlockByLineNumber(line_number - 1));
  text_cursor.select(QTextCursor::LineUnderCursor);
  this->setTextCursor(text_cursor);
}


void ScriptEditor::insertCompletion(QString const& completion)
{
  if (completer_->widget() != this) { return; }
  QTextCursor tc = textCursor();
  int extra = completion.length() - completer_->completionPrefix().length();
  tc.movePosition(QTextCursor::Left);
  tc.movePosition(QTextCursor::EndOfWord);
  tc.insertText(completion.right(extra));
  setTextCursor(tc);
}


QString ScriptEditor::textUnderCursor() const
{
  QTextCursor tc = textCursor();
  tc.select(QTextCursor::WordUnderCursor);
  return tc.selectedText();
}


void ScriptEditor::focusInEvent(QFocusEvent* e)
{
  if (completer_) { completer_->setWidget(this); }
  QPlainTextEdit::focusInEvent(e);
}


void ScriptEditor::keyPressEvent(QKeyEvent* e)
{
  if (completer_ and completer_->popup()->isVisible()) {
    // The following keys are forwarded by the completer to the widget
    switch (e->key()) {
      case Qt::Key_Enter:
      case Qt::Key_Return:
      case Qt::Key_Escape:
      case Qt::Key_Tab:
      case Qt::Key_Backtab:
        e->ignore();
        return; // let the completer do default behavior
      default:
        break;
    }
  }

  bool isShortcut = ((e->modifiers() & Qt::ControlModifier) and e->key()==Qt::Key_E);
  if (not completer_ or not isShortcut) { // don't process the shortcut when we have a completer
    QPlainTextEdit::keyPressEvent(e);
  }

  const bool ctrlOrShift = e->modifiers()
      & (Qt::ControlModifier | Qt::ShiftModifier);
  if (!completer_ or (ctrlOrShift and e->text().isEmpty())) {
    return;
  }

  static QString eow("~!@#$%^&*()_+{}|:\"<>?,./;'[]\\-="); // e-o-word, spaces++
  bool hasModifier = (e->modifiers() != Qt::NoModifier) and not ctrlOrShift;
  QString completionPrefix = textUnderCursor();

  if (not isShortcut and (hasModifier or e->text().isEmpty()
      or completionPrefix.length() < 5 // "color".size()
      or eow.contains(e->text().right(1)))) {
    completer_->popup()->hide();
    return;
  }

  if (completionPrefix != completer_->completionPrefix()) {
    completer_->setCompletionPrefix(completionPrefix);
    completer_->popup()->setCurrentIndex(completer_->completionModel()->index(0, 0));
  }
  QRect cr = cursorRect();
  cr.setWidth(completer_->popup()->sizeHintForColumn(0)
              + completer_->popup()->verticalScrollBar()->sizeHint().width());
  completer_->complete(cr); // pop-up!
}
