// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "ScriptHighlighter.hpp"
#include "GUI/Windows/MainWindow.hpp"
#include <QtGui>


ScriptHighlighter::ScriptHighlighter(QTextDocument *parent)
  : QSyntaxHighlighter(parent)
{
  QString const delim{"\\b"};

  HighlightingRule rule;

  // -- primary keywords
  primaryKeywordFormat.setForeground(QColor{248,128,0});
  // primaryKeywordFormat.setFontWeight(QFont::Bold);
  QStringList primaryKeywordPatterns;
  primaryKeywordPatterns
      << delim + "break"    + delim
      << delim + "case"     + delim
      << delim + "catch"    + delim
      << delim + "const"    + delim
      << delim + "continue" + delim
      << delim + "default"  + delim
      << delim + "delete"   + delim
      << delim + "do"       + delim
      << delim + "else"     + delim
      << delim + "finally"  + delim
      << delim + "for"      + delim
      << delim + "function" + delim
      << delim + "if"       + delim
      << delim + "in"       + delim
      << delim + "instanceof"+delim
      << delim + "new"      + delim
      << delim + "return"   + delim
      << delim + "switch"   + delim
      << delim + "this"     + delim
      << delim + "throw"    + delim
      << delim + "try"      + delim
      << delim + "typeof"   + delim
      << delim + "var"      + delim
      << delim + "void"     + delim
      << delim + "while"    + delim
      << delim + "with"     + delim
      << delim + "for\\s+each\\s+frame" + delim; //(?=\\{)"
  for (QString const& pattern: qAsConst(primaryKeywordPatterns)) {
    rule.pattern = QRegularExpression(pattern);
    rule.format = primaryKeywordFormat;
    highlightingRules.append(rule);
  }

  // -- secondary keywords
  secondaryKeywordFormat.setForeground(Qt::white);
  // secondaryKeywordFormat.setFontWeight(QFont::Bold);
  QStringList secondaryKeywordPatterns;
  auto const api_funcs = MainWindow::api_functions.keys();
  for (auto const& fun_name: qAsConst(api_funcs)) {
    secondaryKeywordPatterns << delim + fun_name + delim;
  }
  secondaryKeywordPatterns << delim + "this" + delim;
  for (QString const& pattern: qAsConst(secondaryKeywordPatterns)) {
    rule.pattern = QRegularExpression(pattern);
    rule.format = secondaryKeywordFormat;
    highlightingRules.append(rule);
  }

  // -- tertiary keywords
  tertiaryKeywordFormat.setForeground(QColor{255,192,32}.darker(85));
  QStringList tertiaryKeywordPatterns;
  for (auto const& param_name: qAsConst(MainWindow::api_parameters)) {
    tertiaryKeywordPatterns << delim + param_name + delim;
  }
  for (QString const& pattern: qAsConst(tertiaryKeywordPatterns)) {
    rule.pattern = QRegularExpression(pattern);
    rule.format = tertiaryKeywordFormat;
    highlightingRules.append(rule);
  }

  // -- quaternary keywords
  quaternaryKeywordFormat.setForeground(QColor{128,224,192});
  QStringList quaternaryKeywordPatterns;
  for (auto const& param_name: qAsConst(MainWindow::api_specials)) {
    quaternaryKeywordPatterns << delim + param_name + delim;
  }
  for (QString const& pattern: qAsConst(quaternaryKeywordPatterns)) {
    rule.pattern = QRegularExpression(pattern);
    rule.format = quaternaryKeywordFormat;
    highlightingRules.append(rule);
  }

  // -- quotations
  quotationFormat.setForeground(QColor{128,248,0});
  rule.pattern = QRegularExpression{"\"[^\"]*\""};
  rule.format = quotationFormat;
  highlightingRules.append(rule);

  // -- comments
  QColor commentColor = QColor{96,144,144};
  singleLineCommentFormat.setForeground(commentColor);
  rule.pattern = QRegularExpression{"//[^\n\r]*"};
  rule.format = singleLineCommentFormat;
  highlightingRules.append(rule);
  //
  multiLineCommentFormat.setForeground(commentColor);
  commentStartExpression = QRegularExpression{"/\\*"};
  commentEndExpression   = QRegularExpression{"\\*/"};
}


void ScriptHighlighter::highlightBlock(QString const& text)
{
  for (HighlightingRule const& rule: qAsConst(highlightingRules)) {
    QRegularExpressionMatchIterator mit = rule.pattern.globalMatch(text);
    while (mit.hasNext()) {
      QRegularExpressionMatch match = mit.next();
      setFormat(match.capturedStart(), match.capturedLength(), rule.format);
  } }

  setCurrentBlockState(0);

  int startIndex = 0;
  if (previousBlockState() != 1) {
    startIndex = text.indexOf(commentStartExpression);
  }

  while (startIndex >= 0) {
    QRegularExpressionMatch match = commentEndExpression.match(text, startIndex);
    int endIndex = match.capturedStart();
    int commentLength = 0;
    if (endIndex == -1) {
      setCurrentBlockState(1);
      commentLength = text.length() - startIndex;
    } else {
      commentLength = endIndex - startIndex + match.capturedLength();
    }
    setFormat(startIndex, commentLength, multiLineCommentFormat);
    startIndex = text.indexOf(commentStartExpression, startIndex + commentLength);
  }
}
