// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "Track.hpp"
#include <cstdlib> // abs()


TrackSegment::TrackSegment( Track const* parent,  Outline *outline,
    int frame_index,  TrackSegment const* previous_segment,
    Outline const* previous_outline )
: parent_{parent}, outline_{outline}, frame_index_{frame_index}
{
  assert(parent_);
  assert(outline_);
  assert(previous_segment or previous_outline);
  auto const com = st::centerOfMass(outline_->cv_contour());
  auto const prev_com = (previous_segment)
      ? st::centerOfMass(previous_segment->outline_->cv_contour())
      : st::centerOfMass(previous_outline->cv_contour());
  line_ = QLineF{static_cast<float>(prev_com.x), static_cast<float>(prev_com.y),
                 static_cast<float>(     com.x), static_cast<float>(     com.y)};
}


Track::Track( \
    QList<TrackSegment>::Iterator begin, QList<TrackSegment>::Iterator end,
    Outline* initial_outline, int initial_frame_index)
: initial_outline_{initial_outline},
  initial_frame_index_{initial_frame_index},
  self_cleanup_{false} // this would be likely a temporary copy
{
  assert(initial_outline_);
  assert(segments_.empty());
  for (auto it = begin; it != end; ++it) { segments_.push_back(*it); }
}


Track::Track(QList<TrackSegment>const& segments1, Outline* initial_outline1,
                 long int initial_frame_index1,
             QList<TrackSegment>const& segments2, Outline* initial_outline2,
                 long int initial_frame_index2)
: initial_outline_{initial_outline1},
  initial_frame_index_{initial_frame_index1},
  self_cleanup_{false} // this would be likely a temporary copy
{
  assert(initial_outline_);
  assert(segments_.empty());
  for (auto const& s1: segments1) { segments_.push_back(s1); }
  segments_.push_back( \
      TrackSegment(this,initial_outline2,initial_frame_index2,
                      (not segments_.empty())
                          ? &segments_.back() : nullptr,
                      (not segments_.empty())
                          ? segments_.back().outline_ : initial_outline1 ) );
  for (auto const& s2: segments2) { segments_.push_back(s2); }
}


bool Track::updateAt(Outline const* o)
{
  assert(segments_.empty() != marks_.empty());
  assert(o);

  auto const com = st::centerOfMass(o->cv_contour());

  if (o == initial_outline_) {

    if (segments_.empty()) {
      // displace stub marks:
      for (auto& m: marks_) {
        qreal const w = 3.0; // PARAM 2b779541-389f-47df-9e6a-03d9e1976042
        m->setPos(com.x-w/2, com.y-w/2);
      }
    } else {
      // displace the initial point of the first segment:
      assert(not segments_.empty());
      segments_.front().line_.setPoints( QPointF(com.x, com.y),
                                         segments_.front().line_.p1() );
      for (auto& gxl: segments_.front().gxlines_) {
        gxl->setLine(segments_.front().line_);
      }
    }
    return true;

  } else { // o is not an initial outline

    int seg_i = 0;
    int seg_o_i = -1;
    for (auto& s: segments_) {
      if (o == s.outline()) {
        // displace the segment point:
        s.line_.setPoints( s.line_.p1(), QPointF(com.x, com.y) );
        for (auto& gxl: s.gxlines_) { gxl->setLine(s.line_); }
        seg_o_i = seg_i;
        break;
      }
      seg_i++;
    }

    bool const found = (seg_o_i >= 0);
    if (found) {
      bool const found_last_outline_in_track = (seg_o_i+1 == segments_.size());
      if (not found_last_outline_in_track) {
        assert(seg_o_i + 1 < segments_.size());
        auto& s = segments_[seg_o_i + 1];
        s.line_.setPoints( QPointF(com.x, com.y), s.line_.p2() );
        for (auto& gxl: s.gxlines_) { gxl->setLine(s.line_); }
      }
    }

    return found;
  }
}


//static
bool Track::splitAt(Track* track, TrackSegment* cut_segment,
                     QList<Track>* tracks)
{
  assert(track);
  assert(cut_segment);
  assert(tracks);

  // -- splitting a solitary segment
  //
  if (track->nodesCount() == 2) {
    track->splitSingle(tracks);
    return true;
  }


  int segment_i = 0;
  for (TrackSegment& s: track->segments()) {
    if (&s == cut_segment) { break; }
    segment_i++;
  }
  size_t const cut_segment_i = segment_i; // const alias

  // -- trimming at the head
  //
  if (cut_segment_i == 0) {
    track->trimHead(tracks);
    return false; // not split, just trimmed

  // -- trimming at the tail
  //
  } else if (cut_segment_i == static_cast<size_t>(track->segments().size() - 1)) {
    track->trimTail(tracks);
    return false; // not split, just trimmed

  // -- cutting somewhere in the middle
  //
  } else {
    track->split(segment_i,tracks);
    return true;
  }

  assert(false); // should never reach this place
  return false;
}


void Track::trimHead(QList<Track>* tracks)
{
  assert(segments_.size() >= 2);
  tracks->push_back( Track(initial_outline_,initial_frame_index_,false) );
  tracks->back().setSelfCleanup(true);
  tracks->back().initialOutline()->setDepictionsWidth(0.5);//PARAM:track_nuc_del

  initial_frame_index_ = segments_.front().frameIndex();
  initial_outline_ = segments_.front().outline();
  initial_outline_->setDepictionsWidth(0.5); //PARAM:track_nuc_del

  segments_.removeAt(0);
}


void Track::trimTail(QList<Track>* tracks)
{
  assert(segments_.size() >= 2);

  segments_.back().outline()->setDepictionsWidth(0.5); //PARAM:track_nuc_del

  tracks->push_back( Track(segments().back().outline(),
                           segments().back().frameIndex(), false) );
  tracks->back().setSelfCleanup(true);
  tracks->back().initialOutline()->setDepictionsWidth(0.5);//PARAM:track_nuc_del

  segments_.removeAt(segments_.size() - 1);
}


void Track::split(int cut_segment_i, QList<Track>* tracks)
{
  assert(segments_.size() >= 2);

  setTrackedNucleiContourWidth(0.5); // PARAM CONSISTENCY:track_nuc_del

  auto it = segments_.begin();
  it += cut_segment_i;

  // -- copy at head
  tracks->push_back( Track(segments_.begin(),it, initial_outline_,
                           initial_frame_index_) );
  tracks->back().setSelfCleanup(true);

  // -- copy at tail
  auto itpp = it; ++itpp;
  tracks->push_back(Track(itpp,segments_.end(),it->outline_,it->frameIndex()));
  tracks->back().setSelfCleanup(true);

  // -- remove split track
  detachDepictions();       // prevent deallocs
  int ti = 0;
  for (auto& t: *tracks) {
    if (&t == this) {
      tracks->removeAt(ti);  // (deallocs won't be performed)
      break;
    }
    ti++;
  } // for
}


void Track::splitSingle(QList<Track>* tracks)
{
  assert(segments_.size() == 1);

  setTrackedNucleiContourWidth(0.5); // PARAM CONSISTENCY:track_nuc_del

  tracks->push_back( Track(initial_outline_,initial_frame_index_,false) );
  tracks->back().setSelfCleanup(true);
  tracks->push_back( Track(segments_.back().outline_,
                           segments_.back().frameIndex(),false) );
  tracks->back().setSelfCleanup(true);

  // -- remove split track
  detachDepictions();       // prevent deallocs
  int ti = 0;
  for (auto& t: *tracks) {
    if (&t == this) {
      tracks->removeAt(ti);  // (deallocs won't be performed)
      break;
    }
    ti++;
  } // for
}


//static
std::pair<bool,std::string>
Track::merge(Track* track1, Track* track2, QList<Track>* tracks)
{
  assert(track1);
  assert(track2);

  int frame1i = -1,  frame2i = -1;
  int const frame1i_A = track1->firstFrameIndex();
  int const frame2i_A = track2->firstFrameIndex();
  int const frame1i_B = (not track1->segments().empty())
                          ? track1->segments().back().frameIndex()
                          : frame1i_A;
  int const frame2i_B = (not track2->segments().empty())
                          ? track2->segments().back().frameIndex()
                          : frame2i_A;
  // (1/4): 1A..2A
  if (       abs(frame1i_A - frame2i_A) == 1) {
    frame1i =    frame1i_A;
    frame2i =             frame2i_A;

  // (2/4): 1A..2B
  } else if (abs(frame1i_A - frame2i_B) == 1) {
    frame1i =    frame1i_A;
    frame2i =                frame2i_B;

  // (3/4): 1B..2A
  } else if (abs(frame1i_B - frame2i_A) == 1) {
    frame1i =    frame1i_B;
    frame2i =                frame2i_A;

  // (4/4): 1B..2B
  } else if (abs(frame1i_B - frame2i_B) == 1) {
    frame1i =    frame1i_B;
    frame2i =                frame2i_B;

  } else {
    assert(abs(frame1i - frame2i) != 1);
    if ((frame1i == frame2i) and (frame1i != -1) and (frame2i != -1)) {
      return {false,"Cannot merge tracks at the same time frame"};
    } else {
      return {false,"Cannot merge tracks at non-consecutive time frames"};
    }
  }

  bool const tr1_before_tr2 = (frame1i < frame2i);
  Track* const track_early = tr1_before_tr2 ? track1 : track2;
  Track* const track_late  = tr1_before_tr2 ? track2 : track1;

  tracks->push_back( \
      Track(track_early->segments(), track_early->initialOutline(),
                track_early->firstFrameIndex(),
            track_late->segments(),
               track_late->initialOutline(),track_late->firstFrameIndex() ));
  tracks->back().setSelfCleanup(true);

  for (auto pt: {track_early, track_late}) {
    pt->detachDepictions();   // disown
    int ti = 0;
    for (auto& t: *tracks) {
      if (&t == pt) {
        tracks->removeAt(ti);  // (deallocs won't be performed)
        break;
      }
      ti++;
    } // for each track
  } // for both merged tracks

  return {true,""};
}
