// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#ifndef SHUTTLETRACKER_GUI_SUPPORT_TRACK_HPP
#define SHUTTLETRACKER_GUI_SUPPORT_TRACK_HPP

#include "GUI/GUICommon.hpp"
#include "Image/Quantification.hpp"
#include "GUI/Support/Outline.hpp"
#include <QList>
#include <QHash>
#include <QGraphicsLineItem>

class Outline;
class TrackSegment;
class Track;

class TrackSegment {
public:
  TrackSegment( Track const* parent, Outline* outline, int frameIndex,
                TrackSegment const* previous_segment = nullptr,
                Outline const* previous_outline = nullptr );

  ~TrackSegment() {  clearDepictions();  }

  void setDepictedBy(QGraphicsLineItem* pgl)
  {  gxlines_.push_back(pgl);  }

  bool isDepictedBy(QGraphicsLineItem const* pgl) const
  {  return gxlines_.contains(const_cast<QGraphicsLineItem*>(pgl));  }

  void detachDepictions() {  gxlines_.clear();  }

  void clearDepictions()
  {  for (auto pgl: qAsConst(gxlines_)) { assert(pgl); delete pgl; }  gxlines_.clear();  }

  Outline const* outline() const {  return outline_;  }
  Outline      * outline()       {  return outline_;  }
  int frameIndex() const {  return frame_index_;  }
  QLineF const& line() const {  return line_;  }

  void setGxlinesWidth(double w)
  {
    assert(w > 0);
    for (auto pgl: qAsConst(gxlines_)) {
      assert(pgl);
      QPen pen = pgl->pen();  pen.setWidthF(w);  pgl->setPen(pen);
    }
  }

  void setOutlineWidth(double w)
  {
    assert(w > 0);
    assert(outline_);
    outline_->setDepictionsWidth(w);
  }

protected:
  Track const* parent_;
  Outline* outline_;  // non-const to enable changes in contour width
  int frame_index_;
  QLineF line_;
  QList<QGraphicsLineItem*> gxlines_;

friend class Track;
}; // class TrackSegment


// TODO: Would QGraphicsItemGroup be useful for anything here?
//
class Track {
public:

  Track(Outline* outline, long int frame_index, bool self_cleanup = true)
  : initial_outline_{outline},
    initial_frame_index_{frame_index},
    self_cleanup_{self_cleanup}
  {
    assert(outline);
  }

  ~Track() {  if (self_cleanup_) { clearDepictions(); }  }

  Outline const* last() const
  {
    if (segments_.empty()) {
      assert(initial_outline_);
      return initial_outline_;
    } else {
      return segments_.back().outline_;
    }
  }

  void append(Outline* outline, int frame_index)
  {
    if (segments_.empty()) {
      segments_.push_back( \
          TrackSegment(this, outline, frame_index, nullptr, initial_outline_));
    } else {
      segments_.push_back( \
          TrackSegment(this, outline, frame_index, &segments_.back()));
    }
  }

  void setTraceWidth(double w)
  {
    for (auto& pmr: marks_) {
      QPen pen = pmr->pen();  pen.setWidthF(w);  pmr->setPen(pen);
    }
    for (auto& s: segments_) { s.setGxlinesWidth(w); }
  }

  void setTrackedNucleiContourWidth(double w)
  {
    assert(initial_outline_);
    initial_outline_->setDepictionsWidth(w);
    for (auto& s: segments_) { s.setOutlineWidth(w); }
  }

  void trimHead(QList<Track>* tracks);
  void trimTail(QList<Track>* tracks);
  void split(int cut_segment_i, QList<Track>* tracks);
  void splitSingle(QList<Track>* tracks);

  int nodesCount() const
  {  return segments_.size() + (initial_outline_? 1 : 0);  }

  int firstFrameIndex() const {  return initial_frame_index_;  }

  int lastFrameIndex() const
  {
    if (segments_.empty()) {
      assert(initial_outline_); // hopefully initial_frame_index_ also init'd
      return initial_frame_index_;
    } else {
      return segments_.back().frameIndex();
    }
  }

  QList<TrackSegment> const& segments() const {  return segments_;  }
  QList<TrackSegment>      & segments()       {  return segments_;  }

  Outline const* initialOutline() const {  return initial_outline_;  }
  Outline      * initialOutline()       {  return initial_outline_;  }


  void setDepictedBy(QGraphicsRectItem* mark) {  marks_.push_back(mark);  }

  bool isDepictedBy(QGraphicsLineItem const* li) const
  {
    for (auto const& s: segments_) {
      if (s.isDepictedBy(li)) { return true; }
    }
    return false;
  }

  bool isDepictedBy(QGraphicsRectItem const* ri) const
  {
    for (auto const& pgr: marks_) {
      if (pgr == ri) { return true; }
    }
    return false;
  }

  bool isDepictedBy(QGraphicsItem const* gi) const
  {
    if (       gi->type() == QGraphicsRectItem::Type) {
      return isDepictedBy(static_cast<QGraphicsRectItem const*>(gi));
    } else if (gi->type() == QGraphicsLineItem::Type) {
      return isDepictedBy(static_cast<QGraphicsLineItem const*>(gi));
    } else {
      return false;
    }
  }

  bool hasDepictions() const
  {
    if (not segments().empty()) {
      return (not segments().front().gxlines_.empty());
    } else {
      return (not marks_.empty());
    }
  }

  void clearDepictions()
  {
    clearMarks_();
    for (auto& s: segments_) {  s.clearDepictions();  }
  }

  void detachDepictions()
  {
    marks_.clear();
    for (auto& s: segments_) {  s.detachDepictions();  }
  }

  // @return true if outline found in the track == geometry was updated
  bool updateAt(Outline const* o);

  void setSelfCleanup(bool selfcl) {  self_cleanup_ = selfcl;  }


  void increaseConfidence(double v) {  confidence_ += v;  }

  double confidence() const {  return confidence_;  }


  void setRevised(bool r) {  revised_ = r;  }

  bool isRevised() const {  return revised_;  }

  static
  bool splitAt(Track* track, TrackSegment* segment, QList<Track>* tracks);

  static
  std::pair<bool, std::string>
  merge(Track* track1, Track* track2, QList<Track>* tracks);

protected:

  void clearMarks_()
  {  for (auto pgr: qAsConst(marks_)) { assert(pgr); delete pgr; }  marks_.clear();  }

  Track( QList<TrackSegment>::Iterator begin, QList<TrackSegment>::Iterator end,
         Outline* initialOutline, int initial_frame_index );

  Track( QList<TrackSegment>const& segments1, Outline* initial_outline1,
             long int initial_frame_index1,
         QList<TrackSegment>const& segments2, Outline* initial_outline2,
             long int initial_frame_index2 );

  QList<TrackSegment> segments_ = {};

  Outline* initial_outline_ = nullptr;  // non-const => can change contour width
  long int initial_frame_index_ = -1;

  // Properties:
  double confidence_ = 0.0;
  bool revised_ = false;

  QList<QGraphicsRectItem*> marks_ = {};

  bool self_cleanup_ = true;
}; // class Track


#endif // !defined(SHUTTLETRACKER_GUI_SUPPORT_TRACK_HPP)
