// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "ZoomableGraphicsView.hpp"
#include <QWheelEvent>
#include <QScrollBar>
#include <QGraphicsItem>
#include <cassert>


// static (shared)
QVector<QPoint> ZoomableGraphicsView::contouri_;
QVector<QPointF> ZoomableGraphicsView::contourf_;


ZoomableGraphicsView::ZoomableGraphicsView(QGraphicsScene* scene,
                                           QWidget* parent)
  : QGraphicsView(scene, parent)
{ }


ZoomableGraphicsView::~ZoomableGraphicsView()
{
  // ~ZoomableGraphicsView may indirectly call call mouseMoveEvent(), which sets
  // sliders of other ZoomableGraphicsView objects, which may be already
  // destroyed, so all mutual dependencies should cease:
  co_scrolling_.clear();
}


void ZoomableGraphicsView::wheelEvent(QWheelEvent* event)
{
  // if e.g. Ctrl pressed, use an original functionality:
  if(event->modifiers() & Qt::ControlModifier) {
    QGraphicsView::wheelEvent(event);
  // else override:
  } else {
    if (event->angleDelta().y() > 0) {
      scale(1./scaling_factor_, 1./scaling_factor_);
      foreach (ZoomableGraphicsView *pz, co_scrolling_) {
        pz->scale(1./scaling_factor_, 1./scaling_factor_);
      }
    } else {
      scale(scaling_factor_, scaling_factor_);
      foreach (ZoomableGraphicsView *pz, co_scrolling_) {
        pz->scale(scaling_factor_, scaling_factor_);
  } } }

  //qDebug() << sceneRect();
  //qDebug() << geometry();
  //qDebug() << matrix();
  //qDebug() << transform();
  //qDebug() << viewportTransform(); // contains info on scrolling
  //qDebug() << visibleRegion();
  //qDebug() << this->viewport()->geometry();
  //qDebug() << x() << " " << y();
  //qDebug() << width() << "x" << height();
}


void ZoomableGraphicsView::keyPressEvent(QKeyEvent* event)
{
  if (event->key() == Qt::Key_1) {
    resetTransform();
    foreach (ZoomableGraphicsView *pz, co_scrolling_) { pz->resetTransform(); }
    event->accept();
  } else if (   (event->key() == Qt::Key_Plus)
             or (event->key() == Qt::Key_Equal)) {
    scale(1./scaling_factor_, 1./scaling_factor_);
    foreach (ZoomableGraphicsView*pz, co_scrolling_) {
      pz->scale(1./scaling_factor_, 1./scaling_factor_);
    }
    event->accept();
  } else if (   (event->key() == Qt::Key_Minus)
             or (event->key() == Qt::Key_Underscore)) {
    scale(scaling_factor_, scaling_factor_);
    foreach (ZoomableGraphicsView *pz, co_scrolling_) {
      pz->scale(scaling_factor_, scaling_factor_);
    }
    event->accept();
  } else {
    event->ignore();
  }
}


void ZoomableGraphicsView::mousePressEvent(QMouseEvent* event)
{
#if 0
  if (event->modifiers() & Qt::ControlModifier) {
    if (scene()->selectedItems().size() > 2) {
      auto sel_item_type = scene()->selectedItems().front()->type();
      bool const trace_selected =    (sel_item_type == QGraphicsLineItem::Type)
                                  or (sel_item_type == QGraphicsRectItem::Type);
      if (not trace_selected) { // then: prevent multiple selection
        event->accept();
        return;
      } else {
        //
      }
    } // if # selected items
  }  // if Ctrl
#endif

  QGraphicsView::mousePressEvent(event);
  if (drawing_) {
    viewport()->setCursor(Qt::CrossCursor);
  }
  event->ignore();
}


void ZoomableGraphicsView::mouseReleaseEvent(QMouseEvent* event)
{
  QGraphicsView::mouseReleaseEvent(event);
  event->ignore();

  if (         (event->button() == Qt::RightButton)
       and not (event->buttons() & Qt::LeftButton )) {
    //
    // Removing graphic items or entering the drawing mode.
    //
    if (not scene()->selectedItems().empty()) {
      if (not erasable_.empty()) {
        auto const& items = scene()->selectedItems();
        for (auto pi: items) {
          assert(scene()->items().contains(pi));
          if (erasable_.contains(pi->type())) {
            //scene()->removeItem(pi);  // handled by slot
            emit itemRemoved(pi);
          }
      } }

      scene()->clearSelection();

      if (drawing_) {
        viewport()->setCursor(Qt::CrossCursor);
      }
    } else {
      assert(scene()->selectedItems().empty());
      if (not insertable_.empty()) {
        drawing_ = true;
        viewport()->setCursor(Qt::CrossCursor);
      }
    }

  } else if (        (event->button() == Qt::LeftButton)
             and not (event->buttons() & Qt::RightButton)) {

    if (scene()->selectedItems().size() == 1) {

      // helpful for manual nuclei corrections
      last_selected_ = scene()->selectedItems().front();

      // manual nuclei highlighting
      if (event->modifiers() & Qt::ShiftModifier) {
        auto const& p = scene()->selectedItems().front();
        if (p->type() == QGraphicsPolygonItem::Type) {
          auto pp = static_cast<QGraphicsPolygonItem*>(p);
          emit polygonToExternalHighlight(pp);
    } } }

    if (drawing_) {
      //
      // Finished drawing
      //
      drawing_ = false;
      viewport()->setCursor(Qt::OpenHandCursor);

      if ((not replacing_) and (not splitting_) // then: adding
          and insertable_.contains(QGraphicsPolygonItem::Type)) {
        emit itemAdded(contouri_);
      } else if (replacing_ and swappable_.contains(QGraphicsPolygonItem::Type)
                 and not splitting_){
        assert(selected_);
        if (not contouri_.empty()) {
          emit itemReplaced(selected_, contouri_);
        } else {
          selected_->show();
          viewport()->setCursor(Qt::OpenHandCursor);
        }
        selected_ = nullptr;
        replacing_ = false;
      } else if (splitting_) {
        assert(erasable_.contains(QGraphicsPolygonItem::Type));
        assert(insertable_.contains(QGraphicsPolygonItem::Type));
        // If LMB click was on a contour, then selection size is 1;
        // otherwise, we have to rely on a previos recorded selection:
        if (   (scene()->selectedItems().empty() and last_selected_)
            or (scene()->selectedItems().size() == 1)) {
          emit itemSplit(last_selected_, contouri_);
        }
        splitting_ = false;
      }

      this->clearContourLines_();
      foreach (ZoomableGraphicsView* pz, co_scrolling_) {
        assert(pz);
        pz->clearContourLines_();
      }
      contouri_.clear();       contourf_.clear();
      contouri_.reserve(640);  contourf_.reserve(640);
    }

    if (quantifying_) {
      if (scene()->selectedItems().size() == 1) {
        auto const& p = scene()->selectedItems().front();
        if (p->type() == QGraphicsPolygonItem::Type) {
          auto pp = static_cast<QGraphicsPolygonItem*>(p);
          emit polygonToQuantSelected(pp);
      } }
    }

    // unused
    //selected_ = (scene()->selectedItems().size() == 1)
    //      ? scene()->selectedItems().front() : nullptr;

    if (scene()->selectedItems().size() == 1) {
      if (scene()->selectedItems().front()->type() == QGraphicsLineItem::Type) {
        emit lineSelected(static_cast<QGraphicsLineItem*>(
                                             scene()->selectedItems().front()));
      } else
      if (scene()->selectedItems().front()->type() == QGraphicsRectItem::Type) {
        emit rectSelected(static_cast<QGraphicsRectItem*>(
                                             scene()->selectedItems().front()));
      }
    }
    if (event->modifiers() & Qt::ControlModifier) {
      if (scene()->selectedItems().size() == 2) {
        auto const sel1_item_type = scene()->selectedItems().front()->type();
        auto const sel2_item_type = scene()->selectedItems().back() ->type();
        using L_t = QGraphicsLineItem const*;
        using R_t = QGraphicsRectItem const*;
        if  (     (sel1_item_type == QGraphicsLineItem::Type)
              and (sel2_item_type == QGraphicsLineItem::Type)) {
          emit lineLineSelected(static_cast<L_t>(scene()->selectedItems().front()),
                                static_cast<L_t>(scene()->selectedItems().back()));
        } else if  (    (sel1_item_type == QGraphicsLineItem::Type)
                    and (sel2_item_type == QGraphicsRectItem::Type)) {
          emit lineRectSelected(static_cast<L_t>(scene()->selectedItems().front()),
                                static_cast<R_t>(scene()->selectedItems().back()));
        } else if  (    (sel1_item_type == QGraphicsRectItem::Type)
                    and (sel2_item_type == QGraphicsLineItem::Type)) {
          emit rectLineSelected(static_cast<R_t>(scene()->selectedItems().front()),
                                static_cast<L_t>(scene()->selectedItems().back()));
        } else if  (    (sel1_item_type == QGraphicsRectItem::Type)
                    and (sel2_item_type == QGraphicsRectItem::Type)) {
          emit rectRectSelected(static_cast<R_t>(scene()->selectedItems().front()),
                                static_cast<R_t>(scene()->selectedItems().back()));
        } else {
          // such selection cannot be handled
          scene()->clearSelection();
        }
        scene()->clearSelection();
      } // if 2 items selected
    } // if Ctrl

    for (auto& co: co_scrolling_) { co->scene()->clearSelection(); }

  } else if (    (event->button() == Qt::LeftButton )
             and (event->buttons() & Qt::RightButton)) {
    //
    // Will draw after removal is done. This is for replacing a contour.
    //
    if ((not swappable_.empty()) and (scene()->selectedItems().size() == 1)) {

      selected_ = scene()->selectedItems().front();
      assert(selected_);
      assert (scene()->items().contains(selected_));
      selected_->hide();

      viewport()->setCursor(Qt::CrossCursor);
      drawing_  = true;
      replacing_= true;
    }

  } else if (    (event->button() == Qt::RightButton)
             and (event->buttons() & Qt::LeftButton )) {

    if ((not erasable_.empty()) and (not insertable_.empty())) {
      splitting_ = true;
      drawing_ = true;
      viewport()->setCursor(Qt::CrossCursor);
    }

  } else if (    (event->button() == Qt::MiddleButton)) {

    if (scene()->selectedItems().size() == 1) {
      selected_ = scene()->selectedItems().front();
      if (selected_->type() == QGraphicsPolygonItem::Type){
        emit itemExtended(static_cast<QGraphicsPolygonItem*>(selected_));
    } }

  } else {

    // nothing here

  }

  event->ignore();
}


void ZoomableGraphicsView::mouseDoubleClickEvent(QMouseEvent* event)
{
  if (scene()->selectedItems().size() == 1) {
    QGraphicsItem* gi = scene()->selectedItems().front();
    auto const gi_type = gi->type();
    if (togglable_.contains(gi_type)) {
      if (gi_type != QGraphicsPolygonItem::Type) { return; }  //don't know how to
      QPen pen(static_cast<QGraphicsPolygonItem*>(gi)->pen());//toggle non-contour
      if (pen.style() == Qt::SolidLine) {
        emit itemToggled(gi);
        // pen.setStyle(Qt::DotLine);  // handled by slot
      } else {
        emit itemToggled(gi);
        //pen.setStyle(Qt::SolidLine);  // handled by slot
      }
      //static_cast<QGraphicsPolygonItem*>(gi)->setPen(pen);  // handled by slot
      scene()->clearSelection();
  } }
  event->ignore();
}


void ZoomableGraphicsView::mouseMoveEvent(QMouseEvent* event)
{
  //qDebug() << "MOVE " << this << event->type();
  //qDebug() << QString::number(event->pos().x());
  //qDebug() << QString::number(event->pos().y());

  if (not drawing_) {

    // Moving around.

    QGraphicsView::mouseMoveEvent(event);

    foreach (ZoomableGraphicsView* pz, co_scrolling_) {  assert(pz);
      pz->horizontalScrollBar()->setValue(horizontalScrollBar()->value());
      pz->verticalScrollBar()->setValue(verticalScrollBar()->value());
    }

    event->ignore();

  } else if ((event->buttons() == Qt::LeftButton)) {

    // Drawing a path.

    viewport()->setCursor(Qt::CrossCursor);

    QPointF const pf = this->mapToScene(QPoint(event->pos()));
    QPoint const pt = pf.toPoint();

    qreal const dx = pf.x() - pt.x(),
                dy = pf.y() - pt.y();
    QPointF const pf_mid{ pt.x() + ((dx > 0.) ? 0.5 : -0.5) ,
                          pt.y() + ((dy > 0.) ? 0.5 : -0.5) };
    QPoint const ptt{ static_cast<int>( pt.x() + ((dx > 0.) ? 0.5 : -0.5) ),
                      static_cast<int>( pt.y() + ((dy > 0.) ? 0.5 : -0.5) )};

    if (contourf_.empty()) {
      contourf_.push_back(pf_mid);
      contouri_ .push_back(ptt);
    } else if (ptt != contouri_.back()) { // uniq
      contourf_.push_back(pf_mid);
      contouri_.push_back(ptt);
    }

    drawContourf_();
    foreach (ZoomableGraphicsView *pz, co_scrolling_) { pz->drawContourf_(); }
  }
}


void ZoomableGraphicsView::drawContourf_(bool close)
{
  contour_lines_.reserve(contourf_.size());
  for (int i=0; i<contourf_.size() - 1; ++i) {
    contour_lines_.push_back(
        scene()->addLine(contourf_.at(i  ).x(), contourf_.at(i  ).y(),
                         contourf_.at(i+1).x(), contourf_.at(i+1).y(),
                         contour_pen_));
  }

  if (close and (contouri_.size() > 1)) {
    contour_lines_.push_back(
        scene()->addLine(contourf_.at(contourf_.size()-1).x(),
                         contourf_.at(contourf_.size()-1 ).y(),
                         contourf_.at(0  ).x(), contourf_.at(0  ).y(),
                         contour_pen_));
  }
}


void ZoomableGraphicsView::clearContourLines_()
{
  for (auto pline: qAsConst(contour_lines_)) {  assert(pline);
    scene()->removeItem(static_cast<QGraphicsItem*>(pline));
  }
  contour_lines_.clear();
}


void ZoomableGraphicsView::maybeChangePen_()
{
  contour_pen_ = QPen((insertable_.contains(QGraphicsPolygonItem::Type)
                          ? QColor(Qt::blue).lighter(145)
                          : QColor(Qt::gray).darker(300)),
                       0.375,
                       Qt::SolidLine, Qt::RoundCap, Qt::RoundJoin);
}


// This may be useful only for preview/debugging.
//
#if 0
void ZoomableGraphicsView::DEBUG_draw_contour_(bool close)
{
  contour_lines_.reserve(contouri_.size());
  for (int i=0; i<contouri_.size() - 1; ++i) {
    contour_lines_.push_back(
        scene()->addLine(contouri_.at(i  ).x(), contouri_.at(i  ).y(),
                         contouri_.at(i+1).x(), contouri_.at(i+1).y(), contour_pen_ ) );
  }

  if (close and (contouri_.size() > 1)) {
    contour_lines_.push_back(
        scene()->addLine(contouri_.at(contouri_.size()-1).x(), contouri_.at(contouri_.size()-1 ).y(),
                         contouri_.at(0  ).x(), contouri_.at(0  ).y(), contour_pen_) );
  }
}
#endif

