// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "MainWindow.hpp"
#include "ui_MainWindow.h"
#include "GUI/Windows/MainWindow_aux.hpp"
#include "GUI/Support/ZoomableGraphicsView.hpp"
#include "GUI/Windows/ScriptWindow.hpp"
#include "GUI/Windows/HelpWindow.hpp"
#include "GUI/Windows/AboutDialog.hpp"
#include <QGraphicsScene>
#include <QGraphicsItem>
#include <QPixmap>
#include <QMessageBox>
#include <QDesktopServices>
#include <QDateTime>
#include <QLocale>
#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
#  include <QtConcurrent/QtConcurrent>
#  include <QtConcurrent/QtConcurrentRun>
#else
#  include <QtConcurrentRun>
#endif
#if (QT_VERSION >= QT_VERSION_CHECK(5, 2, 0))
# ifdef Q_OS_WIN
#  include <QWinTaskbarButton>
#  include <QWinTaskbarProgress>
# endif
#endif

// static data initialization ---

QString const MainWindow::sWinGoFull   = "Go full-screen";
QString const MainWindow::sWinExitFull = "Exit full-screen";
QString const MainWindow::sAppVersionLatestUrl = \
    "http://pmbm.ippt.pan.pl/version/software/shuttletracker/latest.txt";


// lifetime ---

MainWindow::MainWindow(QWidget *parent)
: QMainWindow{parent},
  ui_{new Ui::MainWindow},
  images_{new ImageSequence{nullptr, new st::QtReporter}}
#ifdef Q_OS_WIN
 // nothing here, Windows registry is used
#else
#  ifdef SHUTTLETRACKER_SHARED_DIR
    ,settings_file_path_{QString(SHUTTLETRACKER_SHARED_DIR)
      + QDir::separator() + "shuttletracker"
      + QDir::separator() + "config"
      + QDir::separator() + "settings.ini"}
#  else
   ,settings_file_path_{QDir(QApplication::applicationDirPath())
                            .filePath("settings.ini")}
#  endif
#endif
{
#ifdef _MSC_VER
  if (MainWindow::api_functions.empty()) {
    for (auto const& kv: api_functions__) {
      MainWindow::api_functions.insert(kv.first,kv.second);
  } }
  if (MainWindow::api_parameters.empty()) {
    for (auto const& ap: api_parameters__) {
      MainWindow::api_parameters.insert(ap);
  } }
  if (MainWindow::api_specials.empty()) {
    for (auto const& as: api_specials__) {
      MainWindow::api_specials.insert(as);
  } }
#endif

#ifndef Q_OS_WIN
  setupUnixSignalHandling();
#endif

  ui_->setupUi(this);

  // TODO DEVEL FUTURE
  //ui_->statusBar->addPermanentWidget(ui_->progressBar, 0);
  //ui_->progressBar->setMaximumSize(QSize(320,28));
  //ui_->progressBar->setMinimumSize(QSize(320,28));
  //ui_->progressBar->setSizePolicy(QSizePolicy::Fixed,QSizePolicy::Fixed);


   // COMPILER -- VC2013 cannot default-init
  this->draw_frame_nuclei__ = &MainWindow::drawFrameNucleiInWhite_/*drawFrameNucleiByProvenience_*/;

  setupMainToolbar_();

  setupMenuViewShowActions_();
  menuViewEnable_(false);

  setupTools_();
  setupMenuToolsActions_();
  menuToolsEnable_(false);

  setupScripting_();
  menuScriptUpdate_();
  //menu_script_enable_(false); // commented out <= scripted initial loading of images

  setupNavigationWidgets_();

  ui_->progressBar->setVisible(false);

#if (QT_VERSION >= QT_VERSION_CHECK(5, 2, 0))
# ifdef Q_OS_WIN
  win_taskbar_button_.reset(new QWinTaskbarButton(this));
  win_taskbar_progress_.reset(win_taskbar_button_->progress());
# endif
#endif

  setupKeyboardShortcuts_();

#ifdef Q_OS_WIN  // use a font with required coverage of unicode support
  ui_->pushButtonNucEditSwitchOnBorders->setFont(QFont("Cambria",13));
#endif

  loadSettings_(true);

  if (ui_->actionPreferenceCheckLatestVersionUponStartup->isChecked()) {
    checkLatestVersion_();
  }

  // DEVEL: hide widgets of features in preparation
  ui_->actionReloadSettings->setVisible(false);
  ui_->checkBoxNucDetRecoverFromDebris->setVisible(false);
  ui_->actionPreferenceParamScanPrecomputeShared->setVisible(false);
  ui_->checkBoxPerinucQuantSavePng->setVisible(false);
  ui_->hspacerPerinucQuantSaveFormatsM->setGeometry(QRect{0, 0, 0, 0});

  // right mouse button click handling
  ui_->tableViewTracksEdit->setContextMenuPolicy(Qt::CustomContextMenu);
  connect(ui_->tableViewTracksEdit,
                SIGNAL(customContextMenuRequested(QPoint)),
          this, SLOT(tracksViewRemoveAt(QPoint)));

  // do not start with empty log window
  auto const date = QDateTime::currentDateTime();
  auto const date_s = QLocale{QLocale("en_US")}.toString(date);
  log_ << "<b>Info</b>: ShuttleTracker started on " + date_s + ".";
} // c'tor


MainWindow::~MainWindow()
{
  if (script_window_) {
    disconnect(script_window_.get(), SIGNAL(scriptFilesListModified(QStringList)),
               this,                 SLOT(updateScripts(QStringList)));
  }

  future_int_.cancel();
  future_void_.cancel();

  saveSettings_();

  undrawFrameOutlines_();
  clearFrameAnnotations_();
  undrawTracks_();
  pixmap_cache_.reset();
  for (auto& ps: scenes_) { delete ps; }
  // for (auto& pv: views_)  { delete pv; }

  if (net_manager_) { net_manager_->disconnect(this); }

  delete ui_;
}


#if (QT_VERSION >= QT_VERSION_CHECK(5, 2, 0))
# ifdef Q_OS_WIN
void MainWindow::showEvent(QShowEvent* e)
{
  win_taskbar_button_->setWindow(this->windowHandle());
  e->accept();
}
# endif
#endif


// slot
void MainWindow::openHomepageUrl() const
{
  QString const static www = "http://pmbm.ippt.pan.pl/software/shuttletracker";
  QtConcurrent::run(QDesktopServices::openUrl, QUrl{www})/*.waitForFinished()*/;
}


QString MainWindow::timepointFilenameEndingWith_(QString const& filename_ending,
                                                 QString const& file_purpose)
{
  auto const fn_s = images_->imageFilePathNoChannel(current_frame_i_);
  auto const fn = QString::fromStdString(fn_s);
  auto const fn_ext = QString::fromStdString(images_->imageFileExtension());
  QString fn_rep{fn};
  auto const re_dot_ext = QRegularExpression{QString{"\\."} + fn_ext + "$"};
  fn_rep.replace(re_dot_ext, filename_ending);
  if (fn_rep == fn) {  // replacement failure
    message_('E', "Cannot assemble a name for " + file_purpose + "!");
    return "";
  }
  return fn_rep;
}



// slot
void MainWindow::startProgress(QString msg, int progress_max)
{
  connect(this, SIGNAL(messaging(QChar,QString)),
          this, SLOT(message_(QChar,QString)), Qt::QueuedConnection);
  connect(this, SIGNAL(messaging_done()),
          this, SLOT(messageDone_()), Qt::QueuedConnection);

  if (msg != "") { emit messaging('I', msg); }

  framesNavigationExpose_(false);
  ui_->progressBar->setVisible(true);
  ui_->progressBar->setValue(0);
  ui_->progressBar->setMaximum(static_cast<int>(progress_max));
#if (QT_VERSION >= QT_VERSION_CHECK(5, 2, 0))
# ifdef Q_OS_WIN
  win_taskbar_progress_->setVisible(true);
  win_taskbar_progress_->setMaximum(progress_max);
# endif
#endif
}


// slot
int MainWindow::stopProgress()
{
  QMutexLocker __(&mutex_);
  //qDebug() << "MainWindow got stopProgress()";

  QString problem = "";
  try {
    future_int_.waitForFinished();
  } catch (st::QtInformativeException& e) {
    problem = e.what();
  } catch (...) {
    problem = "Unknown problem (likely related to concurrency).";
  }

  int retcode = EXIT_SUCCESS;
  if (problem != "") {
    emit messaging('E', problem);
    QMessageBox msg;
     msg.setText("Whoops!");
     msg.setInformativeText(problem);
     msg.setStandardButtons(QMessageBox::Ok);
     msg.setDefaultButton(QMessageBox::Ok);
     msg.setIcon(QMessageBox::Critical);
     msg.exec();
    retcode = EXIT_FAILURE;
  } else {
   // emit messaging_done(); -- shows up too late
   retcode = EXIT_SUCCESS;
  }

  disconnect(this, SIGNAL(messaging(QChar, QString const&)),
             this, SLOT  (message_(QChar, QString const&)));
  disconnect(this, SIGNAL(messaging_done()),
             this, SLOT  (messageDone_()));

  ui_->progressBar->setValue(0);
  ui_->progressBar->setVisible(false);
#if (QT_VERSION >= QT_VERSION_CHECK(5, 2, 0))
# ifdef Q_OS_WIN
  win_taskbar_progress_->setVisible(false);
  win_taskbar_progress_->setValue(0);
# endif
#endif
  framesNavigationExpose_(true);

  return retcode;
}


ImageChannel MainWindow::channel(QGraphicsItem const* g)
{
  assert(g);
  int panel_index = 0;
  for (auto const& s: qAsConst(scenes_)) {
    if (g->scene() == s) { break; }
    ++panel_index;
  }
  return channel(panel_index);  // 0 assumed to be a fail-safe default
}


// ============================== CLEAN-UPs ====================================

void MainWindow::clearAllNuclei_()
{
  nuclei_.clear();     // d'tors remove gphx items
  current_nuclei_ = nullptr;
}

void MainWindow::clearAllPerinuclei_()
{
  perinuclei_.clear();     // d'tors remove gphx items
  current_perinuclei_ = nullptr;
}

void MainWindow::clearAllRegions_()
{
  regions_.clear();    // d'tors remove gphx items
  current_regions_ = nullptr;
}

void MainWindow::clearAllTracks_()
{
  //tracks_model_.release();
  tracks_.clear();     // d'tors remove gphx items
  current_track_ = nullptr;
  resetTracksModel_();
}

void MainWindow::clearFrameNuclei_()
{
  if (current_nuclei_) {
    current_nuclei_->clear();  // d'tors remove gphx items
  }

  if (current_perinuclei_) {
    if (not current_perinuclei_->empty()) {
      message_('W', "All perinuclei were removed upon nuclei removal");
      current_perinuclei_->clear();
    }
  }

  if (not tracks_.empty()) {
    clearAllTracks_();
    message_('W', "All tracks were removed upon nuclei removal");
  }
}

void MainWindow::clearFramePerinuclei_()
{
  if (current_perinuclei_) {
    current_perinuclei_->clear();  // d'tors remove gphx items
  }
}

void MainWindow::clearFrameRegions_()
{
  if (current_regions_) {
    current_regions_->clear();  // d'tors remove gphx items
  }
}

// ========================================================================= //
