// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#ifndef SHUTTLETRACKER_GUI_WINDOWS_MAINWINDOW_HPP
#define SHUTTLETRACKER_GUI_WINDOWS_MAINWINDOW_HPP

#include "GUI/GUICommon.hpp"
#include "GUI/Windows/MainWindow_aux.hpp"
#include "GUI/Windows/ScriptWindow.hpp"
#include "GUI/Windows/HelpWindow.hpp"
#include "GUI/Windows/AboutDialog.hpp"
#include "GUI/Support/PixmapCache.hpp"
#include "GUI/Support/Outline.hpp"
#include "GUI/Support/Track.hpp"
#include "Image/NucleiDetection.hpp"
#include <QMainWindow>
#include <QVector>
#include <QSet>
#include <QPair>
#include <QSettings>
#include <QFuture>
#include <QFutureWatcher>
#include <QNetworkAccessManager>
#include <QSystemTrayIcon>
#include <QAbstractTableModel>
#ifndef Q_OS_WIN
#  include <QSocketNotifier>
#endif
#include <memory>

class QGraphicsScene;
class QGraphicsPixmapItem;
class QToolBox;
class QNetworkReply;
class QLineEdit;
#if (QT_VERSION >= QT_VERSION_CHECK(5, 2, 0))
# ifdef Q_OS_WIN
    class QWinTaskbarButton;
    class QWinTaskbarProgress;
# endif
#endif
class ZoomableGraphicsView;
namespace Ui { class MainWindow; }


class MainWindow : public QMainWindow
{
  Q_OBJECT

 public:
  explicit MainWindow(QWidget* parent=nullptr);
  ~MainWindow() OVERRIDE;


  // -- for running from the command line
  void postLoadScript(QString const& script);
  void setQuitUponScriptError(bool whether);
  bool shouldQuitUponScriptError() const;

  void toggleToolboxenEnablednessForScriptExecution(bool on);
  void ensureNotInBatchMode() { batch_processing_ = false; }

#ifdef _MSC_VER
  static /*const*/ QMap<QString,QString> api_functions;
  static const QVector<QPair<QString,QString>> api_functions__;
  static /*const*/ QSet<QString> api_parameters;
  static const QVector<QString> api_parameters__;
  static /*const*/ QSet<QString> api_specials;
  static const QVector<QString> api_specials__;
#else
  static const QMap<QString,QString> api_functions;
  static const QSet<QString> api_parameters;
  static const QSet<QString> api_specials;
#endif
  //
public Q_SLOTS:  // predominantly, below goes scripting API

  void open_directory(QString d);
  void detect_nuclei();
  void detect_nuclei_with_parameter_scan();
  void derive_perinuclei();
  void sample_background();
  void tesselate();
  void remove_nuclei();
  void remove_toggled_nuclei();
  void ___remove_nucleus___(int);
  void remove_split_orphans();
  void remove_debris();
  void remove_border_nuclei();
  void remove_interior_debris();
  void remove_perinuclei();
  void remove_regions();
  void remove_tracks();
  void remove_tracks_nonrevised();
  void toggle_all_nuclei();
  void toggle_border_nuclei();
  void track_nuclei();
  bool has_masking_contours_file();
   int load_masking_contours();
  bool has_nuclei_contours_file();
   int load_nuclei_contours();
  bool has_regions_contours_file();
   int load_regions_contours();
   int save_masking_contours();
   int save_nuclei_contours();
   int save_regions_contours();
   int load_tracks();
   int save_tracks();
  bool has_nuclei_detection_settings_file();
  void load_nuclei_detection_settings();
  void save_nuclei_detection_settings();
  bool has_perinuclei_settings_file();
  void load_perinuclei_settings();
  void save_perinuclei_settings();
  bool has_tracking_settings_file();
  void load_tracking_settings();
  void save_tracking_settings();
  void apply_masking_contours();
  void toggle_overlay_channel(QString channel_name_or_description);
  void highlight_nuclei_from_file();
  void export_all_quantifications();
  void export_nuclei_quantifications();
  void export_perinuclei_quantifications();
  void export_regions_quantifications();
  void export_image_quantifications();
  void rename_per_qfile_as_far_qfile();
  void go_to_frame(long frame_i);
  void set(QString what, QString value);
  QString get(QString what);
  void echo(QString);
  void repaint();
  void millisleep(int msecs) const;
  void panic(QString reason);
  void quit();
  //
  // not exposed in API documentation on purpose
  void ___batch_processing___(bool);

  // The ..._count() const methods declared below have to be among Q_SLOTS (they
  // would be non-invocable by the embedded script interpreter if the were not).
   int nuclei_count()  const;
   int regions_count() const;
  long tracks_count()  const;
  long frames_count()  const;
  long current_frame() const;

private:

#ifndef Q_OS_WIN
  static void sighupForwarder(int unused);
  static void sigtermForwarder(int unused);
  void setupUnixSignalHandling();
#endif

  void saveSettings_();
  void loadSettings_(bool resize_window=false);

  void checkLatestVersion_();

  void setupMenuToolsActions_();
  void setupMenuViewShowActions_();
  void setupKeyboardShortcuts_();
  void setupViewsAndScenes_(bool clear=false);
  void setupMenuTools_();
  void setupNavigationWidgets_();
  void setupMainToolbar_();
  void setupTools_();
  void setupToolboxenBasedOnAvailableChannels_();

  void setupScripting_();
  void menuScriptUpdate_();

  void menuViewUpdate_();
  void on_actionShowAux(bool desaturated, ImageFilter filter=st::identity);
  void showActionsUncheckOthers_(QObject const* sender);
  void regenerateOverlays_();

  void menuViewEnable_ (bool on);
  void menuToolsEnable_(bool on);
  void menuScriptEnable_(bool on);
  void framesNavigationExpose_(bool on) ; // expose = show + enable

  void toolboxToggle_(QToolBox* toolbox, bool on);


  // -- Drawing-related
  //
  void drawFrameNuclei_();  // dispatcher
  void drawFrameNucleiByProvenience_();
  void drawFrameNucleiByProperty_();
  void drawFrameNucleiInWhite_();
  void drawFrameNucleiInGreen_();
  void drawFramePerinuclei_();

  // COMPILER, TEMPORARY -- VC2013 cannot default-init
  void (MainWindow::*draw_frame_nuclei__)() ; // = &MainWindow::draw_frame_nuclei_by_provenience_;
  //
  void redrawNucleus_(NuclearOutline &);
  void drawLastNucleus_();
  void drawLastRegion_();
  void drawExclusions_();
  void drawFrameRegions_();
  //
  void drawFrameOutlines_()
  {
    drawFrameNuclei_();
    drawFramePerinuclei_();
    drawFrameRegions_();
  }
  //
  void undrawFrameOutlines_()
  {
    aux::detachScenesItems<QGraphicsPolygonItem::Type>(scenes_);
    clearFrameOutlinesDepictions_();
  }
  //
  void clearFrameOutlinesDepictions_()
  {
    clearFrameNucleiDepictions_();
    clearFramePerinucleiDepictions_();
    clearFrameRegionsDepictions_();
  }
  //
  void clearFrameNucleiDepictions_()
  { if(current_nuclei_) for(auto& n: *current_nuclei_) n.clearDepictions(); }
  //
  void clearFramePerinucleiDepictions_()
  { if(current_perinuclei_)for(auto&p:*current_perinuclei_)p.clearDepictions();}
  //
  void clearFrameRegionsDepictions_()
  { if(current_regions_) for(auto& r: *current_regions_) r.clearDepictions(); }
  //
  void redrawFrameOutlines_()
  {  undrawFrameOutlines_();  drawFrameOutlines_();  }
  //
  void ifNotHiddenThenRedrawFrameOutlines_();

  void drawTrack__(Track*);
  void drawTracks_() ; // all frames
  void undrawTracks_()
  {
    aux::detachScenesItems<QGraphicsLineItem::Type>(scenes_);
    aux::detachScenesItems<QGraphicsRectItem::Type>(scenes_);
    for (auto& t: tracks_) { t.clearDepictions(); }
  }
  //
  void consolidateTracks_(QList<QPair<Outline const*,Outline const*>>const&,
                          QHash<QPair<Outline const*,Outline const*>,double>const&);
  void resetTracksModel_();
#if 0
  // DEAD code
  bool current_frame_has_tracks_() const;
#endif

  void redrawTracks_() {  undrawTracks_();  drawTracks_();  }
  //
  void ifNotHiddenThenRedrawTracks_();
  //
  void clearFrameNuclei_();
  void clearFramePerinuclei_();
  void clearFrameNucleiByProvenience_(NuclearOutline::Provenience);
  void clearFrameRegions_();
  void clearAllNuclei_();
  void clearAllPerinuclei_();
  void clearAllRegions_();
  void clearAllTracks_();
  //
  void makeFrameAnnotations_(bool an_nuclei = true, bool an_regions = true);
  void clearFrameAnnotations_(bool nuclei_ans = true, bool regions_ans = true);
  void maybeFrameAnnotate_();
  //

  // -- Other/mix
  //
  ImageChannel nuclearImageChannel_() const;
  //
  int nucDetImgPreprocParamScan(cv::Mat const& img_orig);
  //
  QString timepointFilenameEndingWith_(QString const& filename_ending,
                                       QString const& file_purpose);
  QString assembleImageMaskContoursFilename_();
  QString assembleNucDetSettingsFilename_() const;
  QString assembleNucleiContoursFilename_();
  QString assemblePerinucDerivSettingsFilename_() const;
  QString assembleRegionsContoursFilename_();
  QString assembleQuantsFilename_(QString const&, QString const&);
  QString assembleTrackingSettingsFilename_() const;
  //
  bool checkIfImageMaskContoursFileAvailable_();
  bool checkIfNucDetSettingsFileAvailable_();
  bool checkIfNucleiContoursFileAvailable_();
  bool checkIfPerinucDerivSettingsFileAvailable_();
  bool checkIfRegionsContoursFileAvailable_();
  bool checkIfTrackingSettingsFileAvailable_();
  //
  void quantifyNucleiAndSaveCsvFile_();
  void quantifyPerinucleiAndSaveCsvFile_();
  void quantifyRegionsAndSaveCsvFile_();
  void savePixmapWithOptionalMarkings_(QString const& outln_t, bool trim = true);
  //
  void trimScenesToBorder_()
  {  aux::trimToBackgroundPixmap(scenes_);  }
  //
  bool saveOverlayToPixmapFile_(QString const& filename)
  {  return aux::saveToPixmapFile(*scenes_.back(), filename);  }

public:
  using described_images_t = QList<QPair<cv::Mat, QString>>;

private:
  void collectImagesAndDescriptionsInSelectedChannels_(MainWindow::described_images_t*) const;
  std::map<st::ImageChannel, cv::Mat> collectChannelImagesInSelectedChannels_() const;

Q_SIGNALS:
  void busy();
  void idle();
  void messaging(QChar cat, QString const& msg, int timeout = -1);
  void messaging_done();
  void recheck(bool);


public Q_SLOTS:
  void loadImagesFromDirectory(QString dir_hint = "");


private Q_SLOTS:

#ifndef Q_OS_WIN
  void sighupHandler();
  void sigtermHandler();
#endif

  // -- Drawing-related
  //
  void updateEditabilities_();
  void updateEditability_(QGraphicsItem*);
  void updateEditabilityOfNucleus_(QGraphicsPolygonItem*);
  void updateEditabilityOfRegion_(QGraphicsPolygonItem*);
  void itemAdded(QVector<QPoint> const&);
  void itemRemoved(QGraphicsItem const*);
  void itemReplaced(QGraphicsItem const*, QVector<QPoint> const&);
  void itemSplit(QGraphicsItem const*, QVector<QPoint> const&);
  void itemToggled(QGraphicsItem const*);
  void itemExtended(QGraphicsPolygonItem const*);
  //
  void lineSelected(QGraphicsLineItem const*);
  void rectSelected(QGraphicsRectItem const*);
  void lineLineSelected(QGraphicsLineItem const*, QGraphicsLineItem const*);
  void lineRectSelected(QGraphicsLineItem const*, QGraphicsRectItem const*);
  void rectLineSelected(QGraphicsRectItem const*, QGraphicsLineItem const*);
  void rectRectSelected(QGraphicsRectItem const*, QGraphicsRectItem const*);
  //
  void tracksViewRemoveAt(QPoint const&);
  void boldenSelectedTrack_(QModelIndex const& sel, QModelIndex const& desel);
  //
  void showQuantifications(QGraphicsPolygonItem const*);
  void toggleExternalHighlight(QGraphicsPolygonItem const*);

  void updateScripts(QStringList scripts_paths);
  void executeScriptFileByName();

  // -- Logging-related
  //
  void message_(QChar cat, QString const& msg, int suggested_timeout = -1);
  void messageDone_();


  void latestVersionReply_(QNetworkReply* reply);
  void openHomepageUrl() const;

  void startProgress(QString msg = "", int progress_max = 100);
  int   stopProgress();
  //
  void startProgressImagesLoading();
  void  stopProgressImagesLoading();
  //
  void startProgressNucDetParamScan();
  void  stopProgressNucDetParamScan();


  int panelIndex(ImageChannel target_channel);
  ImageChannel channel(int panel_index);
  ImageChannel channel(QGraphicsItem const*);

  void showFrame_(long int frame_i);
  void togglePanel_(ImageChannel target_channel);   // common work for on_action#ChannelPanelToggle
  void toggleVisibility_(ImageChannel channel); // common work TODO:doc
  void toggleOverlayComponent_(ImageChannel channel); // common work TODO:doc


 private:

  // -- View-related
  //
  struct {
    QMap<QAction*, bool> channel_pane_available  = {};
    QMap<QAction*, bool> channel_pane_visible    = {};
    QMap<QAction*, bool> overlay_channel_available = {};
    QMap<QAction*, bool> overlay_channel_included  = {};
    qreal scale = 1.;
  } view_settings_;
  //
  QList<QAction *> channelPaneToggleActions_() const;
  QList<QAction *> overlayChannelToggleActions_() const;
  void maybeStoreViewSettings_();
  void maybeRestoreViewSettings_();


private Q_SLOTS:

  // -- Menus

  // menu File:
  //
  void on_actionOpenDirectory_triggered();
  void on_actionReloadSettings_triggered();
  void on_actionQuit_triggered();

  // menu View:
  //
  void on_actionGreenChannelPanelToggle_triggered();
  void on_actionRedChannelPanelToggle_triggered();
  void on_actionBlueChannelPanelToggle_triggered();
  void on_actionCyanChannelPanelToggle_triggered();
  void on_actionMagentaChannelPanelToggle_triggered();
  void on_actionYellowChannelPanelToggle_triggered();
  void on_actionBrightfieldPanelToggle_triggered();
  void on_actionOverlaysPanelToggle_triggered();
  //
  void on_actionGreenChannelVisibilityToggle_triggered();
  void on_actionRedChannelVisibilityToggle_triggered();
  void on_actionBlueChannelVisibilityToggle_triggered();
  void on_actionCyanChannelVisibilityToggle_triggered();
  void on_actionMagentaChannelVisibilityToggle_triggered();
  void on_actionYellowChannelVisibilityToggle_triggered();
  void on_actionBrightfieldVisibilityToggle_triggered();
  void on_actionOverlaysVisibilityToggle_triggered();
  //
  void on_actionOverlaysGreenChannelInclusionToggle_triggered();
  void on_actionOverlaysRedChannelInclusionToggle_triggered();
  void on_actionOverlaysBlueChannelInclusionToggle_triggered();
  void on_actionOverlaysCyanChannelInclusionToggle_triggered();
  void on_actionOverlaysMagentaChannelInclusionToggle_triggered();
  void on_actionOverlaysYellowChannelInclusionToggle_triggered();
  void on_actionOverlaysBrightfieldInclusionToggle_triggered();
  //
  void on_actionShowOriginal_triggered();
  void on_actionShowOriginalDesaturated_triggered();
  void on_actionShowNormalized_triggered();
  void on_actionShowNormalizedDesaturated_triggered();
  void on_actionShowAutoleveledWeak_triggered();
  void on_actionShowAutoleveledWeakDesaturated_triggered();
  void on_actionShowAutoleveledMedium_triggered();
  void on_actionShowAutoleveledMediumGrayscale_triggered();
  void on_actionShowAutoleveledStrong_triggered();
  void on_actionShowAutoleveledStrongDesaturated_triggered();
  //
  void on_actionHideMarking_triggered(bool checked);
  void on_actionHideImages_triggered(bool);

  // menu Tools:
  //
  void on_actionImageMasking_toggled(bool);
  void on_actionNucleiDetection_toggled(bool);
  void on_actionNucleiEditing_toggled(bool);
  void on_actionPerinuclei_toggled(bool);
  void on_actionRegionsEditing_toggled(bool);
  void on_actionQuantification_toggled(bool);
  void on_actionTracking_toggled(bool);
  void on_actionTracksEditing_toggled(bool);
  //
  void on_actionToolSettingsSave_triggered();
  void on_actionToolSettingsLoad_triggered();

  // menu Script:
  //
  void on_actionScriptEdit_triggered();

  // menu Preferences
  //
  void on_actionAfterOpenNormalizeToCameraBitDepth_toggled(bool);
  void on_actionStoreOriginalImagesInMemory_toggled(bool arg1);
  void on_actionAfterOpenCachePixmaps_toggled(bool);
  void on_actionPreferenceAfterOpenComputeOverlays_toggled(bool arg1);
  void on_actionPreferenceWhenLoadingOmitBrightfieldImages_toggled(bool);
  void on_actionAfterOpenHideBrightfield_toggled(bool);
  void on_actionAfterOpenExcludeBrightfieldFromOverlays_toggled(bool);

  // menu Window:
  //
  void on_actionFullScreenToggle_triggered();

  // menu Help:
  //
  void on_actionManual_triggered();
  void on_actionShowLog_triggered();
  void on_actionAbout_triggered();


  // Bottom slider and associated buttons:
  //
  void on_horizontalSlider_valueChanged(int value);
  void on_pushButtonFrameNext_clicked();
  void on_pushButtonFramePrev_clicked();


  // -- Toolbox: image masking
  //
  void on_pushButtonImgMaskReset_clicked();
  void on_pushButtonImgMaskApply_clicked();
  //
  void on_pushButtonImgMaskSave_clicked();
  void on_pushButtonImgMaskLoad_clicked();


  // -- Toolbox: nuclei detection
  //
private:
  bool nucleiDetectionParameterScanSettingsAreCorrect() /*const*/;
  st::ImagePreprocessingParameters getImagePreprocessingParameters() const;
  st::NucleiPreselectionParameters getNucleiPreselectionParameters() const;
  st::stained::NucleiDetectionParameters getStainedNucleiDetectionParameters()const;
  st::nonstained::NucleiDetectionParameters getNonstainedNucleiDetectionParameters()const;
  //
private Q_SLOTS:
  //
  // image preprocessing:
  //
  void on_comboBoxNucDetChannelSelection_currentIndexChanged(int);
  void on_radioButtonNucDetChannelSelectionTypeNuclear_toggled(bool);
  //
  void on_radioButtonNucDetParamsAuto_toggled(bool);
  void on_radioButtonNucDetParamsManual_toggled(bool);
  //
  void on_checkBoxNucDetNormalization_toggled(bool);
  //
  void on_checkBoxNucDetDenoising_toggled(bool);
  void on_dspinBoxNucDetDenoising_editingFinished();
  void on_dspinBoxNucDetDenoising_valueChanged(QString const&);
  //
  void on_checkBoxNucDetSmoothing_toggled(bool);
  void on_spinBoxNucDetSmoothing_valueChanged(QString const&);
  void on_spinBoxNucDetSmoothing_editingFinished();
  //
  void on_checkBoxNucDetThresholding_toggled(bool);
  void on_spinBoxNucDetThresholdingBlocksize_valueChanged(QString const&);
  void on_spinBoxNucDetThresholdingBlocksize_editingFinished();
  void on_spinBoxNucDetThresholdingBaseline_valueChanged(QString const&);
  void on_spinBoxNucDetThresholdingBaseline_editingFinished();
  //
  void on_checkBoxNucDetOpening_toggled(bool);
  void on_spinBoxNucDetOpeningTimes_editingFinished();
  void on_spinBoxNucDetOpeningTimes_valueChanged(QString const&);
  void on_spinBoxNucDetOpeningErosion_editingFinished();
  void on_spinBoxNucDetOpeningErosion_valueChanged(QString const&);
  void on_spinBoxNucDetOpeningDilation_editingFinished();
  void on_spinBoxNucDetOpeningDilation_valueChanged(QString const&);
  //
  void on_checkBoxNucDetClosing_toggled(bool);
  void on_spinBoxNucDetClosingDilation_editingFinished();
  void on_spinBoxNucDetClosingDilation_valueChanged(QString const&);
  void on_spinBoxNucDetClosingErosion_editingFinished();
  void on_spinBoxNucDetClosingErosion_valueChanged(QString const&);
  void on_spinBoxNucDetClosingTimes_editingFinished();
  void on_spinBoxNucDetClosingTimes_valueChanged(QString const&);
  //
  void on_checkBoxNucDetEqualization_toggled(bool checked);
  void on_spinBoxNucDetEqualizationThreshold_valueChanged(QString const&);
  void on_spinBoxNucDetEqualizationThreshold_editingFinished();
  //
  // contour detection:
  //
  void on_spinBoxNucDetContourDetThresholdLow_editingFinished();
  void on_spinBoxNucDetContourDetThresholdLow_valueChanged(QString const&);
  void on_dspinBoxNucDetContourDetThresholdsRatio_editingFinished();
  void on_dspinBoxNucDetContourDetThresholdsRatio_valueChanged(QString const&);
  void on_dspinBoxNucDetContourDetTolerance_editingFinished();
  void on_dspinBoxNucDetContourDetTolerance_valueChanged(QString const&);
  void on_checkBoxNucDetContourDetReconstruct_toggled(bool checked);
  //
  // nuclei preselection:
  //
  void on_dspinBoxNucDetSolidity_editingFinished();
  void on_dspinBoxNucDetSolidity_valueChanged(QString const& arg);
  //
  void on_dspinBoxNucDetMinAreaFrac_editingFinished();
  void on_dspinBoxNucDetMinAreaFrac_valueChanged(QString const& arg);
  //
  void on_dspinBoxNucDetMaxAreaFrac_editingFinished();
  void on_dspinBoxNucDetMaxAreaFrac_valueChanged(QString const& arg);
  //
  void on_checkBoxNucDetSplitAdjacent_toggled(bool checked);
  //
  void on_spinBoxNucDetMinPocketArea_editingFinished();
  void on_spinBoxNucDetMinPocketArea_valueChanged(QString const& arg);
  //
  // actions:
  //
  void on_checkBoxNucDetDetectAutoup_toggled(bool);
  //
  void on_pushButtonNucDetDetect_clicked();
  //
  void maybe_NucDetDetect_autoclick();
  //
  void on_pushButtonNucDetSettingsSave_clicked();
  void on_pushButtonNucDetSettingsLoad_clicked();


  // -- Toolbox: nuclei editing
  //
  void on_dspinBoxNucEditContourWidth_editingFinished();
  void on_dspinBoxNucEditContourWidth_valueChanged(QString const& arg);
  void on_comboBoxNucEditColorBy_currentIndexChanged(QString const& arg);
  //
  void on_checkBoxNucEditShowAnnotations_toggled(bool checked);
  //
  void on_checkBoxNucEditMarkNucleiSolo_toggled(bool checked);
  void on_checkBoxNucEditMarkNucleiSplit_toggled(bool checked);
  void on_checkBoxNucEditMarkNucleiManual_toggled(bool checked);
  void on_checkBoxNucEditMarkNonsplittable_toggled(bool checked);
  void on_checkBoxNucEditMarkSplitOrphans_toggled(bool checked);
  void on_checkBoxNucEditMarkDebris_toggled(bool checked);
  //
  void on_pushButtonNucEditToggleAll_clicked();
  void on_pushButtonNucEditClearToggled_clicked();
  void on_pushButtonNucEditClearAll_clicked();
  void on_pushButtonNucEditRemoveSplitOrphans_clicked();
  void on_pushButtonNucEditRemoveDebris_clicked();
  //
  void on_pushButtonNucEditSwitchOnBorders_clicked();
  void on_pushButtonNucEditExcludeInteriorOrphans_clicked();
  void on_pushButtonNucEditExcludeOnBorders_clicked();
  void on_pushButtonNucEditExcludeInteriorDebris_clicked();
  //
  void on_comboBoxNucEditRefineNuccutChannelAux1_currentIndexChanged(int index);
  void on_comboBoxNucEditRefineNuccutChannelAux2_currentIndexChanged(int index);
  void on_pushButtonNucEditRefine_clicked();
  //
  void on_lineEditNucEditGotoNuc_returnPressed();
  void on_pushButtonNucEditMarkFromFile_clicked();
  //
  void on_pushButtonNucEditNucleiContoursSave_clicked();
  void on_pushButtonNucEditNucleiContoursLoad_clicked();


  // -- Toolbox: perinuclei
  //
  void on_pushButtonPerinucCreate_clicked();
  void maybe_PerinucDerive_autoclick();
  //
  void on_pushButtonPerinucClearAll_clicked();
  void on_dspinBoxPerinucContourWidth_valueChanged(double);
  //
  void on_spinBoxPerinucSettingInnerOffset_valueChanged(QString const&);
  void on_spinBoxPerinucSettingInnerOffset_editingFinished();
  void on_spinBoxPerinucSettingWidth_valueChanged(QString const&);
  void on_spinBoxPerinucSettingWidth_editingFinished();
  void on_spinBoxPerinucSettingNeighAvoidance_valueChanged(QString const&);
  void on_spinBoxPerinucSettingNeighAvoidance_editingFinished();
  void on_spinBoxPerinucSettingOverlapAvoidance_valueChanged(QString const&);
  void on_spinBoxPerinucSettingOverlapAvoidance_editingFinished();
  void on_spinBoxPerinucSettingMinArea_valueChanged(QString const&);
  void on_spinBoxPerinucSettingMinArea_editingFinished();
  void on_checkBoxPerinucSettingsBgAvoid_toggled(bool);
  void on_comboBoxPerinucSettingsBgAvoidChannel_currentTextChanged(QString const&);
  void on_dspinBoxPerinucSettingBgAvoidExpansion_valueChanged(QString const&);
  void on_dspinBoxPerinucSettingBgAvoidExpansion_editingFinished();
  void on_spinBoxPerinucSettingBgAvoidSmoothing_valueChanged(QString const&);
  void on_spinBoxPerinucSettingBgAvoidSmoothing_editingFinished();
  //
  void on_pushButtonPerinucSettingsSave_clicked();
  void on_pushButtonPerinucSettingsLoad_clicked();


  // -- Toolbox: regions editing
  //
  void on_checkBoxRgnEditVoronoiScaling_toggled(bool checked);
  void on_checkBoxRgnEditVoronoiErosion_toggled(bool checked);
  void on_pushButtonRgnEditTesselate_clicked();
  //
  void on_pushButtonRgnEditBackgroundSample_clicked();
  //
  void on_pushButtonRgnEditReset_clicked();
  //
  void on_checkBoxRgnEditShowAnnotations_toggled(bool checked);
  //
  void on_pushButtonRgnEditRegionsSave_clicked();
  void on_pushButtonRgnEditRegionsLoad_clicked();

  // -- Toolbox: quantification
  void on_actionQuantification_triggered();
  //
  void on_lineEditQuantNucChannelG_editingFinished();
  void on_lineEditQuantNucChannelR_editingFinished();
  void on_lineEditQuantNucChannelB_editingFinished();
  void on_lineEditQuantNucChannelC_editingFinished();
  void on_lineEditQuantNucChannelM_editingFinished();
  void on_lineEditQuantNucChannelY_editingFinished();
  void convertChannelDescriptionToId_(QLineEdit*);
  //
private Q_SLOTS:  // continued
  //
  void on_checkBoxQuantImgFilesDat_toggled(bool checked);
  void on_pushButtonQuantImgSaveAll_clicked();
  //
  void on_checkBoxNucQuantSaveCsv_toggled(bool checked);
  void on_checkBoxNucQuantSavePng_toggled(bool checked);
  void on_pushButtonNucQuantSaveAll_clicked();
  //
  void on_pushButtonQuantPerinucSaveAll_clicked();
  //
  void on_checkBoxQuantRgnSavePng_toggled(bool checked);
  void on_checkBoxQuantRgnSaveDat_toggled(bool checked);
  void on_pushButtonQuantRgnSaveAll_clicked();
  //
  void on_pushButtonQuantSaveAll_clicked();


  // -- Toolbox: tracking
  //
  void on_pushButtonTrackingStep_clicked();
  //
  void on_pushButtonTrackingSettingsLoad_clicked();
  void on_pushButtonTrackingSettingsSave_clicked();


  // -- Toolbox: tracks editing
  //
  void on_pushButtonTracksEditRemoveAll_clicked();
  //
  void on_checkBoxTracksEditTracesShowOnlySelected_toggled(bool checked);
  //
  void on_comboBoxTracksEditTraceColorBy_activated(int index);
  //
  void on_dspinBoxTracksEditTraceWidthSelected_editingFinished();
  void on_dspinBoxTracksEditTraceWidthSelected_valueChanged(double width);
  void on_dspinBoxTracksEditTraceWidthNonselected_editingFinished();
  void on_dspinBoxTracksEditTraceWidthNonselected_valueChanged(double width);
  //
  void on_checkBoxTracksEditTraceClippingLookBack_toggled(bool checked);
  void on_checkBoxTracksEditTraceClippingLookAhead_toggled(bool checked);
  void on_spinBoxTracksEditTraceClippingLookBack_editingFinished();
  void on_spinBoxTracksEditTraceClippingLookBack_valueChanged(int back);
  void on_spinBoxTracksEditTraceClippingLookAhead_editingFinished();
  void on_spinBoxTracksEditTraceClippingLookAhead_valueChanged(int ahead);
  void on_checkBoxTracksEditTracesShowStubs_toggled(bool checked);
  //
  void on_pushButtonTracksEditSave_clicked();
  void on_pushButtonTracksEditLoad_clicked();
  //
  void on_pushButtonTracksEditRemoveNonrevised_clicked();

  // ====

protected:

  // middle vector: timepoints, inner vector: channels
  using pixmapitems_container_t = QVector<QGraphicsPixmapItem*>;

  // vector of channels
  using scenes_container_t = QVector<QGraphicsScene*>;
  using views_container_t = QVector<ZoomableGraphicsView*>;

  using QuantContourFeatures = st::QuantFeatures;

  QString const static sWinGoFull;
  QString const static sWinExitFull;
  QString const static sAppVersionLatestUrl;

  Ui::MainWindow* ui_;
  QList<QAction*> menu_tools_actions_ = {};
  QMap<QToolBox*, QAction*> tools_to_actions_ = {};
  QList<QAction*> menu_view_show_actions_  = {};
  QMap<QToolBox*, QList<QWidget*>> toolboxen_external_widgets_ = {};
  QList<QWidget*> navigation_widgets_ = {};
  QMap<ImageChannel, QAction*> channel_toggle_;

  std::unique_ptr<ImageSequence> images_;
  ImageFilter display_filter_ = st::identity;

  QString post_images_load_hook_script_ = "";

  QMutex mutex_;
  QFuture<int>  future_int_;
  QFuture<void> future_void_;
  bool batch_processing_ = false;
  bool should_quit_upon_script_error_ = false;

  QString settings_file_path_;

  QStringList log_;

  std::unique_ptr<ScriptWindow> script_window_ = nullptr;
  std::unique_ptr<HelpWindow>   help_window_   = nullptr;
  std::unique_ptr<AboutDialog>  about_dialog_  = nullptr;
  // Note: LogWindow is derived from QMessageBox, hence no need to store a ptr.

  std::unique_ptr<QNetworkAccessManager> net_manager_ = nullptr;

  std::unique_ptr<QSystemTrayIcon> tray_ = nullptr;
  std::unique_ptr<QMenu> tray_menu_;
  std::unique_ptr<QAction> tray_show_home_webpage_;
  std::unique_ptr<QAction> tray_latest_version_info_;

#if (QT_VERSION >= QT_VERSION_CHECK(5, 2, 0))
# ifdef Q_OS_WIN
    std::unique_ptr<QWinTaskbarButton>   win_taskbar_button_;
    std::unique_ptr<QWinTaskbarProgress> win_taskbar_progress_;
public:
    void showEvent(QShowEvent*);
private:
# endif
#endif

  //QGraphicsScene* current_scene_;
  //ZoomableQGraphicsView* current_view_;

  std::unique_ptr<PixmapCache> pixmap_cache_ = nullptr;
  pixmapitems_container_t pixmapitems_ = {};
  scenes_container_t scenes_ = {};
  views_container_t views_ = {};

  // -- Persistence:
  QString images_directory_ = "";

  // Used in nucDetImgPreprocParamScan():
  using img_preproc_scored_params_t = std::pair<double, st::ImagePreprocessingParameters const*>;
  st::NucleiPreselectionAssessmentParameters nuc_presel_assess_params_;
  QFuture<img_preproc_scored_params_t> future_img_preproc_;
  QFutureWatcher<img_preproc_scored_params_t> watcher_img_preproc_;
  st::ImagePreprocessingParameters img_preproc_best_params_;
  //

  // Data to draw
  //
  QList<QPolygon>              current_exclusions_ = {}; // a.k.a mask; polygons for drawing in cv images
  QList<QGraphicsPolygonItem*> current_exclusions_polys_ = {}; // polygons for drawing in all scenes
  //
  QList<QList<NuclearOutline>>      nuclei_     = {};
  QList<QList<PerinuclearOutlines>> perinuclei_ = {};
  QList<QList<Outline>>             regions_    = {}; // TODO: consider using Polygon, as in exclusions

  QList<NuclearOutline>*      current_nuclei_     = nullptr;
  QList<PerinuclearOutlines>* current_perinuclei_ = nullptr;
  QList<Outline>*             current_regions_    = nullptr;
  //
  QSet<QGraphicsTextItem const*> current_nuclei_labels_ = {};
  QSet<QGraphicsTextItem const*> current_regions_labels_= {};
  //
  // (annotations are not stored, but computed when changing the time-frame)

  QList<Track> tracks_ = {};
  std::unique_ptr<QAbstractTableModel> tracks_model_ = nullptr;
  Track* current_track_ = nullptr;

  // Toolbox-dependent editability in panels
  //
  // list of selectable, insertable, erasable, swappable, togglable and a pair
  // of bools for finer modification of selectability of nuclei and regions
  using panel_editability_props_t = QPair<QList<QSet<int>>,QPair<bool,bool>>;
  using toolbox_to_panel_editability_props_t = QMap<QToolBox*,
                                                    panel_editability_props_t>;

  toolbox_to_panel_editability_props_t editablity_ = {};
  QToolBox* current_toolbox_ = nullptr;

  // Drawing properties
  //
  double nucleus_contour_width_ = 1.0;
  double perinucleus_contour_width_ = 0.8*nucleus_contour_width_;
  double outline_id_font_scaling_ = 1.0;
  double trace_width_ = 1.6;
  //
  // Other properties
  //
  QString highlited_nuclei_file_path_ = "";

  // --
  //
  int current_stage_i_ = 0;
  //
  long int current_frame_i_ = 0;
  long int n_frames_ = 0;

#ifndef Q_OS_WIN
  static int sighupFd[2];
  static int sigtermFd[2];
  std::unique_ptr<QSocketNotifier> snHup = nullptr;
  std::unique_ptr<QSocketNotifier> snTerm = nullptr;
#endif
}; // class MainWindow

#endif // !define(SHUTTLETRACKER_GUI_WINDOWS_MAINWINDOW_HPP)
