// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#ifndef SHUTTLETRACKER_GUI_WINDOWS_MAINWINDOW_AUX_HPP
#define SHUTTLETRACKER_GUI_WINDOWS_MAINWINDOW_AUX_HPP

#include <QSet>
#include <QString>
#include <QGraphicsItem>
#include <QGraphicsPolygonItem>
#include <QGraphicsRectItem>
#include <QGraphicsLineItem>
#include <QGraphicsPixmapItem>
#include <QGraphicsScene>
#include <QPainter>
#include <QImage>
#include <QCursor>
#include <cassert>

namespace aux {

static constexpr double kZNum = 1000.;

// if "less recent" or any error, returns false
//
bool isVersionMoreRecent(QString const& query_v, QString const& ref_v);

bool saveToPixmapFile(QGraphicsScene& scene, QString const& filename);

QVector<QPoint> trimmedToSceneBorder(QVector<QPoint>const& src, QSize const& sz);

// Assuming that the background pixmap is the one and only pixmap.
template <class ScenesContainerType> inline
void trimToBackgroundPixmap(ScenesContainerType& scenes)
{
  for (auto& scene: scenes) {
    for (auto pi: scene->items()) {
      if (pi->type() == QGraphicsPixmapItem::Type) { // then found
        auto gpi = static_cast<QGraphicsPixmapItem*>(pi);
        scene->setSceneRect(gpi->boundingRect());
        break;  // there will be no other pixmap
  } } }
}


// Assuming that {polygons} == {outlines} in the way it's used.
template <int GraphicsItemTypeEnum, class ScenesContainerType> inline
void detachScenesItems(ScenesContainerType& scenes)
{
  for (auto& scene: scenes) {
    for (auto& it : scene->items()) {
      if (it->type() == GraphicsItemTypeEnum) {
        scene->removeItem(it);
  } } }
}


template <int GraphicsItemTypeEnum, class ScenesContainerType> inline
void setScenesItemsSelectable(ScenesContainerType& scenes, bool able)
{
  for (auto& scene: scenes) {
    for (auto& it : scene->items()) {
      if (it->type() == GraphicsItemTypeEnum) {
        if (able) {
          it->setFlags(it->flags() | QGraphicsItem::ItemIsSelectable);
          it->setCursor(QCursor(Qt::PointingHandCursor));
        } else {
          it->setFlags(it->flags() & (~QGraphicsItem::ItemIsSelectable));
          it->unsetCursor();
        }
  } } }
}


template <class ScenesContainerType, class OutlineContainerType>
void showContourNumbers(ScenesContainerType& scenes,
                        OutlineContainerType* outlines,
                        QSet<QGraphicsTextItem const*>& labels,
                        double outline_id_font_scaling_,
                        QString prefix = "")
{
  int ni = 1;
  for (auto& ol: *outlines) {
    QPointF mid{0, 0};
    for (auto const& pt: ol.contour()) { mid += pt; }
    mid /= ol.contour().size();
    QString const label = prefix + QString::number(ni);
    int const annot_shift =  ol.contour().size()/10.f       // PARAM
                           - ((prefix == "R") ? 3.f : 0.f); // PARAM
    for (auto& s: scenes) {
      QFont f;
      f.setPointSize(10.5*outline_id_font_scaling_);
      QGraphicsTextItem* t0 = s->addText(label);  // shadow
      t0->setZValue(kZNum);
      t0->setDefaultTextColor(QColor{0, 0, 0, 158});
      t0->setFont(f);
      t0->setPos(mid.toPoint() - QPoint{annot_shift-1, annot_shift-1});
      QGraphicsTextItem* t = s->addText(label);
      t->setZValue(kZNum + 1);
      t->setDefaultTextColor(QColor{255, 255, 255, 192});
      t->setFont(f);   // TODO? -- configurable annot_shift
      t->setPos(mid.toPoint() - QPoint{annot_shift, annot_shift});
      labels.insert(t0);
      labels.insert(t);
    } // for each scene
    ni++;
  } // for each contour
} // showContourNumbers_()


template <class OutlineContainerType>
bool removeFromOutlines(QGraphicsItem const* gi,
                        OutlineContainerType* outlines_container)
{
  assert(gi->type() == QGraphicsPolygonItem::Type);
  auto*gpi = static_cast<QGraphicsPolygonItem*>(const_cast<QGraphicsItem*>(gi));
  int i = 0;
  for (auto& n: *outlines_container) {
    if (n.isDepictedBy(gpi)) {
      outlines_container->removeAt(i);
      return true;
    }
    i++;
  }
  return false;
}

} // namespace aux

#endif // !defined(SHUTTLETRACKER_GUI_WINDOWS_MAINWINDOW_AUX_HPP)
