// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "MainWindow.hpp"
#include "ui_MainWindow.h"
#include "MainWindow_aux.hpp"
#include <cassert>

static const QColor kWhite{255, 255, 255,  255};
static const QColor kGreen{  8, 242,  16,  248};

static const QColor kNucSolo  {222, 222, 255,  242};
static const QColor kNucSplit {  0, 255,   0,  242};
static const QColor kNucManual{  0, 192, 255,  242};
static const QColor kNucNonsp {255,   0,   0,  242};
static const QColor kNucOrphan{222, 196,   0,  242};
static const QColor kNucDebris{144 , 96 ,162,  242};

static const QColor kRegion  = QColor{255, 0,  255, 242};
static const QColor kPerinuc = QColor{255, 160, 32, 242};

static const Qt::PenStyle kNontoggledStyle = Qt::SolidLine;
static const Qt::PenStyle kToggledStyle = Qt::DotLine;

static const double kZTrk = 200.;  // front-most
static const double kZNuc = 100.;
static const double kZPer =  50.;
static const double kZReg =  20.;


void MainWindow::ifNotHiddenThenRedrawFrameOutlines_()
{  if (not ui_->actionHideMarking->isChecked()) { redrawFrameOutlines_(); }  }


void MainWindow::ifNotHiddenThenRedrawTracks_()
{  if (not ui_->actionHideMarking->isChecked()) { redrawTracks_(); }  }


void MainWindow::drawExclusions_()
{
  for (auto& excl_poly: current_exclusions_polys_) { delete excl_poly; }
  current_exclusions_polys_.clear();

  for (auto const& excl: qAsConst(current_exclusions_)) {
    for (auto& scene: scenes_) {
      QGraphicsPolygonItem* poly =
          scene->addPolygon(excl, QPen{Qt::gray},
                            QBrush{Qt::darkGray, Qt::BrushStyle::BDiagPattern});
      current_exclusions_polys_.push_back(poly);
  } }
  updateEditabilities_();
}


void MainWindow::drawFrameNuclei_()
{
  assert(draw_frame_nuclei__);
  (this->*draw_frame_nuclei__)();
}


void MainWindow::drawFrameNucleiByProvenience_()
{
  for (auto& nuc: *current_nuclei_) {
    QPen pen{kWhite, nucleus_contour_width_,
             Qt::SolidLine, Qt::RoundCap, Qt::RoundJoin};
    switch (nuc.provenience()) {
      case NuclearOutline::Provenience::kAutoSolo:
        if (ui_->checkBoxNucEditMarkNucleiSolo->isChecked()) {
          pen.setColor(kNucSolo);  break;
        } else { continue; }
      case NuclearOutline::Provenience::kAutoFromSplit:
        if (ui_->checkBoxNucEditMarkNucleiSplit->isChecked()) {
          pen.setColor(kNucSplit);  break;
        } else { continue; }
      case NuclearOutline::Provenience::kManual:
        if (ui_->checkBoxNucEditMarkNucleiManual->isChecked()) {
          pen.setColor(kNucManual);  break;
        } else { continue; }
      case NuclearOutline::Provenience::kAutoImproperGroup:
        if (ui_->checkBoxNucEditMarkNonsplittable->isChecked()) {
          pen.setColor(kNucNonsp);  break;
        } else { continue; }
      case NuclearOutline::Provenience::kAutoImproperOrphan:
        if (ui_->checkBoxNucEditMarkSplitOrphans->isChecked()) {
          pen.setColor(kNucOrphan);  break;
        } else { continue; }
      case NuclearOutline::Provenience::kAutoImproperDebris:
        if (ui_->checkBoxNucEditMarkDebris->isChecked()) {
          pen.setColor(kNucDebris);  break;
        } else { continue; }
      case NuclearOutline::Provenience::kUnknown: // fall-through
      default:
        message_('E', "Do not know how to draw a nucleus!");
    }
    pen.setStyle(nuc.isToggled() ? kToggledStyle : kNontoggledStyle);
    nuc.clearDepictions();
    for (auto ps: qAsConst(scenes_)) {
      auto *gp = new QGraphicsPolygonItem(nuc.contour());
      ps->addItem(gp);
      gp->setPen(pen);
      gp->setZValue(kZNuc);
      nuc.setDepictedBy(gp);
      // OBSOLETE
      //if (ui_->actionNucleiEditing->isChecked()) {
      //  gp->setFlag(QGraphicsItem::ItemIsSelectable);
      //  gp->setCursor(QCursor(Qt::PointingHandCursor));
      //}
    } // for each scene
  } // for each nucleus

  updateEditabilities_();
}


void MainWindow::drawFrameNucleiByProperty_()
{
  QVector<NuclearOutline*> sorted;
  for (auto& nuc: *current_nuclei_) { sorted.push_back(&nuc); }

  auto order_by_area = [](NuclearOutline* a, NuclearOutline* b)
      { return a->area() > b->area(); };

  auto order_by_solidity = [](NuclearOutline* a, NuclearOutline* b)
      { return a->solidity() < b->solidity(); };

  if (       ui_->comboBoxNucEditColorBy->currentText() == "by area") {
    std::sort(sorted.begin(),sorted.end(), order_by_area);
  } else if (ui_->comboBoxNucEditColorBy->currentText() == "by solidity") {
    std::sort(sorted.begin(),sorted.end(), order_by_solidity);
  } else {
    message_('W', "Cannot order contours "
                  + ui_->comboBoxNucEditColorBy->currentText(), 6000);
    return;
  }

  int pxi = 0;
  for (auto& pxnuc: sorted) {
    QColor const color(
        static_cast<float>(sorted.size() - pxi)/sorted.size()*200 + 55,
        112,
        200 - static_cast<float>(sorted.size() - pxi)/sorted.size()*200 + 55,
        248);
    QPen pen(color, nucleus_contour_width_, Qt::SolidLine, Qt::RoundCap,
             Qt::RoundJoin);
    NuclearOutline::Provenience prov = pxnuc->provenience();
    if (   (    prov == NuclearOutline::Provenience::kAutoSolo
            and ui_->checkBoxNucEditMarkNucleiSolo->isChecked())
        or (    prov == NuclearOutline::Provenience::kAutoFromSplit
            and ui_->checkBoxNucEditMarkNucleiSplit->isChecked())
        or (    prov == NuclearOutline::Provenience::kManual
            and ui_->checkBoxNucEditMarkNucleiManual->isChecked())
        or (    prov == NuclearOutline::Provenience::kAutoImproperGroup
            and ui_->checkBoxNucEditMarkNonsplittable->isChecked())
        or (    prov == NuclearOutline::Provenience::kAutoImproperOrphan
            and ui_->checkBoxNucEditMarkSplitOrphans->isChecked())
        or (    prov == NuclearOutline::Provenience::kAutoImproperDebris
            and ui_->checkBoxNucEditMarkDebris->isChecked())) {
      pen.setStyle(pxnuc->isToggled() ? kToggledStyle : kNontoggledStyle);
      pxnuc->clearDepictions();
      for (auto& ps: scenes_) {
        auto *gp = new QGraphicsPolygonItem(pxnuc->contour());
        ps->addItem(gp);
        gp->setPen(pen);
        gp->setZValue(kZNuc);
        // OBSOLETE
        //if (ui_->actionNucleiEditing->isChecked()) {
        //  gp->setFlag(QGraphicsItem::ItemIsSelectable);
        //  gp->setCursor(QCursor(Qt::PointingHandCursor));
        //}
        pxnuc->setDepictedBy(gp);
      } // for each scene
    } // if
    pxi++;
  } // for each contour

  updateEditabilities_();
}


void MainWindow::drawFrameNucleiInWhite_()
{
  for (auto& nuc: *current_nuclei_) {
    QPen const pen{kWhite, nucleus_contour_width_,
                   nuc.isToggled() ? kToggledStyle : kNontoggledStyle,
                   Qt::RoundCap, Qt::RoundJoin};
    nuc.clearDepictions();
    for (auto ps: qAsConst(scenes_)) {
      auto *gp = new QGraphicsPolygonItem(nuc.contour());
      ps->addItem(gp);
      gp->setPen(pen);
      gp->setZValue(kZNuc);
      nuc.setDepictedBy(gp);
      // OBSOLETE
      //if (ui_->actionNucleiEditing->isChecked()) {
      //  gp->setFlag(QGraphicsItem::ItemIsSelectable);
      //  gp->setCursor(QCursor(Qt::PointingHandCursor));
      //}
    } // for each scene
  } // for each nucleus

  updateEditabilities_();
}



void MainWindow::drawFrameNucleiInGreen_()
{
  for (auto& nuc: *current_nuclei_) {
    QPen const pen{kGreen, nucleus_contour_width_,
                   nuc.isToggled() ? kToggledStyle : kNontoggledStyle,
                   Qt::RoundCap, Qt::RoundJoin};
    nuc.clearDepictions();
    for (auto ps: qAsConst(scenes_)) {
      auto *gp = new QGraphicsPolygonItem(nuc.contour());
      ps->addItem(gp);
      gp->setPen(pen);
      gp->setZValue(kZNuc);
      nuc.setDepictedBy(gp);
      // OBSOLETE
      //if (ui_->actionNucleiEditing->isChecked()) {
      //  gp->setFlag(QGraphicsItem::ItemIsSelectable);
      //  gp->setCursor(QCursor(Qt::PointingHandCursor));
      //}
    } // for each scene
  } // for each nucleus

  updateEditabilities_();
}



void MainWindow::drawFramePerinuclei_()
{
  perinucleus_contour_width_ = ui_->dspinBoxPerinucContourWidth->value();
  QPen const pen{kPerinuc, perinucleus_contour_width_,
                 Qt::SolidLine, Qt::RoundCap, Qt::RoundJoin};
  for (auto& pn: *current_perinuclei_) {
    pn.clearDepictions();
    for (auto ps: qAsConst(scenes_)) {
      if (pn.kind() == PerinuclearOutlines::Kind::kCompleteO) {
        for (auto i: {2, 1}) {
          auto& outline = pn.outlines()[i];
          auto* gp = new QGraphicsPolygonItem(outline.contour());
          ps->addItem(gp);
          gp->setPen(pen);
          gp->setZValue(kZPer);
          outline.setDepictedBy(gp);
        }
      } else {
        for (auto& outline: pn.outlines()) {
          auto* gp = new QGraphicsPolygonItem(outline.contour());
          ps->addItem(gp);
          gp->setPen(pen);
          gp->setZValue(kZPer);
          outline.setDepictedBy(gp);
        } // for each outline
      } // if of kind
    } // for each scene
  } // for each perinucleus
}


void MainWindow::redrawNucleus_(NuclearOutline& nuc)
{
  nuc.clearDepictions();

  QColor color = Qt::black;  // shuo
  NuclearOutline::Provenience prov = nuc.provenience();
  if (    prov == NuclearOutline::Provenience::kAutoSolo
      and ui_->checkBoxNucEditMarkNucleiSolo->isChecked()) {  color = kNucSolo; }
  else if (prov == NuclearOutline::Provenience::kAutoFromSplit
      and ui_->checkBoxNucEditMarkNucleiSplit->isChecked()){  color = kNucSplit; }
  else if (prov == NuclearOutline::Provenience::kManual
      and ui_->checkBoxNucEditMarkNucleiManual->isChecked()){ color = kNucManual; }
  else if (prov == NuclearOutline::Provenience::kAutoImproperGroup
      and ui_->checkBoxNucEditMarkNonsplittable->isChecked()){color = kNucNonsp; }
  else if (prov == NuclearOutline::Provenience::kAutoImproperOrphan
      and ui_->checkBoxNucEditMarkSplitOrphans->isChecked()){ color = kNucOrphan; }
  else if (prov == NuclearOutline::Provenience::kAutoImproperDebris
      and ui_->checkBoxNucEditMarkDebris->isChecked())      { color = kNucDebris; }
  if (color == Qt::black) { return; }

  QPen const pen(color, nucleus_contour_width_,
                nuc.isToggled() ? kToggledStyle : kNontoggledStyle,
                Qt::RoundCap, Qt::RoundJoin);
  for (auto ps: qAsConst(scenes_)) {
    auto *gp = new QGraphicsPolygonItem(nuc.contour());
    ps->addItem(gp);
    gp->setPen(pen);
    gp->setZValue(kZNuc);
    nuc.setDepictedBy(gp);
    updateEditabilityOfNucleus_(gp);
  }
}


void MainWindow::drawLastNucleus_()
{
  assert(draw_frame_nuclei__ != &MainWindow::drawFrameNucleiByProperty_);
  assert(current_nuclei_);
  assert(not current_nuclei_->empty());
  NuclearOutline& nuc = current_nuclei_->last();
//assert(nuc.provenience() == NuclearOutline::Provenience::kManual);
  redrawNucleus_(nuc);
}


void MainWindow::drawLastRegion_()
{
  assert(current_regions_);
  assert(not current_regions_->empty());

  auto& r = current_regions_->last();
  r.clearDepictions();

  auto const& selectable_nuc_reg = editablity_[current_toolbox_].second;
  assert(selectable_nuc_reg.first or selectable_nuc_reg.second);

  // CONSISTENCY ac46f46e
  QPen const pen(kRegion, nucleus_contour_width_,Qt::SolidLine,Qt::RoundCap,
                 Qt::RoundJoin);

  for (auto& ps: scenes_) {
    auto *gp = new QGraphicsPolygonItem(r.contour());
    ps->addItem(gp);
    gp->setPen(pen);
    gp->setZValue(kZReg);
    // update editability:
    if (selectable_nuc_reg.second) {
      gp->setFlags(gp->flags() | QGraphicsItem::ItemIsSelectable);
      gp->setCursor(QCursor(Qt::PointingHandCursor));
    } else {
      gp->setFlags(gp->flags() & (~QGraphicsItem::ItemIsSelectable));
      gp->unsetCursor();
    }
    r.setDepictedBy(gp);

  } // for each scene
}


void MainWindow::drawFrameRegions_()
{
  QPen const pen{kRegion, nucleus_contour_width_, Qt::SolidLine, Qt::RoundCap,
                 Qt::RoundJoin};

  for (auto& r: *current_regions_) {
    r.clearDepictions();
    for (auto& ps: scenes_) {
      auto *gp = new QGraphicsPolygonItem(r.contour());
      ps->addItem(gp);
      gp->setPen(pen);
      gp->setZValue(kZReg);
      // OBSOLETE
      //if (ui_->actionRegionsEditing->isChecked()) {
      //  gp->setFlag(QGraphicsItem::ItemIsSelectable);
      //  gp->setCursor(QCursor(Qt::PointingHandCursor));
      //}
      r.setDepictedBy(gp);
    } // for each scene
  } // for each region

  updateEditabilities_();
}


// TODO: When there is one track point, the center of mass is refreshed;
//       when there are more points, it's not. That's inconsistent.
//
void MainWindow::drawTrack__(Track* track)
{
  assert(track->nodesCount() > 0);

  bool const drawing_current_track = (track == current_track_);
  double const width = drawing_current_track
    ? ui_->dspinBoxTracksEditTraceWidthSelected->value()
    : ui_->dspinBoxTracksEditTraceWidthNonselected->value();
  auto const rand_max = static_cast<double>(RAND_MAX);
  QColor const rcolor{static_cast<int>(255*(0.333 + 0.667*rand()/rand_max)),
                      static_cast<int>(255*(0.5   + 0.5  *rand()/rand_max)),
                      static_cast<int>(255*(0.5   + 0.5  *rand()/rand_max)), 242};
  QPen pen{rcolor, width, Qt::SolidLine, Qt::RoundCap};

  // -- no segments
  if (track->nodesCount() == 1) {
    if (ui_->checkBoxTracksEditTracesShowStubs->isChecked()) {
      for (auto ps: qAsConst(scenes_)) {
        assert(track->initialOutline());
        auto const com = st::centerOfMass(track->initialOutline()->cv_contour());
        //auto* circle = new QGraphicsEllipseItem(com.x,com.y,3,3);
        qreal const w = 3.0; // PARAM 2b779541-389f-47df-9e6a-03d9e1976042
        auto* mark = new QGraphicsRectItem{com.x - w/2.f, com.y - w/2.f,  w, w};
        mark->setZValue(kZTrk);
        mark->setPen(pen);
        // OBSOLETE
        //if (ui_->actionTracksEditing->isChecked()) {
        //  mark->setFlag(QGraphicsItem::ItemIsSelectable);
        //  mark->setCursor(QCursor(Qt::PointingHandCursor));
        //}
        ps->addItem(mark);
        track->setDepictedBy(mark);

        updateEditability_(mark);
    } }

  // -- at least one segment
  } else {
    for (TrackSegment& segment: track->segments()) {
      if (ui_->checkBoxTracksEditTraceClippingLookBack->isChecked()) {
        long int const n_back = current_frame_i_ - segment.frameIndex();
        if (n_back >= 0) {
          if (n_back <= ui_->spinBoxTracksEditTraceClippingLookBack->value()) {
            pen.setColor(rcolor.darker(100 + 150 *\
                n_back/(ui_->spinBoxTracksEditTraceClippingLookBack
                           ->value() + 1)));
          } else { continue; }
      } }
      if (ui_->checkBoxTracksEditTraceClippingLookAhead->isChecked()) {
        long int const n_ahead = segment.frameIndex() - current_frame_i_;
        if (n_ahead >= 0) {
          if (n_ahead <= ui_->spinBoxTracksEditTraceClippingLookAhead->value()) {
            pen.setColor(rcolor.lighter(100 + 75 *\
                n_ahead/(ui_->spinBoxTracksEditTraceClippingLookAhead->value()+1)));
          } else { continue; }
      } }
      for (auto ps: qAsConst(scenes_)) {
        auto* line = new QGraphicsLineItem(segment.line());
        line->setPen(pen);
        line->setZValue(kZTrk);
        // OBSOLETE
        //if (ui_->actionTracksEditing->isChecked()) {
        //  line->setFlag(QGraphicsItem::ItemIsSelectable);
        //  line->setCursor(QCursor(Qt::PointingHandCursor));
        //}
        ps->addItem(line);
        segment.setDepictedBy(line);

        updateEditability_(line);
      }
    } // for each segment
  } // if # segments
} // drawTrack__()


void MainWindow::drawTracks_()
{
  srand(42);  // enforce (some) coloring consistency

  for (Track& track: tracks_) {

    bool const drawing_current_track = (&track == current_track_);
    if (not drawing_current_track) {

      // unbolden track nucleus
      track.setTrackedNucleiContourWidth(ui_->dspinBoxNucEditContourWidth->value());
      if (ui_->checkBoxTracksEditTracesShowOnlySelected->isChecked()) {
        continue;
      }
    }

    drawTrack__(&track);

    // unbolden track nucleus
    track.setTrackedNucleiContourWidth(ui_->dspinBoxNucEditContourWidth->value());

  } // for each track

  // bolden current track nucleus
  if (current_track_)  {
#ifndef NDEBUG
    bool found = false;
    for (auto const& t: qAsConst(tracks_)) {if (&t==current_track_) {found=true; break;}}
    assert(found);
#endif
    current_track_->setTrackedNucleiContourWidth(
        ui_->dspinBoxTracksEditTraceWidthSelected->value() );
  }
  // update_editabilities_(); // redundant -- draw_track__ does this already
}



// -- Annotations -------------------------------------------------------------


void  MainWindow::makeFrameAnnotations_(bool an_nuclei, bool an_regions)
{
  if (an_nuclei) {
    aux::showContourNumbers(scenes_, current_nuclei_, current_nuclei_labels_,
                            outline_id_font_scaling_, "");
  }

  if (an_regions) {
    aux::showContourNumbers(scenes_, current_regions_,current_regions_labels_,
                            outline_id_font_scaling_, "R");
  }
}


void MainWindow::clearFrameAnnotations_(bool nuclei_ans, bool regions_ans)
{
  for (auto& scene: scenes_) {
    auto const& items = scene->items();
    for (auto it: items) {
      if (it->type() == QGraphicsTextItem::Type) {

        auto const* gti = static_cast<QGraphicsTextItem const*>(it);
        if (nuclei_ans) {
          if (current_nuclei_labels_.contains(gti)) {
            delete it;
            current_nuclei_labels_.remove(gti);
            continue; // cannot be a region annotation, so skip remaining checks
        } }

        if (regions_ans) {
          if (current_regions_labels_.contains(gti)) {
            current_regions_labels_.remove(gti);
            delete it;
        } }

      } // if
  } } // for, for
}


void MainWindow::maybeFrameAnnotate_()
{
  makeFrameAnnotations_(ui_->checkBoxNucEditShowAnnotations->isChecked(),
                        ui_->checkBoxRgnEditShowAnnotations->isChecked());
}
