// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "MainWindow.hpp"
#include "ui_MainWindow.h"
#include "MainWindow_aux.hpp"
#include <cassert>


void MainWindow::updateEditabilities_()
{
  if (not current_toolbox_) {
    aux::setScenesItemsSelectable<QGraphicsPolygonItem::Type>(scenes_,false);
    aux::setScenesItemsSelectable<QGraphicsRectItem::Type>   (scenes_,false);
    aux::setScenesItemsSelectable<QGraphicsLineItem::Type>   (scenes_,false);
    return;
  }

  assert(current_toolbox_);
  assert(editablity_.contains(current_toolbox_));
  assert(not editablity_[current_toolbox_].first.empty());
  auto const& selectable = editablity_[current_toolbox_].first[0];

  // "selectability":
  aux::setScenesItemsSelectable<QGraphicsPolygonItem::Type>(scenes_,
      selectable.contains(QGraphicsPolygonItem::Type));
  aux::setScenesItemsSelectable<QGraphicsRectItem::Type>(scenes_,
     selectable.contains(QGraphicsRectItem::Type));
  aux::setScenesItemsSelectable<QGraphicsLineItem::Type>(scenes_,
     selectable.contains(QGraphicsLineItem::Type));

  // partial override: make regions or regions selectable or non-selectable:
  //
  if (selectable.contains(QGraphicsPolygonItem::Type)) {
    //
    auto const& selectable_nuc_reg = editablity_[current_toolbox_].second;
    assert(selectable_nuc_reg.first or selectable_nuc_reg.second);
    assert(current_nuclei_);
    for (Outline& n: *current_nuclei_) {
      for (auto& pp: n.depictions()) {
        assert(pp);
        if (selectable_nuc_reg.first) {
          pp->setFlags(pp->flags()  |  QGraphicsItem::ItemIsSelectable);
          pp->setCursor(QCursor(Qt::PointingHandCursor));
        } else {
          pp->setFlags(pp->flags() & (~QGraphicsItem::ItemIsSelectable));
          pp->unsetCursor();
    } } }
    assert(current_regions_);
    for (Outline& r: *current_regions_) {
      for (auto& pp: r.depictions()) {
        assert(pp);
        if (selectable_nuc_reg.second) {
          pp->setFlags(pp->flags()  |  QGraphicsItem::ItemIsSelectable);
          pp->setCursor(QCursor(Qt::PointingHandCursor));
        } else {
          pp->setFlags(pp->flags() & (~QGraphicsItem::ItemIsSelectable));
          pp->unsetCursor();
        }
    } }
  }
}


void MainWindow::updateEditability_(QGraphicsItem* gi)
{
  if (current_toolbox_) {
    auto const& editability = editablity_[current_toolbox_];
    if (editability.first[0].contains(gi->type())) {
      gi->setFlag(QGraphicsItem::ItemIsSelectable);
      gi->setCursor(QCursor(Qt::PointingHandCursor));
      return;
  } }
  gi->setFlags(gi->flags() & (~QGraphicsItem::ItemIsSelectable));
  gi->unsetCursor();
}


void MainWindow::updateEditabilityOfNucleus_(QGraphicsPolygonItem* pi)
{
  if (current_toolbox_) {
    auto const& editability = editablity_[current_toolbox_];
    if (editability.first[0].contains(QGraphicsPolygonItem::Type)) {
      pi->setFlag(QGraphicsItem::ItemIsSelectable);
      pi->setCursor(QCursor(Qt::PointingHandCursor));
      return;
  } }
  pi->setFlags(pi->flags() & (~QGraphicsItem::ItemIsSelectable));
  pi->unsetCursor();
}


void MainWindow::updateEditabilityOfRegion_(QGraphicsPolygonItem* pi)
{
  if (current_toolbox_) {
    auto const& editability = editablity_[current_toolbox_];
    if (    editability.first[0].contains(QGraphicsPolygonItem::Type)
        and editability.second.second                                ) {
      pi->setFlag(QGraphicsItem::ItemIsSelectable);
      pi->setCursor(QCursor(Qt::PointingHandCursor));
      return;
  } }
  pi->setFlags(pi->flags() & (~QGraphicsItem::ItemIsSelectable));
  pi->unsetCursor();
}


// TODO: redrawing all scenes after every new contour is a performance burden
// slot
void MainWindow::itemAdded(QVector<QPoint> const& vp)
{
  try {

    if (ui_->actionImageMasking->isChecked()) { // then: exclusion

      if (vp.size() >= 3) { // kind of PARAM CONSISTENCY:min_contour_length
        current_exclusions_.push_back({vp});
        drawExclusions_();
      } else {
        message_('W', "Empty contour ignored.");
      }
      ui_->pushButtonImgMaskReset->setEnabled(not current_exclusions_.empty());
      ui_->pushButtonImgMaskApply->setEnabled(not current_exclusions_.empty());

    } else if (ui_->actionNucleiEditing->isChecked()) { // then: nucleus

      auto const vp_trimmed = aux::trimmedToSceneBorder(vp,
                                            QSize{images_->imageSize().width,
                                                  images_->imageSize().height});
      if (vp_trimmed.size() >= 3) { // kind of PARAM CONSISTENCY:min_contour_length

        current_nuclei_->push_back({vp_trimmed,
                                    NuclearOutline::Provenience::kManual});

        if (       draw_frame_nuclei__==&MainWindow::drawFrameNucleiByProvenience_
             or    draw_frame_nuclei__==&MainWindow::drawFrameNucleiInWhite_
             or    draw_frame_nuclei__==&MainWindow::drawFrameNucleiInGreen_) {
          // full refresh is not necesary:
          drawLastNucleus_();
        } else if (draw_frame_nuclei__==&MainWindow::drawFrameNucleiByProperty_) {
          // -- must recalculate the distribution, full refresh is needed:
          redrawFrameOutlines_();
        } else {
          assert(false); // implausible
        }

        if (ui_->checkBoxNucEditShowAnnotations->isChecked()) {
          clearFrameAnnotations_(true, false);
          makeFrameAnnotations_(true, false);
        }
        clearFramePerinuclei_();

      } else {
        message_('W', "Contour too small, not added!");
        return;
      }

    } else if (ui_->actionRegionsEditing->isChecked()) { // then: region

      auto const vp_trimmed = aux::trimmedToSceneBorder(vp,
                                            QSize{images_->imageSize().width,
                                                  images_->imageSize().height});
      if (vp_trimmed.size() >= 3) { // kind of PARAM CONSISTENCY:min_contour_length

        current_regions_->push_back( Outline{vp_trimmed} );
        // full redraw is not necessary:
        //redraw_frame_outlines_();
        drawLastRegion_();
        if (ui_->checkBoxRgnEditShowAnnotations->isChecked()) {
          clearFrameAnnotations_(false, true);
          makeFrameAnnotations_(false, true);
        }

      } else {
        message_('W', "Contour too small, not added!");
        return;
      }

    } else if (ui_->actionNucleiDetection->isChecked()) { // then nucleus
      message_('W', "Cannot draw new nuclei "
                    "in the nuclei detection mode", 9000);
    } else {
      message_('W', "Don't know what to do with the just drawn contour."
                    " (Hint: Activate an editing tool.)", 9000);
    }

  } catch (...) {
    message_('W', "Contour not added!");
  }
}


// slot
void MainWindow::itemRemoved(QGraphicsItem const* gi)
{
  bool outline_removed = false;
  bool track_cut = false;

  if (ui_->actionNucleiEditing->isChecked()) { // then: nucleus or non-nucleus
    if (not tracks_.empty()) {
      message_('W', "Cannot remove nuclei while having tracks.");
    } else if (not current_perinuclei_->empty()) {
      message_('W', "Cannot remove nuclei while having perinuclei.");
    } else {
      outline_removed = aux::removeFromOutlines(gi, current_nuclei_);
      if (outline_removed) {
        if (ui_->checkBoxNucEditShowAnnotations->isChecked()) {
          clearFrameAnnotations_(true, false);
          makeFrameAnnotations_(true, false);
  } } } }

  else if (ui_->actionRegionsEditing->isChecked()) { // then: region
    outline_removed = aux::removeFromOutlines(gi, current_regions_);
    if (outline_removed and ui_->checkBoxRgnEditShowAnnotations->isChecked()) {
      clearFrameAnnotations_(false, true);
      makeFrameAnnotations_(false, true);
  } }

  else if (ui_->actionTracksEditing->isChecked()) { // then: cutting a track
    if (gi->type() == QGraphicsLineItem::Type) {
      auto*li = static_cast<QGraphicsLineItem*>(const_cast<QGraphicsItem*>(gi));
      for (auto& t: tracks_) {
        if (track_cut) { break; } // use flag
        for (auto& s: t.segments()) {
          if (s.isDepictedBy(li)) {
            bool const mid_split = Track::splitAt(&t,&s,&tracks_);
            for (auto& n: *current_nuclei_) { // TODO: costly: affects only 1 nuc
              auto const w = ui_->dspinBoxNucEditContourWidth->value();
              n.setDepictionsWidth(w);
            }
            resetTracksModel_(); // TODO: partial update if mid_split == false
            if (mid_split) { current_track_ = nullptr; }
            track_cut = true; // set flag
            break;
        } }
      } // for
    } else if (gi->type() == QGraphicsRectItem::Type) {
      auto*ri = static_cast<QGraphicsRectItem*>(const_cast<QGraphicsItem*>(gi));
#ifndef NDEBUG
      bool found_depiction = false;
#endif
      QMutableListIterator<Track> it(tracks_);
      while (it.hasNext()) {
        if (it.next().isDepictedBy(ri)) {
#ifndef NDEBUG
          found_depiction = true;
#endif
          assert(it.value().nodesCount() == 1);
          it.remove();
          resetTracksModel_();
          current_track_ = nullptr;
          break;
        }
      } // for each track
#ifndef NDEBUG
      assert(found_depiction);
#endif
    }
  } // if toolbox active

  // A graphics polygon can be removed in ZoomableGraphicsView, but if it was
  // not removed above, it will be redrawn:
  //
  if (outline_removed) { redrawFrameOutlines_(); }
  if (track_cut)       { redrawTracks_();        }
}


void MainWindow::itemReplaced(QGraphicsItem const* pi, QVector<QPoint>const& vp)
{
  assert(ui_->actionTracksEditing->isChecked());
  assert(pi->type() == QGraphicsPolygonItem::Type);
  auto* pl = static_cast<QGraphicsPolygonItem*>(const_cast<QGraphicsItem*>(pi));

  // -- (1/2) replace outline in current_nuclei_
  assert(current_nuclei_);

  QSize const image_size{images_->imageSize().width,
                         images_->imageSize().height};
#ifndef NDEBUG
  bool to_be_replaced_found = false;
#endif
  for (NuclearOutline& nuc_otl: *current_nuclei_) {
    if (nuc_otl.isDepictedBy(pl)) {
#ifndef NDEBUG
      to_be_replaced_found = true;
#endif
      // detach
      // why not just "nuc_otl.clear_depictions();" below?
      for (auto& scene: scenes_) {
        for (auto& it: scene->items()) {
          if (it->type() == QGraphicsPolygonItem::Type) {
            auto* spl = static_cast<QGraphicsPolygonItem*>(it);
            if (nuc_otl.isDepictedBy(spl)) {
              scene->removeItem(it);
      } } } }

      auto const vp_trimmed = aux::trimmedToSceneBorder(vp, image_size);
      nuc_otl = NuclearOutline{vp_trimmed, NuclearOutline::Provenience::kManual};

      if (       draw_frame_nuclei__==&MainWindow::drawFrameNucleiByProvenience_
          or     draw_frame_nuclei__==&MainWindow::drawFrameNucleiInGreen_) {
        // full refresh is not necesary:
        redrawNucleus_(nuc_otl);  // which is already new_o
      } else if (draw_frame_nuclei__==&MainWindow::drawFrameNucleiByProperty_) {
        // must recalculate the distribution, full refresh is needed:
        redrawFrameOutlines_();
      } else {
        assert(false);  // implausible
      }

      if (ui_->checkBoxNucEditShowAnnotations->isChecked()) {
        clearFrameAnnotations_(true, false);
        makeFrameAnnotations_(true, false);
      }
      clearFramePerinuclei_();

      if (not tracks_.empty()) {
        if (not (current_track_ and current_track_->updateAt(&nuc_otl))) {
          for (auto& t: tracks_) {
            if (t.updateAt(&nuc_otl)) { break; }
      } } }
      redrawTracks_();

      break;
    } // if depicted
  } // for each nucleus in current frame

#ifndef NDEBUG
  assert(to_be_replaced_found);
#endif
}


void MainWindow::itemSplit(QGraphicsItem const* gi, QVector<QPoint>const& vp)
{
  assert(ui_->actionNucleiEditing->isChecked());
  assert(gi->type() == QGraphicsPolygonItem::Type);

  for (int nuc_i=0, nuc_s=SSIZE_PTR(current_nuclei_); nuc_i < nuc_s; ++nuc_i) {
    auto const& nuc = current_nuclei_->at(nuc_i);
    if (nuc.isDepictedBy(static_cast<QGraphicsPolygonItem const*>(gi))) {
      st::contour_t splitter;
      for (auto const& p: vp) { splitter.emplace_back(p.x(), p.y()); }
      auto const split_nuclei = st::splitNucleus(nuc.cv_contour(), splitter,
                                                 images_->imageSize());
      if (split_nuclei.size() > 1) {
        current_nuclei_->removeAt(nuc_i);
        for (auto const& sn: split_nuclei) {
          current_nuclei_->push_back({sn, NuclearOutline::Provenience::kManual});
          redrawNucleus_(current_nuclei_->back());
        }
        if (ui_->checkBoxNucEditShowAnnotations->isChecked()) {
          clearFrameAnnotations_(true, false);
          makeFrameAnnotations_(true, false);
        }
      } else {
        message_('E', "Could not split the nucleus");
      }
      return;
  } }
  message_('E', "Could not split the nucleus");
}


void MainWindow::itemToggled(QGraphicsItem const* cpi)
{
  assert(cpi->type() == QGraphicsPolygonItem::Type);
  auto* pi = static_cast<QGraphicsPolygonItem*>(const_cast<QGraphicsItem*>(cpi));

  auto tggld = [&](QList<NuclearOutline>& nuc_container) -> bool {
    for (auto& n: nuc_container) {
      if (n.isDepictedBy(pi)) { n.toggle(); return true; }
    }
    return false;
  };

  bool const tgd = tggld(*current_nuclei_);
  if (not tgd) { message_('E', "Cannot toggle this contour"); }
}


// TODO: test more thoroughly
void MainWindow::itemExtended(QGraphicsPolygonItem const* pi)
{
  // -- Do not add a stub if a nucleus is already a part of a track.
  auto const warning = "Cannot create a stub for a nucleus "
                       "that is already part of a track.";
  for (auto const& t: qAsConst(tracks_)) {
    if (t.initialOutline()->isDepictedBy(pi)) {
      message_('W', warning, 9000);
      return;
    }
    for (auto const& s: t.segments()) {
      if (s.outline()->isDepictedBy(pi)) {
        message_('W', warning, 9000);
        return;
  } } }

  for (auto& curr_nuc: *current_nuclei_) {
    if (curr_nuc.isDepictedBy(pi)) {
      tracks_.push_back( Track{&curr_nuc,current_frame_i_,false} );
      tracks_.back().setSelfCleanup(true);
  } }

  resetTracksModel_(); // TODO: partial update
  current_track_ = &tracks_.back();
  redrawTracks_();     // TODO: partial redraw
}
