// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "MainWindow.hpp"
#include "ui_MainWindow.h"
#include <QFileDialog>
#include <QPixmap>
#include <QGraphicsPixmapItem>
#include <QImage>
#include <QMessageBox>
#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
#  include <QtConcurrent/QtConcurrent>
#  include <QtConcurrent/QtConcurrentRun>
#else
#  include <QtConcurrentRun>
#endif
#if (QT_VERSION >= QT_VERSION_CHECK(5, 2, 0))
#  ifdef Q_OS_WIN
#    include <QWinTaskbarButton>
#    include <QWinTaskbarProgress>
#  endif
#endif
#include <limits>
#include <cassert>


// slot
void MainWindow::loadImagesFromDirectory(QString dir_hint)
{
  QMutexLocker __(&mutex_);

  if (not images_->empty()) {
    QMessageBox::StandardButton reply = QMessageBox::question(
        this, "Opening a directory...",
        "Unsaved data will be lost. Continue?",
        QMessageBox::Yes|QMessageBox::No);
    if (reply == QMessageBox::No) { return; }
  }

  QString const dir_s = (not dir_hint.isEmpty() and not dir_hint.isNull())
      ? dir_hint
      : QFileDialog::getExistingDirectory(this, "Open a directory with images",
                                          images_directory_,
                                          QFileDialog::ShowDirsOnly);
  if (dir_s.isEmpty() or dir_s.isNull()) { return; }

  clearFrameAnnotations_();
  clearAllNuclei_();
  clearAllPerinuclei_();
  clearAllRegions_();
  clearAllTracks_();

  QDir const dir{dir_s};
  images_directory_ = dir.absolutePath();

  images_.reset(new ImageSequence{nullptr, new st::QtReporter});
  assert(images_->empty());
  if (ui_->actionStoreOriginalImagesInMemory->isChecked()) {
    images_->set(st::ImageSequence::StorageStrategy::RAM);
  } else {
    images_->set(st::ImageSequence::StorageStrategy::DISK);
    assert(not ui_->actionPreferenceAfterOpenComputeOverlays->isChecked());
  }
  auto progress_reporter = static_cast<st::QtReporter*>(images_->progress());
  connect(progress_reporter, SIGNAL(busy()),
          this,              SLOT(startProgressImagesLoading()),
          Qt::QueuedConnection);
  connect(progress_reporter, SIGNAL(idle()),
          this,              SLOT(stopProgressImagesLoading()),
           Qt::QueuedConnection);
  connect(progress_reporter, SIGNAL(progressMax(int)),
          ui_->progressBar,  SLOT(setMaximum(int)),
          Qt::QueuedConnection);
  connect(progress_reporter, SIGNAL(progressValue(int)),
          ui_->progressBar,  SLOT(setValue(int)),
          Qt::QueuedConnection);
#if (QT_VERSION >= QT_VERSION_CHECK(5, 2, 0))
# ifdef Q_OS_WIN
  connect(progress_reporter,          SIGNAL(progressMax(int)),
          win_taskbar_progress_.get(),SLOT(setMaximum(int)),Qt::QueuedConnection);
  connect(progress_reporter,           SIGNAL(progressValue(int)),
          win_taskbar_progress_.get(), SLOT(setValue(int)), Qt::QueuedConnection);
# endif
#endif

  ImageSequence::Settings settings;
  settings.setNormalizeToBitDepth(ui_->actionAfterOpenNormalizeToCameraBitDepth
                                     ->isChecked());
  settings.setComputeOverlays(ui_->actionPreferenceAfterOpenComputeOverlays
                                 ->isChecked());
  images_->set(std::move(settings));

  std::vector<ImageSequence::Channel> excluded_channels;
  if (ui_->actionPreferenceWhenLoadingOmitBrightfieldImages->isChecked()) {
    excluded_channels.push_back(ImageSequence::Channel::BF);
  }

  std::vector<ImageSequence::Channel> overlay_excluded_channels;
  if (ui_->actionAfterOpenExcludeBrightfieldFromOverlays->isChecked()) {
    overlay_excluded_channels.push_back(ImageSequence::Channel::BF);
  }

  future_int_ = QtConcurrent::run(
      images_.get(), &ImageSequence::loadDirectory, dir.path().toStdString(),
                                                    excluded_channels,
                                                    overlay_excluded_channels);
}


// slot
void MainWindow::startProgressImagesLoading()
{
  startProgress("Loading image files from " + images_directory_ + "...");
  framesNavigationExpose_(false);

  ui_->menuFile->setEnabled(false);
  ui_->menuView->setEnabled(false);
  ui_->menuTools->setEnabled(false);
  ui_->menuScript->setEnabled(false);
}


// slot
void MainWindow::stopProgressImagesLoading()
{
  int const retcode = stopProgress();

  auto progress_reporter = static_cast<st::QtReporter*>(images_->progress());
  disconnect(progress_reporter, SIGNAL(busy()),
             this,              SLOT(startProgressImagesLoading()));
  disconnect(progress_reporter, SIGNAL(idle()),
             this,              SLOT(stopProgressImagesLoading()));
  disconnect(progress_reporter, SIGNAL(progressMax(int)),
             ui_->progressBar,  SLOT(setMaximum(int)));
  disconnect(progress_reporter, SIGNAL(progressValue(int)),
             ui_->progressBar,  SLOT(setValue(int)));
#if (QT_VERSION >= QT_VERSION_CHECK(5, 2, 0))
# ifdef Q_OS_WIN
  disconnect(progress_reporter,          SIGNAL(progressMax(int)),
             win_taskbar_progress_.get(),SLOT(setMaximum(int)));
  disconnect(progress_reporter,           SIGNAL(progressValue(int)),
             win_taskbar_progress_.get(), SLOT(setValue(int)));
# endif
#endif

  if (retcode == EXIT_SUCCESS) {
    messageDone_();

    assert(images_->timepointsCount() > 0l);
    assert(images_->timepointsCount() <= std::numeric_limits<int>::max());
    n_frames_ = static_cast<int>(images_->timepointsCount()); // n_frames_:int
    current_stage_i_ = 0;
    current_frame_i_ = 0;

    assert(nuclei_.empty());
    assert(perinuclei_.empty());
    assert(regions_.empty());
    assert(tracks_.empty());
    for (int frame_i=0; frame_i<n_frames_; ++frame_i) {
      nuclei_.push_back({});
      perinuclei_.push_back({});
      regions_.push_back({});
    }

    setWindowTitle(QString{APP_NAME} + " - " + images_directory_);

    // Prepare enhanced image for viewing:
    try {

      if (ui_->actionAfterOpenCachePixmaps->isChecked()) {
        message_('I', "Caching images in fast display buffers...");
      } else {
        message_('I', "Preparing view...");
      }
      repaint();  // pathetic enforcement of msg display


      maybeStoreViewSettings_();

      setupViewsAndScenes_(true);
      messageDone_();

      menuViewUpdate_();
      if (ui_->actionAfterOpenExcludeBrightfieldFromOverlays->isChecked()) {
        ui_->actionOverlaysBrightfieldInclusionToggle
           ->setText("Include bright-field");
      }

      maybeRestoreViewSettings_();

      setupMenuTools_();

      framesNavigationExpose_(not images_->empty());
      ui_->horizontalSlider->setMaximum(static_cast<int>(n_frames_ - 1));
      ui_->horizontalSlider->setValue(0);
      ui_->pushButtonFramePrev->setEnabled(false);
      ui_->pushButtonFrameNext->setEnabled(n_frames_ > 1);
      showFrame_(0);

      menuScriptEnable_(true);

      if (post_images_load_hook_script_.length() > 0) {
        script_window_->execute(post_images_load_hook_script_);
      }

    } catch (std::runtime_error& e) {
      QMessageBox msg;
      QString const desc = "Cannot display image files: " + QString(e.what());
      message_('E', desc);
      msg.setText("Whoops!");
      msg.setInformativeText(desc);
      msg.setStandardButtons(QMessageBox::Ok);
      msg.setDefaultButton(QMessageBox::Ok);
      msg.setIcon(QMessageBox::Critical);
      msg.exec();
    }

  } else { // then nothing read in:

    // -- Rollback/clean-up:
    if (images_->empty()) {
      framesNavigationExpose_(false);
    }

  }

  ui_->menuFile->setEnabled(true);
  ui_->menuView->setEnabled(true);
  ui_->menuTools->setEnabled(true);
  ui_->menuScript->setEnabled(true);
}
