// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "MainWindow.hpp"
#include "ui_MainWindow.h"
#include "GUI/Windows/MainWindow_aux.hpp"
#include "GUI/Support/ZoomableGraphicsView.hpp"
#include <cassert>


void MainWindow::menuToolsEnable_(bool on)
{
  ui_->menuTools->setEnabled(on);
  for (QAction* pa: qAsConst(menu_tools_actions_)) { pa->setEnabled(on); }
}


void MainWindow::toolboxToggle_(QToolBox* toolbox, bool on)
{
  // visibility of this toolbox and associated widgets:
  toolbox->setVisible(on);
  for (auto pw: qAsConst(toolboxen_external_widgets_[toolbox])) {  assert(pw);
    pw->setVisible(on);
  }

  if (on) {
    QAction* tool_action = tools_to_actions_[toolbox];

    // un-toggle other toolboxen:
    for (QAction* pa: qAsConst(menu_tools_actions_)) {  assert(pa);
      if ((pa != tool_action) and (pa->isChecked())) { pa->toggle(); }
    }

    current_toolbox_ = toolbox; // note order: untoggline others would 0x0-ify

    // "contour-ability":
    assert(editablity_.contains(toolbox));
    auto const& props = editablity_[toolbox].first;
    assert(props.size() == 5);
    auto const& insertable = props[1];
    auto const& erasable   = props[2];
    auto const& swappable  = props[3];
    auto const& togglable  = props[4];
    //
    for (auto pv: qAsConst(views_)) {  assert(pv);
      pv->canInsert(insertable);
      pv->canErase (erasable);
      pv->canSwap  (swappable);
      pv->canToggle(togglable);
    }

  } else { // (when "off")
    current_toolbox_ = nullptr;
  }
  updateEditabilities_();

  for (auto pv: qAsConst(views_)) {  assert(pv);
    pv->quantifies(current_toolbox_ == ui_->toolBoxTool5); // 5 is Quantification
  }
}


void MainWindow::on_actionImageMasking_toggled(bool on)
{ toolboxToggle_(ui_->toolBoxTool0, on); }

void MainWindow::on_actionNucleiDetection_toggled(bool on)
{ toolboxToggle_(ui_->toolBoxTool1, on); }

void MainWindow::on_actionNucleiEditing_toggled(bool on)
{ toolboxToggle_(ui_->toolBoxTool2, on); }

void MainWindow::on_actionPerinuclei_toggled(bool on)
{
  toolboxToggle_(ui_->toolBoxTool3, on);
  if (on) {
    auto const twain = QList<QWidget*>{ui_->checkBoxPerinucDeriveAutoup,
                                       ui_->pushButtonPerinucCreate};
    for (auto w: twain) {
      w->setEnabled(not current_nuclei_->empty());
  } }
}

void MainWindow::on_actionRegionsEditing_toggled(bool on)
{
  toolboxToggle_(ui_->toolBoxTool4, on);
  if (on) {
    ui_->groupBoxRgnEditVoronoiArea->setEnabled(not current_nuclei_->empty());
//  ui_->pushButtonRgnEditTesselate->setEnabled(not current_nuclei_->empty());
  }
}

void MainWindow::on_actionQuantification_toggled(bool on)
{
  toolboxToggle_(ui_->toolBoxTool5, on);
  if (on) {
    ui_->pushButtonNucQuantSaveAll->setEnabled(not current_nuclei_->empty());
    ui_->pushButtonQuantRgnSaveAll->setEnabled(not current_regions_->empty());
    ui_->checkBoxQuantImgOnlyNonMasked->setEnabled(images_
                                                   ->hasMask(current_frame_i_));
  }
}

void MainWindow::on_actionTracking_toggled(bool on)
{ toolboxToggle_(ui_->toolBoxTool6, on); }

void MainWindow::on_actionTracksEditing_toggled(bool on)
{ toolboxToggle_(ui_->toolBoxTool7, on); }



void MainWindow::on_actionToolSettingsSave_triggered()
{
  if (ui_->actionImageMasking->isChecked()) {
   on_pushButtonImgMaskSave_clicked();
  } else if (ui_->actionNucleiDetection->isChecked()) {
    on_pushButtonNucDetSettingsSave_clicked();
  } else if (ui_->actionNucleiEditing->isChecked()) {
    on_pushButtonNucEditNucleiContoursSave_clicked();
  } else if (ui_->actionPerinuclei->isChecked()) {
    on_pushButtonPerinucSettingsSave_clicked();
  } else if (ui_->actionRegionsEditing->isChecked()) {
    on_pushButtonRgnEditRegionsSave_clicked();
  } else if (ui_->actionTracking->isChecked()) {
    on_pushButtonTrackingSettingsSave_clicked();
  } else if (ui_->actionTracksEditing->isChecked()) {
    on_pushButtonTracksEditSave_clicked();
  } else {
    message_('W', "Cannot save settings or data when there is no active tool");
  }
}

void MainWindow::on_actionToolSettingsLoad_triggered()
{
  if (ui_->actionImageMasking->isChecked()) {
   on_pushButtonImgMaskLoad_clicked();
  } else if (ui_->actionNucleiDetection->isChecked()) {
    on_pushButtonNucDetSettingsLoad_clicked();
  } else if (ui_->actionNucleiEditing->isChecked()) {
    on_pushButtonNucEditNucleiContoursLoad_clicked();
  } else if (ui_->actionPerinuclei->isChecked()) {
    on_pushButtonPerinucSettingsLoad_clicked();
  } else if (ui_->actionRegionsEditing->isChecked()) {
    on_pushButtonRgnEditRegionsLoad_clicked();
  } else if (ui_->actionTracking->isChecked()) {
    on_pushButtonTrackingSettingsLoad_clicked();
  } else if (ui_->actionTracksEditing->isChecked()) {
    on_pushButtonTracksEditLoad_clicked();
  } else {
    message_('W', "Cannot load settings or data when there is no active tool");
  }
}
