// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "MainWindow.hpp"
#include "ui_MainWindow.h"
#if (QT_VERSION >= QT_VERSION_CHECK(5, 2, 0))
# ifdef Q_OS_WIN
#  include <QWinTaskbarButton>
#  include <QWinTaskbarProgress>
# endif
#endif


void MainWindow::framesNavigationExpose_(bool on)
{
  for (auto nw: qAsConst(navigation_widgets_)) {
    nw->setEnabled(on); nw->setVisible(on);
  }
}


// slot
void MainWindow::showFrame_(long int frame_i)
{
  assert(0l <= frame_i);

  // -- set pixmaps
  int panel_i = 0;
  for (auto channel: st::AllChannels) {
    if (images_->has(channel)) {
      pixmapitems_[panel_i]->setPixmap(
          pixmap_cache_->at(static_cast<int>(frame_i), channel));
      panel_i++;
  } }
  pixmapitems_[panel_i]->setPixmap
      (pixmap_cache_->at(static_cast<int>(frame_i), st::ImageChannel::UNDETERMINED));

  current_exclusions_polys_.clear();
  current_exclusions_.clear();
  ui_->pushButtonImgMaskApply->setEnabled(false);
  ui_->pushButtonImgMaskReset->setEnabled(false);

  // -- set outlines' indices
  current_nuclei_    = &nuclei_ [frame_i];
  current_perinuclei_= &perinuclei_ [frame_i];
  current_regions_   = &regions_[frame_i];
  //
  ifNotHiddenThenRedrawFrameOutlines_();
  //
  clearFrameAnnotations_();
  maybeFrameAnnotate_();

  if (   ui_->checkBoxTracksEditTraceClippingLookBack->isChecked()
      or ui_->checkBoxTracksEditTraceClippingLookAhead->isChecked()
      or (    ui_->tableViewTracksEdit->selectionModel()
          and ui_->tableViewTracksEdit->selectionModel()->hasSelection())) {
    ifNotHiddenThenRedrawTracks_();
  }

  // -- set frame number (index is 0-based)
  ui_->labelFrameNumber->setText(QString::number(frame_i));
  ui_->labelFramesNumbersRange->setText("[0.."+
      QString::number(n_frames_ - 1) +"]");

  // -- set frame time
  // Note: despite the recording time can be different for every channel,
  //       only one (earliest) time is printed.
  QString secs = QString::fromStdString(images_->time(frame_i)[0]);
  if (secs != "?") { // "?" is in the "protocol"...
    auto fsecs = static_cast<unsigned int>(secs.replace(",", ".").toFloat());
    QString hrs  = QString::number(fsecs / 3600);
    fsecs -= (fsecs / 3600)*3600;
    QString minsr = QString::number(fsecs / 60);
    QString secsr= QString::number(fsecs % 60);
    if (minsr.length() < 2) { minsr = "0" + minsr; }
    if (secsr.length() < 2) { secsr = "0" + secsr; }
    ui_->labelFrameTime->setText(hrs + ":" + minsr + ":" + secsr);
    ui_->labelFrameTime->setVisible(true);
  } else {
    ui_->labelFrameTime->setText("--:--:--");
    ui_->labelFrameTime->setVisible(false);
  }

  // -- set load/save push buttons enabledness
  // if (not batch_processing_) { // commented out <= may break scripts
    checkIfImageMaskContoursFileAvailable_();
    checkIfNucDetSettingsFileAvailable_();
    checkIfNucleiContoursFileAvailable_();
    checkIfPerinucDerivSettingsFileAvailable_();
    checkIfRegionsContoursFileAvailable_();
    checkIfTrackingSettingsFileAvailable_();
  //}

  // -- set global progress
#if (QT_VERSION >= QT_VERSION_CHECK(5, 2, 0))
# ifdef Q_OS_WIN
  if (batch_processing_) {
    win_taskbar_progress_->setVisible(true);
    win_taskbar_progress_->setRange(0, n_frames_ - 1);
    win_taskbar_progress_->setValue(frame_i);
  } else {
    win_taskbar_progress_->setValue(0);
    win_taskbar_progress_->setVisible(false);
  }
# endif
#endif
}


// slot
void MainWindow::on_horizontalSlider_valueChanged(int frame_i)
{
  assert(frame_i >= 0);
  current_frame_i_ = frame_i;
  showFrame_(current_frame_i_);
  ui_->pushButtonFramePrev->setEnabled(current_frame_i_ > 0);
  ui_->pushButtonFrameNext->setEnabled(current_frame_i_ < n_frames_ - 1);
}


// slot
void MainWindow::on_pushButtonFrameNext_clicked()
{
  current_frame_i_ = (current_frame_i_ + 1 < n_frames_)
                     ? (current_frame_i_ + 1)
                     :  current_frame_i_;
  assert((0 <= current_frame_i_) and (current_frame_i_ <= n_frames_ - 1));
  showFrame_(current_frame_i_);
  ui_->pushButtonFramePrev->setEnabled(current_frame_i_ > 0);
  ui_->pushButtonFrameNext->setEnabled(current_frame_i_ < n_frames_ - 1);
  ui_->horizontalSlider->setValue(static_cast<int>(current_frame_i_));
  ui_->horizontalSlider->setFocus();
}

// slot
void MainWindow::on_pushButtonFramePrev_clicked()
{
  current_frame_i_ = (static_cast<long int>(current_frame_i_) - 1 >= 0)
                     ? (current_frame_i_ - 1)
                     :  current_frame_i_;
  assert((0 <= current_frame_i_) and (current_frame_i_ <= n_frames_ - 1));
  showFrame_(current_frame_i_);
  ui_->pushButtonFramePrev->setEnabled(current_frame_i_ > 0);
  ui_->pushButtonFrameNext->setEnabled(current_frame_i_ < n_frames_ - 1);
  ui_->horizontalSlider->setValue(static_cast<int>(current_frame_i_));
  ui_->horizontalSlider->setFocus();
}
