// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "MainWindow.hpp"
#include "ui_MainWindow.h"

namespace {
struct SettingsStrings {
  using S_t = QString const;

  S_t windowGrp  = "Window";
  S_t mainWinFull= "main_window_fullscreen";
  S_t mainWinX   = "main_window_position_x";
  S_t mainWinY   = "main_window_position_y";
  S_t mainWinW   = "main_window_size_width";
  S_t mainWinH   = "main_window_size_height";
  //
  S_t nucDetImgPreprocParamScanGrp = "NucleiDetectionImagePreprocessingParameterScan";
  //S_t sSmSzLo   = "smoothing_size_low";
  //S_t sSmSzHi   = "smoothing_size_high";
  //S_t sSmSzSt   = "smoothing_size_step";
  S_t thBlckSzLo = "thresholding_block_size_low";
  S_t thBlckSzHi = "thresholding_block_size_high";
  S_t thBlckSzSt = "thresholding_block_size_steps";
  S_t thBslnLo   = "thresholding_baseline_low";
  S_t thBslnHi   = "thresholding_baseline_high";
  S_t thBslnSt   = "thresholding_baseline_step";
  //S_t opSzLo   = "opening_low";
  //S_t opSzHi   = "opening_high";
  //S_t opSzSt   = "opening_step";

  S_t nucDetPreselScoringGrp = "NucleiDetectionPreselectionScoringCriterion";
  S_t areaCovWeight     = "area_cov";
  S_t solidityCovWeight = "solidity";
  //
  S_t quantificationGrp = "Quantification";
  S_t quantImageExportDAT     = "image_quantification_export_dat";
  S_t quantRegionsExportPNG   = "regions_quantification_export_png";
  S_t quantRegionsExportDAT   = "regions_quantification_export_dat";
  S_t quantNucleiExportPNG    = "nuclei_quantification_export_png";
  S_t quantNucleiExportDAT    = "nuclei_quantification_export_dat";
  S_t quantPerinucleiExportDAT= "perinuclei_quantification_export_dat";

  S_t drawingGrp = "Drawing";
  S_t nucleusContourWidth     = "nucleus_contour_width";
  S_t perinucleusContourWidth = "perinucleus_contour_width";
  S_t nucleusNumberFontScaling= "nucleus_number_font_scaling";

  S_t workspaceGrp  = "Workspace";
  S_t lastImagesDir = "last_images_directory";

  S_t preferencesGrp  = "Preferences";
  S_t afterOpenNormalizeToCameraBitDepth = "normalize_to_camera_bit_depth_if_available";
  S_t storeOriginalImagesInMemory = "store_original_images_in_memory";
  S_t afterOpenCachePixmapsForFastDisplay = "after_opening_cache_pixmaps_to_speed_up_display";
  S_t afterOpeningGenerateAndStoreOverlays = "when_opening_generate_and_store_overlays";
  S_t afterOpeningOmitBF = "after_opening_hide_brightfield_images_panel";
  S_t afterOpeningHideBF = "after_opening_exclude_brightfield_images_from_overlays";
  S_t afterOpenExcludeBFFromOverlays = "when_opening_omit_brightfield_images";
  //S_t prefParamScanPrecomputeShared = "parameter_scan_precompute_shared_stages";
  S_t retainViewForNextOpening = "retain_channel_panes_and_overlay_composition_for_a_next_opened_folder";
  S_t checkLatestVersionUponStartup = "check_lastest_version_upon_startup";

  S_t highlitedNucleiFilePath = "path_to_file_with_nuclei_to_be_highlighted";
  S_t defaultHighlitedNucleiFilePath = "";
};


struct SettingsDefaultValues {
  using V_t = double const;
  V_t areaCovWeight     = 0.;
  V_t solidityCovWeight = 1.;

  V_t nucleusContourWidth      = 1.;
  V_t perinucleusContourWidth  = 1.;
  V_t nucleusNumberFontScaling = 1.;
};
} // namespace


void MainWindow::loadSettings_(bool resize_window)
{
  SettingsStrings const s;

#ifdef Q_OS_WIN
  QSettings setts;
#else
  if (not QFileInfo::exists(settings_file_path_)) { return; }
  QSettings setts(settings_file_path_, QSettings::IniFormat);
#endif

  setts.beginGroup(s.workspaceGrp);
    if (setts.contains(s.lastImagesDir)) {
      images_directory_ = setts.value(s.lastImagesDir,"").toString();
    }
  setts.endGroup();

  setts.beginGroup(s.preferencesGrp);
    if (setts.contains(s.afterOpenNormalizeToCameraBitDepth)) {
      ui_->actionAfterOpenNormalizeToCameraBitDepth->setChecked(
          setts.value(s.afterOpenNormalizeToCameraBitDepth,"").toBool());
    }
    if (setts.contains(s.storeOriginalImagesInMemory)) {
      ui_->actionStoreOriginalImagesInMemory->setChecked(
          setts.value(s.storeOriginalImagesInMemory,"").toBool());
    }
    if (setts.contains(s.afterOpenCachePixmapsForFastDisplay)) {
      ui_->actionAfterOpenCachePixmaps->setChecked(
          setts.value(s.afterOpenCachePixmapsForFastDisplay,"").toBool());
    }
    if (setts.contains(s.afterOpeningGenerateAndStoreOverlays)) {
      ui_->actionPreferenceAfterOpenComputeOverlays->setChecked(
          setts.value(s.afterOpeningGenerateAndStoreOverlays,"").toBool());
    }
    if (setts.contains(s.afterOpeningOmitBF)) {
      ui_->actionPreferenceWhenLoadingOmitBrightfieldImages->setChecked(
          setts.value(s.afterOpeningOmitBF,"").toBool());
    }
    if (setts.contains(s.afterOpeningOmitBF)) {
      ui_->actionAfterOpenHideBrightfield->setChecked(
          setts.value(s.afterOpeningOmitBF,"").toBool());
    }
    if (setts.contains(s.afterOpenExcludeBFFromOverlays)) {
      ui_->actionAfterOpenExcludeBrightfieldFromOverlays->setChecked(
          setts.value(s.afterOpenExcludeBFFromOverlays,"").toBool());
    }
    if (ui_->actionPreferenceWhenLoadingOmitBrightfieldImages->isChecked()) {
      ui_->actionAfterOpenHideBrightfield->setEnabled(false);
      ui_->actionAfterOpenExcludeBrightfieldFromOverlays->setEnabled(false);
    }
    if (setts.contains(s.retainViewForNextOpening)) {
      ui_->actionPreferenceRetainViewForNextOpening->setChecked(
          setts.value(s.retainViewForNextOpening,"").toBool());
    }
    if (setts.contains(s.checkLatestVersionUponStartup)) {
      ui_->actionPreferenceCheckLatestVersionUponStartup->setChecked(
          setts.value(s.checkLatestVersionUponStartup,"").toBool());
    }
  //ui_->actionPreferenceParamScanPrecomputeShared->setChecked(
  //    setts.value(sPrefParamScanPrecomputeShared,"").toBool());
  setts.endGroup();

  setts.beginGroup(s.windowGrp);
    if (resize_window) {
      if (setts.value(s.mainWinFull,"").toInt() == 1) {
        showFullScreen();
        ui_->actionFullScreenToggle->setText(MainWindow::sWinExitFull);
      } else {
        showNormal();
        if (setts.contains(s.mainWinX) and setts.contains(s.mainWinY) and
            setts.contains(s.mainWinW) and setts.contains(s.mainWinH)    ) {
          setGeometry(
              setts.value(s.mainWinX,"").toInt(), setts.value(s.mainWinY,"").toInt(),
              setts.value(s.mainWinW,"").toInt(), setts.value(s.mainWinH,"").toInt());
        }
        ui_->actionFullScreenToggle->setText(MainWindow::sWinGoFull);
    } }
  setts.endGroup();

  setts.beginGroup(s.drawingGrp);
    if (setts.contains(s.nucleusContourWidth)) {
      nucleus_contour_width_ = setts.value(s.nucleusContourWidth).toDouble();
      ui_->dspinBoxNucEditContourWidth->setValue(nucleus_contour_width_);
    }
    if (setts.contains(s.perinucleusContourWidth)) {
      perinucleus_contour_width_=setts.value(s.perinucleusContourWidth).toDouble();
      ui_->dspinBoxPerinucContourWidth->setValue(perinucleus_contour_width_);
    }
    if (setts.contains(s.nucleusNumberFontScaling)) {
      outline_id_font_scaling_=setts.value(s.nucleusNumberFontScaling).toDouble();
    }
    if (setts.contains(s.highlitedNucleiFilePath)) {
      highlited_nuclei_file_path_ = setts.value(s.highlitedNucleiFilePath).toString();
    }
  setts.endGroup();

  setts.beginGroup(s.quantificationGrp);
    if (setts.contains(s.quantImageExportDAT)) {
      ui_->checkBoxQuantImgFilesDat ->setChecked(
          setts.value(s.quantImageExportDAT).toBool());
    }
    if (setts.contains(s.quantRegionsExportDAT)) {
      ui_->checkBoxQuantRgnSaveDat ->setChecked(
          setts.value(s.quantRegionsExportDAT).toBool());
    }
    if (setts.contains(s.quantRegionsExportPNG)) {
      ui_->checkBoxQuantRgnSavePng->setChecked(
          setts.value(s.quantRegionsExportPNG).toBool());
    }
    if (setts.contains(s.quantNucleiExportDAT)) {
      ui_->checkBoxNucQuantSaveCsv->setChecked(
          setts.value(s.quantNucleiExportDAT).toBool());
    }
    if (setts.contains(s.quantNucleiExportPNG)) {
      ui_->checkBoxNucQuantSavePng->setChecked(
          setts.value(s.quantNucleiExportPNG).toBool());
    }
    if (setts.contains(s.quantPerinucleiExportDAT)) {
      ui_->checkBoxQuantPerinucSaveCsv->setChecked(
          setts.value(s.quantPerinucleiExportDAT).toBool());
    }
  setts.endGroup();


  QMap<QString,int> image_preproc_param_scan_ranges;
  //
  setts.beginGroup(s.nucDetImgPreprocParamScanGrp);
  for (auto& qs: {/*s.sSmSzLo,    s.sSmSzHi,    s.sSmSzSt,*/
                    s.thBlckSzLo, s.thBlckSzHi, s.thBlckSzSt,
                    s.thBslnLo,   s.thBslnHi,   s.thBslnSt //,
                  /*s.sOpSzLo,    s.sOpSzHi,    s.sOpSzSt*/}) {
    if (setts.contains(qs)) {
      image_preproc_param_scan_ranges[qs] =  setts.value(qs,"").toInt();
    } else {
      // This message scares worries users on first launch, commented out.
      //message_to_user_('W',"Cannot find '"+qs+"' in settings!");
  } }
  setts.endGroup();

  if (image_preproc_param_scan_ranges.contains(s.thBlckSzLo)) {
    ui_->spinBoxNucDetThresholdingBlocksizeAutoscanFrom->setValue(
        image_preproc_param_scan_ranges[s.thBlckSzLo]);
  }
  if (image_preproc_param_scan_ranges.contains(s.thBlckSzHi)) {
    ui_->spinBoxNucDetThresholdingBlocksizeAutoscanUpto->setValue(
        image_preproc_param_scan_ranges[s.thBlckSzHi]);
  }
  if (image_preproc_param_scan_ranges.contains(s.thBlckSzSt)) {
    ui_->spinBoxNucDetThresholdingBlocksizeAutoscanNsteps->setValue(
        image_preproc_param_scan_ranges[s.thBlckSzSt]);
  }

  if (image_preproc_param_scan_ranges.contains(s.thBslnLo)) {
    ui_->spinBoxNucDetThresholdingBaselineAutoscanFrom->setValue(
        image_preproc_param_scan_ranges[s.thBslnLo]);
  }
  if (image_preproc_param_scan_ranges.contains(s.thBslnHi)) {
    ui_->spinBoxNucDetThresholdingBaselineAutoscanUpto->setValue(
        image_preproc_param_scan_ranges[s.thBslnHi]);
  }
  if (image_preproc_param_scan_ranges.contains(s.thBslnSt)) {
    ui_->spinBoxNucDetThresholdingBaselineAutoscanStep->setValue(
        image_preproc_param_scan_ranges[s.thBslnSt]);
  }

  setts.beginGroup(s.nucDetPreselScoringGrp);
    nuc_presel_assess_params_ = {
        setts.value(s.areaCovWeight,    "").toDouble(),
        setts.value(s.solidityCovWeight,"").toDouble() };
  setts.endGroup();
}



// If a user did not change the default (which would be read in), write defaults.
//
void MainWindow::saveSettings_()
{
#ifdef Q_OS_WIN
  QSettings setts;
#else
  QSettings setts(settings_file_path_, QSettings::IniFormat);
#endif

  SettingsStrings const s;
  SettingsDefaultValues const k;

  setts.beginGroup(s.workspaceGrp);
    setts.setValue(s.lastImagesDir, images_directory_);
  setts.endGroup();

  setts.beginGroup(s.preferencesGrp);
    setts.setValue(s.afterOpenNormalizeToCameraBitDepth,
        ui_->actionAfterOpenNormalizeToCameraBitDepth->isChecked());
    setts.setValue(s.storeOriginalImagesInMemory,
        ui_->actionStoreOriginalImagesInMemory->isChecked());
    setts.setValue(s.afterOpenCachePixmapsForFastDisplay,
        ui_->actionAfterOpenCachePixmaps->isChecked());
    setts.setValue(s.afterOpeningGenerateAndStoreOverlays,
        ui_->actionPreferenceAfterOpenComputeOverlays->isChecked());
    setts.setValue(s.afterOpeningHideBF,
        ui_->actionAfterOpenHideBrightfield->isChecked());
    setts.setValue(s.afterOpenExcludeBFFromOverlays,
        ui_->actionAfterOpenExcludeBrightfieldFromOverlays->isChecked());
    setts.setValue(s.afterOpeningOmitBF,
        ui_->actionPreferenceWhenLoadingOmitBrightfieldImages->isChecked());
    setts.setValue(s.retainViewForNextOpening,
        ui_->actionPreferenceRetainViewForNextOpening->isChecked());
    setts.setValue(s.checkLatestVersionUponStartup,
        ui_->actionPreferenceCheckLatestVersionUponStartup->isChecked());
  //setts.setValue(sPrefParamScanPrecomputeShared,
  //    ui_->actionPreferenceParamScanPrecomputeShared->isChecked());
  setts.endGroup();

  setts.beginGroup(s.windowGrp);
    setts.setValue(s.mainWinFull, QString::number(isFullScreen()));
    if (not isFullScreen()) {
      setts.setValue(s.mainWinX, QString::number(geometry().x()));
      setts.setValue(s.mainWinY, QString::number(geometry().y()));
      setts.setValue(s.mainWinW, QString::number(geometry().width()));
      setts.setValue(s.mainWinH, QString::number(geometry().height()));
    }
  setts.endGroup();

  setts.beginGroup(s.drawingGrp);
    if (not setts.contains(s.nucleusContourWidth)) {
      setts.setValue(s.nucleusContourWidth, k.nucleusContourWidth);
    } else {
      setts.setValue(s.nucleusContourWidth, nucleus_contour_width_);
    }
    if (not setts.contains(s.perinucleusContourWidth)) {
      setts.setValue(s.perinucleusContourWidth, k.perinucleusContourWidth);
    } else {
      setts.setValue(s.perinucleusContourWidth, perinucleus_contour_width_);
    }
    if (not setts.contains(s.nucleusNumberFontScaling)) {
      setts.setValue(s.nucleusNumberFontScaling, k.nucleusNumberFontScaling);
    } else {
      setts.setValue(s.nucleusNumberFontScaling, outline_id_font_scaling_);
    }
    if (not setts.contains(s.highlitedNucleiFilePath)) {
      setts.setValue(s.highlitedNucleiFilePath, s.defaultHighlitedNucleiFilePath);
    } else {
      setts.setValue(s.highlitedNucleiFilePath, highlited_nuclei_file_path_);
    }
  setts.endGroup();

  setts.beginGroup(s.quantificationGrp);
    if (not setts.contains(s.quantImageExportDAT)) {
      setts.setValue(s.quantImageExportDAT,
                     ui_->checkBoxQuantImgFilesDat->isChecked());
    }
    if (not setts.contains(s.quantRegionsExportDAT)) {
      setts.setValue(s.quantRegionsExportDAT,
                     ui_->checkBoxQuantRgnSaveDat->isChecked());
    }
    if (not setts.contains(s.quantRegionsExportPNG)) {
      setts.setValue(s.quantRegionsExportPNG,
                     ui_->checkBoxQuantRgnSavePng->isChecked());
    }
    if (not setts.contains(s.quantNucleiExportDAT)) {
      setts.setValue(s.quantNucleiExportDAT,
                     ui_->checkBoxNucQuantSaveCsv->isChecked());
    }
    if (not setts.contains(s.quantNucleiExportPNG)) {
      setts.setValue(s.quantNucleiExportPNG,
                     ui_->checkBoxNucQuantSavePng->isChecked());
    }
    if (not setts.contains(s.quantPerinucleiExportDAT)) {
      setts.setValue(s.quantPerinucleiExportDAT,
                     ui_->checkBoxQuantPerinucSaveCsv->isChecked());
    }
  setts.endGroup();


  setts.beginGroup(s.nucDetImgPreprocParamScanGrp);
    if (true /*!setts.contains(sThBlckSzLo)*/) { setts.setValue(s.thBlckSzLo,
        QString::number(ui_->spinBoxNucDetThresholdingBlocksizeAutoscanFrom->value())); }
    if (true /*!setts.contains(sThBlckSzHi)*/) { setts.setValue(s.thBlckSzHi,
        QString::number(ui_->spinBoxNucDetThresholdingBlocksizeAutoscanUpto->value())); }
    if (true /*!setts.contains(sThBlckSzSt)*/) { setts.setValue(s.thBlckSzSt,
        QString::number(ui_->spinBoxNucDetThresholdingBlocksizeAutoscanNsteps->value()));}
    if (true /*!setts.contains(sThBslnLo  )*/) { setts.setValue(s.thBslnLo  ,
        QString::number(ui_->spinBoxNucDetThresholdingBaselineAutoscanFrom->value())); }
    if (true /*!setts.contains(sThBslnHi  )*/) { setts.setValue(s.thBslnHi  ,
        QString::number(ui_->spinBoxNucDetThresholdingBaselineAutoscanUpto->value())); }
    if (true /*!setts.contains(sThBslnSt  )*/) { setts.setValue(s.thBslnSt  ,
        QString::number(ui_->spinBoxNucDetThresholdingBaselineAutoscanStep->value())); }
  setts.endGroup();

  setts.beginGroup(s.nucDetPreselScoringGrp);
   if (!setts.contains(s.areaCovWeight)) {
     setts.setValue(s.areaCovWeight, k.areaCovWeight);
   }
   if (!setts.contains(s.solidityCovWeight)) {
     setts.setValue(s.solidityCovWeight, k.solidityCovWeight);
   }
  setts.endGroup();

  setts.sync();
}
