// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "MainWindow.hpp"
#include "ui_MainWindow.h"
#include "GUI/Support/ZoomableGraphicsView.hpp"
#include <QNetworkRequest>
#include <QNetworkReply>
#include <QMessageBox>
//#include <QTimer>


void MainWindow::checkLatestVersion_()
{
  net_manager_.reset(new QNetworkAccessManager(this));
  connect(net_manager_.get(), SIGNAL(finished(QNetworkReply*)),
          this, SLOT(latestVersionReply_(QNetworkReply*)),
          Qt::QueuedConnection);
  net_manager_->get(QNetworkRequest(QUrl(sAppVersionLatestUrl)));
}


void MainWindow::latestVersionReply_(QNetworkReply* reply)
{
  disconnect(net_manager_.get(), SIGNAL(finished(QNetworkReply*)),
             this, SLOT(latestVersionReply_(QNetworkReply*)));
  if (reply->canReadLine()) {
    auto const reply_all = reply->readAll();
    QString const latest_version = QString(reply_all.data()).replace('\n',"");
  //qDebug() << latest_version;
    if (latest_version.contains("404")) { // NONGENERIC
      return;
    } else if (aux::isVersionMoreRecent(latest_version,QString(APP_VERSION))) {
      if (QSystemTrayIcon::isSystemTrayAvailable()) {
        tray_.reset(new QSystemTrayIcon(this));
        tray_menu_.reset(new QMenu("ShuttleTracker update", this));
        tray_show_home_webpage_.reset(new QAction{"Visit ShuttleTracker homepage",
                                                  tray_menu_.get()});
        connect(tray_show_home_webpage_.get(), SIGNAL(triggered()),
                this, SLOT(openHomepageUrl()));
        tray_latest_version_info_.reset(new QAction{QString("Latest version is ")
                                                    + latest_version + "."});
        tray_latest_version_info_->setEnabled(false);
        tray_menu_->addActions(QList<QAction*>()
                               << tray_latest_version_info_.get()
                               << tray_show_home_webpage_.get());
        tray_->setContextMenu(tray_menu_.get());
        tray_->setIcon(QIcon(":/icons/shuttletracker_tray_icon.png"));
        tray_->show(); // may look crappy on linux (https://bugreports.qt.io/browse/QTBUG-17823)
        tray_->showMessage("New version of ShuttleTracker is available",
                           QString("You can upgrade from ")
                               + QString(APP_VERSION) +" to "+ latest_version +".",
                           QSystemTrayIcon::Information, 8000 );
        // -- do not hide, menu can be useful
        // QTimer::singleShot(8000, tray_.get(), SLOT(hide()));
      } else {
        QMessageBox m;
        m.setWindowModality(Qt::NonModal);
        m.setWindowTitle("New version");
        m.setText("You can upgrade ShuttleTracker:");
        m.setInformativeText("    "+ QString(APP_VERSION) + "  -->  "
                             + latest_version +".");
        m.exec();
  } } }
  reply->deleteLater();
}


void MainWindow::setupKeyboardShortcuts_() { }


// slot
void MainWindow::setupViewsAndScenes_(bool clear)
{
  {
    QLayoutItem* w;
    while ((w = ui_->horizontalMainLayout->takeAt(0)) != nullptr) {
      ui_->horizontalMainLayout->removeItem(w);
      delete w;
    }
  } // scope

  // -- Clean-up:
  //
  if (clear) {
    for (auto& vp: views_)  { delete vp; }  views_.clear();
    for (auto& sp: scenes_) { sp->clear(); delete sp; }  scenes_.clear();
    if (pixmap_cache_) { pixmap_cache_->clear(); }
    pixmapitems_.clear();
  }

  assert(views_.empty());
  assert(scenes_.empty());
  assert(pixmapitems_.empty());
  if (pixmap_cache_) { assert(pixmap_cache_->empty()); }

  // -- Construction:
  //
  {
    bool const pixmap_caching_on = ui_->actionAfterOpenCachePixmaps->isChecked();
    pixmap_cache_.reset(
        new PixmapCache(*images_.get(), display_filter_, pixmap_caching_on));
  }

  int const n_panels = static_cast<int>(images_->channelsCount() + 1); // +1 for overlays
  for (int i=0; i<n_panels; i++) {
    scenes_.push_back(new QGraphicsScene());
    views_.push_back(new ZoomableGraphicsView(scenes_.back(),this));
    views_.back()->setDragMode(ZoomableGraphicsView::ScrollHandDrag);
    views_.back()->setMouseTracking(true);
    views_.back()->setInteractive(true);
//  views_.back()->setBackgroundBrush(Qt::darkGray);
    views_.back()->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    views_.back()->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    views_.back()->setResizeAnchor(ZoomableGraphicsView::AnchorUnderMouse);
    ui_->horizontalMainLayout->addWidget(views_.back(), 1000);

    connect(views_.back(),SIGNAL(itemAdded(QVector<QPoint>)),
            this,         SLOT  (itemAdded(QVector<QPoint>)) );

    connect(views_.back(),SIGNAL(itemRemoved(const QGraphicsItem*)),
            this,         SLOT  (itemRemoved(const QGraphicsItem*)) );

    connect(views_.back(),SIGNAL(itemReplaced(const QGraphicsItem*,QVector<QPoint>)),
            this,         SLOT  (itemReplaced(const QGraphicsItem*,QVector<QPoint>)) );

    connect(views_.back(),SIGNAL(itemSplit(const QGraphicsItem*,QVector<QPoint>)),
            this,         SLOT  (itemSplit(const QGraphicsItem*,QVector<QPoint>)) );

    connect(views_.back(),SIGNAL(itemToggled(const QGraphicsItem*)),
            this,         SLOT  (itemToggled(const QGraphicsItem*)) );

    connect(views_.back(),SIGNAL(lineSelected(const QGraphicsLineItem*)),
            this,         SLOT  (lineSelected(const QGraphicsLineItem*)) );

    connect(views_.back(),SIGNAL(rectSelected(const QGraphicsRectItem*)),
            this,         SLOT  (rectSelected(const QGraphicsRectItem*)) );

    connect(views_.back(),SIGNAL(lineLineSelected(const QGraphicsLineItem*,const QGraphicsLineItem*)),
            this,         SLOT  (lineLineSelected(const QGraphicsLineItem*,const QGraphicsLineItem*)));

    connect(views_.back(),SIGNAL(lineRectSelected(const QGraphicsLineItem*,const QGraphicsRectItem*)),
            this,         SLOT  (lineRectSelected(const QGraphicsLineItem*,const QGraphicsRectItem*)));

    connect(views_.back(),SIGNAL(rectLineSelected(const QGraphicsRectItem*,const QGraphicsLineItem*)),
            this,         SLOT  (rectLineSelected(const QGraphicsRectItem*,const QGraphicsLineItem*)));

    connect(views_.back(),SIGNAL(rectRectSelected(const QGraphicsRectItem*,const QGraphicsRectItem*)),
            this,         SLOT  (rectRectSelected(const QGraphicsRectItem*,const QGraphicsRectItem*)));

    connect(views_.back(),SIGNAL(itemExtended(const QGraphicsPolygonItem*)),
            this,         SLOT  (itemExtended(const QGraphicsPolygonItem*)));

    connect(views_.back(),SIGNAL(polygonToQuantSelected(const QGraphicsPolygonItem*)),
            this,         SLOT  (showQuantifications(const QGraphicsPolygonItem*)));

    connect(views_.back(),SIGNAL(polygonToExternalHighlight(const QGraphicsPolygonItem*)),
            this,         SLOT  (toggleExternalHighlight(const QGraphicsPolygonItem*)));
  }

  for (auto pview: qAsConst(views_)) { pview->setVisible(true); }

  for (int i=0; i<n_panels; i++) {
    for (int j=0; j<n_panels; j++) {
      if (i != j) {
        views_[i]->addCoScrolling(views_[j]);
  } } }

  assert(scenes_.size()           == n_panels);
  assert(pixmap_cache_->rowSize() == n_panels);

  int scene_i = 0;
  for (auto ch: images_->channels()){
    pixmapitems_.push_back(
        scenes_[scene_i++]->addPixmap(pixmap_cache_->at(/*frame_i*/0, ch)));
  }
  pixmapitems_.push_back(
      scenes_[scene_i]->addPixmap(
          pixmap_cache_->at(/*frame_i*/0, ImageChannel::UNDETERMINED)));

  channel_toggle_.clear();
  for (st::ImageChannel channel: st::AllChannels) {
    if (images_->has(channel)) {
      QAction* paction =
            (channel == ImageChannel::G) ? ui_->actionGreenChannelPanelToggle
          : (channel == ImageChannel::R) ? ui_->actionRedChannelPanelToggle
          : (channel == ImageChannel::B) ? ui_->actionBlueChannelPanelToggle
          : (channel == ImageChannel::C) ? ui_->actionCyanChannelPanelToggle
          : (channel == ImageChannel::M) ? ui_->actionMagentaChannelPanelToggle
          : (channel == ImageChannel::Y) ? ui_->actionYellowChannelPanelToggle
          : (channel == ImageChannel::BF)? ui_->actionBrightfieldPanelToggle
          : ui_->actionOverlaysPanelToggle/*<- for coverage, but should never appear*/;
      channel_toggle_.insert(channel, paction);
  } }
  assert(not channel_toggle_.contains(ImageChannel::UNDETERMINED));
  channel_toggle_.insert(ImageChannel::UNDETERMINED,
                         ui_->actionOverlaysPanelToggle);

} // setup_views_and_scenes_()


void MainWindow::setupMenuToolsActions_()
{
  menu_tools_actions_.push_back(ui_->actionImageMasking);
  menu_tools_actions_.push_back(ui_->actionNucleiDetection);
  menu_tools_actions_.push_back(ui_->actionNucleiEditing);
  menu_tools_actions_.push_back(ui_->actionPerinuclei);
  menu_tools_actions_.push_back(ui_->actionRegionsEditing);
  menu_tools_actions_.push_back(ui_->actionQuantification);
  menu_tools_actions_.push_back(ui_->actionTracking);
  menu_tools_actions_.push_back(ui_->actionTracksEditing);
}


void MainWindow::setupMenuViewShowActions_()
{
  menu_view_show_actions_.push_back(ui_->actionShowOriginal);
  menu_view_show_actions_.push_back(ui_->actionShowOriginalDesaturated);
  menu_view_show_actions_.push_back(ui_->actionShowNormalized);
  menu_view_show_actions_.push_back(ui_->actionShowNormalizedDesaturated);
  menu_view_show_actions_.push_back(ui_->actionShowAutoleveledWeak);
  menu_view_show_actions_.push_back(ui_->actionShowAutoleveledWeakDesaturated);
  menu_view_show_actions_.push_back(ui_->actionShowAutoleveledMedium);
  menu_view_show_actions_.push_back(ui_->actionShowAutoleveledMediumGrayscale);
  menu_view_show_actions_.push_back(ui_->actionShowAutoleveledStrong);
  menu_view_show_actions_.push_back(ui_->actionShowAutoleveledStrongDesaturated);
}


void MainWindow::setupNavigationWidgets_()
{
  navigation_widgets_.push_back(ui_->horizontalSlider);
  navigation_widgets_.push_back(ui_->pushButtonFramePrev);
  navigation_widgets_.push_back(ui_->pushButtonFrameNext);
  navigation_widgets_.push_back(ui_->labelFrameNumber);
  navigation_widgets_.push_back(ui_->labelFrameTime);
  navigation_widgets_.push_back(ui_->labelFramesNumbersRange);

  ui_->horizontalSlider->setVisible(false);
  ui_->pushButtonFramePrev->setVisible(false);
  ui_->pushButtonFrameNext->setVisible(false);
  ui_->labelFrameTime->setText("");
  ui_->labelFrameNumber->setText("");
  ui_->labelFramesNumbersRange->setText("");
}


void MainWindow::setupMainToolbar_()
{
  // reverse order:
  ui_->mainToolBar->setVisible(true); // VISUAL DESIGN
  ui_->mainToolBar->insertAction(nullptr,ui_->actionTracksEditing);
  ui_->mainToolBar->insertAction(nullptr,ui_->actionTracking);
  ui_->mainToolBar->insertAction(nullptr,ui_->actionQuantification);
  ui_->mainToolBar->insertAction(nullptr,ui_->actionRegionsEditing);
  ui_->mainToolBar->insertAction(nullptr,ui_->actionPerinuclei);
  ui_->mainToolBar->insertAction(nullptr,ui_->actionNucleiEditing);
  ui_->mainToolBar->insertAction(nullptr,ui_->actionNucleiDetection);
  ui_->mainToolBar->insertAction(nullptr,ui_->actionImageMasking);
}


void MainWindow::setupTools_()
{
  tools_to_actions_.insert(ui_->toolBoxTool0,ui_->actionImageMasking);
  tools_to_actions_.insert(ui_->toolBoxTool1,ui_->actionNucleiDetection);
  tools_to_actions_.insert(ui_->toolBoxTool2,ui_->actionNucleiEditing);
  tools_to_actions_.insert(ui_->toolBoxTool3,ui_->actionPerinuclei);
  tools_to_actions_.insert(ui_->toolBoxTool4,ui_->actionRegionsEditing);
  tools_to_actions_.insert(ui_->toolBoxTool5,ui_->actionQuantification);
  tools_to_actions_.insert(ui_->toolBoxTool6,ui_->actionTracking);
  tools_to_actions_.insert(ui_->toolBoxTool7,ui_->actionTracksEditing);

  toolboxen_external_widgets_.insert(
      ui_->toolBoxTool0, { ui_->labelImgMaskLoadSave,
                           ui_->pushButtonImgMaskSave,
                           ui_->pushButtonImgMaskLoad });
  toolboxen_external_widgets_.insert(
      ui_->toolBoxTool1, { ui_->labelNucDetSettingsLoadSave,
                           ui_->pushButtonNucDetSettingsLoad,
                           ui_->pushButtonNucDetSettingsSave });
  toolboxen_external_widgets_.insert(
      ui_->toolBoxTool2, { ui_->labelNucEditNucleiContoursLoadSave,
                           ui_->pushButtonNucEditNucleiContoursLoad,
                           ui_->pushButtonNucEditNucleiContoursSave });
  toolboxen_external_widgets_.insert(
      ui_->toolBoxTool3, { ui_->labelPerinucSettingsLoadSave,
                           ui_->pushButtonPerinucSettingsLoad,
                           ui_->pushButtonPerinucSettingsSave } );
  toolboxen_external_widgets_.insert(
      ui_->toolBoxTool4, { ui_->labelRgnEditRegionsLoadSave,
                           ui_->pushButtonRgnEditRegionsLoad,
                           ui_->pushButtonRgnEditRegionsSave });
  toolboxen_external_widgets_.insert(
      ui_->toolBoxTool5, { ui_->pushButtonQuantSaveAll });
  toolboxen_external_widgets_.insert(
      ui_->toolBoxTool6, { ui_->labelTrackingSettingsLoadSave,
                           ui_->pushButtonTrackingSettingsLoad,
                           ui_->pushButtonTrackingSettingsSave } );
  toolboxen_external_widgets_.insert(
      ui_->toolBoxTool7, { ui_->labelTracksEditLoadSave,
                           ui_->pushButtonTracksEditLoad,
                           ui_->pushButtonTracksEditSave });

  editablity_.insert( ui_->toolBoxTool0, // @ actionImageMasking
      { {{},
         {QGraphicsPolygonItem::Type},
         {QGraphicsPolygonItem::Type},
         {},
         {}}, {false,false} } );
  editablity_.insert( ui_->toolBoxTool1, // @ actionNucleiDetection
      { {{}, {}, {}, {}, {}}, {false,false} } );
  editablity_.insert( ui_->toolBoxTool2, // @ actionNucleiEditing
      { {{QGraphicsPolygonItem::Type},
         {QGraphicsPolygonItem::Type},
         {QGraphicsPolygonItem::Type}, // TODO: if has tracks, should be empty
         {},
         {QGraphicsPolygonItem::Type}}, {true,false} } );
  editablity_.insert( ui_->toolBoxTool3, // @ actionPerinuclearAnnuli
      { {{}, {}, {}, {}, {}}, {false,false} } );
  editablity_.insert( ui_->toolBoxTool4, // @ actionRegionsEditing
      { {{QGraphicsPolygonItem::Type},
         {QGraphicsPolygonItem::Type},
         {QGraphicsPolygonItem::Type},
         {}, {}}, {false,true} } );
  editablity_.insert( ui_->toolBoxTool5, // @ actionQuantification
      // TODO: In quantification contours could be selectable --> to enable
      //       interactive quantification of single contours.
      { {{QGraphicsPolygonItem::Type},   //
         {}, {}, {}, {}}, {true,true}} );
  editablity_.insert( ui_->toolBoxTool6, // @ actionTracking
      { {{}, {}, {}, {}, {}}, {false,true} } );
  editablity_.insert( ui_->toolBoxTool7, // @ actionTracksEditing
      { {{QGraphicsPolygonItem::Type,
          QGraphicsLineItem::Type,
          QGraphicsRectItem::Type},
         {}, // segments which merge tracks are added not by drawing with mouse
         {QGraphicsLineItem::Type,   //
          QGraphicsRectItem::Type}, //  removal is done with mouse
         {QGraphicsPolygonItem::Type},
         {QGraphicsPolygonItem::Type} }, {true,false} } );

  // initial visibilities ------------------

  // hide all toolboxen:
  auto const& ext_widgets = toolboxen_external_widgets_.keys();
  for (auto pt: ext_widgets) {
    pt->setVisible(false);
    for (auto pw: qAsConst(toolboxen_external_widgets_[pt])) {
      pw->setVisible(false);
  } }

  ui_->groupBoxNucDetBlobDet->setVisible(
      ui_->radioButtonNucDetChannelSelectionTypeNuclear->isChecked());
  ui_->groupBoxNucDetEdgeDet->setVisible(
      ui_->radioButtonNucDetChannelSelectionTypeCytoplasmic->isChecked());

  ui_->labelNucDetThresholdingBaseline->setEnabled(
      ui_->radioButtonNucDetParamsManual->isChecked());
  ui_->labelNucDetThresholdingBlocksize->setEnabled(
      ui_->radioButtonNucDetParamsManual->isChecked());
  ui_->groupBoxNucDetThresholdingBlocksizeAutoscanParams->setVisible(
      ui_->radioButtonNucDetParamsAuto->isChecked());
  ui_->groupBoxNucDetThresholdingBaselineAutoscanParams->setVisible(
      ui_->radioButtonNucDetParamsAuto->isChecked());
  ui_->labelNucDetThresholdingAutoscanObjective->setVisible(
      ui_->radioButtonNucDetParamsAuto->isChecked());
  ui_->comboBoxNucDetThresholdingAutoscanObjective->setVisible(
      ui_->radioButtonNucDetParamsAuto->isChecked());
  ui_->labelNucDetThresholdingThreadTimeout->setVisible(
      ui_->radioButtonNucDetParamsAuto->isChecked());
  ui_->dspinBoxNucDetThresholdingThreadTimeout->setVisible(
      ui_->radioButtonNucDetParamsAuto->isChecked());

  // TODO:REPETITION:4bc8b312
  ui_->checkBoxNucDetClosing->setVisible(
      ui_->radioButtonNucDetChannelSelectionTypeCytoplasmic->isChecked());
  ui_->spinBoxNucDetClosingTimes->setVisible(
      ui_->radioButtonNucDetChannelSelectionTypeCytoplasmic->isChecked());
  ui_->labelNucDetClosingDilation->setVisible(
      ui_->radioButtonNucDetChannelSelectionTypeCytoplasmic->isChecked());
  ui_->spinBoxNucDetClosingDilation->setVisible(
      ui_->radioButtonNucDetChannelSelectionTypeCytoplasmic->isChecked());
  ui_->labelNucDetClosingErosion->setVisible(
      ui_->radioButtonNucDetChannelSelectionTypeCytoplasmic->isChecked());
  ui_->spinBoxNucDetClosingErosion->setVisible(
      ui_->radioButtonNucDetChannelSelectionTypeCytoplasmic->isChecked());

  // Image masking toolbox:
  ui_->pushButtonImgMaskSave->setEnabled(false);

/*
  ui_->checkBoxNucDetOpening->setVisible(
      ui_->radioButtonNucDetChannelSelectionTypeNuclear->isChecked() );
  ui_->spinBoxNucDetOpeningTimes->setVisible(
      ui_->radioButtonNucDetChannelSelectionTypeNuclear->isChecked() );
  ui_->labelNucDetOpeningDilation->setVisible(
      ui_->radioButtonNucDetChannelSelectionTypeNuclear->isChecked() );
  ui_->spinBoxNucDetOpeningDilation->setVisible(
      ui_->radioButtonNucDetChannelSelectionTypeNuclear->isChecked() );
  ui_->labelNucDetOpeningErosion->setVisible(
      ui_->radioButtonNucDetChannelSelectionTypeNuclear->isChecked() );
  ui_->spinBoxNucDetOpeningErosion->setVisible(
      ui_->radioButtonNucDetChannelSelectionTypeNuclear->isChecked() );
  */
  /*
  ui_->spinBoxNucDetContourDetThresholdLow->setVisible(
      ui_->radioButtonNucDetChannelSelectionTypeCytoplasmic->isChecked() );
  ui_->labelNucDetContourDetThresholdLow->setVisible(
      ui_->radioButtonNucDetChannelSelectionTypeCytoplasmic->isChecked() );
  ui_->dspinBoxNucDetContourDetThresholdsRatio->setVisible(
      ui_->radioButtonNucDetChannelSelectionTypeCytoplasmic->isChecked() );
  ui_->labelNucDetContourDetThresholdsRatio->setVisible(
      ui_->radioButtonNucDetChannelSelectionTypeCytoplasmic->isChecked() );
  ui_->dspinBoxNucDetContourDetTolerance->setVisible(
      ui_->radioButtonNucDetChannelSelectionTypeCytoplasmic->isChecked() );
  ui_->labelNucDetContourDetTolerance->setVisible(
      ui_->radioButtonNucDetChannelSelectionTypeCytoplasmic->isChecked() );
  ui_->checkBoxNucDetContourDetReconstruct->setVisible(
      ui_->radioButtonNucDetChannelSelectionTypeCytoplasmic->isChecked() );
   */
}


void MainWindow::setupMenuTools_()
{
  menuToolsEnable_(true);
  setupToolboxenBasedOnAvailableChannels_();
}


void MainWindow::setupToolboxenBasedOnAvailableChannels_()
{
  bool const has_cyan_images   = images_->has(ImageChannel::C);
  bool const has_magenta_images= images_->has(ImageChannel::M);
  bool const has_yellow_images = images_->has(ImageChannel::Y);
  bool const has_green_images  = images_->has(ImageChannel::G);
  bool const has_red_images    = images_->has(ImageChannel::R);
  bool const has_blue_images   = images_->has(ImageChannel::B);
//bool const has_brightfield_images = original_images.has(image_kind::BF);
//bool const has_overlays    = original_images.has_overlays();

  // Nuclei detection toolbox & Nuclei editing toolbox:
  // (first option should be cyan/blue, if available, because the typical
  //  channel for DAPI/Hoechst/H2B staining is often blue or, in our lab, cyan)
  for (auto& combo_box: {ui_->comboBoxNucDetChannelSelection,
                         ui_->comboBoxNucEditRefineNuccutChannelAux1,
                         ui_->comboBoxNucEditRefineNuccutChannelAux2,
                         ui_->comboBoxPerinucSettingsBgAvoidChannel}) {
    combo_box->clear();
    if (   combo_box == ui_->comboBoxNucEditRefineNuccutChannelAux1
        or combo_box == ui_->comboBoxNucEditRefineNuccutChannelAux2) {
      combo_box->addItem("(none)");
    }

    if (has_cyan_images)   { combo_box->addItem("cyan");   }
    if (has_magenta_images){ combo_box->addItem("magenta");}
    if (has_yellow_images) { combo_box->addItem("yellow"); }
    if (has_blue_images)   { combo_box->addItem("blue");   }
    if (has_green_images)  { combo_box->addItem("green");  }
    if (has_red_images)    { combo_box->addItem("red");    }
    if (combo_box->count() > 0) {
      combo_box->setCurrentText(combo_box->itemText(0));
  } }

  // To set availability in nuclei editing toolbox:
  emit ui_->comboBoxNucDetChannelSelection->currentIndexChanged(
    ui_->comboBoxNucDetChannelSelection->currentIndex());

  // Perinuclei toolbox convenience tweak -- 1st usually should not be cyan/blue:
  if (ui_->comboBoxPerinucSettingsBgAvoidChannel->count() > 1) {
    for (int i=0; ui_->comboBoxPerinucSettingsBgAvoidChannel->count(); ++i) {
      auto cur_text = ui_->comboBoxPerinucSettingsBgAvoidChannel->itemText(i);
      if (cur_text != "cyan" and cur_text != "blue") {
        ui_->comboBoxPerinucSettingsBgAvoidChannel->setCurrentText(cur_text);
        break;
  } } }

  // Quantification toolbox:
  ui_->checkBoxNucQuantChannelC->setChecked(has_cyan_images);
  ui_->checkBoxNucQuantChannelC->setVisible(has_cyan_images);
  ui_->lineEditQuantNucChannelC->setVisible(has_cyan_images);
  ui_->checkBoxNucQuantChannelM->setChecked(has_magenta_images);
  ui_->checkBoxNucQuantChannelM->setVisible(has_magenta_images);
  ui_->lineEditQuantNucChannelM->setVisible(has_magenta_images);
  ui_->checkBoxNucQuantChannelY->setChecked(has_yellow_images);
  ui_->checkBoxNucQuantChannelY->setVisible(has_yellow_images);
  ui_->lineEditQuantNucChannelY->setVisible(has_yellow_images);
  ui_->checkBoxNucQuantChannelG->setChecked(has_green_images);
  ui_->checkBoxNucQuantChannelG->setVisible(has_green_images);
  ui_->lineEditQuantNucChannelG->setVisible(has_green_images);
  ui_->checkBoxNucQuantChannelR->setChecked(has_red_images);
  ui_->checkBoxNucQuantChannelR->setVisible(has_red_images);
  ui_->lineEditQuantNucChannelR->setVisible(has_red_images);
  ui_->checkBoxNucQuantChannelB->setChecked(has_blue_images);
  ui_->checkBoxNucQuantChannelB->setVisible(has_blue_images);
  ui_->lineEditQuantNucChannelB->setVisible(has_blue_images);
  //
  if (has_cyan_images) {
    ui_->lineEditQuantNucChannelC->setText(
        QString::fromStdString(images_->description(ImageChannel::C))); }
  if (has_magenta_images){
    ui_->lineEditQuantNucChannelM->setText(
        QString::fromStdString(images_->description(ImageChannel::M))); }
  if (has_yellow_images) {
    ui_->lineEditQuantNucChannelY->setText(
        QString::fromStdString(images_->description(ImageChannel::Y))); }
  if (has_blue_images) {
    ui_->lineEditQuantNucChannelB->setText(
        QString::fromStdString(images_->description(ImageChannel::B))); }
  if (has_green_images) {
    ui_->lineEditQuantNucChannelG->setText(
        QString::fromStdString(images_->description(ImageChannel::G))); }
  if (has_red_images) {
    ui_->lineEditQuantNucChannelR->setText(
        QString::fromStdString(images_->description(ImageChannel::R))); }
}


void MainWindow::setupScripting_()
{
  script_window_.reset(new ScriptWindow(this, this));
  connect(script_window_.get(), SIGNAL(scriptFilesListModified(QStringList)),
          this,                 SLOT  (updateScripts(QStringList)) );

  // -- completer:
  QStringList words;
  auto const& api_func_names = api_functions.keys();
  for (auto const& fn: qAsConst(api_func_names)) { words << fn; }
  for (auto const& pn: qAsConst(api_parameters)) { words << pn; }
  for (auto const& sn: qAsConst(api_specials))   { words << sn; }
  words.sort();
  script_window_->setCompleterWordlist(words);
}

