// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "MainWindow.hpp"
#include "ui_MainWindow.h"
#include "MainWindow_aux.hpp"
#include "GUI/Support/PixmapCache.hpp"


void MainWindow::on_pushButtonImgMaskReset_clicked()
{
  for (auto ppoly: qAsConst(current_exclusions_polys_)) { delete ppoly; }
  current_exclusions_polys_.clear();
  current_exclusions_.clear();
  ui_->pushButtonImgMaskReset->setEnabled(not current_exclusions_.empty());
  ui_->pushButtonImgMaskApply->setEnabled(not current_exclusions_.empty());
}


void MainWindow::on_pushButtonImgMaskApply_clicked()
{
  st::contours_vec_t contours;
  for (auto const& excl: qAsConst(current_exclusions_)) {
    contours.emplace_back(st::contour_t{});
    contours.back().reserve(static_cast<size_t>(excl.size()));
    for (auto const& pt: excl) { contours.back().emplace_back(pt.x(), pt.y()); }
  }

  if (not contours.empty()) {
    //for (auto imgs: {&original_images_, &displayed_images_}) {
    //  imgs->register_and_use_mask(current_frame_i_, contours);
    //}
    images_->setMask(current_frame_i_, contours);
    pixmap_cache_->rebuild(current_frame_i_);
    showFrame_(current_frame_i_);
    switch (images_->maskingMode()) {
      case ImageSequence::Masking::ONLY_STORE:
        message_('I', "Maskin contour(s) have been stored.");
        break;
      case ImageSequence::Masking::STORE_AND_BLACK_OUT:
        message_('I', "Masking has been applied in-place.");
        break;
      default:
        message_('W', "Unrecognized masking mode.");
        break;
    }
    on_pushButtonImgMaskReset_clicked();
    ui_->pushButtonImgMaskSave->setEnabled(true);
  }
}


QString MainWindow::assembleImageMaskContoursFilename_()
{ return timepointFilenameEndingWith_("-mask.txt",
                                      "a file with image mask contours"); }


bool MainWindow::checkIfImageMaskContoursFileAvailable_()
{
  QString const fn = assembleImageMaskContoursFilename_();
  if (not fn.isEmpty()) {
    bool const file_exists = QFile(fn).exists();
    ui_->pushButtonImgMaskLoad->setEnabled(file_exists);
    return file_exists;
  } else {
    ui_->pushButtonImgMaskLoad->setEnabled(false);
    ui_->pushButtonImgMaskSave->setEnabled(false);
    return false;
  }
}


void MainWindow::on_pushButtonImgMaskSave_clicked()
{
  QString const fn = assembleImageMaskContoursFilename_();
  if (fn.isEmpty()) {
    message_('E', "Could not save image masking contours!");
    return;
  }
  QFile file(fn);
  if (file.open(QFile::WriteOnly | QFile::Text)) {
    QTextStream out(&file);
    int ci = 0;
    for (auto const& c: images_->maskingContours(current_frame_i_)) {
      for (auto const& pt: c) { out << pt.x << ',' << pt.y << "; "; }
      out << '\n';
      ci++;
    }
    file.close();
    message_('I', "Saved " + QString::number(ci) + " masking contour"
                  + ((ci > 1)? "s": "") + " to " + QFileInfo(file).fileName());
  } // if file.open()

  checkIfImageMaskContoursFileAvailable_();
}


void MainWindow::on_pushButtonImgMaskLoad_clicked()
{
  QString const fn = assembleImageMaskContoursFilename_();
  if (fn.isEmpty()) {
    message_('E', "Could not load image masking contours!");
    return;
  }
  QFile file(fn);
  if (file.open(QFile::ReadOnly | QFile::Text)) {
    QTextStream in(&file);
    int ci = 0;
    st::contours_vec_t contours;
    while (not in.atEnd()) {
      QString const line = in.readLine();
      if (line.length() > 0) {
        contours.push_back({});
        QStringList const xsys =
            line.split("; ", Qt::SkipEmptyParts);
        for (auto const& _xy: xsys) {
          QStringList xy = _xy.split(',');
          contours.back().emplace_back(xy[0].toInt(), xy[1].toInt());
        }
        ci++;
      } // if non-empty line
    } // while not at end
    file.close();
    message_('I', "Loaded " + QString::number(ci) + " mask contour"
                  + ((ci > 1)? "s": "") + " from "+ QFileInfo(file).fileName());

    if (not contours.empty()) {

      // -- draw exclusions
      for (auto& c: contours) {
        QVector<QPoint> vp; for (auto const& cp: c) { vp.push_back({cp.x, cp.y}); }
        current_exclusions_.push_back(vp);
      }
      drawExclusions_();
      ui_->pushButtonImgMaskReset->setEnabled(not current_exclusions_.empty());
      ui_->pushButtonImgMaskApply->setEnabled(not current_exclusions_.empty());

      // -- apply the mask (commented out)
      // images_sequences_[0]->mask(current_frame_i_,contours);
      // setup_pixmaps_();
      // show_frame_(current_frame_i_);
      // auto const masked_area =
      //     images_sequences_[0]->masked_area(current_frame_i_);
      // message_to_user_('I', "Masking excluded "
      //                       + QString::number(masked_area) +" px");
    }

  } else {
    message_('E', "Could not load mask contours from "
                  + QFileInfo(file).fileName());
  }
}
