// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "MainWindow.hpp"
#include "ui_MainWindow.h"
#include "Image/NucleiDetection.hpp"
#include "Image/NucleiDetectionStained.hpp"
#include "Image/NucleiDetectionNonstained.hpp"
#include <QStandardItemModel>
#include <QMessageBox>
#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
#  include <QtConcurrent/QtConcurrent>
#  include <QtConcurrent/QtConcurrentRun>
#  include <QtConcurrent/QtConcurrentMap>
#else
#  include <QtConcurrentRun>
#  include <QtConcurrentMap>
#endif
#if (QT_VERSION >= QT_VERSION_CHECK(5, 2, 0))
# ifdef Q_OS_WIN
#  include <QWinTaskbarButton>
#  include <QWinTaskbarProgress>
# endif
#endif

namespace {
char const* sNucleiDetectionSettingsFilename = "nuclei_detection.ini";

char const* sNucDetCh = "channel";
char const* sStnNuc   = "staining_nuclear";

char const* sNrm     = "normalization";
char const* sDns     = "denoising";
char const* sDnsStrg = "denoising_strength";
char const* sSmt     = "smoothing";
char const* sSmtSz   = "smoothing_block_size";
char const* sThr     = "thresholding";
char const* sThrSz   = "thresholding_block_size";
char const* sThrBsln = "thresholding_baseline_offset";
char const* sOpn     = "opening";
char const* sOpnRpt  = "opening_repeats";
char const* sOpnErSz = "opening_erosion_base_size";
char const* sOpnDiSz = "opening_dilation_base_size";
char const* sCls     = "closing";
char const* sClsRpt  = "closing_repeats";
char const* sClsErSz = "closing_erosion_base_size";
char const* sClsDiSz = "closing_dilation_base_size";

char const* sEql    = "uniformization";
char const* sEqlThr = "uniformization_brightness_threshold";

char const* sConDetThrLo   = "contour_detection_threshold_low";
char const* sConDetThrsRat = "contour_detection_thresholds_ratio";
char const* sConDetTol     = "contour_detection_tolerance";
char const* sConDetRec     = "contour_detection_reconstruct";

char const* sMinSld         = "min_solidity";
char const* sMinAreaMedFrac = "min_area_as_median_fraction";
char const* sMaxAreaMedFrac = "max_area_as_median_fraction";
char const* sSplAdj         = "splitting_adjacent";
char const* sMinPocketArea  = "min_pocket_area";

// "passenger" settings for Nuclei Editing toolbox:
char const* sNuccutChMain       = "nuccut_channel_main";
char const* sNuccutChAux1       = "nuccut_channel_aux1";
char const* sNuccutChAux2       = "nuccut_channel_aux2";
char const* sNuccutChMainWght   = "nuccut_channel_main_weight";
char const* sNuccutChAux1Wght   = "nuccut_channel_aux1_weight";
char const* sNuccutChAux2Wght   = "nuccut_channel_aux2_weight";
char const* sNuccutColorSimil   = "nuccut_color_similarity";
char const* sNuccutEdgeImpassab = "nuccut_edge_impassability";
char const* sNuccutExclToggled  = "nuccut_exclude_toggled";
}


// Settings persistence ---

bool MainWindow::checkIfNucDetSettingsFileAvailable_()
{
  QString const fn = assembleNucDetSettingsFilename_();
  if (fn.size() > 0) {
    bool const file_exists = QFile(fn).exists();
    ui_->pushButtonNucDetSettingsLoad->setEnabled(file_exists);
    return file_exists;
  } else {
    ui_->pushButtonNucDetSettingsLoad->setEnabled(false);
    ui_->pushButtonNucDetSettingsSave->setEnabled(false);
    return false;
  }
}


void MainWindow::on_comboBoxNucDetChannelSelection_currentIndexChanged(int)
{
  // seems that may be called upon filling of combo box lists => count() checks
  if (not ui_->radioButtonNucDetChannelSelectionTypeNuclear->isChecked()) return;
  if (    (   ui_->comboBoxNucEditRefineNuccutChannelAux1->count()
           == ui_->comboBoxNucDetChannelSelection->count() + 1)
      and (   ui_->comboBoxNucEditRefineNuccutChannelAux2->count()
           == ui_->comboBoxNucDetChannelSelection->count()) + 1) {
    ui_->spinBoxNucEditRefineNuccutChannelMain->setEnabled(true);
    QString const nuc_ch_txt = ui_->comboBoxNucDetChannelSelection->currentText();
    ui_->lineEditNucEditRefineNuccutChannelMain->setText(nuc_ch_txt);
    auto const flags_selectable = (Qt::ItemIsSelectable | Qt::ItemIsEnabled);
    for (auto combo_box_aux: {ui_->comboBoxNucEditRefineNuccutChannelAux1,
                              ui_->comboBoxNucEditRefineNuccutChannelAux2}) {
      for (int i = 0; i<combo_box_aux->count(); i++) {
        auto m = qobject_cast<QStandardItemModel const*>(combo_box_aux->model());
        if (combo_box_aux->itemText(i) == nuc_ch_txt) {
          m->item(i)->setFlags(m->item(i)->flags() & ~flags_selectable);
        } else {
          m->item(i)->setFlags(m->item(i)->flags() |  flags_selectable);
  } } } }

  ui_->comboBoxNucEditRefineNuccutChannelAux1->setCurrentIndex(0);
  ui_->comboBoxNucEditRefineNuccutChannelAux2->setCurrentIndex(0);
  ui_->spinBoxNucEditRefineNuccutChannelAux1->setEnabled(false);
  ui_->spinBoxNucEditRefineNuccutChannelAux2->setEnabled(false);
}


QString MainWindow::assembleNucDetSettingsFilename_() const
{
  // -- per time frame settings files:
  // QString const fn = QString::fromStdString(
  //    original_images_.file_path(nuclearChannelImageKind_()));
  // fn = fn.replace(
  //    QRegularExpression(QString("\\.") + QString::fromStdString(
  //        original_images_.image_file_extension()) + "$"),
  //    QString("-nuc_det.ini"));

  // -- single settings file for all time points:
  // QString const fn = QString::fromStdString(
  //      images_->filePathAnyChannel(current_frame_i_));
  //  QString const fn_ext = QString::fromStdString(
  //      images_->imageFileExtension());
  //  QString fn_rep = fn;
  //  fn_rep.replace(QRegularExpression(QString("_?[TtZz][0-9]*_?[Cc][Hh]?[0-9]*\\.")
  //                                    + fn_ext + "$"),
  //                 "-nuc_detect.ini");
  //  if (fn_rep == fn) {  // replacement failure
  //    message_('E', "Could not assemble nuclei detection settings file name!");
  //    return "";
  //  }
  //  return fn_rep;
  assert(not images_directory_.isEmpty());
  return images_directory_ + QDir::separator() + sNucleiDetectionSettingsFilename;
}


void MainWindow::on_pushButtonNucDetSettingsSave_clicked()
{
  QString const fn = assembleNucDetSettingsFilename_();
  if (fn.size() == 0) {
    message_('E', "Could not save nuclei detection settings!");
    return;
  }
  QFile file(fn);
  if (file.open(QFile::WriteOnly | QFile::Text)) {
    assert(   ui_->spinBoxNucDetOpeningErosion ->value()
           == ui_->spinBoxNucDetOpeningDilation->value());
    QTextStream out(&file);
    out << sNucDetCh<<'='<<ui_->comboBoxNucDetChannelSelection->currentText()<<'\n';
    out << sStnNuc <<'='<< ui_->radioButtonNucDetChannelSelectionTypeNuclear
                              ->isChecked()<<'\n';
    out << sNrm    <<'='<< ui_->checkBoxNucDetNormalization->isChecked() <<'\n';
    out << sDns    <<'='<< ui_->checkBoxNucDetDenoising    ->isChecked() <<'\n';
    out << sDnsStrg<<'='<< ui_->dspinBoxNucDetDenoising    ->value()     <<'\n';
    out << sSmt    <<'='<< ui_->checkBoxNucDetSmoothing    ->isChecked() <<'\n';
    out << sSmtSz  <<'='<< ui_->spinBoxNucDetSmoothing     ->value()     <<'\n';
    out << sThr    <<'='<< ui_->checkBoxNucDetThresholding ->isChecked() <<'\n';
    out << sThrSz  <<'='<< ui_->spinBoxNucDetThresholdingBlocksize->value()<<'\n';
    out << sThrBsln<<'='<< ui_->spinBoxNucDetThresholdingBaseline->value()<<'\n';
    out << sOpn    <<'='<< ui_->checkBoxNucDetOpening      ->isChecked() <<'\n';
    out << sOpnRpt <<'='<< ui_->spinBoxNucDetOpeningTimes  ->value()     <<'\n';
    out << sOpnErSz<<'='<< ui_->spinBoxNucDetOpeningErosion->value()     <<'\n';
    out << sOpnDiSz<<'='<< ui_->spinBoxNucDetOpeningDilation->value()    <<'\n';
    out << sCls    <<'='<< ui_->checkBoxNucDetClosing      ->isChecked() <<'\n';
    out << sClsRpt <<'='<< ui_->spinBoxNucDetClosingTimes  ->value()     <<'\n';
    out << sClsErSz<<'='<< ui_->spinBoxNucDetClosingErosion->value()     <<'\n';
    out << sClsDiSz<<'='<< ui_->spinBoxNucDetClosingDilation->value()    <<'\n';
    //
    out << sEql    <<'='<< ui_->checkBoxNucDetEqualization->isChecked()    <<'\n';
    out << sEqlThr <<'='<< ui_->spinBoxNucDetEqualizationThreshold->value()<<'\n';
    //
    out << sConDetThrLo<<'='<< ui_->spinBoxNucDetContourDetThresholdLow->value()<<'\n';
    out << sConDetThrsRat<<'='<<ui_->dspinBoxNucDetContourDetThresholdsRatio->value()<<'\n';
    out << sConDetTol<<'='<< ui_->dspinBoxNucDetContourDetTolerance->value()<<'\n';
    out << sConDetRec<<'='<<ui_->checkBoxNucDetContourDetReconstruct->isChecked()<<'\n';
    //
    out << sMinSld        <<'='<<ui_->dspinBoxNucDetSolidity   ->value()  <<'\n';
    out << sMinAreaMedFrac<<'='<<ui_->dspinBoxNucDetMinAreaFrac->value()  <<'\n';
    out << sMaxAreaMedFrac<<'='<<ui_->dspinBoxNucDetMaxAreaFrac->value()  <<'\n';
    out << sSplAdj        <<'='<<ui_->checkBoxNucDetSplitAdjacent->isChecked()<<'\n';
    out << sMinPocketArea <<'='<<ui_->spinBoxNucDetMinPocketArea->value() <<'\n';
    //
    // "passenger" settings of another toolbox:
    out << sNuccutChMain <<'='<<ui_->lineEditNucEditRefineNuccutChannelMain->text()<<'\n';
    out << sNuccutChAux1 <<'='<<ui_->comboBoxNucEditRefineNuccutChannelAux1->currentText()<<'\n';
    out << sNuccutChAux2 <<'='<<ui_->comboBoxNucEditRefineNuccutChannelAux2->currentText()<<'\n';
    out << sNuccutChMainWght  <<'='<<ui_->spinBoxNucEditRefineNuccutChannelMain->value()<<'\n';
    out << sNuccutChAux1Wght  <<'='<<ui_->spinBoxNucEditRefineNuccutChannelAux1->value()<<'\n';
    out << sNuccutChAux2Wght  <<'='<<ui_->spinBoxNucEditRefineNuccutChannelAux2->value()<<'\n';
    out << sNuccutColorSimil<<'='<<ui_->dspinBoxNucEditRefineNuccutParamColorSimilarity->value()<<'\n';
    out << sNuccutEdgeImpassab<<'='<<ui_->dspinBoxNucEditRefineNuccutParamEdgeImpassability->value()<<'\n';
    out << sNuccutExclToggled <<'='<<ui_->checkBoxNucEditRefineNuccutExcludeToggled->isChecked()<<'\n';

    file.close();
    message_('I', "Nuclei detection settings stored in "
                  + QFileInfo(file).fileName());
  } else {
    message_('E', "Could not store nuclei detection settings in "
                  + QFileInfo(file).fileName());
  }

  checkIfNucDetSettingsFileAvailable_();
}


void MainWindow::on_pushButtonNucDetSettingsLoad_clicked()
{
  bool const should_auto_update = ui_->checkBoxNucDetDetectAutoup->isChecked();
  ui_->checkBoxNucDetDetectAutoup->setChecked(false);

  QString const fn = assembleNucDetSettingsFilename_();
  if (fn.size() == 0) {
    message_('E', "Could not load nuclei detection settings!");
    return;
  }

  QFile file(fn);
  if (file.open(QFile::ReadOnly | QFile::Text)) {
    bool errors_encountered = false;
    QTextStream in(&file);

    while (not in.atEnd()) {
      QString const line = in.readLine();
      QStringList const key_value = line.split('=');
      QString const k = key_value.front(),
                    v = key_value.back();

      if (k == QString{sNucDetCh}) {
        // TODO: DUPLICATION 0x91df8sa
        bool found_and_set = false;
        for (int i=0; i<ui_->comboBoxNucDetChannelSelection->count(); ++i) {
          if (ui_->comboBoxNucDetChannelSelection->itemText(i) == v) {
            ui_->comboBoxNucDetChannelSelection->setCurrentText(v);
            found_and_set = true;
            break;
        } }
        if (not found_and_set) {
          errors_encountered = true;
          message_('E', "Unrecognized or unavailable nuclei detection channel "
                        "'" + v + "'");
        }
      } else if (k == QString{sStnNuc}) {
        ui_->radioButtonNucDetChannelSelectionTypeNuclear
           ->setChecked(    static_cast<bool>(v.toInt()));
        ui_->radioButtonNucDetChannelSelectionTypeCytoplasmic
           ->setChecked(not static_cast<bool>(v.toInt()));
      //
      } else if (k == QString{sNrm}) {
        ui_->checkBoxNucDetNormalization->setChecked(static_cast<bool>(v.toInt()));
      } else if (k == QString{sDns}) {
        ui_->checkBoxNucDetDenoising->setChecked(static_cast<bool>(v.toInt()));
      } else if (k == QString{sDnsStrg}){
        ui_->dspinBoxNucDetDenoising->setValue(v.toDouble());
      } else if (k == QString{sSmt}) {
        ui_->checkBoxNucDetSmoothing->setChecked(static_cast<bool>(v.toInt()));
      } else if (k == QString{sSmtSz}){
        ui_->spinBoxNucDetSmoothing->setValue(v.toInt());
      } else if (k == QString{sThr}) {
        ui_->checkBoxNucDetThresholding->setChecked(static_cast<bool>(v.toInt()));
      } else if (k == QString{sThrSz}) {
        ui_->spinBoxNucDetThresholdingBlocksize->setValue(v.toInt());
      } else if (k == QString{sThrBsln}) {
        ui_->spinBoxNucDetThresholdingBaseline->setValue(v.toInt());
      } else if (k == QString{sOpn}) {
        ui_->checkBoxNucDetOpening->setChecked(static_cast<bool>(v.toInt()));
      } else if (k == QString{sOpnRpt}) {
        ui_->spinBoxNucDetOpeningTimes->setValue(v.toInt());
      } else if (k == QString{sOpnErSz}) {
        ui_->spinBoxNucDetOpeningErosion->setValue(v.toInt());
      } else if (k == QString{sOpnDiSz}) {
        ui_->spinBoxNucDetOpeningDilation->setValue(v.toInt());
      } else if (k == QString{sCls}) {
        ui_->checkBoxNucDetClosing->setChecked(static_cast<bool>(v.toInt()));
      } else if (k == QString{sClsRpt}) {
        ui_->spinBoxNucDetClosingTimes->setValue(v.toInt());
      } else if (k == QString{sClsErSz}) {
        ui_->spinBoxNucDetClosingErosion->setValue(v.toInt());
      } else if (k == QString{sClsDiSz}) {
        ui_->spinBoxNucDetClosingDilation->setValue(v.toInt());
      //
      } else if (k == QString{sEql})   {
          ui_->checkBoxNucDetEqualization->setChecked(static_cast<bool>(v.toInt()));
      } else if (k == QString{sEqlThr}) {
          ui_->spinBoxNucDetEqualizationThreshold->setValue(v.toDouble());
      //
      } else if (k == QString{sConDetThrLo}) {
        ui_->spinBoxNucDetContourDetThresholdLow->setValue(v.toInt());
      } else if (k == QString{sConDetThrsRat}) {
        ui_->dspinBoxNucDetContourDetThresholdsRatio->setValue(v.toDouble());
      } else if (k == QString{sConDetTol}) {
        ui_->dspinBoxNucDetContourDetTolerance->setValue(v.toDouble());
      } else if (k == QString{sConDetRec}) {
        auto const v_as_bool = static_cast<bool>(v.toInt());
        ui_->checkBoxNucDetContourDetReconstruct->setChecked(v_as_bool);
      //
      } else if (k == QString{sMinSld}) {
        ui_->dspinBoxNucDetSolidity->setValue(v.toDouble());
      } else if (k == QString{sMinAreaMedFrac}) {
        ui_->dspinBoxNucDetMinAreaFrac->setValue(v.toDouble());
      } else if (k == QString{sMaxAreaMedFrac}) {
        ui_->dspinBoxNucDetMaxAreaFrac->setValue(v.toDouble());
      } else if (k == QString{sSplAdj}) {
        auto const v_as_bool = static_cast<bool>(v.toInt());
        ui_->checkBoxNucDetSplitAdjacent->setChecked(v_as_bool);
      } else if (k == QString{sMinPocketArea}) {
        ui_->spinBoxNucDetMinPocketArea->setValue(v.toInt());
      //
      //
      // "passenger" settings of another toolbox:
      //
      } else if (k == QString{sNuccutChMain}) {
        if (not images_->has(st::channelFrom(v.toStdString()))) {
          errors_encountered = true;
          message_('E', "Unrecognized or unavailable NucCut main channel '"+v+"'");
        }
        ui_->lineEditNucEditRefineNuccutChannelMain->setText(v.toLower());
      } else if (k == QString{sNuccutChAux1}) {
        // TODO: DUPLICATION 0x91df8sa
        bool found_and_set = false;
        for (int i=0; i<ui_->comboBoxNucEditRefineNuccutChannelAux1->count(); ++i) {
          if (ui_->comboBoxNucEditRefineNuccutChannelAux1->itemText(i) == v) {
            ui_->comboBoxNucEditRefineNuccutChannelAux1->setCurrentText(v);
            found_and_set = true;
            break;
        } }
        if (not found_and_set) {
          errors_encountered = true;
          message_('E', "Unrecognized or unavailable NucCut first auxiliary "
                        "channel '" + v + "'");
        }
      } else if (k == QString{sNuccutChAux2}) {
        // TODO: DUPLICATION 0x91df8sa
        bool found_and_set = false;
        for (int i=0; i<ui_->comboBoxNucEditRefineNuccutChannelAux2->count(); ++i) {
          if (ui_->comboBoxNucEditRefineNuccutChannelAux2->itemText(i) == v) {
            ui_->comboBoxNucEditRefineNuccutChannelAux2->setCurrentText(v);
            found_and_set = true;
            break;
        } }
        if (not found_and_set) {
          errors_encountered = true;
          message_('E', "Unrecognized or unavailable NucCut second auxiliary "
                        "channel '" + v + "'");
        }
      } else if (k == QString{sNuccutChMainWght}) {
        ui_->spinBoxNucEditRefineNuccutChannelMain->setValue(v.toDouble());
      } else if (k == QString{sNuccutChAux1Wght}) {
        ui_->spinBoxNucEditRefineNuccutChannelAux1->setValue(v.toDouble());
      } else if (k == QString{sNuccutChAux2Wght}) {
        ui_->spinBoxNucEditRefineNuccutChannelAux2->setValue(v.toDouble());
      } else if (k == QString{sNuccutColorSimil}) {
        ui_->dspinBoxNucEditRefineNuccutParamColorSimilarity->setValue(v.toDouble());
      } else if (k == QString{sNuccutEdgeImpassab}) {
        ui_->dspinBoxNucEditRefineNuccutParamEdgeImpassability->setValue(v.toDouble());
      } else if (k == QString{sNuccutExclToggled}) {
         ui_->checkBoxNucEditRefineNuccutExcludeToggled
            ->setChecked(static_cast<bool>(v.toInt()));
      } else {
        message_('W', "Unrecognized nuclei detection setting "
                      "'" + k + "' in " + QFileInfo(file).fileName());
        errors_encountered = true;
      }
    } // while reading the file
    file.close();

    if (not errors_encountered) {
      message_('I', "Nuclei detection settings loaded from "
                    + QFileInfo(file).fileName());
    }
  } else {
    message_('E', "Could not load nuclei detection settings from "
                  + QFileInfo(file).fileName() + "!"
                  + ((!QFileInfo(file).exists())? " File does not exist.": ""));
  }

  ui_->checkBoxNucDetDetectAutoup->setChecked(should_auto_update);
  maybe_NucDetDetect_autoclick();
}


// Actions ------

void MainWindow::maybe_NucDetDetect_autoclick()
{
  if (ui_->checkBoxNucDetDetectAutoup->isChecked()) {
    ui_->pushButtonNucDetDetect->click();
  }
}


ImageChannel MainWindow::nuclearImageChannel_() const
{
  QString const nuclear_channel_text =
      ui_->comboBoxNucDetChannelSelection->currentText();
  return (nuclear_channel_text.contains("blue"))    ? ImageChannel::B
       : (nuclear_channel_text.contains("green"))   ? ImageChannel::G
       : (nuclear_channel_text.contains("red"))     ? ImageChannel::R
       : (nuclear_channel_text.contains("cyan"))    ? ImageChannel::C
       : (nuclear_channel_text.contains("magenta")) ? ImageChannel::M
       : (nuclear_channel_text.contains("yellow"))  ? ImageChannel::Y
       : ImageChannel::UNDETERMINED;
}


// aux function used by MainWindow::nucDetImgPreprocParamScan()
//
template <class T> inline
void select_highest_scoring(T& best_as_far, T const& current)
{
  if (current.first > best_as_far.first) { // higher score, outbid!
    best_as_far = current;
  }
}

// This function is intended to run concurrently (in a separate thread).
int MainWindow::nucDetImgPreprocParamScan(cv::Mat const& img_orig)
{
  emit busy();
  disconnect(this, SIGNAL(busy()), this, SLOT(startProgressNucDetParamScan()));

  auto p = getImagePreprocessingParameters();
  assert(ui_->radioButtonNucDetChannelSelectionTypeNuclear->isChecked());
  p.closing = false;
  p.closing_ntimes=0;  p.closing_erosion_size=0;  p.closing_dilation_size=0;

  cv::Mat img_sub_preproc_main, img_preproc_aux;
  if (ui_->actionPreferenceParamScanPrecomputeShared->isChecked()) {
    emit messaging('I', "Parameter scan: precomputing shared image "
                        "preprocessing steps...");
    using Stage = st::stained::ImagePreprocessingStage;
    if (getStainedNucleiDetectionParameters().uniformization) {
      st::stained::preprocessImage(img_orig, p,
          Stage::AFTER_DENOISING, img_preproc_aux,
          Stage::AFTER_SMOOTHING, img_sub_preproc_main, /*preview*/false);
    } else {
      st::stained::preprocessImage(img_orig, p,
          Stage::AFTER_SMOOTHING, img_sub_preproc_main, /*preview*/false);
    }

    // these stages will be no longer necessary:
    p.normalizing = false;
    p.denoising = false;
    p.smoothing = false;
  }

  // This order of loops nestings usually results in a more steady progress of
  // preprocess_and_detect() (called many timeswithin QtConcurrent::mappedReduced).
  using img_preproc_params_t = st::ImagePreprocessingParameters;
  QList<img_preproc_params_t> img_preproc_params_sets;
  for (int thr_bsz  = ui_->spinBoxNucDetThresholdingBlocksizeAutoscanFrom->value();
           thr_bsz <= ui_->spinBoxNucDetThresholdingBlocksizeAutoscanUpto->value();
           thr_bsz += (  ui_->spinBoxNucDetThresholdingBlocksizeAutoscanUpto->value()
                       - ui_->spinBoxNucDetThresholdingBlocksizeAutoscanFrom->value())
                     /(ui_->spinBoxNucDetThresholdingBlocksizeAutoscanNsteps->value()-1)){
    for (int thr_bl  = ui_->spinBoxNucDetThresholdingBaselineAutoscanFrom->value();
             thr_bl <= ui_->spinBoxNucDetThresholdingBaselineAutoscanUpto->value();
             thr_bl += ui_->spinBoxNucDetThresholdingBaselineAutoscanStep->value()) {
      auto p_cur_thr{p};
      p_cur_thr.thresholding_block_size = thr_bsz;
      p_cur_thr.thresholding_baseline = thr_bl;
      img_preproc_params_sets.push_back(p_cur_thr);
  } } // for a grid of thresholding parameters
  emit messaging('I', "Number of parameter sets to scan: "
                      + QString::number(img_preproc_params_sets.size()));

  bool const declump = ui_->checkBoxNucDetSplitAdjacent->isChecked();
  auto const timeout = this->ui_->dspinBoxNucDetThresholdingThreadTimeout->value();
  auto const nucdet_stained_p = getStainedNucleiDetectionParameters();
  auto const nucdet_nonstained_p = getNonstainedNucleiDetectionParameters();
  auto const nuc_presel_params = getNucleiPreselectionParameters();
  auto const&nuc_presel_assess_params = nuc_presel_assess_params_;
  //
  auto preprocess_and_detect = [  // a thread function
      &img_orig, &img_sub_preproc_main, &img_preproc_aux,
      &nucdet_stained_p, &nucdet_nonstained_p,
      &nuc_presel_params, &nuc_presel_assess_params, declump, timeout]
      (img_preproc_params_t const& params_to_score)->img_preproc_scored_params_t
  {
    cv::Mat img_preproc;
    st::stained::preprocessImage(img_sub_preproc_main, params_to_score,
        st::stained::ImagePreprocessingStage::FINAL, img_preproc,/*prevu*/false);
    auto const score = st::detectNuclei_DryWithTimeout(true,
        img_orig, img_preproc, img_preproc_aux,
        params_to_score, nucdet_stained_p, nucdet_nonstained_p,
        nuc_presel_params, nuc_presel_assess_params,
        declump, false, timeout);
    return img_preproc_scored_params_t{score, &params_to_score};
  };

  // std::bind's template parameter provides required traits
  future_img_preproc_ = QtConcurrent::mappedReduced(
      img_preproc_params_sets,
      std::bind<img_preproc_scored_params_t>(preprocess_and_detect,
                                             std::placeholders::_1),
      &select_highest_scoring<img_preproc_scored_params_t>);

  // TODO disconnect these four:
  //
  connect(&watcher_img_preproc_, SIGNAL(progressRangeChanged(int,int)),
          ui_->progressBar,  SLOT(setRange(int,int))/*, Qt::QueuedConnection*/);
  connect(&watcher_img_preproc_, SIGNAL(progressValueChanged(int)),
          ui_->progressBar,  SLOT(setValue(int))/*, Qt::QueuedConnection*/);
  if (not batch_processing_) {
#if (QT_VERSION >= QT_VERSION_CHECK(5, 2, 0))
# ifdef Q_OS_WIN
  connect(&watcher_img_preproc_,       SIGNAL(progressRangeChanged(int,int)),
          win_taskbar_progress_.get(), SLOT(setRange(int,int))/*,
          Qt::QueuedConnection*/);
  connect(&watcher_img_preproc_,       SIGNAL(progressValueChanged(int)),
          win_taskbar_progress_.get(), SLOT(setValue(int))/*,
          Qt::QueuedConnection*/);
# endif
#endif
  }

  watcher_img_preproc_.setFuture(future_img_preproc_);
  watcher_img_preproc_.waitForFinished();
  if (watcher_img_preproc_.result().second) {
    img_preproc_best_params_ = *watcher_img_preproc_.result().second;
  } else {
    QMessageBox msg;
     QString const msg_box_text = "Cannot conclude parameter scan!";
     msg.setText(msg_box_text);
     msg.setInformativeText("Cannot find optimal nuclei detection parameters.");
     msg.setStandardButtons(QMessageBox::Ok);
     msg.setDefaultButton(QMessageBox::Ok);
     msg.setIcon(QMessageBox::Critical);
     msg.exec();
  }

  emit idle();
  disconnect(this, SIGNAL(idle()), this, SLOT(stopProgressNucDetParamScan()));

  emit recheck(true); // may be not connected, that's fine
  disconnect(this, SIGNAL(recheck(bool)),
             ui_->checkBoxNucDetDetectAutoup, SLOT(setChecked(bool)));

  return 0;
} // nucDetImgPreprocParamScan()


// slot
void MainWindow::startProgressNucDetParamScan()
{
  startProgress();
  framesNavigationExpose_(false);
}


// slot
void MainWindow::stopProgressNucDetParamScan()
{
  // redundant -- commented out:
  //disconnect(this, SIGNAL(idle()), this, SLOT(stopProgressNucDetParamScan()));
  if (not batch_processing_) { emit messaging_done(); }
  stopProgress();
  framesNavigationExpose_(true);

  auto const& P = img_preproc_best_params_;
  ui_->spinBoxNucDetThresholdingBaseline ->setValue(P.thresholding_baseline);
  ui_->spinBoxNucDetThresholdingBaseline->setPrefix(
      (ui_->spinBoxNucDetThresholdingBaseline->value() > 0) ? "+" : " ");
  ui_->spinBoxNucDetThresholdingBlocksize->setValue(P.thresholding_block_size);
  if (not ui_->actionPreferenceParamScanPrecomputeShared->isChecked()) {
    ui_->spinBoxNucDetSmoothing            ->setValue(P.smoothing_size);
    ui_->spinBoxNucDetThresholdingBlocksize->setValue(P.thresholding_block_size);
    ui_->spinBoxNucDetOpeningErosion ->setValue(P.opening_erosion_size);
    ui_->spinBoxNucDetOpeningDilation->setValue(P.opening_dilation_size);
  }

  ui_->radioButtonNucDetParamsManual->setChecked(true);  // \_needed for btn click
  ui_->radioButtonNucDetParamsAuto  ->setChecked(false); // /
  on_pushButtonNucDetDetect_clicked();  // finally, detect in the manual mode
}


void MainWindow::on_pushButtonNucDetDetect_clicked()
{
  auto const nuclear_channel = nuclearImageChannel_();
  assert(nuclear_channel != ImageChannel::UNDETERMINED);

  bool const declump = ui_->checkBoxNucDetSplitAdjacent->isChecked();
  bool const preview = ui_->checkBoxNucDetPreviewStages->isChecked();
  bool const detection_channel_is_nuclear =
      ui_->radioButtonNucDetChannelSelectionTypeNuclear->isChecked();
  bool const detection_channel_is_cytoplasmic =
      ui_->radioButtonNucDetChannelSelectionTypeCytoplasmic->isChecked();
  assert(detection_channel_is_nuclear != detection_channel_is_cytoplasmic);

  cv::Mat const img_orig = images_->image(nuclear_channel, current_frame_i_);
  st::contours_vec_t contours_nuclei,
      contours_of_debris,
      contours_nonsegmentable,
      contours_of_nuclei_from_segmentation,
      contours_of_orphans_from_segmentation;

  if (detection_channel_is_nuclear) { // =======================================

    if (ui_->radioButtonNucDetParamsManual->isChecked()) { // ------------------

      auto const t_since = static_cast<double>(cv::getTickCount());

      using Stage = st::stained::ImagePreprocessingStage;
      cv::Mat img_preproc_full;
      cv::Mat img_preproc_partial;
      st::stained::preprocessImage(img_orig,
          getImagePreprocessingParameters(),
          Stage::FINAL,           img_preproc_full,
          Stage::AFTER_DENOISING, img_preproc_partial, preview);

      /*double const score =*/ st::detectNuclei(true,
          img_orig, img_preproc_full, img_preproc_partial,
          &contours_nuclei,
          &contours_of_debris,
          &contours_nonsegmentable,
          &contours_of_nuclei_from_segmentation,
          &contours_of_orphans_from_segmentation,
          getImagePreprocessingParameters(),
          getStainedNucleiDetectionParameters(),
          getNonstainedNucleiDetectionParameters(), //dont matter<=stained==true
          getNucleiPreselectionParameters(),
          nuc_presel_assess_params_,
          declump, preview);

      auto const t_until = static_cast<double>(cv::getTickCount());
      QString const msg = QString("Nuclei detection:")
          + "  N:"  + QString::number(contours_nuclei.size())
          + "  SpN:"+ QString::number(contours_of_nuclei_from_segmentation.size())
          + "  nSp:"+ QString::number(contours_nonsegmentable.size())
          + "  SpO:"+ QString::number(contours_of_orphans_from_segmentation.size())
          + "  D:"  + QString::number(contours_of_debris.size())
        //+ "  (score = "  + QString::number(static_cast<long int>(1000*score)) + ")  "
          + "  ["  + QString::number(static_cast<int>((t_until - t_since)
                                     *1000/cv::getTickFrequency())) +" ms]";
      message_('I', msg, 9000);

    } else if (ui_->radioButtonNucDetParamsAuto->isChecked()) { // -------------

      if (not nucleiDetectionParameterScanSettingsAreCorrect()) { return; }

      // Invoke a separate thread to avoid GUI freeze.
      //
      // Note: connections should be disconnected at stopProgress___()
      //       (after emitting idle()).
      //
      connect(this, SIGNAL(busy()),
              this, SLOT(startProgressNucDetParamScan()), Qt::QueuedConnection);
      connect(this, SIGNAL(idle()),
              this, SLOT(stopProgressNucDetParamScan()), Qt::QueuedConnection);

      if (ui_->checkBoxNucDetDetectAutoup->isChecked()) {
        connect(this, SIGNAL(recheck(bool)),
                ui_->checkBoxNucDetDetectAutoup, SLOT(setChecked(bool)),
                Qt::QueuedConnection);
      }
      // TODO: kill the concurrent thread while running upon the program exit
      future_int_ = QtConcurrent::run(this, &MainWindow::nucDetImgPreprocParamScan,
                                      img_orig);
    } // if Manual, Auto -------------------------------------------------------

  } else if (detection_channel_is_cytoplasmic) { // ============================

    auto const t_since = static_cast<double>(cv::getTickCount());

    cv::Mat img_preproc_complete;
    cv::Mat img_preproc_partial;
    st::nonstained::preprocessImage(img_orig, img_preproc_complete,
        img_preproc_partial, getImagePreprocessingParameters(), preview);

    /*double const score =*/ st::detectNuclei(false,
        img_orig, img_preproc_complete, img_preproc_partial,
        &contours_nuclei,
        &contours_of_debris,
        &contours_nonsegmentable,
        &contours_of_nuclei_from_segmentation,
        &contours_of_orphans_from_segmentation,
        getImagePreprocessingParameters(),
        getStainedNucleiDetectionParameters(),
        getNonstainedNucleiDetectionParameters(),
        getNucleiPreselectionParameters(),
        nuc_presel_assess_params_,
        declump, preview);

    auto const t_until = static_cast<double>(cv::getTickCount());
    QString const msg = QString("Nuclei detection:")
        + "  N:"  + QString::number(contours_nuclei.size())
        + "  SpN:"+ QString::number(contours_of_nuclei_from_segmentation.size())
        + "  nSp:"+ QString::number(contours_nonsegmentable.size())
        + "  SpO:"+ QString::number(contours_of_orphans_from_segmentation.size())
        + "  D:"  + QString::number(contours_of_debris.size())
      //+ "  (score = "  + QString::number(static_cast<long int>(1000*score)) + ")  "
        + "  ["  + QString::number(static_cast<int>((t_until - t_since)
                                   *1000/cv::getTickFrequency())) +" ms]";
    message_('I', msg, 9000);
  }  // if cytoplasmic, nuclear ================================================

  // Adopting results to displayable data structures ---

  // (0/2) Clean-up:
  clearFrameNuclei_();
  if (ui_->checkBoxNucEditShowAnnotations->isChecked()) {
    clearFrameAnnotations_(true, false);
    makeFrameAnnotations_(true, false);
  }

  assert(current_nuclei_);

  // (1/2) Adopt nuclei:
  //
  for (auto const& contour: contours_nuclei) {// TODO LATER QT emplace
    current_nuclei_->push_back(
        NuclearOutline{contour,NuclearOutline::Provenience::kAutoSolo});
  }
  for (auto const& contour: contours_of_nuclei_from_segmentation) {// TODO LATER QT emplace
    current_nuclei_->push_back(
        NuclearOutline{contour,NuclearOutline::Provenience::kAutoFromSplit});
  }

  // (2/2) Adopt non-nuclei:
  //
  for (auto const& contour: contours_nonsegmentable) {
    current_nuclei_->push_back(
        NuclearOutline{contour,NuclearOutline::Provenience::kAutoImproperGroup});
  }
  for (auto const& contour: contours_of_orphans_from_segmentation) {
    current_nuclei_->push_back(
        NuclearOutline{contour,NuclearOutline::Provenience::kAutoImproperOrphan});
  }
  for (auto const& contour: contours_of_debris) {
    current_nuclei_->push_back(
        NuclearOutline{contour,NuclearOutline::Provenience::kAutoImproperDebris});
  }

  drawFrameNuclei_();

  // UI --
  //
  ui_->checkBoxNucEditMarkNonsplittable->setEnabled(true);
  ui_->checkBoxNucEditMarkNonsplittable->setChecked(true);
  ui_->checkBoxNucEditMarkSplitOrphans ->setEnabled(true);
  ui_->checkBoxNucEditMarkSplitOrphans ->setChecked(true);
  ui_->checkBoxNucEditMarkDebris       ->setEnabled(true);
  ui_->checkBoxNucEditMarkDebris       ->setChecked(true);
  //
  ui_->checkBoxNucEditShowAnnotations->setChecked(false);
  //
  if (ui_->actionHideMarking->isChecked()) {
    ui_->actionHideMarking->trigger(); // toggle off
  }
} // on_pushButtonNucDetDetect_clicked()
