// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "MainWindow.hpp"
#include "ui_MainWindow.h"
#include "GUI/Support/ZoomableGraphicsView.hpp"
#include "Image/ImageDecomposition.hpp"
#include <QStandardItemModel>

namespace {
const char* sAutoSolo                = "autodetect-solo";
const char* sAutoFromSplit           = "autodetect-from-split";
const char* sManual                  = "manual";
const char* sAutoImproperSplitOrphan = "autodetect-split-orphan";
const char* sAutoImproperGroup       = "autodetect-nonsplittable";
const char* sAutoImproperDebris      = "autodetect-debris";
}


void MainWindow::on_pushButtonNucEditClearAll_clicked()
{
  clearFrameNuclei_();
  clearFrameAnnotations_(true, false);
  if (current_perinuclei_ and not current_perinuclei_->empty()) {
    clearFramePerinuclei_();
    message_('I', "Removal of nuclei entailed removal of perinuclei");
  }
}


void MainWindow::on_pushButtonNucEditToggleAll_clicked()
{ for (auto& nuc: *current_nuclei_) { nuc.toggle(); } }


void MainWindow::on_pushButtonNucEditClearToggled_clicked()
{
  bool removed_any = false;
  for (int nuc_i=0, nucs_sz=current_nuclei_->size(); nuc_i<nucs_sz; ++nuc_i) {
    if (current_nuclei_->at(nuc_i).isToggled()) {
      current_nuclei_->removeAt(nuc_i);
      removed_any = true;
  } }

  if (removed_any) {
    if (ui_->checkBoxNucEditShowAnnotations->isChecked()) {
      clearFrameAnnotations_(true, false);
      makeFrameAnnotations_(true, false);
    }
  } else {
    message_('W', "No toggled nuclei to remove");
  }
}

void MainWindow::clearFrameNucleiByProvenience_(
                                      NuclearOutline::Provenience prov_to_clear)
{
  QVector<int> to_remove;
  int i = 0;
  for (auto& nuc: *current_nuclei_) {
    if (nuc.provenience() == prov_to_clear) {
      to_remove.push_back(i);
    }
    i++;
  }
  for (int ii=to_remove.size()-1; ii>=0; --ii) {
    current_nuclei_->removeAt(to_remove[ii]);
  }
  if (ui_->checkBoxNucEditShowAnnotations->isChecked()) {
    clearFrameAnnotations_(true, false);
    makeFrameAnnotations_(true, false);
  }
}


void MainWindow::on_pushButtonNucEditRemoveSplitOrphans_clicked()
{
  clearFrameNucleiByProvenience_(
      NuclearOutline::Provenience::kAutoImproperOrphan );
  redrawFrameOutlines_();
  if (ui_->checkBoxNucEditShowAnnotations->isChecked()) {
    clearFrameAnnotations_(true, false);
    makeFrameAnnotations_(true, false);
  }
  clearFramePerinuclei_();

  // UI --
  // disabling this button is not a good idea when a curr frame can be changed
  // => commented out below
  /*
  ui_->checkBoxNucEditMarkSplitOrphans->setChecked(false);
  ui_->checkBoxNucEditMarkSplitOrphans->setEnabled(false);
  ui_->labelNucEditMarkNonnuclei->setEnabled(
           ui_->checkBoxNucEditMarkNonsplittable->isEnabled()
        or ui_->checkBoxNucEditMarkSplitOrphans ->isEnabled()
        or ui_->checkBoxNucEditMarkDebris       ->isEnabled() );
  */
}

void MainWindow::on_pushButtonNucEditRemoveDebris_clicked()
{
  clearFrameNucleiByProvenience_(
      NuclearOutline::Provenience::kAutoImproperDebris );
  redrawFrameOutlines_();
  if (ui_->checkBoxNucEditShowAnnotations->isChecked()) {
    clearFrameAnnotations_(true, false);
    makeFrameAnnotations_(true, false);
  }
  clearFramePerinuclei_();

  // UI --
  // disabling this button is not a good idea when a curr frame can be changed
  // => commented out below
  /*
  ui_->checkBoxNucEditMarkDebris->setChecked(false);
  ui_->checkBoxNucEditMarkDebris->setEnabled(false);
  ui_->labelNucEditMarkNonnuclei->setEnabled(
           ui_->checkBoxNucEditMarkNonsplittable->isEnabled()
        or ui_->checkBoxNucEditMarkSplitOrphans ->isEnabled()
        or ui_->checkBoxNucEditMarkDebris       ->isEnabled() );
  */
}


void MainWindow::on_pushButtonNucEditExcludeOnBorders_clicked()
{
  // TODO DUPLICATION:31874474-7fd5-4cd5-b41a-0ca506c6ab30
  int w = -1, h = -1;
  auto const& items = scenes_.front()->items();
  for (auto it: items) {
    if (it->type() == QGraphicsPixmapItem::Type) {
      auto ppm = static_cast<QGraphicsPixmapItem*>(&*it);
      w = ppm->pixmap().width();
      h = ppm->pixmap().height();
      break;
  } }
  assert(w > 0);  assert(h > 0);

  auto remove_on_border = [&](QList<NuclearOutline>* nuc_container) {
    int i = 0;
    QVector<int> to_remove;
    for (auto& n: *nuc_container) {
      const int margin = 0;
      int on_border_n_pts = 0;
      for (auto const& pt: n.contour()) {
        // Coords have halves -- that's why below are additional ones:
        if (   pt.x() <= 1+margin or pt.x() >= w-1-margin
            or pt.y() <= 1+margin or pt.y() >= h-1-margin ) {
          on_border_n_pts++;
        }
      }
      // (below) PARAM CONSISTENCY:min_num_points_to_classify_as_on_border
      if (on_border_n_pts >= 1) { to_remove.push_back(i); }
      i++;
    } // for each contour

    for (int ii=to_remove.size()-1; ii>=0; --ii) {
      nuc_container->removeAt(to_remove[ii]);
    }
  };

  remove_on_border(current_nuclei_);

  if (ui_->checkBoxNucEditShowAnnotations->isChecked()) {
    clearFrameAnnotations_(true, false);
    makeFrameAnnotations_(true, false);
  }

  clearFramePerinuclei_();
}


void MainWindow::on_pushButtonNucEditSwitchOnBorders_clicked()
{
  // TODO DUPLICATION:31874474-7fd5-4cd5-b41a-0ca506c6ab30
  int w = -1, h = -1;
  auto const& items = scenes_.front()->items();
  for (auto it: items) {
    if (it->type() == QGraphicsPixmapItem::Type) {
      auto ppm = static_cast<QGraphicsPixmapItem*>(&*it);
      w = ppm->pixmap().width();
      h = ppm->pixmap().height();
      break;
  } }
  assert(w > 0);  assert(h > 0);

  auto toggle_on_border = [&](QList<NuclearOutline>* nuc_container) {
    for (auto& n: *nuc_container) {
      const int margin = 0;
      int on_border_n_pts = 0;
      for (auto const& pt: n.contour()) {
        // Coords have halves -- that's why below are additional ones:
        if (   pt.x() <= 1+margin or pt.x() >= w-1-margin
            or pt.y() <= 1+margin or pt.y() >= h-1-margin ) {
          on_border_n_pts++;
        }
      }
      // (below) PARAM CONSISTENCY:min_num_points_to_classify_as_on_border
      if (on_border_n_pts >= 1) { n.toggle(); }
    } // for each contour
  };

  toggle_on_border(current_nuclei_);

  if (ui_->pushButtonNucEditSwitchOnBorders->text() == "---") {
    ui_->pushButtonNucEditSwitchOnBorders->setText("- -");
  } else if (ui_->pushButtonNucEditSwitchOnBorders->text() == "- -") {
    ui_->pushButtonNucEditSwitchOnBorders->setText("---");
  }
}


void MainWindow::on_pushButtonNucEditExcludeInteriorDebris_clicked()
{
  // TODO DUPLICATION:31874474-7fd5-4cd5-b41a-0ca506c6ab30
  int w = -1, h = -1;
  auto const& items = scenes_.front()->items();
  for (auto it: items) {
    if (it->type() == QGraphicsPixmapItem::Type) {
      auto ppm = static_cast<QGraphicsPixmapItem*>(&*it);
      w = ppm->pixmap().width();
      h = ppm->pixmap().height();
      break;
  } }
  assert(w > 0);  assert(h > 0);

  int i = 0;
  QVector<int> to_remove;
  for (auto& n: *current_nuclei_) {

    bool const prov_ok = \
        (n.provenience() == NuclearOutline::Provenience::kAutoImproperDebris);
    if (not prov_ok) { i++; continue; }

    const int margin = 0;
    int on_border_n_pts = 0;
    for (auto const& pt: n.contour()) {
      // Coords have halves -- that's why below are additional ones:
      if (   pt.x() <= 1+margin or pt.x() >= w-1-margin
          or pt.y() <= 1+margin or pt.y() >= h-1-margin ) {
        on_border_n_pts++;
      }
    }
    // (below) PARAM CONSISTENCY:min_num_points_to_classify_as_on_border
    bool const interior = (on_border_n_pts == 0);
    if (not interior) { i++; continue; }

    to_remove.push_back(i);
    i++;
  } // for each contour

  for (int ii=to_remove.size()-1; ii>=0; --ii) {
    current_nuclei_->removeAt(to_remove[ii]);
  }

  if (ui_->checkBoxNucEditShowAnnotations->isChecked()) {
    clearFrameAnnotations_(true, false);
    makeFrameAnnotations_(true, false);
  }

  clearFramePerinuclei_();
}


void MainWindow::on_pushButtonNucEditExcludeInteriorOrphans_clicked()
{
  // TODO DUPLICATION:31874474-7fd5-4cd5-b41a-0ca506c6ab30
  int w = -1, h = -1;
  auto const& items = scenes_.front()->items();
  for (auto it: items) {
    if (it->type() == QGraphicsPixmapItem::Type) {
      auto ppm = static_cast<QGraphicsPixmapItem*>(&*it);
      w = ppm->pixmap().width();
      h = ppm->pixmap().height();
      break;
  } }
  assert(w > 0);  assert(h > 0);

  int i = 0;
  QVector<int> to_remove;
  for (auto& n: *current_nuclei_) {

    bool const prov_ok = \
        (n.provenience() == NuclearOutline::Provenience::kAutoImproperOrphan);
    if (not prov_ok) { i++; continue; }

    const int margin = 0;
    int on_border_n_pts = 0;
    for (auto const& pt: n.contour()) {
      // Coords have halves -- that's why below are additional ones:
      if (   pt.x() <= 1+margin or pt.x() >= w-1-margin
          or pt.y() <= 1+margin or pt.y() >= h-1-margin ) {
        on_border_n_pts++;
      }
    }
    // (below) PARAM CONSISTENCY:min_num_points_to_classify_as_on_border
    bool const interior = (on_border_n_pts == 0);
    if (not interior) { i++; continue; }

    to_remove.push_back(i);
    i++;
  } // for each contour

  for (int ii=to_remove.size()-1; ii>=0; --ii) {
    current_nuclei_->removeAt(to_remove[ii]);
  }

  if (ui_->checkBoxNucEditShowAnnotations->isChecked()) {
    clearFrameAnnotations_(true, false);
    makeFrameAnnotations_(true, false);
  }

  clearFramePerinuclei_();
}


void MainWindow::on_checkBoxNucEditShowAnnotations_toggled(bool checked)
{
  clearFrameAnnotations_(true, false);
  if (checked) { makeFrameAnnotations_(true, false); }
}

void MainWindow::on_checkBoxRgnEditShowAnnotations_toggled(bool checked)
{
  clearFrameAnnotations_(false, true);
  if (checked) { makeFrameAnnotations_(false, true); }
}


void MainWindow::on_checkBoxNucEditMarkNucleiSolo_toggled(bool /*checked*/)
{  redrawFrameOutlines_();  }

void MainWindow::on_checkBoxNucEditMarkNucleiSplit_toggled(bool /*checked*/)
{  redrawFrameOutlines_();  }

void MainWindow::on_checkBoxNucEditMarkDebris_toggled(bool /*checked*/)
{  redrawFrameOutlines_();  }

void MainWindow::on_checkBoxNucEditMarkNonsplittable_toggled(bool /*checked*/)
{  redrawFrameOutlines_();  }

void MainWindow::on_checkBoxNucEditMarkSplitOrphans_toggled(bool /*checked*/)
{  redrawFrameOutlines_();  }

void MainWindow::on_checkBoxNucEditMarkNucleiManual_toggled(bool /*checked*/)
{  redrawFrameOutlines_();  }


// - - - - -

static void combo_box_aux__current_index_changed_(Ui::MainWindow* ui,
                                                  QComboBox* cbox_this,
                                                  int cbox_this_index,
                                                  QSpinBox *sbox_this,
                                                  QComboBox* cbox_other)
{
  if (not (    cbox_this ->itemText(0) == "(none)"
           and cbox_other->itemText(0) == "(none)")) {
    return;
  }

  bool const none_selected = (cbox_this_index == 0);

  // set availability of the weight spin box corresp. to the current combo box
  sbox_this->setEnabled(not none_selected);

  // set availability of the selected channel in the other combo box
  auto const flags_available = (Qt::ItemIsSelectable | Qt::ItemIsEnabled);
  if (none_selected) {
    for (int i = 0; i<cbox_other->count(); i++) {
      bool const same_as_main = cbox_other->itemText(i) ==
                             ui->lineEditNucEditRefineNuccutChannelMain->text();
      auto* m = qobject_cast<QStandardItemModel const *>(cbox_other->model());
      if (same_as_main) {
        m->item(i)->setFlags(m->item(i)->flags() & ~flags_available);
      } else {
        m->item(i)->setFlags(m->item(i)->flags() |  flags_available);
    } }
  } else {
    for (int i = 0; i<cbox_other->count(); i++) {
      bool const same_as_subj = cbox_other->itemText(i)==cbox_this->currentText();
      bool const same_as_main = cbox_other->itemText(i)==
                             ui->lineEditNucEditRefineNuccutChannelMain->text();
      auto m = qobject_cast<QStandardItemModel const*>(cbox_other->model());
      if (same_as_subj or same_as_main) {
        m->item(i)->setFlags(m->item(i)->flags() & ~flags_available);
      } else {
        m->item(i)->setFlags(m->item(i)->flags() |  flags_available);
  } } }
  assert(cbox_this->currentText() != ui->lineEditNucEditRefineNuccutChannelMain
                                       ->text());
}


void MainWindow::on_comboBoxNucEditRefineNuccutChannelAux1_currentIndexChanged(
    int index)
{
  combo_box_aux__current_index_changed_(ui_,
      ui_->comboBoxNucEditRefineNuccutChannelAux1, index,
      ui_->spinBoxNucEditRefineNuccutChannelAux1,
      ui_->comboBoxNucEditRefineNuccutChannelAux2);
}


void MainWindow::on_comboBoxNucEditRefineNuccutChannelAux2_currentIndexChanged(
    int index)
{
  combo_box_aux__current_index_changed_(ui_,
      ui_->comboBoxNucEditRefineNuccutChannelAux2, index,
      ui_->spinBoxNucEditRefineNuccutChannelAux2,
      ui_->comboBoxNucEditRefineNuccutChannelAux1);
}


void MainWindow::on_pushButtonNucEditRefine_clicked()
{
  if (current_nuclei_->empty()) { message_('W', "No nuclei to refine."); return; }
  auto const t_since = static_cast<double>(cv::getTickCount());

  // -- (1/4) create a weighted multi-channel image
  auto const image_main = images_->image(
      ui_->lineEditNucEditRefineNuccutChannelMain->text().toStdString(),
      current_frame_i_);
  double const weight_main = ui_->spinBoxNucEditRefineNuccutChannelMain->value()/100.;
  bool const has_aux1 = ui_->comboBoxNucEditRefineNuccutChannelAux1
                           ->currentText()!="(none)";
  bool const has_aux2 = ui_->comboBoxNucEditRefineNuccutChannelAux2
                           ->currentText()!="(none)";
  auto const image_aux1 = has_aux1
    ? images_->image(ui_->comboBoxNucEditRefineNuccutChannelAux1
                        ->currentText().toStdString(), current_frame_i_)
    : cv::Mat{};
  double const weight_aux1 = has_aux1 ? ui_->spinBoxNucEditRefineNuccutChannelAux1
                                           ->value()/100. : 0.;
  auto const image_aux2 = has_aux2
    ? images_->image(ui_->comboBoxNucEditRefineNuccutChannelAux2
                        ->currentText().toStdString(), current_frame_i_)
    : cv::Mat{};
  double const weight_aux2 = has_aux2 ? ui_->spinBoxNucEditRefineNuccutChannelAux2
                                           ->value()/100. : 0.;
  cv::Mat const image3 = st::mergeWeighted(image_main, weight_main,
                                           image_aux1, weight_aux1,
                                           image_aux2, weight_aux2);
  // -- (2/4) collect nuclei data
  std::vector<st::nuccut::annotated_contour_t> nuclei_data;
  nuclei_data.reserve(current_nuclei_->size());
  std::size_t nuc_i = 0;
  for (auto const& nuc: qAsConst(*current_nuclei_)) {

    // (2a) consider toggledness
    if (    ui_->checkBoxNucEditRefineNuccutExcludeToggled->isChecked()
        and nuc.isToggled()) {
      ++nuc_i;
      continue;
    }

    // (2b) consider provenience
    bool refine = false;
    switch (nuc.provenience()) {
    case NuclearOutline::Provenience::kAutoSolo:
      { refine = ui_->checkBoxNucEditMarkNucleiSolo->isChecked(); }  break;
    case NuclearOutline::Provenience::kAutoFromSplit:
      { refine = ui_->checkBoxNucEditMarkNucleiSplit->isChecked(); }  break;
    case NuclearOutline::Provenience::kManual:
      { refine = ui_->checkBoxNucEditMarkNucleiManual->isChecked(); }  break;
    case NuclearOutline::Provenience::kAutoImproperGroup:
      { refine = ui_->checkBoxNucEditMarkNonsplittable->isChecked(); }  break;
    case NuclearOutline::Provenience::kAutoImproperOrphan:
      { refine = ui_->checkBoxNucEditMarkSplitOrphans->isChecked(); }  break;
    case NuclearOutline::Provenience::kAutoImproperDebris:
      { refine = ui_->checkBoxNucEditMarkDebris->isChecked(); }  break;
    case NuclearOutline::Provenience::kUnknown: { refine = false; }  break;
    } // switch
    assert(not nuc.cv_contour().empty() and nuc.area() > 0.f);
    if (refine) { nuclei_data.emplace_back(nuc_i, nuc.cv_contour(), nuc.area()); }

    ++nuc_i;
  } // for each current nucleus
  assert(nuc_i == AS_SIZE(current_nuclei_->size()));

  // -- (3/4) call refinement procedure
  double const delta = ui_->dspinBoxNucEditRefineNuccutParamEdgeImpassability->value();
  double const gamma = ui_->dspinBoxNucEditRefineNuccutParamColorSimilarity->value();
  bool const preview = ui_->checkBoxNucEditRefineNuccutPreview->isChecked();
  int const canny_low = ui_->spinBoxNucDetContourDetThresholdLow->value();
  float const canny_ratio = ui_->dspinBoxNucDetContourDetThresholdsRatio->value();
  st::contours_vec_t refined_contours = st::nuccut::refine(
      std::move(nuclei_data), image_main, canny_low, canny_ratio,
      image3, {weight_main, weight_aux1, weight_aux2},
      delta, gamma, preview);

  // -- (4/4) seize refinement results
  std::vector<std::size_t> nucs_to_remove;
  nuc_i = -1;
  for (st::contour_t const& c: refined_contours) {
    ++nuc_i;
    if (c.empty()) { nucs_to_remove.push_back(nuc_i); continue; }
    auto& cur_nuc = (*current_nuclei_)[nuclei_data[AS_SIZE(nuc_i)].index];
    cur_nuc.update(c);
    redrawNucleus_(cur_nuc);
  }
  std::sort(nucs_to_remove.begin(), nucs_to_remove.end(),
            std::greater<std::size_t>());
  for (std::size_t xnuc_i: nucs_to_remove) {
    current_nuclei_->erase(current_nuclei_->begin() + xnuc_i);
  }

  auto const t_until = static_cast<double>(cv::getTickCount());
  QString const msg = QString("Nuclei refinement finished")
      + "  ["  + QString::number(static_cast<int>((t_until - t_since)
                                 *1000/cv::getTickFrequency())) +" ms]";
  message_('I', msg, 9000);
}


// - - - - -


void MainWindow::on_comboBoxNucEditColorBy_currentIndexChanged(QString const& arg)
{
  if (arg == "by origin") {
    draw_frame_nuclei__ = &MainWindow::drawFrameNucleiByProvenience_;
  } else if ((arg == "by area") or (arg == "by solidity")) {
    draw_frame_nuclei__ = &MainWindow::drawFrameNucleiByProperty_;
  } else if (arg == "white") {
    draw_frame_nuclei__ = &MainWindow::drawFrameNucleiInWhite_;
  } else if (arg == "green") {
    draw_frame_nuclei__ = &MainWindow::drawFrameNucleiInGreen_;
  } else {
    // This should never happen.
    message_('W', "Cannot color contours by " + arg);
    return;
  }
  redrawFrameOutlines_();
}


void MainWindow::on_dspinBoxNucEditContourWidth_editingFinished()
{
  nucleus_contour_width_ = ui_->dspinBoxNucEditContourWidth->value();
  if (current_nuclei_ and not current_nuclei_->empty()) {
    for (auto& n: *current_nuclei_) {
       n.setDepictionsWidth(nucleus_contour_width_);
  } } // for, if

  /* OBSOLETE (worked OK before pernuclei were introduced)
  for (auto& scene: scenes_) {
    for (auto it : scene->items()) {
      if (it->type() == QGraphicsPolygonItem::Type) {
        QPen p = static_cast<QGraphicsPolygonItem*>(it)->pen();
        p.setWidthF(nucleus_contour_width_);
        static_cast<QGraphicsPolygonItem*>(it)->setPen(p);
  } } }
  */
}

void MainWindow::on_dspinBoxNucEditContourWidth_valueChanged(QString const& arg)
{ if (arg.size() > 0) { on_dspinBoxNucEditContourWidth_editingFinished(); } }


void MainWindow::on_lineEditNucEditGotoNuc_returnPressed()
{
  int const nuc_index = ui_->lineEditNucEditGotoNuc->text().toInt() - 1;
  if ((0 <= nuc_index) and (nuc_index < current_nuclei_->size())) {
    auto const& poly = current_nuclei_->at(nuc_index).contour();
    auto const poly_center = poly.boundingRect().toAlignedRect().center();
    for (auto& view: views_) { view->centerOn(poly_center);  }
  } else {
    message_('W', "No such nucleus");
    return;
  }
}


void MainWindow::toggleExternalHighlight(QGraphicsPolygonItem const* poly)
{
  for (auto& nuc: *current_nuclei_) {
    if (nuc.isDepictedBy(poly)) {
      nuc.setExternallyHighlighted(not nuc.isExternallyHighlighted());
      break;
  } }
}


void MainWindow::on_pushButtonNucEditMarkFromFile_clicked()
{
  if (highlited_nuclei_file_path_.isEmpty()) {
    message_('E', "Path to a file with indices of nuclei to be highlighted not set");
#ifdef Q_OS_WIN
    message_('I', "(Hint: To set the path, edit ShuttleTracker's settings "
                  "in Windows registry.)");
#else
    message_('I', "(Hint: To set the path, edit ShuttleTracker's settings.ini.)");
#endif
    return;
  }

  QSet<int> marked_in_file;
  QFile file{highlited_nuclei_file_path_};
  if (file.open(QFile::ReadOnly | QFile::Text)) {
    QTextStream in{&file};
    while (not in.atEnd()) { marked_in_file.insert(in.readLine().toInt()); }
  } else {
    message_('E', "Could not read nuclei to be highlighted from file "
                  "'" + highlited_nuclei_file_path_ + "'");
    return;
  }
  file.close();

  int n_with_highlight = 0;
  for (int nuc_i=0, s=SSIZE_PTR(current_nuclei_); nuc_i<s; nuc_i++) {
    auto& nuc = (*current_nuclei_)[nuc_i];
    int const nuc_name = nuc_i + 1;
    if (marked_in_file.contains(nuc_name)) {
      ++n_with_highlight;
      nuc.setExternallyHighlighted(true);
      marked_in_file.remove(nuc_name);
    } else {
      nuc.setExternallyHighlighted(false);
  } }
  if (not marked_in_file.empty()) {
    marked_in_file.values();
    if (n_with_highlight) {
      message_('W', "Some nuclei could not be highlighted");
    } else if (current_nuclei_->empty()) {
      message_('W', "No nuclei => no highlighted nuclei");
    } else {
      message_('W', "No highlighted nuclei");
    }
  } else {
    message_('I', "External markings: "
                  "(+): " + QString::number(n_with_highlight) + ", "
                  "(-): " + QString::number(  SSIZE_PTR(current_nuclei_)
                                            - n_with_highlight));
  }
}

// -- Nuclei persistence -------------------------------------------------------

bool MainWindow::checkIfNucleiContoursFileAvailable_()
{
  QString const fn = assembleNucleiContoursFilename_();
  if (fn.size() > 0) {
    bool const file_exists = QFile(fn).exists();
    ui_->pushButtonNucEditNucleiContoursLoad->setEnabled(file_exists);
    return file_exists;
  } else {
    ui_->pushButtonNucEditNucleiContoursLoad->setEnabled(false);
    ui_->pushButtonNucEditNucleiContoursSave->setEnabled(false);
    return false;
  }
}


QString MainWindow::assembleNucleiContoursFilename_()
{ return timepointFilenameEndingWith_("-nuclei.txt",
                                      "a file with nuclei contours"); }


void MainWindow::on_pushButtonNucEditNucleiContoursSave_clicked()
{
  QString const fn = assembleNucleiContoursFilename_();
  if (fn.size() == 0) {
    message_('E', "Could not save nuclei contours!");
    return;
  }
  QFile file(fn);
  if (file.open(QFile::WriteOnly | QFile::Text)) {
    QTextStream out(&file);
    int nuc_i = 0;
    for (auto& nuc: *current_nuclei_) {
      switch (nuc.provenience()) {
        case NuclearOutline::Provenience::kAutoSolo:
          out << "Nucleus " << ++nuc_i << '\n';
          out << "Origin: " << sAutoSolo << '\n';
          break;
        case NuclearOutline::Provenience::kAutoFromSplit:
          out << "Nucleus " << ++nuc_i << '\n';
          out << "Origin: " << sAutoFromSplit << '\n';
          break;
        case NuclearOutline::Provenience::kManual:
          out << "Nucleus " << ++nuc_i << '\n';
          out << "Origin: " << sManual << '\n';
          break;
        case NuclearOutline::Provenience::kAutoImproperGroup:
          out << "Improper-nucleus " << ++nuc_i << '\n';
          out << "Origin: " << sAutoImproperGroup << '\n';
          break;
        case NuclearOutline::Provenience::kAutoImproperOrphan:
          out << "Improper-nucleus " << ++nuc_i << '\n';
          out << "Origin: " << sAutoImproperSplitOrphan << '\n';
          break;
        case NuclearOutline::Provenience::kAutoImproperDebris:
          out << "Improper-nucleus " << ++nuc_i << '\n';
          out << "Origin: " << sAutoImproperDebris << '\n';
          break;
        case NuclearOutline::Provenience::kUnknown: // fall-through
        default: out << "(Error [NuclearOutline.Provenience])\n"; break;
      }
      out << "Toggled: " << (nuc.isToggled() ? "yes" : "no") << '\n';
      out << "Points: ";
      for (auto const& pt: nuc.contour()) {
        QPointF const pt_sh = pt - Outline::contour_shift;
        out << pt_sh.x() << ',' << pt_sh.y() << "; ";
      }
      out << "\n\n";
    } // for each nucleus
    file.close();
    message_('I', QString("Saved ") + QString::number(current_nuclei_->size())
                  + " nuclei contours in " + QFileInfo(file).fileName());
  } else {
    message_('E', "Could not save nuclei contours in "
                  + QFileInfo(file).fileName());
  }

  checkIfNucleiContoursFileAvailable_();
}


void MainWindow::on_pushButtonNucEditNucleiContoursLoad_clicked()
{
  on_pushButtonNucEditClearAll_clicked();

  QString const fn = assembleNucleiContoursFilename_();
  if (fn.isEmpty() or fn.isNull()) {
    message_('E', "Could not load nuclei contours!");
    return;
  }
  QFile file(fn);
  if (file.open(QFile::ReadOnly | QFile::Text)) {
    QTextStream in(&file);
    int nuc_i = 0;  // current nucleus number
    NuclearOutline::Provenience origin = NuclearOutline::Provenience::kUnknown;
    bool toggled = false;
    while (not in.atEnd()) {
      QString const line = in.readLine();
      if (line.length() > 0) {
        if (   line.startsWith("Nucleus ")
            or line.startsWith("Non-nucleus") // BACKWARD COMPATIBILITY
            or line.startsWith("Improper-nucleus") ) {
          nuc_i = line.split(" ").last().toInt();
        } else if (line.startsWith("Origin: ")) {
          QString s_origin = line.split(" ").last();
          if        (s_origin == QString{sAutoSolo}) {
            origin = NuclearOutline::Provenience::kAutoSolo;
          } else if (s_origin == QString{sAutoFromSplit}) {
            origin = NuclearOutline::Provenience::kAutoFromSplit;
          } else if (s_origin == QString{sManual}) {
            origin = NuclearOutline::Provenience::kManual;
          } else if (s_origin == QString{sAutoImproperGroup}) {
            origin=NuclearOutline::Provenience::kAutoImproperGroup;
          } else if (s_origin == QString{sAutoImproperSplitOrphan}) {
            origin = NuclearOutline::Provenience::kAutoImproperOrphan;
          } else if (s_origin == QString{sAutoImproperDebris}) {
            origin = NuclearOutline::Provenience::kAutoImproperDebris;
          } else {
            message_('E', "While reading-in nuclei: at nucleus #"
                          + QString::number(nuc_i) + " encountered "
                          "unexpected origin: " + s_origin);
          }
        } else if (line.startsWith("Toggled: ")) {
          QString line_no_head = line;
          line_no_head.remove(0,QString("Toggled: ").length());
          if (line_no_head.contains("n") or line_no_head.contains("0")) {
            toggled = false;
          } else { toggled = true; }
        } else if (line.startsWith("Points: ")) {
          // at this points all info has been collected, can construct
          QString line_no_head = line;
          line_no_head.remove(0, QString("Points: ").length());
          auto str_pts = line_no_head.split("; ", Qt::SkipEmptyParts);
          QVector<QPoint> pts;
          for (auto const& str_pt: str_pts) {
            int const x = str_pt.split(",").first().toInt();
            int const y = str_pt.split(",").last() .toInt();
            pts.append(QPoint(x,y));
          }

          current_nuclei_->push_back(NuclearOutline{QPolygon{pts}, origin});
          if (toggled) { current_nuclei_->back().toggle(); }

        } // if line starts with...
      } // if line length > 0
    } // while not at end
    file.close();
    message_('I', QString("Loaded ") + QString::number(current_nuclei_->size())
                  + " nuclei contours from " + QFileInfo(file).fileName());
  } else {
    message_('E', "Could not load nuclei contours from "
                  + QFileInfo(file).fileName());
  }

  // PAIRED with code at the top of the block
  drawFrameNuclei_();
  if (ui_->checkBoxNucEditShowAnnotations->isChecked()) {
    makeFrameAnnotations_(true,false);
  }
}
