// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "MainWindow.hpp"
#include "ui_MainWindow.h"
#include "GUI/Support/Outline.hpp"
#include "Image/Perinuclei.hpp"

namespace {
char const* sPerinucleiDerivationSettingsFilename = "perinuclei_derivation.ini";

char const* sInnerOffset  = "inner_offset";
char const* sWidth        = "width";
char const* sNeighAvoid   = "neighbors_avoidance";
char const* sOverlapAvoid = "overlap_avoidance";
char const* sMinPartArea  = "min_part_area";
char const* sBgAvoid          = "background_avoidance";
char const* sBgAvoidChannel   = "background_avoidance_channel";
char const* sBgAvoidExpansion = "background_avoidance_expansion";
char const* sBgAvoidSmoothing = "background_avoidance_smoothing";
}

void MainWindow::on_pushButtonPerinucCreate_clicked()
{
  assert(current_nuclei_);

  if (    current_regions_->empty()
      and ui_->checkBoxPerinucSettingsBgAvoid->isChecked()) {
    message_('W', "No regions that could be used as background samples");
    ui_->checkBoxPerinucSettingsBgAvoid->setChecked(false);
  }

  if (current_nuclei_->empty()) {
    message_('W', "No nuclei, no perinuclei");
    return;
  }

  if (not current_perinuclei_->empty()) { current_perinuclei_->clear(); }

  // tic
  auto t_since = static_cast<double>(cv::getTickCount());

  // extract cv-type nuclear contours for nuclear outlines
  std::vector<st::contour_t const*> nuclear_contours;
  nuclear_contours.reserve(static_cast<std::size_t>(current_nuclei_->size()));
  for (auto const& cn: qAsConst(*current_nuclei_)) {
    nuclear_contours.push_back(&cn.cv_contour());
  }

  // learn about the size of the image to work on
  auto const nuclear_channel = nuclearImageChannel_();
  assert(nuclear_channel != ImageChannel::UNDETERMINED);
  cv::Mat const& img_orig = images_->image(nuclear_channel, current_frame_i_);
  cv::Size const img_size = img_orig.size();

  // set parameters of perinuclei "derivation"
  st::PerinucleiParameters params;
  params.dilation_in      = ui_->spinBoxPerinucSettingInnerOffset->value();
  params.dilation_out     = ui_->spinBoxPerinucSettingWidth->value();
  params.neighbor_evasion = ui_->spinBoxPerinucSettingNeighAvoidance->value();
  params.overlap_evasion  = ui_->spinBoxPerinucSettingOverlapAvoidance->value();
  params.min_part_area    = ui_->spinBoxPerinucSettingMinArea->value();
  params.bg_evasion       = ui_->checkBoxPerinucSettingsBgAvoid->isChecked();
  params.bg_expansion     = static_cast<float>(
                            ui_->dspinBoxPerinucSettingBgAvoidExpansion->value());
  params.bg_smoothing     = ui_->spinBoxPerinucSettingBgAvoidSmoothing->value();
  {
    auto const text = ui_->comboBoxPerinucSettingsBgAvoidChannel->currentText();
    if      (text == "blue")   { params.bg_channel = ImageChannel::B; }
    else if (text == "green")  { params.bg_channel = ImageChannel::G; }
    else if (text == "red")    { params.bg_channel = ImageChannel::R; }
    else if (text == "cyan")   { params.bg_channel = ImageChannel::C; }
    else if (text == "magenta"){ params.bg_channel = ImageChannel::M; }
    else if (text == "yellow") { params.bg_channel = ImageChannel::Y; }
    else if (params.bg_evasion) {
      message_('E', "Unrecognized channel '" + text + "'for perinuclei "
                    "background avoidance");
    }
  } // scope

  // compute background characteristics, if necessary
  auto const meanf = [](std::vector<float> const& v) -> float
    {  return std::accumulate(v.begin(), v.end(), 0.f)/v.size(); };
  cv::Mat img_bg_mask = cv::Mat::zeros(cv::Size{0, 0}, CV_8UC1);
  if (params.bg_evasion) {
    assert(images_->has(params.bg_channel));
    cv::Mat const& img_orig2 = images_->image(params.bg_channel,
                                              current_frame_i_);
    assert(current_regions_ and (not current_regions_->empty()));
    std::size_t const n_regions = current_regions_->size();
    std::vector<float> intens_expects;  intens_expects.reserve(n_regions);
    std::vector<float> intens_deviats;  intens_deviats.reserve(n_regions);
    for (auto& r: *current_regions_) {
      st::QuantResults const q = r.quantify(img_orig2, /*calc_inertia*/false);
      intens_expects.push_back(static_cast<float>(q.median_intensity));
      intens_deviats.push_back(static_cast<float>(q.stddev_intensity));
    }
    st::makeBackgroundMask(img_orig2, &img_bg_mask, params,
        {meanf(intens_expects), std::max(meanf(intens_deviats), 0.1f)});
  } // if avoid background

  // generate perinuclei
  std::vector<st::Perinucleus> perinuclei;
  st::derivePerinuclei(nuclear_contours, &perinuclei, img_bg_mask, img_size, params);
  assert(perinuclei.size() == nuclear_contours.size());

  // toc
  auto const t_until = static_cast<double>(cv::getTickCount());
  auto const delta_t = (t_until - t_since)*1000/cv::getTickFrequency();
  QString const delta_t_s = QString::number(static_cast<int>(delta_t));
  auto const msg = QString("Perinuclei created [") + delta_t_s + " ms]";
  message_('I', msg, 9000);

  for (auto const& pn: perinuclei) {
    switch (pn.kind) {
    case st::Perinucleus::Kind::kCompleteC: {
      auto const kind = PerinuclearOutlines::Kind::kCompleteC;
      current_perinuclei_->push_back(PerinuclearOutlines{pn.contours,kind}); }
      break;
    case st::Perinucleus::Kind::kCompleteO: {
      auto const kind = PerinuclearOutlines::Kind::kCompleteO;
      current_perinuclei_->push_back(PerinuclearOutlines{pn.contours,kind}); }
      break;
    case st::Perinucleus::Kind::kMultipartC: {
      auto const kind = PerinuclearOutlines::Kind::kMultipartC;
      current_perinuclei_->push_back(PerinuclearOutlines{pn.contours,kind}); }
      break;
    case st::Perinucleus::Kind::kEmpty: {
      auto const kind = PerinuclearOutlines::Kind::kEmpty;
      current_perinuclei_->push_back(PerinuclearOutlines{pn.contours,kind}); }
      break;
    } //switch kind
  } // for each perinucleus

  drawFramePerinuclei_();

  // TMP -----------------------------------------------------------
#if 0
  cv::Mat img = cv::Mat::zeros(images_->imageSize(), CV_8UC1);
  for (auto const&pnuc: *current_perinuclei_) {
    if (not pnuc.cv_contours().empty()) {
      //for (pnuc.cv_contours())
      if (pnuc.kind() == PerinuclearOutlines::Kind::kCompleteO) {
        st::drawMask(img, pnuc.cv_contours().at(1));
        cv::Mat img2 = cv::Mat::zeros(images_->imageSize(), CV_8UC1);
        st::drawMask(img2, pnuc.cv_contours().at(0));
        img -= img2;
      } else {
        st::drawMask(img, pnuc.cv_contours().front());
    } }

  }
  for (auto const& nuc: *current_nuclei_) {
    st::drawMask(img, nuc.cv_contour());
    break;
  }

  cv::imshow("M", img);
#endif
}


void MainWindow::on_pushButtonPerinucClearAll_clicked()
{ clearFramePerinuclei_(); }


void MainWindow::on_dspinBoxPerinucContourWidth_valueChanged(double)
{
  perinucleus_contour_width_ = ui_->dspinBoxPerinucContourWidth->value();
  if (current_perinuclei_ and not current_perinuclei_->empty()) {
    for (auto& pn: *current_perinuclei_) {
      for (auto& ol: pn.outlines()) {
        ol.setDepictionsWidth(perinucleus_contour_width_);
  } } } // for, for, if
}


QString MainWindow::assemblePerinucDerivSettingsFilename_() const
{
  assert(not images_directory_.isEmpty());
  return images_directory_ + QDir::separator()
                           + sPerinucleiDerivationSettingsFilename;
}


bool MainWindow::checkIfPerinucDerivSettingsFileAvailable_()
{
  QString const fn = assemblePerinucDerivSettingsFilename_();
  if (fn.size() > 0) {
    bool const file_exists = QFile{fn}.exists();
    ui_->pushButtonPerinucSettingsLoad->setEnabled(file_exists);
    return file_exists;
  } else {
    ui_->pushButtonPerinucSettingsLoad->setEnabled(false);
    ui_->pushButtonPerinucSettingsSave->setEnabled(false);
    return false;
  }
}


void MainWindow::on_pushButtonPerinucSettingsSave_clicked()
{
  QString const fn = assemblePerinucDerivSettingsFilename_();
  if (fn.size() == 0) {
    message_('E', "Could not save perinuclei derivation settings!");
    return;
  }
  QFile file{fn};
  if (file.open(QFile::WriteOnly | QFile::Text)) {
    QTextStream out{&file};
    out << sInnerOffset << '='
        << ui_->spinBoxPerinucSettingInnerOffset->value() << '\n';
    out << sWidth << '='
        << ui_->spinBoxPerinucSettingWidth->value() << '\n';
    out << sNeighAvoid << '='
        << ui_->spinBoxPerinucSettingNeighAvoidance->value() << '\n';
    out << sOverlapAvoid << '='
        << ui_->spinBoxPerinucSettingOverlapAvoidance->value() <<'\n';
    out << sMinPartArea << '='
        << ui_->spinBoxPerinucSettingMinArea->value() << '\n';
    out << sBgAvoid << '='
        << static_cast<int>(ui_->checkBoxPerinucSettingsBgAvoid->isChecked())
        << '\n';
    out << sBgAvoidChannel << '='
        << ui_->comboBoxPerinucSettingsBgAvoidChannel->currentText() << '\n';
    out << sBgAvoidExpansion << '='
        << ui_->dspinBoxPerinucSettingBgAvoidExpansion->value() << '\n';
    out << sBgAvoidSmoothing << '='
        << ui_->spinBoxPerinucSettingBgAvoidSmoothing->value() << '\n';
    file.close();
    message_('I', "Saved perinuclei derivation settings to "
                  + QFileInfo{file}.fileName());
  }
  checkIfPerinucDerivSettingsFileAvailable_();
}


void MainWindow::on_pushButtonPerinucSettingsLoad_clicked()
{
  bool const should_auto_update = ui_->checkBoxPerinucDeriveAutoup->isChecked();
  ui_->checkBoxPerinucDeriveAutoup->setChecked(false);

  QString const fn = assemblePerinucDerivSettingsFilename_();
  if (fn.size() == 0) {
    message_('E', "Could not load perinuclei derivation settings!");
    return;
  }
  QFile file{fn};
  if (file.open(QFile::ReadOnly | QFile::Text)) {
    bool errors_encountered = false;
    QTextStream in{&file};
    while (not in.atEnd()) {
      QStringList const split_line = in.readLine().split('=');
      assert(split_line.size() == 2);
      QString const key = split_line.front();
      QString const value = split_line.back();
      if (key == QString{sInnerOffset}) {
        ui_->spinBoxPerinucSettingInnerOffset->setValue(value.toInt());
      } else if (key == QString{sWidth}) {
        ui_->spinBoxPerinucSettingWidth->setValue(value.toInt());
      } else if (key == QString{sNeighAvoid}) {
        ui_->spinBoxPerinucSettingNeighAvoidance->setValue(value.toInt());
      } else if (key == QString{sOverlapAvoid}) {
        ui_->spinBoxPerinucSettingOverlapAvoidance->setValue(value.toInt());
      } else if (key == QString{sMinPartArea}) {
        ui_->spinBoxPerinucSettingMinArea->setValue(value.toInt());
      } else if (key == QString{sBgAvoid}) {
        auto const value_as_bool = static_cast<bool>(value.toInt());
        ui_->checkBoxPerinucSettingsBgAvoid->setChecked(value_as_bool);
      } else if (key == QString{sBgAvoidChannel}) {
        bool found_and_set = false;
        for (int i=0; i<ui_->comboBoxPerinucSettingsBgAvoidChannel->count(); ++i) {
          if (ui_->comboBoxPerinucSettingsBgAvoidChannel->itemText(i) == value) {
            ui_->comboBoxPerinucSettingsBgAvoidChannel->setCurrentText(value);
            found_and_set = true;
            break;
        } }
        if (not found_and_set) {
          errors_encountered = true;
          message_('E', "Unrecognized or unavailable perinuclei "
                        "derivation channel '" + value + "'");
        }
      } else if (key == QString{sBgAvoidExpansion}) {
        ui_->dspinBoxPerinucSettingBgAvoidExpansion->setValue(value.toDouble());
      } else if (key == QString{sBgAvoidSmoothing}) {
        ui_->spinBoxPerinucSettingBgAvoidSmoothing->setValue(value.toInt());
      } else {
        message_('E', "Unrecognized perinuclei derivation setting '" + key + "'"
                      " in " + QFileInfo{file}.fileName());
      }
    } // while reading file
    file.close();

    if (not errors_encountered) {
      message_('I', "Perinuclei derivation settings loaded from "
                    + QFileInfo{file}.fileName());
    }
  } else {
    message_('E', "Could not load perinuclei derivation settings from "
                  + QFileInfo{file}.fileName() + "!"
                  + ((!QFileInfo{file}.exists())? " File does not exist.": ""));
  }

  ui_->checkBoxPerinucDeriveAutoup->setChecked(should_auto_update);
  maybe_PerinucDerive_autoclick();
}


void MainWindow::maybe_PerinucDerive_autoclick()
{
  if (ui_->checkBoxPerinucDeriveAutoup->isChecked()) {
    ui_->pushButtonPerinucCreate->click();
  }
}


void MainWindow::on_spinBoxPerinucSettingInnerOffset_editingFinished()
{  maybe_PerinucDerive_autoclick();  }
//
void MainWindow::on_spinBoxPerinucSettingInnerOffset_valueChanged(QString const& s)
{  if (s.size() > 0) {on_spinBoxPerinucSettingInnerOffset_editingFinished();}  }


void MainWindow::on_spinBoxPerinucSettingWidth_editingFinished()
{  maybe_PerinucDerive_autoclick();  }
//
void MainWindow::on_spinBoxPerinucSettingWidth_valueChanged(QString const& s)
{  if (s.size() > 0) { on_spinBoxPerinucSettingWidth_editingFinished(); }  }


void MainWindow::on_spinBoxPerinucSettingNeighAvoidance_editingFinished()
{  maybe_PerinucDerive_autoclick();  }
//
void MainWindow::on_spinBoxPerinucSettingNeighAvoidance_valueChanged(QString const& s)
{  if (s.size()>0) {on_spinBoxPerinucSettingNeighAvoidance_editingFinished();}  }


void MainWindow::on_spinBoxPerinucSettingOverlapAvoidance_editingFinished()
{  maybe_PerinucDerive_autoclick();  }
//
void MainWindow::on_spinBoxPerinucSettingOverlapAvoidance_valueChanged(QString const& s)
{  if(s.size()>0) {on_spinBoxPerinucSettingOverlapAvoidance_editingFinished();}  }


void MainWindow::on_spinBoxPerinucSettingMinArea_editingFinished()
{  maybe_PerinucDerive_autoclick();  }
//
void MainWindow::on_spinBoxPerinucSettingMinArea_valueChanged(QString const& s)
{  if (s.size() > 0) { on_spinBoxPerinucSettingMinArea_editingFinished(); }  }



void MainWindow::on_checkBoxPerinucSettingsBgAvoid_toggled(bool)
{  maybe_PerinucDerive_autoclick();  }

void MainWindow::on_comboBoxPerinucSettingsBgAvoidChannel_currentTextChanged(QString const&)
{  maybe_PerinucDerive_autoclick();  }

void MainWindow::on_dspinBoxPerinucSettingBgAvoidExpansion_valueChanged(QString const& s)
{  if (s.size()>0) {on_dspinBoxPerinucSettingBgAvoidExpansion_editingFinished();}  }
//
void MainWindow::on_dspinBoxPerinucSettingBgAvoidExpansion_editingFinished()
{  maybe_PerinucDerive_autoclick();  }

void MainWindow::on_spinBoxPerinucSettingBgAvoidSmoothing_valueChanged(QString const& s)
{  if(s.size()>0) { on_spinBoxPerinucSettingBgAvoidSmoothing_editingFinished();}  }
//
void MainWindow::on_spinBoxPerinucSettingBgAvoidSmoothing_editingFinished()
{  maybe_PerinucDerive_autoclick();  }
