// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "MainWindow.hpp"
#include "ui_MainWindow.h"
#include <boost/format.hpp>
#include <QRegularExpression>
#include <QLineEdit>
#include <limits>

namespace {
char const* sQuantFileHeaderInitialNuc = "nucleus_id,area,area_masked,"
    "eccentricity,is_toggled,center_x,center_y";
char const* sQuantFileHeaderInitialPer = "corresp_nucleus_id,area";
char const* sQuantFileHeaderInitialHal = "corresp_nucleus_id"; /* ,sum_weights */ // TODO
char const* sQuantFileHeaderInitialReg = "region_id,area";
char const* sQuantFileHeaderInitialImg = "area,area_masked";
}

void write_quant_file_header__(QTextStream& out,
                               char const* header_initial,
                               MainWindow::described_images_t const& desc_oimgs)
{
  out << header_initial;
  for (auto const& img_desc: desc_oimgs) {
    auto const& ch_desc = img_desc.second;
    out << "," << ch_desc << "_intensity_min";
    out << "," << ch_desc << "_intensity_max";
    out << "," << ch_desc << "_intensity_median";
    out << "," << ch_desc << "_intensity_quartile1";
    out << "," << ch_desc << "_intensity_quartile3";
    out << "," << ch_desc << "_intensity_mean";
    out << "," << ch_desc << "_intensity_stddev";
    out << "," << ch_desc << "_intensity_80_mean";
    out << "," << ch_desc << "_intensity_80_stddev";
    out << "," << ch_desc << "_intensity_q1_mean";
    out << "," << ch_desc << "_intensity_q2_mean";
    out << "," << ch_desc << "_intensity_q3_mean";
    out << "," << ch_desc << "_intensity_q4_mean";
    out << "," << ch_desc << "_intensity_t1_mean";
    out << "," << ch_desc << "_intensity_t2_mean";
    out << "," << ch_desc << "_intensity_t3_mean";
    out << "," << ch_desc << "_intensity_h1_mean";
    out << "," << ch_desc << "_intensity_h2_mean";
    out << "," << ch_desc << "_intensity_inertia";
    out << "," << ch_desc << "_intensity_sum";
  }
  out << '\n';
}


void write_quant_file_header_for_halos__(
    QTextStream& out,
    MainWindow::described_images_t const& desc_oimgs)
{
  out << sQuantFileHeaderInitialHal;
  for (auto const& [_, channel_description]: desc_oimgs) {
    for (auto const& scaling_name: st::distanceScalingsNames()) {
    out << "," << channel_description << '_'
               << QString::fromStdString(scaling_name);
  } }
  out << '\n';
}



bool write_quant_file_line__(QTextStream& out,
                             st::QuantResults const& q,
                             bool valid = true,
                             bool include_inertia = true)
{
  if (valid and q.median_intensity >= 0) {
    auto const formatted_line =
      include_inertia ? boost::format(",%d,%d,%d,%d,%d"
                                        ",%.3f,%.3f,%.3f,%.3f,%.3f"
                                          ",%.3f,%.3f,%.3f,%.3f,%.3f"
                                            ",%.3f,%.3f,%.3f,%.5f,%d")
                        % q.min_intensity
                        % q.max_intensity
                        % q.median_intensity
                        % q.quartile1_intensity
                        % q.quartile3_intensity
                          % q.mean_intensity
                          % q.stddev_intensity
                          % q.mean80_intensity
                          % q.stddev80_intensity
                          % q.q1_mean_intensity
                            % q.q2_mean_intensity
                            % q.q3_mean_intensity
                            % q.q4_mean_intensity
                            % q.t1_mean_intensity
                            % q.t2_mean_intensity
                              % q.t3_mean_intensity
                              % q.h1_mean_intensity
                              % q.h2_mean_intensity
                              % q.inertia
                              % q.sum_of_intensities
                      : boost::format(",%d,%d,%d,%d,%d"
                                        ",%.3f,%.3f,%.3f,%.3f,%.3f"
                                          ",%.3f,%.3f,%.3f,%.3f,%.3f"
                                            ",%.3f,%.3f,%.3f,NaN,%d")
                        % q.min_intensity
                        % q.max_intensity
                        % q.median_intensity
                        % q.quartile1_intensity
                        % q.quartile3_intensity
                          % q.mean_intensity
                          % q.stddev_intensity
                          % q.mean80_intensity
                          % q.stddev80_intensity
                          % q.q1_mean_intensity
                            % q.q2_mean_intensity
                            % q.q3_mean_intensity
                            % q.q4_mean_intensity
                            % q.t1_mean_intensity
                            % q.t2_mean_intensity
                              % q.t3_mean_intensity
                              % q.h1_mean_intensity
                              % q.h2_mean_intensity
                            //% q.inertia
                              % q.sum_of_intensities;
    out << QString::fromStdString(formatted_line.str());
    return true;
  } else {
    for (int i=0; i<15; ++i) { out << ',' << "NaN"; }
    return false;
  }
}

void MainWindow::on_actionQuantification_triggered()
{
  ui_->checkBoxNucQuantSaveCsv->setEnabled(not current_nuclei_->empty());
  ui_->checkBoxNucQuantSavePng->setEnabled(not current_nuclei_->empty());
  ui_->pushButtonNucQuantSaveAll->setEnabled(not current_nuclei_->empty());

  ui_->checkBoxQuantRgnSaveDat->setEnabled(not current_regions_->empty());
  ui_->pushButtonQuantRgnSaveAll->setEnabled(not current_regions_->empty());

  ui_->checkBoxQuantRgnSavePng->setEnabled(not current_regions_->empty());
  ui_->checkBoxQuantPerinucSaveCsv->setEnabled(not current_perinuclei_->empty());
  ui_->pushButtonQuantPerinucSaveAll->setEnabled(not current_perinuclei_->empty());
}

void MainWindow::convertChannelDescriptionToId_(QLineEdit* line_edit)
{ line_edit->setText(line_edit->text().replace(QRegularExpression{"[^a-zA-Z0-9]"}, "")); }

void MainWindow::on_lineEditQuantNucChannelG_editingFinished()
{ convertChannelDescriptionToId_(ui_->lineEditQuantNucChannelG);
  images_->setDescription(ImageChannel::G,
                          ui_->lineEditQuantNucChannelG->text().toStdString());
}

void MainWindow::on_lineEditQuantNucChannelR_editingFinished()
{ convertChannelDescriptionToId_(ui_->lineEditQuantNucChannelR);
  images_->setDescription(ImageChannel::R,
                          ui_->lineEditQuantNucChannelR->text().toStdString());
}

void MainWindow::on_lineEditQuantNucChannelB_editingFinished()
{ convertChannelDescriptionToId_(ui_->lineEditQuantNucChannelB);
  images_->setDescription(ImageChannel::B,
                          ui_->lineEditQuantNucChannelB->text().toStdString());
}

void MainWindow::on_lineEditQuantNucChannelC_editingFinished()
{ convertChannelDescriptionToId_(ui_->lineEditQuantNucChannelC);
  images_->setDescription(ImageChannel::C,
                          ui_->lineEditQuantNucChannelC->text().toStdString());
}

void MainWindow::on_lineEditQuantNucChannelM_editingFinished()
{ convertChannelDescriptionToId_(ui_->lineEditQuantNucChannelM);
  images_->setDescription(ImageChannel::M,
                          ui_->lineEditQuantNucChannelM->text().toStdString());
}

void MainWindow::on_lineEditQuantNucChannelY_editingFinished()
{ convertChannelDescriptionToId_(ui_->lineEditQuantNucChannelY);
  images_->setDescription(ImageChannel::Y,
                          ui_->lineEditQuantNucChannelY->text().toStdString());
}


std::map<st::ImageChannel, cv::Mat>
MainWindow::collectChannelImagesInSelectedChannels_() const
{
  QList<QPair<ImageChannel,QCheckBox*>> const selection = {
    { ImageChannel::G, ui_->checkBoxNucQuantChannelG },
    { ImageChannel::R, ui_->checkBoxNucQuantChannelR },
    { ImageChannel::B, ui_->checkBoxNucQuantChannelB },
    { ImageChannel::C, ui_->checkBoxNucQuantChannelC },
    { ImageChannel::M, ui_->checkBoxNucQuantChannelM },
    { ImageChannel::Y, ui_->checkBoxNucQuantChannelY }
  };

  std::map<st::ImageChannel, cv::Mat> collected;
  for (auto const& [channel, checkbox]: selection) {
    if (checkbox->isChecked()) {
      collected.emplace(channel, images_->image(channel, current_frame_i_));
  } } // if, for
  return collected;
}


void MainWindow::collectImagesAndDescriptionsInSelectedChannels_(
    described_images_t* orignal_images_and_their_descriptions) const
{
  assert(orignal_images_and_their_descriptions);

  QList<QPair<ImageChannel,QPair<QCheckBox*,QLineEdit*>>> const kinds_and_descs =
  {{ImageChannel::G,{ui_->checkBoxNucQuantChannelG,ui_->lineEditQuantNucChannelG}},
   {ImageChannel::R,{ui_->checkBoxNucQuantChannelR,ui_->lineEditQuantNucChannelR}},
   {ImageChannel::B,{ui_->checkBoxNucQuantChannelB,ui_->lineEditQuantNucChannelB}},
   {ImageChannel::C,{ui_->checkBoxNucQuantChannelC,ui_->lineEditQuantNucChannelC}},
   {ImageChannel::M,{ui_->checkBoxNucQuantChannelM,ui_->lineEditQuantNucChannelM}},
   {ImageChannel::Y,{ui_->checkBoxNucQuantChannelY,ui_->lineEditQuantNucChannelY}}};

  for (auto const& k_d: kinds_and_descs) {
    auto const kind = k_d.first;
    auto const selected = k_d.second.first->isChecked();
    if (selected) {
      auto const description = k_d.second.second->text();
      cv::Mat const image = images_->image(kind, current_frame_i_);
      orignal_images_and_their_descriptions->push_back({image, description});
  } } // if, for
}


QString MainWindow::assembleQuantsFilename_(QString const& core_prefix,
                                            QString const& file_ext)
{
  QString const ending =  "-" + ((core_prefix != "") ? (core_prefix + "_") : "")
                         + "quant." + file_ext.toLower();
  return timepointFilenameEndingWith_(ending, "a file with quantifications");
}


void MainWindow::quantifyNucleiAndSaveCsvFile_()
{
  if (current_nuclei_->empty()) {
    message_('E', "No nuclei to quantify");
    return;
  }

  described_images_t described_images;
  collectImagesAndDescriptionsInSelectedChannels_(&described_images);

  { // --- nuclei ---
    message_('I', "Nuclei to quantify: " + QString::number(current_nuclei_->size()));

    float const kDefaultNucleusMaskedAreaPlaceholder = 0.f;  // DEVEL:81b50b88

    QString const fn_dat = assembleQuantsFilename_("nuc", "CSV");
    if (fn_dat.isEmpty()) {
      message_('E', "Nuclei quantifications were not saved!");
      return;
    }
    QFile file(fn_dat);
    if (file.open(QFile::WriteOnly | QFile::Text)) {
      QTextStream out(&file);
      write_quant_file_header__(out, sQuantFileHeaderInitialNuc, described_images);
      //
      int ni = 1;
      assert(current_nuclei_);
      for (auto const& n: qAsConst(*current_nuclei_)) {
        auto const center = n.center();
        auto const vals = boost::format("%d,%.1f,%.1f,%.3f,%1d,%.2f,%.2f")
            % ni
            % n.area()
            % kDefaultNucleusMaskedAreaPlaceholder  // DEVEL:81b50b88
            % n.eccentricity()
            % static_cast<int>(n.isToggled())
            % center.x
            % center.y;
        out << QString::fromStdString(vals.str());
        bool could_not_quantify = false;
        for (auto const& imgo_desc: qAsConst(described_images)) {
          cv::Mat const& img_orig = imgo_desc.first;
          st::QuantResults const q = n.quantify(img_orig);
          could_not_quantify = not write_quant_file_line__(out, q);
        } // for each channel
        if (could_not_quantify) {
          message_('W', "Nucleus " + QString::number(ni) + " "
                        "cannot be quantified!");
        }
        out << '\n';
        ni++;
      } // for each nucleus
      //
      file.close();
      message_('I', "Quantifications of nuclei were saved to "
                    + QFileInfo(fn_dat).fileName());
    } else {
      message_('E', "Cannot save quantifications of nuclei to "
                    + QFileInfo(fn_dat).fileName());
    }
  } // nuclei

  { // --- halos ---
    if (ui_->checkBoxQuantNucSaveHalos->isChecked()) {

      QString const fn_dat = assembleQuantsFilename_("hal", "CSV");
      if (fn_dat.isEmpty()) {
        message_('E', "Perinuclear halos quantifications were not saved!");
        return;
      }
      QFile file(fn_dat);
      if (file.open(QFile::WriteOnly | QFile::Text)) {
        QTextStream out(&file);
        write_quant_file_header_for_halos__(out, described_images);

        std::vector<st::contour_t const*> nuclear_contours;
        nuclear_contours.reserve(current_nuclei_->size());
        for (auto const&n: qAsConst(*current_nuclei_)) {
          nuclear_contours.push_back(&n.cv_contour());
        }

        auto const t_since = static_cast<double>(cv::getTickCount());

        auto const all_vquants = st::quantifyInCorrespWeightedVoronois(
            collectChannelImagesInSelectedChannels_(), nuclear_contours);

        int ni = 0;
        for (auto const& nuc_vquants: all_vquants) {
          out << ++ni;
          for (auto const& [/*channel*/_, fnamed_vquants]: nuc_vquants) {
            for (auto const& [/*fname*/__, value]: fnamed_vquants) {
            auto const value_s = boost::format(",%.3f") % value;
            out << QString::fromStdString(value_s.str());
          } }
          out << '\n';
        }

        auto const t_until = static_cast<double>(cv::getTickCount());
        auto const dur_ms = QString::number(static_cast<int>((t_until - t_since)
                                            *1000/cv::getTickFrequency()));
        file.close();
        message_('I', "Quantifications of halos were saved to "
                      + QFileInfo(fn_dat).fileName() + "  ["  + dur_ms + " ms]");
      } else {
        message_('E', "Cannot save quantifications of halos to "
                      + QFileInfo(fn_dat).fileName());
      }
    } // if box checked
  } // halos
}


void MainWindow::quantifyPerinucleiAndSaveCsvFile_()
{
  if (current_perinuclei_->empty()) {
    message_('E', "No perinuclei to quantify");
    return;
  }

  described_images_t described_images;
  collectImagesAndDescriptionsInSelectedChannels_(&described_images);

  message_('I', "Perinuclei to quantify: "
                + QString::number(current_perinuclei_->size()));

  QString const fn_dat = assembleQuantsFilename_("per", "CSV");
  if (fn_dat.isEmpty()) {
    message_('E', "Perinuclei quantifications were not saved!");
    return;
  }
  QFile file(fn_dat);
  if (file.open(QFile::WriteOnly | QFile::Text)) {
    QTextStream out(&file);
    write_quant_file_header__(out, sQuantFileHeaderInitialPer, described_images);
    //
    int pni = 1;
    assert(current_perinuclei_);
    for (auto const& pn: qAsConst(*current_perinuclei_)) {
      auto const area = pn.area();
      auto id_area = boost::format("%d,%.1f") % pni % area;
      out << QString::fromStdString(id_area.str());
      bool could_not_quantify = false;
      for (auto const& imgo_desc: qAsConst(described_images)) {
        cv::Mat const& img_orig = imgo_desc.first;
        st::QuantResults const q = pn.quantify(img_orig);
        could_not_quantify = not write_quant_file_line__(out, q,
                                          /*valid*/area > 0.f, /*inertia*/false);
      } // for each channel
      if (could_not_quantify) {
        message_('W', "Perinucleus " + QString::number(pni) + " "
                      "cannot be quantified!");
      }
      out << '\n';
      pni++;
    } // for each perinucleus
    //
    file.close();
    message_('I', "Quantifications of perinuclei were saved to "
                  + QFileInfo(fn_dat).fileName());
  } else {
    message_('E', "Cannot save quantifications of perinuclei to "
                  + QFileInfo(fn_dat).fileName());
  }
}


void MainWindow::quantifyRegionsAndSaveCsvFile_()
{
  if (current_regions_->empty()) {
    message_('E', "No regions to quantify");
    return;
  }

  message_('I', "Regions to quantify: "
                + QString::number(current_regions_->size()));
  described_images_t described_images;
  collectImagesAndDescriptionsInSelectedChannels_(&described_images);

  QString const fn_dat = assembleQuantsFilename_("reg", "CSV");
  if (fn_dat.isEmpty()) {
    message_('E', "Region(s) quantifications were not saved!");
    return;
  }
  QFile file(fn_dat);
  if (file.open(QFile::WriteOnly | QFile::Text)) {
    QTextStream out(&file);
    write_quant_file_header__(out, sQuantFileHeaderInitialReg, described_images);
    //
    int ri = 1;
    assert(current_regions_);
    for (auto const& r: qAsConst(*current_regions_)) {
      auto const regid = QString("R") + QString::number(ri);
      auto id_area = boost::format("%s,%.1f") % regid.toStdString() % r.area();
      out << QString::fromStdString(id_area.str());
      bool could_not_quantify = false;
      for (auto const& imgo_desc: qAsConst(described_images)) {
        cv::Mat const& img_orig = imgo_desc.first;
        st::QuantResults const q = r.quantify(img_orig);
        could_not_quantify = not write_quant_file_line__(out, q);
      } // for each channel
      if (could_not_quantify) {
        message_('W', "Region " + regid + " cannot be quantified!");
      }
      out << '\n';
      ri++;
    } // for each region
    //
    file.close();
    message_('I', "Quantifications of regions were saved to "
                  + QFileInfo(fn_dat).fileName());
  } else {
    message_('E', "Cannot save quantifications of regions to "
                  + QFileInfo(fn_dat).fileName());
  }
}


void MainWindow::on_pushButtonQuantImgSaveAll_clicked()
{
  described_images_t described_images;
  collectImagesAndDescriptionsInSelectedChannels_(&described_images);

  long int const area = images_->imageSize().area(),
          masked_area = images_->calcMaskedArea(current_frame_i_);
  bool const only_non_masked = ui_->checkBoxQuantImgOnlyNonMasked->isChecked();
  auto const& masking_contours = images_->maskingContours(current_frame_i_);

  QString const fn_dat = assembleQuantsFilename_("img", "CSV");
  if (fn_dat.isEmpty()) {
    message_('E', "Image quantifications were not saved!");
    return;
  }
  QFile file(fn_dat);
  if (file.open(QFile::WriteOnly | QFile::Text)) {
    QTextStream out(&file);
    write_quant_file_header__(out, sQuantFileHeaderInitialImg, described_images);
    auto const vals = boost::format("%d,%d") % area % masked_area;
    out << QString::fromStdString(vals.str());
    for (auto const& imgo_desc: qAsConst(described_images)) {
    //char ch_lett = lett_imgo.first;
      cv::Mat const& img = imgo_desc.first;
      st::QuantResults const q = \
          only_non_masked ? st::quantifyImage(img, masking_contours)
                          : st::quantifyImage(img);
      write_quant_file_line__(out, q);
    }
    out << '\n';
    file.close();
    message_('I', "Quantifications of the image were saved to "
                  + QFileInfo(fn_dat).fileName());
  } else {
    message_('E', "Cannot save quantifications of the image to "
                        + QFileInfo(fn_dat).fileName());
  }
}


void MainWindow::savePixmapWithOptionalMarkings_(QString const& outline_type,
                                                 bool trim)
{
  if (trim) { trimScenesToBorder_(); }

  // set outlines
  bool has_external_highlight = false;
  if (outline_type == "nuclei" ) {
    for (auto const& nuc: const_cast<QList<NuclearOutline> const&>(*current_nuclei_)) {
      if (nuc.isExternallyHighlighted()) {
        has_external_highlight = true;
        break;
    } }
    // if (not has_external_highlight) { // commented out => do not re-draw
    //   drawFrameNuclei_ ();            //                  when no nuclei
    // }                                 //                  are marked
  } else if (outline_type == "regions") {
    drawFrameRegions_(); }
  else {
    message_('E', "[Internal] Don't know how to export the pixmap");
    return;
  }

  // set annotations
  clearFrameAnnotations_();
  if (outline_type == "nuclei") {
    makeFrameAnnotations_(ui_->checkBoxNucEditShowAnnotations->isChecked(),
                          false);
  } else if (outline_type == "regions") {
    makeFrameAnnotations_(false,
                          ui_->checkBoxRgnEditShowAnnotations->isChecked());
  }

  // export current scene of the overlay to a pixmap file
  QString const fn = assembleQuantsFilename_(outline_type.left(3), "PNG");
  if (fn.isEmpty()) {
    message_('E', "Image with annotations was not saved!");
  } else {
    if (saveOverlayToPixmapFile_(fn)) {
      message_('I', "Pixmap was exported to " + QFileInfo(fn).fileName());
    } else {
      message_('E', "Cannot export pixmap to " + QFileInfo(fn).fileName());
  } }

  // reset
  if (not (outline_type == "nuclei" and has_external_highlight)) {
    redrawFrameOutlines_();
  }
  maybeFrameAnnotate_();
}


void MainWindow::on_pushButtonNucQuantSaveAll_clicked()
{
  if (ui_->checkBoxNucQuantSaveCsv->isChecked()) {
    quantifyNucleiAndSaveCsvFile_();
  }
  if (ui_->checkBoxNucQuantSavePng->isChecked()) {
    savePixmapWithOptionalMarkings_("nuclei");
  }
}

void MainWindow::on_pushButtonQuantPerinucSaveAll_clicked()
{
  if (ui_->checkBoxQuantPerinucSaveCsv->isChecked()) {
    quantifyPerinucleiAndSaveCsvFile_();
  }
}

void MainWindow::on_pushButtonQuantRgnSaveAll_clicked()
{
  if (ui_->checkBoxQuantRgnSaveDat->isChecked()) {
    quantifyRegionsAndSaveCsvFile_();
  }
  if (ui_->checkBoxQuantRgnSavePng->isChecked()) {
    savePixmapWithOptionalMarkings_("regions");
  }
}


void MainWindow::on_checkBoxQuantImgFilesDat_toggled(bool checked)
{ ui_->pushButtonQuantImgSaveAll->setEnabled(checked); }

void MainWindow::on_checkBoxNucQuantSaveCsv_toggled(bool checked)
{
  bool const other_checked = ui_->checkBoxNucQuantSavePng->isChecked();
  ui_->pushButtonNucQuantSaveAll->setEnabled(
      (checked or other_checked) and (not current_nuclei_->empty()));
}

void MainWindow::on_checkBoxNucQuantSavePng_toggled(bool checked)
{
  bool const other_checked = ui_->checkBoxNucQuantSaveCsv->isChecked();
  ui_->pushButtonNucQuantSaveAll->setEnabled(
      (checked or other_checked) and (not current_nuclei_->empty()));
}


void MainWindow::on_checkBoxQuantRgnSavePng_toggled(bool checked)
{
  bool const other_checked = ui_->checkBoxQuantRgnSaveDat->isChecked();
  ui_->pushButtonQuantRgnSaveAll->setEnabled(
      (checked or other_checked) and (not current_regions_->empty()));
}

void MainWindow::on_checkBoxQuantRgnSaveDat_toggled(bool checked)
{
  bool const other_checked = ui_->checkBoxQuantRgnSavePng->isChecked();
  ui_->pushButtonQuantRgnSaveAll->setEnabled(
      (checked or other_checked) and (not current_regions_->empty()));
}


void MainWindow::on_pushButtonQuantSaveAll_clicked()
{
  using MW = MainWindow;
  std::map<QPushButton const*, void (MainWindow::*)()> const buttons_methods = {
      {ui_->pushButtonQuantImgSaveAll,    &MW::on_pushButtonQuantImgSaveAll_clicked},
      {ui_->pushButtonNucQuantSaveAll,    &MW::on_pushButtonNucQuantSaveAll_clicked},
      {ui_->pushButtonQuantPerinucSaveAll,&MW::on_pushButtonQuantPerinucSaveAll_clicked},
      {ui_->pushButtonQuantRgnSaveAll,    &MW::on_pushButtonQuantRgnSaveAll_clicked}};
  for (auto& bm: buttons_methods) {
    //if (bm.first->isEnabled() or batch_processing_) { // commented out => uncoditionally
      (this->*bm.second)();                             // because GUI, incl. buttons, is
    //}                                                 // disabled upon script execution
  }
}


// slot
void MainWindow::showQuantifications(QGraphicsPolygonItem const* poly)
{
  auto const quantify = [this](auto const& otl, auto const ch)
  {
    auto const& image = images_->image(ch, current_frame_i_);
    auto const q = otl.quantify(image, true);
    auto su = QString::number(q.sum_of_intensities);
    if (q.sum_of_intensities >= 10000)   { su = su.insert(su.length()-3, ' '); }
    if (q.sum_of_intensities >= 1000000) { su = su.insert(su.length()-7, ' '); }
    if (q.sum_of_intensities >= 1000000000){su= su.insert(su.length()-11,' '); }
    auto const quant_str = QString{}.asprintf(
        "Intensity in %s [%s]:   "
        "(μ, σ) = %.1f, %.1f    "
        "(q1, q2, q3) = %d, %d, %d    "
        "(min, max) = %d, %d    "
        "inertia = %.2f    "
        "sum = %s.    "
        "Geometry:   area = %.1f.",
        images_->description(ch).c_str(), st::toString(ch).c_str(),
        q.mean_intensity, q.stddev_intensity,
        q.quartile1_intensity, q.median_intensity, q.quartile3_intensity,
        q.min_intensity,  q.max_intensity,
        q.inertia,
        su.toStdString().c_str(),
        otl.area());
    message_('I', quant_str) ;
  };

  auto const find_and_quantify = [this, poly, quantify](auto const* container) -> bool
  {
    for (auto const& outline: *container) {
      if (outline.isDepictedBy(poly)) {
        auto const ch = channel(poly);
        if (ch != st::ImageChannel::UNDETERMINED) {
          quantify(outline, ch);
          return true;
        }
    } }
    return false;
  };

     find_and_quantify(current_regions_)
  or find_and_quantify(current_nuclei_)
  or find_and_quantify(current_perinuclei_);
}
