// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "MainWindow.hpp"
#include "ui_MainWindow.h"
#include <Image/ImageDecomposition.hpp>


void MainWindow::on_pushButtonRgnEditTesselate_clicked()
{
  assert(current_nuclei_);
  if (current_nuclei_->empty()) {
    message_('W', "No nuclei, no tesselation.");
    return;
  }

  on_pushButtonRgnEditReset_clicked();

  auto const img_sz = images_->imageSize();

  std::vector<cv::Point2f> points;
  for (auto const& n: qAsConst(*current_nuclei_)) {
    points.emplace_back(n.center());
  }
  st::contours_vec_t voronoi_contours;
  std::vector<std::pair<int, int>> voronoi_neighborhoods;
  st::partitionIntoVoronoiCellsCoarse(voronoi_contours, voronoi_neighborhoods,
      points, img_sz,
      (ui_->dspinBoxRgnEditVoronoiScaling->isEnabled())
          ? static_cast<float>(ui_->dspinBoxRgnEditVoronoiScaling->value()):1.f,
      (ui_->spinBoxRgnEditVoronoiErosion->isEnabled())
          ? ui_->spinBoxRgnEditVoronoiErosion->value() : 0);

  std::size_t vci = 0;
  for (auto& vc: voronoi_contours) {
    if (not vc.empty()) {
      current_regions_->push_back(Outline{vc});
    } else {
      int const x = static_cast<int>(points[vci].x),
                y = static_cast<int>(points[vci].y);
      st::contour_t const fix{{MAX(x - 1, 0),  y},
                              {x,  MAX(y - 1, 0)},
                              {((x + 1) <= img_sz.width)? x + 1 : x,  y},
                              {x,  ((y + 1) <= img_sz.height)? y + 1 : y} };
      current_regions_->push_back(Outline{fix});
      message_('W', "A tiny quick-fix Voronoi region has been inserted "
                    "for nucleus " + QString::number(vci + 1) + ". "
                    "Hint: decrease tesselation erosion parameter.", 9000);
    }
    ++vci;
  }
  message_('I', "Tesselation resulted in "
                + QString::number(current_regions_->size())+" Voronoi regions");

  if (ui_->checkBoxRgnEditVoronoiSaveNeighs->isChecked()) {
    if (current_nuclei_->size() == current_regions_->size()) {
      QString fn = assembleRegionsContoursFilename_();
      fn.replace("regions", "regions_neighborhoods");
      QFile file(fn);
      if (file.open(QFile::WriteOnly | QFile::Text)) {
        QTextStream out(&file);
        out << "region1,region2\n";
        for (auto const& nuclei_indices: voronoi_neighborhoods) {
          out << 'R' << nuclei_indices.first  + 1 << ','
              << 'R' << nuclei_indices.second + 1 << '\n';
      } }
      file.close();
      message_('I', "Neighborhoods of Voronoi regions have been saved "
                    "to " + QFileInfo(file).fileName());
    } else {
      message_('E', "Number of Voronoi regions is different than the number "
                    "of nuclei! (Not saving neighborhoods to a file.)");
  } } // if, if

  drawFrameRegions_();
  maybeFrameAnnotate_();
}


void MainWindow::on_checkBoxRgnEditVoronoiScaling_toggled(bool checked)
{  ui_->dspinBoxRgnEditVoronoiScaling->setEnabled(checked);  }

void MainWindow::on_checkBoxRgnEditVoronoiErosion_toggled(bool checked)
{  ui_->spinBoxRgnEditVoronoiErosion->setEnabled(checked);  }


QString MainWindow::assembleRegionsContoursFilename_()
{ return timepointFilenameEndingWith_("-regions.txt",
                                      "a file with image region contours"); }


void MainWindow::on_pushButtonRgnEditBackgroundSample_clicked()
{
  on_pushButtonRgnEditReset_clicked();

  auto const cs = images_->suggestBackgroundRegions(current_frame_i_,
      ui_->spinBoxRgnEditBackgroundSubdivs->value(),
      ui_->spinBoxRgnEditBackgroundCount->value(),
      ui_->checkBoxRgnEditBackgroundMergeAdjacent->isChecked());
  for (auto& c: cs) {
    assert(not c.empty());
    current_regions_->push_back(Outline{c});
  }

  drawFrameRegions_();
  maybeFrameAnnotate_();
  message_('I', QString("Background sampled in ") + QString::number(cs.size())
                + " place" + QString((cs.size() > 1) ? "s" : ""));
}


void MainWindow::on_pushButtonRgnEditRegionsSave_clicked()
{
  QString const fn = assembleRegionsContoursFilename_();
  if (fn.size() == 0) {
    message_('E', "Could not save image region contour(s)!");
    return;
  }
  QFile file(fn);
  if (file.open(QFile::WriteOnly | QFile::Text)) {
    QTextStream out(&file);
    int rgn_i = 0;
    for (auto const& r: qAsConst(*current_regions_)) {
      out << "Region R" << ++rgn_i << '\n';
      out << "Origin: manual\n";
      out << "Points: ";
      auto const& c = r.contour();
      for (auto const& pt: c) {
        QPointF const pt_sh = pt - Outline::contour_shift;
        out << pt_sh.x() << ',' << pt_sh.y() << "; ";
      }
      out << "\n\n";
    }
    file.close();
    message_('I', "Saved " + QString::number(rgn_i) + " image region contours"
                  " to " + QFileInfo(fn).fileName());
  }
  checkIfRegionsContoursFileAvailable_();
}


bool MainWindow::checkIfRegionsContoursFileAvailable_()
{
  auto fn = assembleRegionsContoursFilename_();
  if (fn.size() > 0) {
    bool const file_exists = QFile(fn).exists();
    ui_->pushButtonRgnEditRegionsLoad->setEnabled(file_exists);
    return file_exists;
  } else {
    ui_->pushButtonRgnEditRegionsLoad->setEnabled(false);
    ui_->pushButtonRgnEditRegionsSave->setEnabled(false);
    return false;
  }
}


void MainWindow::on_pushButtonRgnEditRegionsLoad_clicked()
{
  clearFrameRegions_();
  clearFrameAnnotations_( // TODO: create a "maybe" function for clearing and making annotz
      ui_->checkBoxNucEditShowAnnotations->isChecked(),
      ui_->checkBoxRgnEditShowAnnotations->isChecked());

  QString const fn = assembleRegionsContoursFilename_();
  if (fn.size() == 0) {
    message_('E', "Could not load image region contour(s)!");
    return;
  }
  QFile file(fn);
  if (file.open(QFile::ReadOnly | QFile::Text)) {
    QTextStream in(&file);
    int ci = 0;
    st::contours_vec_t contours;
    while (not in.atEnd()) {
      QString const line = in.readLine();
      if (line.length() > 0) {
        if (line.startsWith("Region ")) { continue; }
        if (line.startsWith("Origin: ")) { continue; }
        contours.push_back({});
        QString line_no_head = line;
        line_no_head.remove(0,QString("Points: ").length());
        QStringList const xsys = line_no_head.split("; ", Qt::SkipEmptyParts);
        for (auto const& _xy: xsys) {
          QStringList xy = _xy.split(',');
          contours.back().emplace_back(xy[0].toInt(),xy[1].toInt());
        }
        ci++;
      } // if non-empty line
    } // while not at end
    message_('I', "Loaded " + QString::number(ci) + " region contour"
                  + ((ci > 1) ? "s" : "") + " from " + QFileInfo(fn).fileName(),
             12000);

    if (not contours.empty()) {
      for (auto& c: contours) {
        QVector<QPoint> vp; for (auto const& cp: c) { vp.push_back({cp.x, cp.y}); }
        current_regions_->push_back(Outline{vp});
      }
      drawFrameRegions_();
      maybeFrameAnnotate_();
    }

  } else {
    message_('E', "Could not load image region contours from "
                  + QFileInfo(fn).fileName());
  }
}


void MainWindow::on_pushButtonRgnEditReset_clicked()
{
  clearFrameAnnotations_(false, true);
  clearFrameRegions_();
}
