// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "MainWindow.hpp"
#include "ui_MainWindow.h"
#include "GUI/Support/TracksModel.hpp"
#include "Image/Quantification.hpp"
#include <QGraphicsRectItem>
#include <QGraphicsLineItem>
#include <algorithm>
#include <unordered_set>
#include <cmath>

namespace {
char const* sTrackingSettingsFilename = "tracking.ini";

char const* sProx  = "similiarity_proximity_weight";
char const* sArea  = "similiarity_area_weight";
char const* sEccen = "similiarity_eccentricity_weight";
char const* sOrntn = "similiarity_orientation_weight";
char const* sISum  = "similiarity_intensity_sum_weight";
char const* sINrt  = "similiarity_intensity_inertia_weight";
char const* sIDstr = "similiarity_intensity_distribution_weight";
char const* sNormE = "norm_exponent";
char const* sRunMd = "running_median";
char const* sAreaRelDrop = "area_relative_drop";
}


// DEAD code
#if 0
bool MainWindow::current_frame_has_tracks_() const
{
  for (auto const& t: tracks_) {
    if (   (t.firstFrameIndex() <= static_cast<int>(current_frame_i_))
        and(static_cast<int>(current_frame_i_) <= t.lastFrameIndex())) {
      return true;
  } }
  return false;
}
#endif


void MainWindow::on_pushButtonTrackingStep_clicked()
{
  auto maybe_draw_partial_tracks = [this]() -> void
  {
    // inter-toolbox crosstalk:
    if (ui_->checkBoxTrackingAutoredraw->isChecked()) {
      ui_->checkBoxTracksEditTracesShowOnlySelected->setChecked(false);
      ui_->checkBoxTracksEditTracesShowStubs->setChecked(true);
      drawTracks_();
    //auto state1 = ui_->checkBoxTracksEditTracesShowOnlySelected->checkState();
    //auto state2 = ui_->checkBoxTracksEditTracesShowStubs->checkState();
    //ui_->checkBoxTracksEditTracesShowOnlySelected->setCheckState(state1); // emits!
    //ui_->checkBoxTracksEditTracesShowStubs->setCheckState(state2); // emits!
    } // if
  }; // lambda


  // -- Find nuclei already engaged in tracks
  //
  auto collect_nuclei_in_tracks = [this](long int              frame_i,
                                         QSet<Outline const*>* in_tracks,
                                         bool     include_stubs_and_last)
  {
    for (auto const& t: qAsConst(tracks_)) {
      if (t.firstFrameIndex() == frame_i) {
        if (t.nodesCount() > 1) {
          in_tracks->insert(t.initialOutline());
          continue;
        } else if (include_stubs_and_last) {
          assert(t.nodesCount() == 1);
          in_tracks->insert(t.initialOutline());
          continue;
        }
      } else if (    (frame_i > t.firstFrameIndex())
                 and (frame_i < t.lastFrameIndex()
                      or (    include_stubs_and_last
                          and (frame_i == t.lastFrameIndex())))) {
        for (auto& s: t.segments()) {
          if (s.frameIndex() == frame_i) {
            in_tracks->insert(s.outline());
            break;
        } }
      }
    } // for each track
  }; // lambda


  // k = 2 : weights = {1/2, 1/4}            * 2^2/(2^2 - 1)
  // k = 3 : weights = {1/2, 1/4, 1/8}       * 2^3/(2^3 - 1)
  // k = 4 : weights = {1/2, 1/4, 1/8, 1/16} * 2^4/(2^4 - 1), and so on
  auto predict_coms = [this](QHash<Outline const*,QPointF>* predicted_coms,     \
                             int n_frames_back,  double conservation_complement)
  {
    double const lambda = 5.;
    auto const previous_frame_index = current_frame_i_ - 1;
    for (auto const& t: qAsConst(tracks_)) {
      if (    (t.lastFrameIndex() == previous_frame_index)
          and (t.nodesCount() > 1)) {
        int const n_segments = t.segments().size();
        assert(t.nodesCount() == n_segments + 1);
        int const n_steps_back = std::min(n_frames_back, n_segments);
        auto seg_it = t.segments().crbegin();
        QPointF pred_com_shift{0., 0.};
        auto sum_weights = 0.;
        for (int i=0; i<n_steps_back; ++i) {
          auto const weight = std::exp(-lambda*i*conservation_complement
                                        /n_steps_back);
          auto const line = seg_it->line();
          pred_com_shift += QPointF{weight*line.dx(),weight*line.dy()};
          ++seg_it;
          sum_weights += weight;
        }
        QPointF const pred_com = t.segments().back().line().p2()+
                                 pred_com_shift*(1./sum_weights);
        assert(t.last());
        predicted_coms->insert(t.last(), pred_com);
      } // if track suitable
    } // for each track
  }; // lambda


  auto com_extrapolate = [](cv::Point2f const& com_orig,
                            QPointF const&     com_pred,
                            double             pred_contrib) -> cv::Point2f
  {
    auto const orig_contrib= 1 - pred_contrib;
    return cv::Point2f(com_orig.x*orig_contrib + com_pred.x()*pred_contrib,
                       com_orig.y*orig_contrib + com_pred.y()*pred_contrib);
  };


  // == Function body

  undrawTracks_();

  auto const t_since = static_cast<double>(cv::getTickCount());

  QSet<Outline const*> nuclei_curr_in_tracks;
  collect_nuclei_in_tracks(current_frame_i_, &nuclei_curr_in_tracks, true);


  // == Case: first frame

  if (current_frame_i_ == 0) {
    QList<QPair<Outline const*, Outline const*>> current_adj_frames_trackings;
    for (auto const& nuc: qAsConst(*current_nuclei_)) {
      if (not nuclei_curr_in_tracks.contains(&nuc)) {
        current_adj_frames_trackings.push_back(
            QPair<Outline const*, Outline const*>(nullptr,&nuc));
    } }
    if (current_adj_frames_trackings.empty()) {
      message_('W', "Cannot track: There are no available nuclear contours "
                    "in the current frame.");
      return;
    }
    QHash<QPair<Outline const*, Outline const*>, double>
        current_adj_frames_trackings_confidencies_empty;
    consolidateTracks_(current_adj_frames_trackings,
                        current_adj_frames_trackings_confidencies_empty);

    maybe_draw_partial_tracks();
    return;
  }


  // == Case: non-first frame

  assert(current_frame_i_ > 0);

  // -- get previous & current frame nuclei and images
  //
  auto const nuc_ch = nuclearImageChannel_();
  assert(nuc_ch != ImageChannel::UNDETERMINED);
  auto      & nuclei_prev= nuclei_[              current_frame_i_- 1];
  auto const& image_prev = images_->image(nuc_ch,current_frame_i_- 1);
  auto      & nuclei_curr= nuclei_[              current_frame_i_   ];
  auto const& image_curr = images_->image(nuc_ch,current_frame_i_   );

  if (nuclei_prev.empty()) {
    message_('W', "Cannot track: Nuclei contours in the previous frame "
                  "are missing.");
    return;
  }
  if (nuclei_curr.empty()) {
    message_('W', "Cannot track: Nuclei contours in the current frame "
                  "are missing.");
    return;
  }

  // -- calculate features of nuclei, in previous & current frame
  //
  bool const use_com_prediction =
      ui_->groupBoxTrackingNucPositionPrediction->isChecked();
  int const com_pred_memory =
      ui_->spinBoxTrackingNucPositionPredictionHistory->value();
  double const com_pred_conservation_cpt =
      1.0 - ui_->dspinBoxTrackingNucPositionPredictionConservation->value();
  double const com_pred_contribution =
      ui_->dspinBoxTrackingNucPositionPredictionContribution->value();
  QHash<Outline const*, QPointF> predicted_coms;
  if (use_com_prediction) {
    predict_coms(&predicted_coms, com_pred_memory, com_pred_conservation_cpt);
  }
  //
  QSet<Outline const*> nuclei_prev_in_tracks;
  collect_nuclei_in_tracks(current_frame_i_ - 1, &nuclei_prev_in_tracks, false);
  // 3rd arg above is false, because -- by not including tracks ends into
  // nuclei_prev_in_tracks -- we do not exclude them from similarity analysis
  // and consolidation.
  //
  std::map<Outline const*, QuantContourFeatures> feats_prev;
  for (auto const& pnuc: nuclei_prev) {
    if (not nuclei_prev_in_tracks.contains(&pnuc)) {
      auto pfeats = st::calculateFeatures(pnuc.cv_contour(), image_prev);
      pfeats.toggled = pnuc.isToggled();
      if (use_com_prediction and predicted_coms.contains(&pnuc)) {
        pfeats.center_of_mass = com_extrapolate(pfeats.center_of_mass,
            predicted_coms[&pnuc], com_pred_contribution);
      }
      feats_prev.emplace(&pnuc, std::move(pfeats));
  } }
  if (feats_prev.empty()) {
    message_('W', "Tracking: Previous time frame has no free nuclei.");
    return;
  }
  //
  std::map<Outline const*, QuantContourFeatures> feats_curr;
  for (auto const& cnuc: nuclei_curr) {
    if (not nuclei_curr_in_tracks.contains(&cnuc)) {
      auto cfeats = st::calculateFeatures(cnuc.cv_contour(), image_curr);
      cfeats.toggled = cnuc.isToggled();
      feats_curr.emplace(&cnuc, std::move(cfeats));
  } }
  //

  // -- calculate inter-frame nuclei similarities
  //
  st::SimilarityParameters w;  // relative _w_eights
  w.center_of_the_mass = ui_->dspinBoxTrackingNucSimilarityWeightProximity->value();
  w.surface_area = ui_->dspinBoxTrackingNucSimilarityWeightArea->value();
  w.eccentricity = ui_->dspinBoxTrackingNucSimilarityWeightEccentricity->value();
  w.orientation = ui_->dspinBoxTrackingNucSimilarityWeightOrientation->value();
  w.intensities_sum = ui_->dspinBoxTrackingNucSimilarityWeightIntensitySum->value();
  w.intensities_inertia = ui_->dspinBoxTrackingNucSimilarityWeightIntensityInertia->value();
  w.intensities_distribution=ui_->dspinBoxTrackingNucSimilarityWeightIntensityDistrib->value();

  double const dst_sc = ui_->dspinBoxTrackingNucDistanceScaling->value();
  std::vector<std::vector<double>> const similarities =
      st::calculateSimilarities(feats_prev, feats_curr, w, dst_sc);
  assert(not similarities.empty());

  // -- order nuclei pairs by similarity
  //
#if 0
  using scored_match_t = std::pair<double,std::pair<unsigned int,unsigned int>>;
  std::vector<scored_match_t> scored_matches;
  scored_matches.reserve(similarities.size()*similarities.front().size());
  for (int i1=0, s1=similarities.size(); i1<s1; ++i1) {
    for (int i2=0u, s2=similarities[i1].size(); i2<s2; ++i2) {
      scored_matches.push_back({similarities[i1][i2], {i1,i2}});
  } }
  std::sort( scored_matches.begin(),scored_matches.end(),
             [](scored_match_t const& ms1, scored_match_t const& ms2) -> bool
                 { return ms1.first > ms2.first; });  // better to the front
#else
  using match_t = std::pair<Outline const*, Outline const*>;
  using scored_match_t = std::pair<double, match_t>;
  std::vector<scored_match_t> scored_matches;
  scored_matches.reserve(feats_prev.size()*feats_curr.size());
  // TODO: This unwrapping is very clumsy (st::calculate_similarities
  //       should return a map {O*,O*} :-> score -- current vector is for
  //       performance reasons).
  unsigned int i1=0;
  for (auto const& featp: feats_prev) {
    unsigned int i2=0;
    for (auto const& featc: feats_curr) {
      auto const similarity = similarities[i1][i2];
      scored_matches.emplace_back(similarity, match_t{featp.first,featc.first});
      ++i2;
    }
    ++i1;
  }
  std::sort(scored_matches.begin(), scored_matches.end(),
            [](scored_match_t const& ms1, scored_match_t const& ms2) -> bool
              { return ms1.first > ms2.first; });  // better to the front
#endif
  QHash<QPair<Outline const*, Outline const*>, double>
      current_adj_frames_trackings_confidencies;

  // -- find corresponding nuclei (greedily)

  std::unordered_set<Outline const*> matched_prev_nucs,
                                     matched_curr_nucs;
  QList<QPair<Outline const*, Outline const*>> current_adj_frames_trackings;

  constexpr double min_median_d_sq = 10.*std::numeric_limits<double>::epsilon()
                                    *10.*std::numeric_limits<double>::epsilon() ;
  double const max_n_median_dists =
      ui_->dspinBoxTrackingNucDisplacementCutoffMedian->value();
  double const nuc_area_rel_drop_threshold =
      ui_->dspinBoxTrackingTrackBreakUponNucAreaDropRelative->value() / 100.;

  std::vector<double> sq_displacements;
  sq_displacements.reserve(std::max(feats_prev.size(),
                                      feats_curr.size()));


  auto calc_sq_displacement = [](cv::Point2f const& prev_com,
                                 cv::Point2f const& curr_com) -> double {
    auto const dx = prev_com.x - curr_com.x,
               dy = prev_com.y - curr_com.y;
    return static_cast<double>(dx*dx + dy*dy);
  };

  auto find_median_sq_displacement = [&sq_displacements]() -> double {
    assert(not sq_displacements.empty());
    std::sort(sq_displacements.begin(), sq_displacements.end());
    return sq_displacements[sq_displacements.size()/2];
  };


  // A fraction of the best-scoring matches is taken for initial estimates
  // of (the square of) the median displacement.

  // PARAM
  const int n_warmup_displacements = static_cast<int>(0.75*sqrt(scored_matches.size()));

  std::unordered_set<Outline const*> pre_matched_prev_nucs,
                                     pre_matched_curr_nucs;
  for (auto const& sm: scored_matches) {
    Outline const* prev_nuc = sm.second.first,
                 * curr_nuc = sm.second.second;
    if (   pre_matched_prev_nucs.count(prev_nuc)
        or pre_matched_curr_nucs.count(curr_nuc)) { continue; }
    double const sq_d = calc_sq_displacement(feats_prev[prev_nuc].center_of_mass,
                                             curr_nuc->center());
    sq_displacements.push_back(sq_d);
    if (sq_displacements.size() >= AS_SIZE(n_warmup_displacements)) {
      break;
  } }
  double median_sq_d = find_median_sq_displacement();


  for (int sm_i=0, sm_sz=scored_matches.size(); sm_i < sm_sz; ++sm_i) {

    auto const& sm = scored_matches[sm_i];

    Outline const* prev_nuc = sm.second.first,
                 * curr_nuc = sm.second.second;
    if (   matched_prev_nucs.count(prev_nuc)
        or matched_curr_nucs.count(curr_nuc)) {
      continue;  // already matched
    }

    assert(feats_prev.count(prev_nuc));
    double const d_sq = calc_sq_displacement(feats_prev[prev_nuc].center_of_mass,
                                             curr_nuc->center());

    // consider updating the median displacement
    if (not (sq_displacements.size() % 10)) { // PARAM => median update freq
      median_sq_d = find_median_sq_displacement();
    }

    if ((median_sq_d > min_median_d_sq) and (d_sq > median_sq_d*max_n_median_dists
                                                   *median_sq_d*max_n_median_dists)) {
      continue; // too far away => current pair will not be matched
    }

    // do not to re-insert the squared displacements collected during warm-up
    if (sm_i >= n_warmup_displacements) {
      sq_displacements.push_back(d_sq);
    }

    if (ui_->groupBoxTrackingTrackBreakUponNucAreaDropRelative->isChecked()) {
      assert(feats_prev[prev_nuc].surface_area > 0.f);
      double const nuc_area_rel_drop =
            (static_cast<double>(  feats_curr[curr_nuc].surface_area
                                 - feats_prev[prev_nuc].surface_area)
            /static_cast<double>(  feats_prev[prev_nuc].surface_area));
      if (nuc_area_rel_drop >= nuc_area_rel_drop_threshold) {
        continue;
    } }

    matched_prev_nucs.insert(prev_nuc);
    matched_curr_nucs.insert(curr_nuc);

    auto const& confidence = sm.first;
    current_adj_frames_trackings.push_back({prev_nuc, curr_nuc});
    current_adj_frames_trackings_confidencies.insert(
        current_adj_frames_trackings.back(), confidence);
  } // for each scored match

  // -- mark track beginnings and ends with a nullptr component

  for (auto const& featp: feats_prev) {
    if (not matched_prev_nucs.count(featp.first)) {
      current_adj_frames_trackings.push_back({featp.first, nullptr});
      current_adj_frames_trackings_confidencies.insert(
          current_adj_frames_trackings.back(), 0.);
  } }

  for (auto const& featc: feats_curr) {
    if (not matched_curr_nucs.count(featc.first)) {
      current_adj_frames_trackings.push_back({nullptr, featc.first});
      current_adj_frames_trackings_confidencies.insert(
          current_adj_frames_trackings.back(), 0.);
  } }


  consolidateTracks_(current_adj_frames_trackings,
                     current_adj_frames_trackings_confidencies);

  auto const t_until = static_cast<double>(cv::getTickCount());
  auto const dur_ms = QString::number(static_cast<int>((t_until - t_since)
                                      *1000/cv::getTickFrequency()));
  message_('I', "Tracking step has finished  [" + dur_ms + " ms]", 9000);

  maybe_draw_partial_tracks();

#if 0
  for (auto& prev_curr: current_adj_frames_trackings) {

    assert((prev_curr.first != nullptr) or (prev_curr.second != nullptr));

    // -- 1/2 -- matched pairs
    if ( (prev_curr.first != nullptr) and (prev_curr.second != nullptr)) {

      Outline& prev_nuc = *prev_curr.first;
      Outline& curr_nuc = *prev_curr.second;
      cv::Point2f const \
          curr_com = st::center_of_mass_(curr_nuc.cv_contour()),
          prev_com = st::center_of_mass_(prev_nuc.cv_contour());

      QColor color( (int)(255*(0.333 + 0.667*rand()/(double)RAND_MAX)),
                    (int)(255*(0.333 + 0.667*rand()/(double)RAND_MAX)),
                    (int)(255*(0.333 + 0.667*rand()/(double)RAND_MAX)), 242);
      {
        QPen pen( color, nucleus_contour_width_, \
                  curr_nuc.is_complete() ? Qt::SolidLine : Qt::DotLine,
                  Qt::RoundCap, Qt::RoundJoin);
        for (auto ps: scenes_) {
          auto *gp = new QGraphicsPolygonItem(curr_nuc.contour());
          ps->addItem(gp);
          gp->setPen(pen);
        //gp->setFlag(QGraphicsItem::ItemIsSelectable);
        //gp->setCursor(QCursor(Qt::PointingHandCursor));
          curr_nuc.depicted_by(gp);
        }
      }{
        QPen pen(color.darker(133), nucleus_contour_width_,
                 prev_nuc.is_complete() ? Qt::SolidLine : Qt::DotLine,
                 Qt::RoundCap, Qt::RoundJoin);
        for (auto ps: scenes_) {
          auto *gp = new QGraphicsPolygonItem(prev_nuc.contour());
          ps->addItem(gp);
          gp->setPen(pen);
        //gp->setFlag(QGraphicsItem::ItemIsSelectable);
        //gp->setCursor(QCursor(Qt::PointingHandCursor));
          prev_nuc.depicted_by(gp);
        }
      }{
       QPen pen(color, nucleus_contour_width_,
                Qt::SolidLine, Qt::RoundCap, Qt::RoundJoin);
       for (auto ps: scenes_) {
         auto* line = new QGraphicsLineItem( prev_com.x,prev_com.y,
                                             curr_com.x,curr_com.y);
         line->setPen(pen);
         line->setFlag(QGraphicsItem::ItemIsSelectable);
         ps->addItem(line);
       }
      }

    // 2a/2 -- beginning
    } else if ((prev_curr.first == nullptr) and (prev_curr.second != nullptr)) {

      Outline& curr_nuc = *prev_curr.second;
      QPen pen(Qt::green,
               // PARAM consistency TRACK_BEGINNING_END_CONTOUR_SCALING
               1.75*nucleus_contour_width_,
               curr_nuc.is_complete() ? Qt::SolidLine : Qt::DotLine,
               Qt::RoundCap, Qt::RoundJoin);
      for (auto ps: scenes_) {
        auto *gp = new QGraphicsPolygonItem(curr_nuc.contour());
        ps->addItem(gp);
        gp->setPen(pen);
        gp->setFlag(QGraphicsItem::ItemIsSelectable);
      //gp->setCursor(QCursor(Qt::PointingHandCursor));
        curr_nuc.depicted_by(gp);
      }

    // 2b/2 -- end
    } else if ((prev_curr.first != nullptr) and (prev_curr.second == nullptr)) {

      Outline& prev_nuc = *prev_curr.first;
      QPen pen(Qt::red,
               // PARAM consistency TRACK_BEGIN_END_CONTOUR_SCALING
               1.75*nucleus_contour_width_,
               prev_nuc.is_complete() ? Qt::SolidLine : Qt::DotLine,
               Qt::RoundCap, Qt::RoundJoin);
      for (auto ps: scenes_) {
        auto *gp = new QGraphicsPolygonItem(prev_nuc.contour());
        ps->addItem(gp);
        gp->setPen(pen);
        gp->setFlag(QGraphicsItem::ItemIsSelectable);
      //gp->setCursor(QCursor(Qt::PointingHandCursor));
        prev_nuc.depicted_by(gp);
      }
    } else {
      std::cerr << "Warning: Do not know how to handle a NULL--NULL "
                   "tracking segment." << std::endl;
    }

  } // for each tracking
#endif
} // on_pushButtonTrackingStep_clicked()



void MainWindow::consolidateTracks_(
    QList<QPair<Outline const*, Outline const*>>        const& adj_trs,
    QHash<QPair<Outline const*, Outline const*>, double> const& confidencies)
{
  QHash<Outline const*, size_t> lasts;
  size_t track_i = 0;
  for (auto const& track: qAsConst(tracks_)) {
    auto const p_last = track.last();
    if (p_last != nullptr) {
      lasts.insert(p_last, track_i);
    }
    ++track_i;
  }

  // QHash<Track const*, QPair<Outline*,Outline*>const*> imap; // intermediate map

  for (auto const& tr: adj_trs) {
    Outline const* prev_nuc = tr.first;
    Outline const* curr_nuc = tr.second;
    if (prev_nuc != nullptr and lasts.contains(prev_nuc)) {
      if (curr_nuc != nullptr) {
        assert(lasts[prev_nuc] < static_cast<size_t>(tracks_.size()));
        Track& track = tracks_[static_cast<int>(lasts[prev_nuc])];
        track.append( const_cast<Outline*>(curr_nuc),
                      static_cast<int>(current_frame_i_));
        if (confidencies.contains(tr)) {
          track.increaseConfidence(confidencies[tr]);
      } }
    } else {
      assert((prev_nuc == nullptr) or (not lasts.contains(prev_nuc)));
      if (curr_nuc != nullptr) {
        // if: discontinuity then: prev_nuc != nullptr
        tracks_.push_back(Track(const_cast<Outline*>(curr_nuc),
                                current_frame_i_, false));
        tracks_.back().setSelfCleanup(true);
    } }
  } // for

  if (not confidencies.empty()) { // then: sort by confidence, in descend~ order
    std::sort(tracks_.begin(), tracks_.end(),
        [](Track const& ta, Track const& tb)
          { if (ta.nodesCount() == tb.nodesCount()) {
              return ta.confidence() > tb.confidence();
            } else {
              return (ta.nodesCount() > tb.nodesCount());
          }});
  } else { // then: sort by length, in descending order
    std::sort( tracks_.begin(),tracks_.end(),
        [](Track const& ta, Track const& tb)
          { return (ta.nodesCount() > tb.nodesCount()); });
  }

  resetTracksModel_();
}


void MainWindow::resetTracksModel_()
{
  if (tracks_model_) {
    QItemSelectionModel *slxn_model =
        ui_->tableViewTracksEdit->selectionModel();
    disconnect(slxn_model,SIGNAL(currentChanged(QModelIndex,QModelIndex)),
               this,      SLOT(boldenSelectedTrack_(QModelIndex,QModelIndex)));
  }

  tracks_model_.reset(new TracksModel(&tracks_, n_frames_));
  assert(tracks_model_.get());
  assert(ui_->tableViewTracksEdit);
  ui_->tableViewTracksEdit->setModel(tracks_model_.get());

  QItemSelectionModel* slxn_model = ui_->tableViewTracksEdit->selectionModel();
  assert(slxn_model);
  connect(slxn_model,SIGNAL(currentChanged(QModelIndex,QModelIndex)),
          this, SLOT(boldenSelectedTrack_(QModelIndex,QModelIndex)));

  ui_->tableViewTracksEdit->resizeColumnsToContents();
  ui_->tableViewTracksEdit->resizeRowsToContents();
}


// -- File I/O -------------------------------------------------------------

QString MainWindow::assembleTrackingSettingsFilename_() const
{
  assert(not images_directory_.isEmpty());
  return images_directory_ + QDir::separator() + sTrackingSettingsFilename;
}


bool MainWindow::checkIfTrackingSettingsFileAvailable_()
{
  QString const fn = assembleTrackingSettingsFilename_();
  if (fn.size() > 0) {
    bool const file_exists = QFile(fn).exists();
    ui_->pushButtonTrackingSettingsLoad->setEnabled(file_exists);
    return file_exists;
  } else {
    ui_->pushButtonTrackingSettingsLoad->setEnabled(false);
    ui_->pushButtonTrackingSettingsSave->setEnabled(false);
    return false;
  }
}


void MainWindow::on_pushButtonTrackingSettingsLoad_clicked()
{
  QString const fn = assembleTrackingSettingsFilename_();
  if (fn.size() == 0) {
    message_('E', "Could not load tracking settings!");
    return;
  }
  QFile file(fn);
  if (file.open(QFile::ReadOnly | QFile::Text)) {
    QTextStream in(&file);
    while (not in.atEnd()) {
      QString const line = in.readLine();
      QStringList const key_value = line.split('=');
      QString const k = key_value.front(),  v = key_value.back();
      double const dv = v.toDouble();
      if (       k == QString{sProx})  {
        ui_->dspinBoxTrackingNucSimilarityWeightProximity->setValue(dv);
      } else if (k == QString{sArea})  {
        ui_->dspinBoxTrackingNucSimilarityWeightArea->setValue(dv);
      } else if (k == QString{sEccen}) {
        ui_->dspinBoxTrackingNucSimilarityWeightEccentricity->setValue(dv);
      } else if (k == QString{sOrntn}) {
        ui_->dspinBoxTrackingNucSimilarityWeightOrientation->setValue(dv);
      } else if (k == QString{sISum})  {
        ui_->dspinBoxTrackingNucSimilarityWeightIntensitySum->setValue(dv);
      } else if (k == QString{sINrt})  {
        ui_->dspinBoxTrackingNucSimilarityWeightIntensityInertia->setValue(dv);
      } else if (k == QString{sIDstr}) {
       ui_->dspinBoxTrackingNucSimilarityWeightIntensityDistrib->setValue(dv);
      } else if (k == QString{sNormE}) {
        ui_->dspinBoxTrackingNucDistanceScaling->setValue(dv);
      } else if (k == QString{sRunMd}) {
        ui_->dspinBoxTrackingNucDisplacementCutoffMedian->setValue(dv);
      } else if (k == QString{sAreaRelDrop}) {
        ui_->dspinBoxTrackingTrackBreakUponNucAreaDropRelative->setValue(dv);
      } else {
        message_('W', "Unrecognized tracking setting: '" + k + "'");
      }
    } // while reading the file
    message_('I', "Tracking settings loaded from "+ QFileInfo(file).fileName());
  } else {
    message_('E', "Could not load tracking settings from "
                  + QFileInfo(file).fileName() + "!"
                  + ((!QFileInfo(file).exists())? " File does not exist.": ""));
  }
}


void MainWindow::on_pushButtonTrackingSettingsSave_clicked()
{
  QString const fn = assembleTrackingSettingsFilename_();
  if (fn.size() == 0) {
    message_('E', "Could not save tracking settings!");
    return;
  }
  QFile file(fn);
  if (file.open(QFile::WriteOnly | QFile::Text)) {
    QTextStream out(&file);
    out << sProx <<'='<< ui_->dspinBoxTrackingNucSimilarityWeightProximity
                            ->value() <<'\n';
    out << sArea <<'='<< ui_->dspinBoxTrackingNucSimilarityWeightArea
                            ->value() <<'\n';
    out << sEccen<<'='<< ui_->dspinBoxTrackingNucSimilarityWeightEccentricity
                            ->value() <<'\n';
    out << sOrntn<<'='<< ui_->dspinBoxTrackingNucSimilarityWeightOrientation
                            ->value() <<'\n';
    out << sISum <<'='<< ui_->dspinBoxTrackingNucSimilarityWeightIntensitySum
                            ->value() <<'\n';
    out << sINrt <<'='<< ui_->dspinBoxTrackingNucSimilarityWeightIntensityInertia
                            ->value() <<'\n';
    out << sIDstr<<'='<< ui_->dspinBoxTrackingNucSimilarityWeightIntensityDistrib
                            ->value() <<'\n';
    out << sNormE<<'='<< ui_->dspinBoxTrackingNucDistanceScaling
                            ->value() <<'\n';
    out << sRunMd<<'='<< ui_->dspinBoxTrackingNucDisplacementCutoffMedian
                            ->value() <<'\n';
    out << sAreaRelDrop << '='
        << ui_->dspinBoxTrackingTrackBreakUponNucAreaDropRelative ->value()
        <<'\n';
    file.close();

    message_('I', "Tracking settings stored in " + QFileInfo(file).fileName());

  } else {
    message_('E', "Could not store tracking settings in "
                  + QFileInfo(file).fileName());
  }

  checkIfTrackingSettingsFileAvailable_();
}
