// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "MainWindow.hpp"
#include "ui_MainWindow.h"
#include <QDirIterator>
#include <algorithm>

namespace {
char const* sTracksFileName = "tracks.csv";
char const* sTracksMetaFileName = "tracks_meta.csv";
}

// Note tracks splitting is handled by item_removed() (--> MW_editing.cpp).


// -- Trace Properties box ----------------------------------------------------


// (Unless there are more color-by properties, that's all here.)
void MainWindow::on_comboBoxTracksEditTraceColorBy_activated(int)
{  redrawTracks_();  }

void MainWindow::on_dspinBoxTracksEditTraceWidthSelected_editingFinished()
{  redrawTracks_();  }

void MainWindow::on_dspinBoxTracksEditTraceWidthNonselected_editingFinished()
{  redrawTracks_();  }

void MainWindow::on_dspinBoxTracksEditTraceWidthSelected_valueChanged(double)
{  redrawTracks_();  }

void MainWindow::on_dspinBoxTracksEditTraceWidthNonselected_valueChanged(double)
{  redrawTracks_();  }

void MainWindow::on_checkBoxTracksEditTracesShowOnlySelected_toggled(bool)
{  redrawTracks_();  }


// -- Trace Clipping box ------------------------------------------------------

void MainWindow::on_checkBoxTracksEditTraceClippingLookBack_toggled(bool checked)
{
  ui_->spinBoxTracksEditTraceClippingLookBack->setEnabled(checked);
  redrawTracks_();
}

void MainWindow::on_checkBoxTracksEditTraceClippingLookAhead_toggled(bool checked)
{
  ui_->spinBoxTracksEditTraceClippingLookAhead->setEnabled(checked);
  redrawTracks_();
}


void MainWindow::on_spinBoxTracksEditTraceClippingLookBack_editingFinished()
{  redrawTracks_();  }

void MainWindow::on_spinBoxTracksEditTraceClippingLookBack_valueChanged(int)
{  redrawTracks_();  }

void MainWindow::on_spinBoxTracksEditTraceClippingLookAhead_editingFinished()
{  redrawTracks_();  }

void MainWindow::on_spinBoxTracksEditTraceClippingLookAhead_valueChanged(int)
{  redrawTracks_();  }

void MainWindow::on_checkBoxTracksEditTracesShowStubs_toggled(bool)
{  redrawTracks_();  }



// ----------------------------------------------------------------------------


// slot
void MainWindow::boldenSelectedTrack_(QModelIndex const& selected,
                                      QModelIndex const& deselected)
{
  if (selected.isValid()) {
    assert(selected.row() < tracks_.size());
    current_track_ = &tracks_[selected.row()];
    assert(current_track_);
    current_track_->setTraceWidth(
        ui_->dspinBoxTracksEditTraceWidthSelected->value());
    current_track_->setTrackedNucleiContourWidth(
        ui_->dspinBoxTracksEditTraceWidthSelected->value());

    if (ui_->checkBoxTracksEditTracesShowOnlySelected->isChecked()) {
      if (not current_track_->hasDepictions()) {  // draw
        drawTrack__(current_track_);
    } }
  }
  if (deselected.isValid()) {
    assert(deselected.row() < tracks_.size());
    tracks_[deselected.row()].setTraceWidth(
        ui_->dspinBoxTracksEditTraceWidthNonselected->value());
    tracks_[deselected.row()].setTrackedNucleiContourWidth(
        ui_->dspinBoxNucEditContourWidth->value());

    if (ui_->checkBoxTracksEditTracesShowOnlySelected->isChecked()) {
      tracks_[deselected.row()].clearDepictions();  // undraw
    }
  }
}


// slot
void MainWindow::tracksViewRemoveAt(QPoint const& p)
{
  auto const tracks_model_i = ui_->tableViewTracksEdit->indexAt(p);
  if (tracks_model_i.isValid()) {
    auto const track_i = tracks_model_i.row();
    assert(track_i < tracks_.size());
    tracks_[track_i].setTrackedNucleiContourWidth(
        ui_->dspinBoxNucEditContourWidth->value());  // unbolden nucleus
    current_track_ = nullptr;
    tracks_.removeAt(track_i);
    resetTracksModel_();
  }
}


// slot
void MainWindow::lineSelected(QGraphicsLineItem const* li)
{
  int ti = 0;
  for (auto const& t: qAsConst(tracks_)) {
    if (t.isDepictedBy(li)) {
      ui_->tableViewTracksEdit->selectRow(ti);
      return;
    }
    ti++;
  } // for tracks
}


// slot
void MainWindow::rectSelected(QGraphicsRectItem const* ri)
{
  int ti = 0;
  for (auto const& t: qAsConst(tracks_)) {
    if (t.isDepictedBy(ri)) {
      ui_->tableViewTracksEdit->selectRow(ti);
      return;
    }
    ti++;
  } // for tracks
}


void MainWindow::lineLineSelected(QGraphicsLineItem const* line1,
                                  QGraphicsLineItem const* line2)
{
  Track *pt1 = nullptr, *pt2 = nullptr;
  for (auto& t: tracks_) {
    if (not pt1) { if (t.isDepictedBy(line1)) { pt1 = &t; } }
    if (not pt2) { if (t.isDepictedBy(line2)) { pt2 = &t; } }
    if (pt1 and pt2) { break; }
  }
  if (pt1 == nullptr or pt2 == nullptr) {
    message_('E', "[Internal] Cannnot find tracks to which selected line(s) "
                  "belong!");
    return;
  }
  assert(pt1);
  assert(pt2);

  if (pt1 == pt2) {
    message_('W', "Cannot merge segments of the same track");
    return;
  }
  assert(pt1 != pt2);

  assert(pt1->nodesCount() > 1);
  assert(pt2->nodesCount() > 1);

  int t1si = 0;
  auto const& pt1_segments = pt1->segments();
  for (auto const& s: pt1_segments)
    { if (s.isDepictedBy(line1)) { break; } else { ++t1si; } }
  int t2si = 0;
  auto const& pt2_segments = pt2->segments();
  for (auto const& s: pt2_segments)
    { if (s.isDepictedBy(line2)) { break; } else { ++t2si; } }
  //
  if ((t1si > 0) and (t1si < pt1->segments().size() - 1)) {
    message_('W', "Cannot merge a middle track segment");
    return;
  }
  if ((t2si > 0) and (t2si < pt2->segments().size() - 1)) {
    message_('W', "Cannot merge a middle track segment");
    return;
  }

  auto const merged_msg = Track::merge(pt1,pt2,&tracks_);
  if (merged_msg.first) {
    resetTracksModel_(); // TODO: partial update
    current_track_ = &tracks_.back();
    redrawTracks_();      // TODO: partial redraw
  } else {
    message_('W', QString::fromStdString(merged_msg.second));
  }
}


void MainWindow::lineRectSelected(QGraphicsLineItem const* line,
                                  QGraphicsRectItem const* rect)
{
  Track *pt1 = nullptr, *pt2 = nullptr;
  for (auto& t: tracks_) {
    if (not pt1) { if (t.isDepictedBy(line)) { pt1 = &t; } }
    if (not pt2) { if (t.isDepictedBy(rect)) { pt2 = &t; } }
    if (pt1 and pt2) { break; }
  }

  if (pt1 == nullptr or pt2 == nullptr) {
    message_('E', "[Internal] Cannnot find tracks to which selected line(s) "
                  "belong!");
    return;
  }
  assert(pt1);
  assert(pt2);

  if (pt1 == pt2) {
    message_('E', "[Internal == this should never happen]: "
                  "Cannot merge a segment and an element of the same track");
    return;
  }
  assert(pt1 != pt2);
  assert(pt1->nodesCount()  > 1);
  assert(pt2->nodesCount() == 1);

  int t1si = 0;
  auto const& pt1_segments = pt1->segments();
  for (auto const& s: pt1_segments)
    { if (s.isDepictedBy(line)) { break; } else { ++t1si; } }
  if ((t1si > 0) and (t1si < pt1->segments().size() - 1)) {
    message_('W', "Cannot merge a middle track segment");
    return;
  }

  auto const merged_msg = Track::merge(pt1,pt2,&tracks_);
  if (merged_msg.first) {
    resetTracksModel_();  // TODO: partial update
    current_track_ = &tracks_.back();
    redrawTracks_();      // TODO: partial redraw
  } else {
    message_('W', QString::fromStdString(merged_msg.second));
  }
}


void MainWindow::rectLineSelected(QGraphicsRectItem const* rect,
                                  QGraphicsLineItem const* line)
{
  lineRectSelected(line, rect);
}


void MainWindow::rectRectSelected(QGraphicsRectItem const* rect1,
                                  QGraphicsRectItem const* rect2)
{
  Track *pt1 = nullptr, *pt2 = nullptr;
  for (auto & t: tracks_) {
    if (not pt1) { if (t.isDepictedBy(rect1)) { pt1 = &t; } }
    if (not pt2) { if (t.isDepictedBy(rect2)) { pt2 = &t; } }
    if (pt1 and pt2) { break; }
  }

  if (pt1 == nullptr or pt2 == nullptr) {
    message_('E', "[Internal] Cannnot find tracks to which selected line(s) "
                  "belong!");
    return;
  }
  assert(pt1);
  assert(pt2);

  if (pt1 == pt2) {
    message_('E', "[Internal == this should never happen]: "
                  "Cannot merge elements of the same track");
    return;
  }
  assert(pt1 != pt2);
  assert(pt1->nodesCount() == 1);
  assert(pt2->nodesCount() == 1);

  auto const merged_msg = Track::merge(pt1,pt2,&tracks_);
  if (merged_msg.first) {
    resetTracksModel_();  // TODO: partial update
    current_track_ = &tracks_.back();
    redrawTracks_();      // TODO: partial redraw
  } else {
    message_('W', QString::fromStdString(merged_msg.second));
  }
}


// slot
void MainWindow::on_pushButtonTracksEditRemoveAll_clicked()
{
  clearAllTracks_();

  redrawFrameOutlines_(); // a bit of overshot to just unbolden a potentially
                          // selected nucleus
}


void MainWindow::on_pushButtonTracksEditRemoveNonrevised_clicked()
{
  QMutableListIterator<Track> it(tracks_);
  while (it.hasNext()) {  if (not it.next().isRevised()) { it.remove(); }  }
  current_track_ = nullptr;
  resetTracksModel_();

  redrawFrameOutlines_(); // a bit overshot, just to unbolden a nucleus
  redrawTracks_();
}


void MainWindow::on_pushButtonTracksEditSave_clicked()
{
  // Create a bimap: outline ptr <-> outline index.
  QList<QMap<Outline const*,int>> frames_o2i; // outline-to-index
  QList<QMap<int,Outline const*>> frames_i2o; // index-to-outline
  for (int frame_i=0; frame_i<n_frames_; ++frame_i) {
    frames_o2i.push_back(QMap<Outline const*,int>());
    frames_i2o.push_back(QMap<int,Outline const*>());
    int nuc_i = 0;
    int const nuc_i_offset = 1;
    for (auto const& n: qAsConst(nuclei_[frame_i])) {
      frames_o2i.back().insert(&n,nuc_i + nuc_i_offset);
      frames_i2o.back().insert(nuc_i + nuc_i_offset,&n);
      ++nuc_i;
  } }

  // Create the tracks directory, if does not exists. If it exists, clean it.
  QString const tracks_file_name = images_directory_ + QDir::separator()
                                   + sTracksFileName;
  QFile tracks_file(tracks_file_name);
  if (not tracks_file.open(QFile::WriteOnly | QFile::Text)) {
    message_('E', "Could not save tracks to "
                  + QFileInfo(tracks_file).fileName() + ".");
    return;
  }

  QString const tracks_meta_file_name = images_directory_ + QDir::separator()
                                        + sTracksMetaFileName;
  QFile tracks_meta_file(tracks_meta_file_name);
  if (not tracks_meta_file.open(QFile::WriteOnly | QFile::Text)) {
    message_('E', "Could not save tracks metadata to "
                  + QFileInfo(tracks_meta_file).fileName() + ".");
    return;
  }

  QTextStream out(&tracks_file);
  QTextStream out_meta(&tracks_meta_file);

  out      << "track_index,time_point_index,nucleus_index\n";
  out_meta << "track_index,track_revision_status\n";

  int track_i = 0;
  for (Track const& track: qAsConst(tracks_)) {

    if (ui_->checkBoxTracksEditSaveOnlyRevisedTracks->isChecked())
      { if (not track.isRevised()) { continue; } }

    auto const frame_i = track.firstFrameIndex();
    out << track_i <<','<< frame_i <<','
        << frames_o2i[frame_i][track.initialOutline()] <<'\n';
    for (auto const& s: track.segments()) {
      auto const frame_ii = s.frameIndex();
      out << track_i <<','<< frame_ii <<','
          << frames_o2i[frame_ii][s.outline()] << '\n';
    }

    out_meta << track_i << ',' << (track.isRevised()? '1': '0') << '\n';

    ++track_i;
  } // for each track

  tracks_file.close();
  tracks_meta_file.close();

  message_('I', QString("Saved ") + QString::number(track_i) +" tracks.");
}



void MainWindow::on_pushButtonTracksEditLoad_clicked()
{
  // Check 1 of 2: existence of nuclei outlines for all time frames
  //
  if (nuclei_.empty()) {
    message_('E', "Cannot load tracks when nuclei contours are missing");
    return;
  }
  if (nuclei_.size() != n_frames_) {
    message_('E', "Cannot load tracks when nuclei contours are missing "
                  "in any frame");
    return;
  }
  assert(nuclei_.size() == n_frames_);
  for (auto frame_i=0; frame_i<static_cast<int>(n_frames_); ++frame_i) {
    if (nuclei_[frame_i].empty()) {
      message_('E', "Cannot load tracks when nuclei contours are missing: "
                    "Nuclei lack in frame " + QString::number(frame_i));
      return;
  } }

  // Check 2 of 2: existence of the tracks files
  //
  QString const tracks_file_name = images_directory_ + QDir::separator()
                                   + "tracks.csv";
  QFileInfo const tracks_file_info(tracks_file_name);
  if (not tracks_file_info.exists()) {
    message_('E', "Tracks file " + tracks_file_name + " does not exist.");
    return;
  }
  QFile tracks_file(tracks_file_name);
  if (not tracks_file.open(QFile::ReadOnly | QFile::Text)) {
    message_('E', "Tracks file " + tracks_file_name + " cannot be open.");
    return;
  }
  //
  QString const tracks_meta_file_name = images_directory_ + QDir::separator()
                                        + "tracks_meta.csv";
  QFileInfo const tracks_meta_file_info(tracks_meta_file_name);
  if (not tracks_meta_file_info.exists()) {
    message_('E', "Tracks metadata file " + tracks_meta_file_name
                  + " does not exist.");
    return;
  }
  QFile tracks_meta_file(tracks_meta_file_name);
  if (not tracks_meta_file.open(QFile::ReadOnly | QFile::Text)) {
    message_('E', "Tracks metadata file " + tracks_meta_file_name
                  + " cannot be open.");
    return;
  }

  // Clean-up
  //
  clearAllTracks_();

  QTextStream in(&tracks_file);
  QTextStream in_meta(&tracks_meta_file);

  // Note: in files to be read traj_index, nucleus_index are 1-base numbered,
  //       only frame_index is 0-base numbered.

  auto frame_has_nuc = [this](long int frame_i, long int nuc_i) -> bool
  {
    if (nuclei_.size() < frame_i) { return false; }
    return nuc_i < nuclei_[AS_SIZE(frame_i)].size();
  };

  QMap<int, int> revision_status;
  in_meta.readLine(); // omit header
  while (not in_meta.atEnd()) {
    QString const line = in_meta.readLine();
    if (line.length() > 0) {
      QStringList const entries = line.split(',');
      revision_status.insert(entries.first().toInt(),
                             entries.last().toInt());
  } }

  in.readLine(); // omit header
  int last_track_i = -1;
  while (not in.atEnd()) {
    QString const line = in.readLine();
    if (line.length() == 0) { break; }
    QStringList const entries = line.split(',');
    auto const track_i = entries[0].toInt();
    auto const frame_i = entries[1].toInt();   // no shift
    auto const nuc_i = entries[2].toInt() - 1; // SHIFT!
    if (not frame_has_nuc(frame_i, nuc_i)) {
      message_('E', "Track " + QString::number(track_i + 1) + " does not have "
                    + "nucleus " + QString::number(nuc_i + 1) + ". "
                    + "Loading tracks has not been completed.");
      return;
    }
    if (track_i != last_track_i) {
      assert(track_i - last_track_i == 1);
      assert(frame_i < nuclei_.size());
      assert(nuc_i < nuclei_[frame_i].size());
      tracks_.push_back(Track(&nuclei_[frame_i][nuc_i], frame_i, false));
      tracks_.back().setSelfCleanup(true);
      assert(revision_status.count(track_i));
      tracks_.back().setRevised(revision_status[track_i]);
    } else {
      tracks_.back().append(&nuclei_[frame_i][nuc_i],frame_i);
    }
    last_track_i = track_i;
  } // while not at the end of the tracks file

//std::sort( tracks_.begin(),tracks_.end(),
//    [this](Track const& ta, Track const& tb) {
//        if (ta.nodesCount() == tb.nodesCount()) {
//          return ta.confidence() > tb.confidence();
//        } else {
//          return (ta.nodesCount() > tb.nodesCount());
//        } });
  resetTracksModel_();
  current_track_ = nullptr;
  message_('I', "Loaded "+ QString::number(tracks_.size()) +" tracks");
}


// The OBSOLETE functions below saved/loaded tracks to/from separate CSV
// files in 'tracks/' subdirectory.
#if 0
void MainWindow::on_pushButtonTracksEditLoad_clicked_OBSOLETE()
{
  // Check 1 of 2: existence of nuclei outlines for all time frames
  //
  if (nuclei_.empty()) {
    message_to_user_('E',"Cannot load tracks when nuclei contours are missing",
                     12000);
    return;
  }
  if (nuclei_.size() != n_frames_) {
    message_to_user_('E',"Cannot load tracks when nuclei contours are missing "
                         "for any frame",12000);
    return;
  }
  assert(nuclei_.size() == n_frames_);
  for (auto frame_i=0; frame_i<static_cast<int>(n_frames_); ++frame_i) {
    if (nuclei_[frame_i].empty()) {
      message_to_user_('E',QString("Cannot load tracks when nuclei contours ") +
                           "are missing: Nuclei lack in frame " +
                           QString::number(frame_i), 12000);
      return;
  } }

  // Check 2 of 2: existence of the tracks directory
  //
  QString const imgs_dir = last_images_directory_;
  QDir tracks_dir(imgs_dir + QDir::separator() + "tracks");
  if (not tracks_dir.exists()) {
    message_to_user_('E',QString("Tracks directory ") + tracks_dir.path() +
                         " does not exist.", 12000);
    return;
  }

  // Clean-up
  //
  clear_all_tracks_();

  // Note: in files to be read traj_index, nucleus_index are 1-base numbered,
  //       only frame_index is 0-base numbered.

  auto frame_has_nuc = [this](long int frame_i, long int nuc_i) -> bool
  {
    if (nuclei_.size() < frame_i) { return false; }
    return nuc_i < nuclei_[frame_i].size();
  };

  auto track_i = 0;
  //QDirIterator dit(tracks_dir);
  //while (dit.hasNext()) {
  for (auto const& fn : tracks_dir.entryList(QStringList("*"), QDir::Files)) {
    //auto const& fit = dit.next();
    //auto const fp = QFileInfo(fit).absoluteFilePath();
    //auto const fn = QFileInfo(fit).fileName();
    auto const fp = QFileInfo(tracks_dir.path() + QDir::separator() + fn)
                        .absoluteFilePath();
  //qDebug() << fn << "  " << fp;
    if (    fn.endsWith(".csv", Qt::CaseInsensitive)
        and fn.startsWith("track-", Qt::CaseInsensitive)) {
      QFile track_file(fp);
      if (track_file.open(QFile::ReadOnly | QFile::Text)) {
        QTextStream in(&track_file);
        QString const line = in.readLine();
        bool revised = false;
        if (line.length() > 0) {
          if (line.startsWith("#revised: ")) { // then: header
            auto const line_split_last = line.split(" ").last();
            revised =   (line_split_last == "yes" )
                     or (line_split_last == "Y"   )
                     or (line_split_last == "true")
                     or (line_split_last == "T"   );
        } }
        in.readLine(); // skip CSV header
        bool track_initd = false;
        while (not in.atEnd()) {  // reading past the header
          QString const line = in.readLine();
          if (line.length() > 0) {
            auto const frame_i = line.split(",").first().toInt();    // no shift
            auto const nuc_i   = line.split(",").last().toInt() - 1; // SHIFT!
            if (not frame_has_nuc(frame_i,nuc_i)) {
              message_to_user_('E', QString("Track ")+QString::number(track_i+1)
                      +" does not have a nucleus "+ QString::number(nuc_i+1) +
                      ". Loading tracks has not been completed.",12000);
              return;
            }
            if (not track_initd) {
              assert(frame_i < nuclei_.size());
              assert(nuc_i < nuclei_[frame_i].size());
              tracks_.push_back(Track(&nuclei_[frame_i][nuc_i],frame_i,false));
              tracks_.back().setSelfCleanup(true);
              tracks_.back().setRevised(revised);
              track_initd = true;
            } else {
              tracks_.back().append(&nuclei_[frame_i][nuc_i],frame_i);
            }
      } } }
      ++track_i;
    } // if file name conforms to the track file naming convention
  } // for each tracks directory file

//std::sort( tracks_.begin(),tracks_.end(),
//    [this](Track const& ta, Track const& tb) {
//        if (ta.nodesCount() == tb.nodesCount()) {
//          return ta.confidence() > tb.confidence();
//        } else {
//          return (ta.nodesCount() > tb.nodesCount());
//        } });
  reset_tracks_model_();
  current_track_ = nullptr;
  message_to_user_('I', QString("Loaded ") + QString::number(tracks_.size()) +
                        " tracks", 6000);
}

QString MainWindow::assemble_track_file_name_OBSOLETE_(
        QString const& tracks_dir, int i) const
{
  QString idx = QString::number(i);
  if (i < 10)    { idx = QString("0") + idx; }
  if (i < 100)   { idx = QString("0") + idx; }
  if (i < 1000)  { idx = QString("0") + idx; }
  if (i < 10000) { idx = QString("0") + idx; }
  if (i < 100000){ idx = QString("0") + idx; }
  return QString(tracks_dir + QDir::separator() +"track-"+ idx +".csv");
}

void MainWindow::on_pushButtonTracksEditSave_clicked_OBSOLETE()
{
  bool CONSTEXPR clean_nonempty_tracks_dir = true;

  // Create a bimap: outline ptr <-> outline index.
  QList<QMap<Outline const*,int>> frames_o2i; // outline-to-index
  QList<QMap<int,Outline const*>> frames_i2o; // index-to-outline
  for (int frame_i=0; frame_i<n_frames_; ++frame_i) {
    frames_o2i.push_back(QMap<Outline const*,int>());
    frames_i2o.push_back(QMap<int,Outline const*>());
    int nuc_i = 0;
    int const nuc_i_offset = 1;
    for (auto const& n: nuclei_[frame_i]) {
      frames_o2i.back().insert(&n,nuc_i + nuc_i_offset);
      frames_i2o.back().insert(nuc_i + nuc_i_offset,&n);
      ++nuc_i;
  } }

  // Create the tracks directory, if does not exists. If it exists, clean it.
  QString const imgs_dir = last_images_directory_;
  QDir tracks_dir(imgs_dir + QDir::separator() + "tracks");
  if (not tracks_dir.exists()) {
    tracks_dir.mkdir(tracks_dir.path());
  } else {  // directory clean-up required:
    if (clean_nonempty_tracks_dir) {
      QDirIterator dit(tracks_dir);
      while (dit.hasNext()) {
        auto const& fit = dit.next();
        auto const fp = QFileInfo(fit).absoluteFilePath();
        auto const fn = QFileInfo(fit).fileName();
        if (    fn.endsWith(".csv", Qt::CaseInsensitive)
            and fn.startsWith("track-", Qt::CaseInsensitive)) {
          tracks_dir.remove(fp);
  } } } }

  int track_i = 0;
  for (Track const& track: tracks_) {

    if (ui_->checkBoxTracksEditSaveOnlyRevisedTracks->isChecked())
      { if (not track.isRevised()) { continue; } }

    auto fn = assemble_track_file_name_(tracks_dir.absolutePath(), track_i + 1);
    QFile track_file(fn);
    if (track_file.open(QFile::WriteOnly | QFile::Text)) {
      QTextStream out(&track_file);
      auto const frame_i = track.firstFrameIndex();
      out << "#revised: " << (track.isRevised()? "yes":"no") << '\n'
          << "time_point_index,nucleus_id\n";
      out << frame_i <<','<< frames_o2i[frame_i][track.initialOutline()] <<'\n';
      for (auto const& s: track.segments()) {
        auto const frame_i = s.frameIndex();
        out << frame_i << ',' << frames_o2i[frame_i][s.outline()] << '\n';
      }
      track_file.close();
    } else {
      message_to_user_('E',"Could not save a track in "
          + QFileInfo(track_file).fileName() +"."
          " Saving tracks has not been completed.", 12000);
      return;
    }

    ++track_i;
  } // for each track

  message_to_user_('I',QString("Saved ") + QString::number(track_i) +" tracks.",
                   6000);
}
#endif
