// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "MainWindow.hpp"

#ifndef Q_OS_WIN

// Handle UNIX signals using "self-pipes".

#include <sys/types.h>
#include <sys/socket.h>  // AF_UNIX, SOCK_STREAM
#include <csignal>

int MainWindow::sighupFd[2] = {0, 0};
int MainWindow::sigtermFd[2] = {0, 0};

void MainWindow::setupUnixSignalHandling()
{
  if (::socketpair(AF_UNIX, SOCK_STREAM, 0, sighupFd)) {
    fprintf(stderr, "Error: Couldn't create a SIGHUP socketpair!");
    return;
  }

  if (::socketpair(AF_UNIX, SOCK_STREAM, 0, sigtermFd)) {
    fprintf(stderr, "Error: Couldn't create a SIGTERM socketpair!");
    return;
  }

  snHup.reset(new QSocketNotifier(sighupFd[1], QSocketNotifier::Read, this));
  connect(snHup.get(), SIGNAL(activated(int)), this, SLOT(sighupHandler()));
  snTerm.reset(new QSocketNotifier(sigtermFd[1], QSocketNotifier::Read, this));
  connect(snTerm.get(), SIGNAL(activated(int)), this, SLOT(sigtermHandler()));

  struct sigaction hup, term;
  //
  hup.sa_handler = MainWindow::sighupForwarder;
  sigemptyset(&hup.sa_mask);
  hup.sa_flags = 0;
  hup.sa_flags |= SA_RESTART;
  if (sigaction(SIGHUP, &hup, 0)) {
    fprintf(stderr, "Error: Couldn't set a SIGHUP handler!");
  }
  //
  term.sa_handler = MainWindow::sigtermForwarder;
  sigemptyset(&term.sa_mask);
  term.sa_flags = 0;
  term.sa_flags |= SA_RESTART;
  if (sigaction(SIGTERM, &term, 0)) {
    fprintf(stderr, "Error: Couldn't set a SIGTERM handler!");
  }
}

void MainWindow::sighupForwarder(int)
{
  char a = 1;
  if (::write(sighupFd[0], &a, sizeof(a)) <= 0) {
    fprintf(stderr, "Warning: Could not forward signal HUP.\n");
  }
}

void MainWindow::sigtermForwarder(int)
{
  char a = 1;
  if (::write(sigtermFd[0], &a, sizeof(a)) <= 0) {
    fprintf(stderr, "Warning: Could not forward signal TERM.");
  }
}

void MainWindow::sigtermHandler()
{
  snTerm->setEnabled(false);
  char tmp;
  if (::read(sigtermFd[1], &tmp, sizeof(tmp)) > 0) {
    message_('W', "Received SIGTERM.");
    quit();
    // snTerm->setEnabled(true);
  }
}

void MainWindow::sighupHandler()
{
  snHup->setEnabled(false);
  char tmp;
  if (::read(sighupFd[1], &tmp, sizeof(tmp)) > 0) {
    message_('W', "Received SIGHUP.");
    snHup->setEnabled(true);
  }
}

#endif // !defined(Q_OS_WIN)
