// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "ScriptWindow.hpp"
#include "ui_ScriptWindow.h"
#include "MainWindow.hpp"
#include "GUI/Windows/MainWindow.hpp"
#include <QQmlEngine>
#include <QRegularExpression>
#include <QDir>
#include <QMessageBox>
#include <QInputDialog>
#include <QModelIndex>

#ifdef Q_OS_WIN
#  if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
#    include <QStandardPaths>
#  else
#    include <QDesktopServices>
#  endif
#endif
#include <cstdio>

ScriptWindow::ScriptWindow(MainWindow* main_window, QWidget* parent)
: QMainWindow{parent},
  main_window_{main_window},
  ui_{new Ui::ScriptWindow},
  scripts_dir_(
#ifdef Q_OS_WIN
#  if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
      QStandardPaths::locate( QStandardPaths::DocumentsLocation, "STScripts",
                              QStandardPaths::LocateDirectory                )
#  else // never tested:
      QDesktopServices::storageLocation(QDesktopServices::DocumentsLocation)
          + "\\STScripts\\"
#  endif
#else
#  ifdef SHUTTLETRACKER_SHARED_DIR
    QDir(SHUTTLETRACKER_SHARED_DIR).absolutePath()
      + QDir::separator() + "shuttletracker"
      + QDir::separator() + "scripts" + QDir::separator()
#  else
      QApplication::applicationDirPath() + "/STScripts/"
#  endif
#endif
  ) // scripts_dir_
{
#ifdef Q_OS_WIN
#  if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))

  // fallback:
  if (not scripts_dir_.absolutePath().contains("STScripts")
       or scripts_dir_.absolutePath().contains("Program Files")) {
    scripts_dir_.setPath(
        QStandardPaths::locate(QStandardPaths::HomeLocation,
                               "Documents/STScripts",
                               QStandardPaths::LocateDirectory));
  }

  // fallback:
  if (not scripts_dir_.absolutePath().contains("STScripts")
       or scripts_dir_.absolutePath().contains("Program Files")) {
    scripts_dir_ = QDir(QDir::homePath()
                        + QDir::separator() + "Documents"
                        + QDir::separator() + "STScripts");
  }

#  endif
#endif
  ui_->setupUi(this);

  assert(ui_->scriptSnippets);
  ui_->textEditScriptErrors->setVisible(false);
  ui_->scriptSnippets->hideLineNumbersArea();
  ui_->scriptSnippets->setVisible(false);

  auto g = this->geometry();
  g.setWidth(1150);
  g.setX(50);
  g.setY(100);
  this->setGeometry(g);

  Qt::WindowFlags flags = this->windowFlags();
  flags &= ~Qt::WindowMaximizeButtonHint;
  flags |=  Qt::WindowCloseButtonHint;
//flags |=  Qt::WindowCancelButtonHint; // not available since Qt 5.8
  setWindowFlags(flags);

  setupSnippets_();
  listAvailableScriptFiles_();

  QQmlEngine::setObjectOwnership(main_window_, QQmlEngine::CppOwnership);
  QJSValue mw = interpreter_.newQObject(main_window_);
  QJSValue gl = interpreter_.globalObject();
  gl.setProperty(main_window_str_, mw);
}


ScriptWindow::~ScriptWindow()
{
  interpreter_.collectGarbage();
  delete ui_;
}


QStringList ScriptWindow::listAvailableScriptFiles_()
{
  QStringList scripts;
  ui_->listWidgetScriptFiles->clear();
  auto script_files = scripts_dir_.entryList({"*" + script_file_ext_});
  for (auto const& sf: script_files) {
    scripts.push_back(sf);
    ui_->listWidgetScriptFiles->addItem(sf);
  }
  return scripts;
}


void ScriptWindow::setCompleterWordlist(QStringList const& words)
{  ui_->scriptEditor->setCompleterWordlist(words);  }


void ScriptWindow::setupSnippets_()
{
  ui_->scriptSnippets->appendHtml("<br><i>Actions</i>:"/*<br>*/);
  auto const& api_func_names = MainWindow::api_functions.keys();
  for (auto const& fun_sntx: api_func_names) {
    ui_->scriptSnippets->appendPlainText(  QString(" ")
                                         + MainWindow::api_functions[fun_sntx]);
  }

  auto params_list = MainWindow::api_parameters.values();
  params_list.sort();
  ui_->scriptSnippets->appendHtml("<br><br><i>Parameters</i>:<br>");
  for (auto const& param: params_list) {
    ui_->scriptSnippets->appendPlainText(QString(" ") + param);
  }

  auto specs_list = MainWindow::api_specials.values();
  specs_list.sort();
  ui_->scriptSnippets->appendHtml("<br><br><i>Predefined arguments</i>:<br>");
  for (auto const& spec: specs_list) {
    ui_->scriptSnippets->appendPlainText(QString(" ") + spec);
  }

  ui_->scriptSnippets->appendHtml("<br><br><i>General syntax</i>:<br><br>"
                                  "&nbsp;JavaScript syntax.");
  ui_->scriptSnippets->appendHtml("<br><br><i>Additional macros</i>:<br>");
  ui_->scriptSnippets->appendPlainText(" for each frame { }");
}


void ScriptWindow::preprocess_(QString& script) const
{
  // for-each loop: enable batch processing just before the loop
  QStringList captures; // to reinject possibly multiline break between
                        // for-loop-boundaries & for-loop-body-opening-bracket
  QRegularExpression const for_each_frame{"\\bfor\\s+each\\s+frame([\\s\\n\\r]*)\\{"};
  {
    QRegularExpressionMatchIterator mit = for_each_frame.globalMatch(script);
    while (mit.hasNext()) { captures << mit.next().captured(1); }
  }
  {
    QRegularExpressionMatchIterator mit = for_each_frame.globalMatch(script);
    int repl_i = 0;
    while (mit.hasNext()) {
      auto const match = mit.next();
      auto const repl = QString{"___batch_processing___(true); "
           "for (var ___i___=0; ___i___ < frames_count(); ___i___++)"}
           + captures[repl_i] + "{"
           "  go_to_frame(___i___);";
      script.replace(match.capturedStart(), match.capturedLength(), repl);
      int const offset = match.capturedStart() + match.capturedLength();
      mit = for_each_frame.globalMatch(script, offset);
      ++repl_i;
    }
  }
  {
    QRegularExpression const batch_on{
        "___batch_processing___\\(true\\);.*[\\s\\r\\n]*.*(?=\\{)",
         QRegularExpression::InvertedGreedinessOption
       | QRegularExpression::MultilineOption};
    QRegularExpressionMatchIterator mit = batch_on.globalMatch(script);
    while (mit.hasNext()) {
      auto const match = mit.next();
      QString const cap = match.captured();
      int brace_open_pos = match.capturedStart() + cap.length(); // batch begin found,
      assert(script.at(brace_open_pos) == '{');                  // search for a place for end
      int n_next_brace_open = 0;
      int brace_close_pos = -1;
      for (int pos2 = brace_open_pos + 1; pos2 < script.length(); ++pos2) {
        if (script.at(pos2) == '{') { n_next_brace_open++; }
        if (script.at(pos2) == '}') {
          if (n_next_brace_open > 0) { n_next_brace_open--; }
          else                       { brace_close_pos = pos2;  break; }
      } }
      if (brace_close_pos > 0) {
        script.replace(brace_close_pos, 1,
                       QString{"repaint(); } ___batch_processing___(false);"});
      } else { }  // likely a syntax error, do nothing
      int const offset = match.capturedStart() + match.capturedLength();
      mit = batch_on.globalMatch(script, offset);
    }
  }

  // free functions -> member slots
  QStringList func_names = MainWindow::api_functions.keys();
  func_names.sort();
  for (int i=func_names.size()-1; i>=0; --i) {
    QRegularExpression const fn(QString{"\\b"} + func_names[i] + QString{"\\b"});
    script.replace(fn, main_window_str_ + "." + func_names[i]);
  }
} // preprocess_()


void ScriptWindow::toggleMainWindowControl(bool on)
{
  main_window_->toggleToolboxenEnablednessForScriptExecution(on);
}


void ScriptWindow::on_pushButtonScriptRun_clicked()
{
  ui_->textEditScriptErrors->setVisible(false);

  toggleMainWindowControl(false);

  QString script = ui_->scriptEditor->document()->toPlainText();
  //qDebug() << "RAW:\n" << script << '[]\n';
  preprocess_(script);
  //qDebug() << "PREPROCESSED:\n" << script << '[]\n';

  // no need for quotes for set()-able parameter names and colors:
  for (auto const& param_name: MainWindow::api_parameters) {
    interpreter_.evaluate(QString("")+ param_name+" = "+"\""+ param_name+"\";");
  }
  for (auto const& spec_name: MainWindow::api_specials) {
    interpreter_.evaluate(QString("")+ spec_name +" = "+"\""+ spec_name +"\";");
  }

#if SHUTTLETRACKER_GUI_EVAL_SCRIPT_IN_A_SEPARATE_THREAD
  {
    using T = QScriptValue(QScriptEngine::*)(QString const&,QString const&,int);
    QString const a2 = "";  int const a3 = 1;  // default args
    script_future_ = QtConcurrent::run(&interpreter_,
                                       static_cast<T>(&QScriptEngine::evaluate),
                                       script, a2, a3                         );
  }
  //script_future_.waitForFinished();
  auto const result = script_future_.result();
#else
  QJSValue result = interpreter_.evaluate(script);
#endif

  if ((last_evaluation_had_error_ = result.isError())) {
    int const line_no = result.property("lineNumber").toInt();
    QString const err_msg = QString("Error: Line ") + QString::number(line_no)
        +": " + result.toString()
     /* + "\n-------------------------------------------\n" + script */ ;
    fprintf(stderr, "Script error: %s", err_msg.toStdString().c_str());
    if (main_window_->shouldQuitUponScriptError()) {
      QApplication::quit();
    } else {
      ui_->textEditScriptErrors->setText(err_msg);
      ui_->scriptEditor->goToLine(line_no);
      ui_->textEditScriptErrors->setTextColor(QColor{Qt::red}.lighter(140));
      ui_->textEditScriptErrors->setVisible(true);
    }
  } else {
    ui_->textEditScriptErrors->setTextColor(Qt::green);
    if (result.toString() == "undefined") {
      ui_->textEditScriptErrors->setVisible(false);
      ui_->textEditScriptErrors->setText("OK");
    } else {
      ui_->textEditScriptErrors->setVisible(true);
      ui_->textEditScriptErrors->setText(result.toString() );
  } }

  toggleMainWindowControl(true);
  main_window_->ensureNotInBatchMode();
}


void ScriptWindow::on_pushButtonSnippetsToggle_toggled(bool checked)
{
  ui_->pushButtonSnippetsToggle->setText(checked ? "Hide &API" : "Show &API");
  if (checked) {
    ui_->scriptSnippets->setVisible(true);
    ui_->horizontalSpacerDynamic->changeSize(60, 1, QSizePolicy::Preferred,
                                             QSizePolicy::Fixed);
  } else {
    ui_->scriptSnippets->setVisible(false);
    ui_->horizontalSpacerDynamic->changeSize(120, 1, QSizePolicy::Preferred,
                                             QSizePolicy::Fixed);
  }
}


void ScriptWindow::on_pushButtonScriptFileLoad_clicked()
{
  if (ui_->listWidgetScriptFiles->selectedItems().empty()) { return; }

  auto const fn =  scripts_dir_.absolutePath() + "/"
                 + ui_->listWidgetScriptFiles->selectedItems().at(0)->text();
//qDebug() << fn;
  QFile file(fn);
  if (file.open(QFile::ReadOnly | QFile::Text)) {
    QTextStream in(&file);
    ui_->scriptEditor->setPlainText( in.readAll() );
  }
}


void ScriptWindow::on_pushButtonScriptFileSave_clicked()
{
  if (ui_->listWidgetScriptFiles->selectedItems().empty()) {
    QMessageBox::critical(this, "Error", "Cannot save: No file selected.");
    return;
  }

  auto const fn =  scripts_dir_.absolutePath() + "/"
                 + ui_->listWidgetScriptFiles->selectedItems().at(0)->text();
  QFile file(fn);
  if (file.open(QFile::WriteOnly | QFile::Text)) {
    QTextStream out(&file);
    out << ui_->scriptEditor->document()->toPlainText();
  }
}


void ScriptWindow::on_pushButtonScriptFileNew_clicked()
{
  bool ok;
  QString const text = QInputDialog::getText(this, tr("New script file"),
      tr("File name:"), QLineEdit::Normal, "my_new" + script_file_ext_, &ok);

  if (ok && !text.isEmpty()) {
    if (not text.endsWith(script_file_ext_)) {
      QMessageBox::warning(this, "Wrong file name extension",
                           "File name should end with '.stscript'.");
      on_pushButtonScriptFileNew_clicked(); // another chance
      return;
    }
    auto const fn = scripts_dir_.absolutePath() + "/" + text;
    QFile file(fn);
    if (not file.open(QFile::WriteOnly | QFile::Text)) {
      QMessageBox::critical(this, "Error",
                            "Cannot create an empty file " + fn + ".");
      return;
  } }
  auto const scripts = listAvailableScriptFiles_();
  emit scriptFilesListModified(scripts);
}


void ScriptWindow::on_pushButtonScriptFileDelete_clicked()
{
  if (ui_->listWidgetScriptFiles->selectedItems().empty()) {
    QMessageBox msg;
    msg.setText("Cannot delete: No file selected.");
    msg.exec();
    return;
  }
  auto const fn =   scripts_dir_.absolutePath() + "/"
                  + ui_->listWidgetScriptFiles->selectedItems().at(0)->text();

  QMessageBox::StandardButton reply = QMessageBox::question( this,
      "Confirm file deletion", QString("Do you really want to delete ")
       + ui_->listWidgetScriptFiles->selectedItems().at(0)->text() + "?",
      QMessageBox::Yes|QMessageBox::No );
  if (reply == QMessageBox::Yes) {
    QFile file(fn);
    try {
      file.remove();
    } catch (...) {
      QMessageBox::warning(this, "Error", "Cannot delete the file, sorry.",
                           QMessageBox::Ok);
      return;
  } }

  auto const scripts = listAvailableScriptFiles_();
  emit scriptFilesListModified(scripts);
}


int ScriptWindow::loadAndExecute(QString const& script_name)
{
  auto const fn =  scripts_dir_.absolutePath() + "/" + script_name
                 + script_file_ext_;

  // select file
  for(int i = 0; i < ui_->listWidgetScriptFiles->count(); ++i) {
    auto pi = ui_->listWidgetScriptFiles->item(i);
    if (pi->text() == script_name + script_file_ext_) {
      pi->setSelected(true);  break;
  } }

  // read in file
  QFile file(fn);
  if (file.open(QFile::ReadOnly | QFile::Text)) {
    QTextStream in(&file);
    return execute(in.readAll());
  } else {
    return 1;
  }

  //return last_evaluation_had_error_;
}


int ScriptWindow::execute(QString const& script)
{
  ui_->scriptEditor->setPlainText(script);
  on_pushButtonScriptRun_clicked();
  return last_evaluation_had_error_;
}


void ScriptWindow::abortExecution()
{  interpreter_.setInterrupted(true);  }


QDir const& ScriptWindow::scriptsDir() const
{  return scripts_dir_;  }


QString const& ScriptWindow::scriptFileNameExtension() const
{  return script_file_ext_;  }
