// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "GUI/Windows/MainWindow.hpp"
#include <QApplication>
#include <QCommandLineParser>
#include <QString>
#include <QStringList>
#include <QDir>
#include <QFile>
#include <QStyleFactory>
#include <QtGlobal> // qgetenv
#include <opencv2/core/utility.hpp> // setNumThreads()
#include <cstdio>
#include <clocale>


void set_visual_styling(QApplication& a)
{
#if defined(Q_OS_WIN)
  a.setWindowIcon(QIcon(":/icons/shuttletracker.ico"));
#endif

#if defined(Q_OS_WIN) || (defined(Q_OS_LINUX) && (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))) || defined(Q_OS_MACOS)

  QColor const disabled  = QColor(64,64,64),
               invisible = QColor( 0,0,0,0),
               bg        = QColor(32,32,32),
               box_bg    = QColor(32,32,32),
               text      = Qt::lightGray;
  QPalette palette = a.palette();
  palette.setColor(QPalette::Base,       box_bg); // bg of boxes
  palette.setColor(QPalette::Window,     bg);    // window bg
  palette.setColor(QPalette::WindowText, text);
  palette.setColor(QPalette::Text,       text);
  palette.setColor(QPalette::Button,     bg); // frames too
  palette.setColor(QPalette::ButtonText, Qt::lightGray);
  palette.setColor(QPalette::ToolTipBase,QColor(202,232,242,224));
  palette.setColor(QPalette::ToolTipText,Qt::black);
  palette.setColor(QPalette::Disabled,   QPalette::ButtonText, disabled);
  palette.setColor(QPalette::Disabled,   QPalette::WindowText, disabled);
  palette.setColor(QPalette::Disabled,   QPalette::Text, disabled);
  palette.setColor(QPalette::Disabled,   QPalette::Base, disabled);
  palette.setColor(QPalette::Disabled,   QPalette::Light,invisible); // text shadow
//palette.setColor(QPalette::Highlight,  QColor(142,122,100).lighter());
//palette.setColor(QPalette::HighlightedText, Qt::black);
//palette.setColor(QPalette::AlternateBase, QColor(128,255,128));
//palette.setColor(QPalette::BrightText, Qt::yellow);
  a.setPalette(palette);

  a.setStyle(QStyleFactory::create("fusion"));

#ifdef Q_OS_WIN
  QString const css_file_path = "fixfusion.css";
#else
#  ifdef SHUTTLETRACKER_SHARED_DIR
    QString const css_file_path =
    QDir(QString(SHUTTLETRACKER_SHARED_DIR)).absolutePath()
         + QDir::separator() + "shuttletracker"
         + QDir::separator() + "gui"
         + QDir::separator() + "fixfusion.css";
#  else
    QString const css_file_path = "fixfusion.css";
#  endif
#endif

  QFile f{QDir(QApplication::applicationDirPath()).filePath(css_file_path)};
  if (f.open(QIODevice::ReadOnly | QIODevice::Text)) {
    QTextStream in(&f);
    QString css = in.readAll();
    a.setStyleSheet(css);
  }

  // TODO -- is this needed?
  a.setStyle("QToolBox::tab { border: 1pt solid black; }");

#else // tested under Linux, Qt 4.8.3

  QColor const invisible = QColor(0,0,0,0);
  QPalette palette = a.palette();
  palette.setColor(QPalette::Disabled,   QPalette::Light, invisible); // text shadow
  a.setPalette(palette);

  QFile f{QDir(QApplication::applicationDirPath()).filePath("darkorange.css")};
  if (f.open(QIODevice::ReadOnly | QIODevice::Text)) {
    QTextStream in(&f);
    QString css = in.readAll();
    a.setStyleSheet(css);
  }

  a.setStyle("plastique");
#endif
}


int main(int argc, char *argv[])
{
  std::locale::global(std::locale::classic());

  // apply settings enforced by environmental variables
  QString const num_threads_s = qgetenv("SHUTTLETRACKER_OPENCV_NUM_THREADS");
  if (not num_threads_s.isEmpty()) {
    int num_threads = num_threads_s.toInt();
    cv::setNumThreads(num_threads);
  }

  // set application metadata
  QApplication a(argc, argv);
  a.setApplicationName("ShuttleTracker");
  a.setApplicationVersion(APP_VERSION);
  a.setOrganizationName("Lipniacki Lab");
  a.setOrganizationDomain("http://pmbm.ippt.pan.pl");

  // display main window
  set_visual_styling(a);
  MainWindow w;
  w.show();

  // set up a parser
  QCommandLineParser p;
  p.setSingleDashWordOptionMode(QCommandLineParser::ParseAsLongOptions);
  p.addPositionalArgument("directory",
                          "A directory with images and metadata (optional).");
  p.addHelpOption();
  p.addVersionOption();
  QCommandLineOption const script_file_option(
      QStringList() << "s" << "script",
      "Execute a script from <file>.", "file");
  p.addOption(script_file_option);
  QCommandLineOption const script_auto_quit_option(
      QStringList() << "q" << "quit-after-script-finished",
      "Quit after a script has been executed.");
  p.addOption(script_auto_quit_option);


  // -- parse command line and perform some consistency checking

  p.process(a);
  QStringList const pos_args = p.positionalArguments();
  if (pos_args.isEmpty()) {
    // that's legal, do not complain
  } else if (pos_args.size() > 1) {
    fprintf(stderr, "Error: Can read only a single input folder at a time.\n");
  }

  bool script_auto_quits = false;
  if (p.isSet(script_auto_quit_option) and not p.isSet(script_file_option)) {
    fprintf(stderr, "Warning: Cannot quit after executing a non-specified script.\n");
  } else {
    script_auto_quits = p.isSet(script_auto_quit_option);
  }


  // -- use parsed arguments

  // input folder, if given
  if (pos_args.size() == 1) {
    QString const input_dir = pos_args.first();
    if (QDir(input_dir).exists() and QFileInfo(input_dir).isDir()) {
      w.loadImagesFromDirectory(input_dir);
    } else {
      fprintf(stderr, "Error: Cannot load images from directory %s.\n", argv[1]);
  } }

  // script to be executed, if given
  if (p.isSet(script_file_option) and not p.value(script_file_option).isEmpty()) {
    QString const script_file_path = p.value(script_file_option);
    if (QFileInfo(script_file_path).isFile()) {
      QFile script_file(script_file_path);
      if (script_file.open(QFile::ReadOnly | QFile::Text)) {
        QString const script = QTextStream(&script_file).readAll()
            + QString(script_auto_quits ? QString("\n") + "quit();" : "");
        w.postLoadScript(script);
        w.setQuitUponScriptError(false);
      } else {
        fprintf(stderr, "Error: Cannot read script file %s.\n",
            script_file_path.toStdString().c_str());
      }
    } else {
      fprintf(stderr, "Error: Apparently, %s is not a file.\n",
          script_file_path.toStdString().c_str());
  } }

  return a.exec();
}
