// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "ImageDecomposition.hpp"
#include "ImageOperations.hpp"
#include "SharedUtilities.hpp"
#include <gtest/FloatingPoint.hpp>
#include <vector>
#include <type_traits>  // is_same
#include <cmath>
#include <cstdio>
#include <opencv2/imgproc.hpp>
#include <opencv2/highgui.hpp>

namespace st {

void partitionIntoVoronoiCellsCoarse(st::contours_vec_t& voronoi_contours,
    std::vector<std::pair<int, int>>& voronoi_neighbors,
    std::vector<cv::Point2f> const& points,
    cv::Size sz, float homothetic_scaling, int erosion_width)
{
  cv::Rect rect{0, 0, sz.width, sz.height};
  cv::Subdiv2D subdiv(rect);
#ifndef NDEBUG
  for (auto const&p: points) {
    assert(0 <= p.x and p.x < sz.width);
    assert(0 <= p.y and p.y < sz.height);
  }
#endif
  subdiv.insert(points);

  std::vector<std::vector<cv::Point2f>> facets;
  std::vector<cv::Point2f> centers;
  subdiv.getVoronoiFacetList(contouri_t(), facets, centers);

  std::vector<cv::Vec4f> edges;
  subdiv.getEdgeList(edges);
  for (auto const& e: edges) {
    cv::Point2f const p1{e[0], e[1]}, p2{e[2], e[3]};
    if (rect.contains(p1) and rect.contains(p2)) {
      int p1i=0, p2i=0;
      for (auto const& p: points) { if (p == p1) { break; } p1i++; }
      for (auto const& p: points) { if (p == p2) { break; } p2i++; }
      voronoi_neighbors.emplace_back(p1i, p2i);
  } }

// -- PREVIEW (1/3)
//cv::Mat img = cv::Mat::zeros(sz.height, sz.width, CV_8UC3);

  contour_t      ifacet;
  contours_vec_t ifacets(1);
  cv::Mat mask = cv::Mat::zeros(sz.height, sz.width, CV_8UC1);
  contours_vec_t single_contour;

  for (size_t i=0; i<facets.size(); i++) {

    ifacet.resize(facets[i].size());
    for (size_t j=0; j<facets[i].size(); j++) {
      ifacet[j] = centers[i];
      ifacet[j].x += static_cast<int>(homothetic_scaling*(   facets[i][j].x
                                                          - centers[i].x));
      ifacet[j].y += static_cast<int>(homothetic_scaling*(   facets[i][j].y
                                                          - centers[i].y));
    }

    mask.setTo(cv::Scalar{0});
    cv::fillConvexPoly(mask, ifacet, cv::Scalar{255}, 8, 0);

    if (erosion_width > 0) {
      int const size_base = 2*erosion_width;
      auto const elem_erode = cv::getStructuringElement(cv::MORPH_ELLIPSE,
                                      cv::Size{size_base+1, size_base+1},
                                      cv::Point{size_base/2, size_base/2});
      cv::erode(mask, mask, elem_erode);
    }

    single_contour.clear();
    cv::findContours(mask, single_contour,
                     cv::RETR_EXTERNAL, cv::CHAIN_APPROX_NONE);
    if (single_contour.size() == 1) {
      voronoi_contours.push_back(single_contour.front());
    } else if (single_contour.size() == 0) {
      voronoi_contours.push_back(contour_t{});
    } else {
      // Erosion of a convex shape that results in more than 1 contour
      // is an artifact. Anyway, pick the largest contour:
      double max_area = -1.;
      int max_area_i = 0, ci =0;
      for (auto const& c: single_contour) {
        if (c.size() > 0) {
          auto const area = cv::contourArea(c);
          if (area > max_area) { max_area = area;  max_area_i = ci; }
          ci++;
        }
      } // for each detected contour
      if (max_area > 0.) {
        voronoi_contours.push_back(single_contour[AS_SIZE(max_area_i)]);
      } else {
        voronoi_contours.push_back(contour_t{});
      }
    }

// -- PREVIEW (2/3)
//  cv::Scalar const color(rand() & 255, rand() & 255, rand() & 255);
//  cv::fillConvexPoly(img, ifacet, color, 8, 0);
//  ifacets[0] = ifacet;
//  cv::polylines(img, ifacets, true, cv::Scalar(), 1, CV_AA, 0);
//  cv::circle(img, centers[i], 3, cv::Scalar(), CV_FILLED, CV_AA, 0);

  } // for each facet

  assert(voronoi_contours.size() == points.size());

// -- PREVIEW (3/3)
//  cv::imshow( "V", img);
} // partition_into_voronoi_cells()


void suggestBackgroundTiles(std::map<ImageChannel, cv::Mat> const& images,
                            cv::Mat const& xmask_img, contours_vec_t*bg_contours,
                            int n_subdivs, int n_contours, int n_subshifts)
{
  assert(bg_contours);
  assert(not images.empty());
  assert(n_subdivs > 0);
  assert(n_contours > 0);
  assert(n_subshifts >= 0);
  assert((n_subshifts > 0) ? n_contours == 1: true);

  cv::Size const images_size = images.begin()->second.size();
  cv::Size const block_sz = images_size/n_subdivs;
  assert(block_sz.area() > 0);

  // -- (1/3) calculate mean intensity in each rectangular tile, in each channel
  //
  // (1a) prepare data structures
  using rect_props_t = std::tuple<std::size_t, float>; //rect_idx, avg_intensity
  std::map<ImageChannel, std::vector<rect_props_t>> rects_intensities;
  auto const approx_n_rectangles = AS_SIZE(n_subdivs*n_subdivs*(n_subshifts + 1));
  for (auto const& ch_im: images) {  // channel description & ptr to image
    rects_intensities.emplace(ch_im.first, std::vector<rect_props_t>{});
    rects_intensities.find(ch_im.first)->second.reserve(approx_n_rectangles);
  }
  //
  // 1b) create rectangles and analyze intensities in all channels
  std::vector<contour_t> rectangles;
  rectangles.reserve(approx_n_rectangles);
  int const incr_i = block_sz.width/(n_subshifts + 1);
  int const incr_j = block_sz.height/(n_subshifts + 1);
  long rect_i = 0;
  for (int i=0; i<images_size.width-1; i+=incr_i) {
    for (int j=0; j<images_size.height-1; j+=incr_j) {
      int const next_i = std::min(i + block_sz.width,  images_size.width - 1),
                next_j = std::min(j + block_sz.height, images_size.height - 1);
      if ((next_i - i +1 <= incr_i) or (next_j - j +1 <= incr_j)) { continue; }
      rectangles.push_back(st::contour_t{
           cv::Point{     i,      j}, cv::Point{next_i,      j},
           cv::Point{next_i, next_j}, cv::Point{     i, next_j}});
      long const n_pixels_complete = ((next_i - i + 1)*(next_j - j + 1));
      assert(n_pixels_complete > 0);
      for (auto const& ch_im: images) {
        auto const ch  = ch_im.first;
        cv::Mat const& im = ch_im.second;
        long intensities_sum = 0;
        long n_pixels = 0;
        for (int roi_i=i; roi_i<=next_i; ++roi_i) {
          for (int roi_j=j; roi_j<=next_j; ++roi_j) {
            if (xmask_img.at<unsigned char>(roi_j, roi_i) == 255) { continue; }
            auto const v = static_cast<int>(im.at<unsigned char>(roi_j, roi_i));
            intensities_sum += v;
            n_pixels++;
        } }
        float const avg_intensity =
            ((n_pixels > 0) and (n_pixels == n_pixels_complete))
                ? static_cast<float>(intensities_sum)/n_pixels
                : std::numeric_limits<float>::max();
        rects_intensities.find(ch)->second.emplace_back(rect_i, avg_intensity);
      } // for ch_im
      ++rect_i;
  } } // for i, for j

  // -- (2/3) create combined rank based on partial (single-channel) rankings
  struct ranking_entry_t { std::size_t rect_index, rect_combined_rank; };
  std::vector<ranking_entry_t> ranking(rectangles.size(), {0, 1});
  for (auto const& ch_im: images) {
    auto const ch = ch_im.first;
    auto& rcts_ntns = rects_intensities.find(ch)->second;
    std::sort(rcts_ntns.begin(), rcts_ntns.end(),
              [](rect_props_t const& rp1, rect_props_t const& rp2)
                { return std::get<1>(rp1) < std::get<1>(rp2); });
    for (std::size_t rank_i=0; rank_i<AS_SIZE(rectangles.size()); ++rank_i) {
      auto const rect_ii = std::get<0>(rcts_ntns[rank_i]);
      ranking[rect_ii].rect_index = rect_ii;
      ranking[rect_ii].rect_combined_rank *= (1 + rank_i);
  } }
  std::sort(ranking.begin(), ranking.end(),
            [](ranking_entry_t const& A, ranking_entry_t const& B)
              { return A.rect_combined_rank < B.rect_combined_rank; });

  // -- (3/3) pick the highest ranked contours
  for (std::size_t contour_i=0; contour_i<AS_SIZE(n_contours); ++contour_i) {
    auto const top_rectangle_index = ranking[contour_i].rect_index;
    bg_contours->push_back(rectangles[top_rectangle_index]);
  }
} // suggestBackgroundTiles()


std::pair<corresp_voronois_points_distances_t, float>
getCorrespVoronoisPointsDistances_(cv::Size const& img_size,
                                   pcontours_vec_t const& contours_ptrs
#ifdef DEBUG
                                  , float (*scaling_fun)(float, float)
#endif
                                  )
{
  auto const black = cv::Scalar{  0,   0  , 0};
  auto const white = cv::Scalar{255, 255, 255};

  auto const n_nuclei = contours_ptrs.size();

  cv::Mat masks = cv::Mat::zeros(img_size, CV_8UC1);
  masks = white;  // background

  // (using std::pair instead of cv::Point because pairs are sortable)
  std::map<std::pair<int, int>, std::size_t> contour_point_to_contour_index;

  std::vector<float> areas;
  areas.reserve(n_nuclei);

  std::size_t ci = 0;
  for (contour_t const* pc: contours_ptrs) {

    // filled black nuclear contours
    cv::drawContours(masks, contours_vec_t{*pc}, 0, black, -1);

    // map each contour point -> contour index
    for (auto const& p: *pc) {
      contour_point_to_contour_index.emplace(std::make_pair(p.x, p.y), ci);
    }

    // collect areas
    areas.push_back(static_cast<float>(cv::contourArea(*pc)));

    ci++;
  }


  // -- statistics of areas -> reasonable filter kernel sizes

  std::sort(areas.begin(), areas.end());
  float const median_area = areas[areas.size() / 2];
  int const diameter         = 2*(static_cast<int>(std::sqrt(median_area/3.141592f))/2) + 1;
  int const diameter_onehalf = 2*(static_cast<int>(diameter*1.5f)/2) + 1;
  int const diameter_twice   = 2* static_cast<int>(diameter) + 3;
  assert((diameter         % 2) == 1);
  assert((diameter_onehalf % 2) == 1);
  assert((diameter_twice   % 2) == 1);


  // -- image pre-processing
  //
  // The aim of these section is to additionally softly discard regions,
  // where the nuclei lie next to each other. This is achieved by processing
  // the image with gaussian filter: narrow passages between nuclei are
  // dimmed more by accumulating contributions from the neighboring nuclei.
  // We however do not want to downplay the importance of pixels close to
  // the contour of a well isolated (part of the) nucleus. Hence, median blur
  // comes to help.

  cv::Mat masks_shrunk, masks_shrunk_blurred, masks_shrunk_blurred_filtered;
  cv::Mat const elem3 = cv::getStructuringElement(cv::MORPH_ELLIPSE, cv::Size{3, 3});
  cv::dilate(masks, masks_shrunk, elem3);
  cv::GaussianBlur(masks_shrunk, masks_shrunk_blurred,
                   cv::Size{diameter_onehalf, diameter_onehalf}, 0, 0);
  cv::medianBlur(masks_shrunk_blurred, masks_shrunk_blurred_filtered,
                 diameter_twice);
//cv::imshow("0", masks);
//cv::imshow("1", masks_shrunk);
//cv::imshow("2", masks_shrunk_blurred);
//cv::imshow("3", masks_shrunk_blurred_filtered);


  // -- computation of distance map and voronoi
  //
  // We follow the precise procedure (than may be executed in parallel
  // if OpenCV has been compiled with TBB support).

  cv::Mat distances, labels;
  cv::distanceTransform(masks, distances, labels, cv::DIST_L2,
                        cv::DIST_MASK_PRECISE, cv::DIST_LABEL_PIXEL);


  // -- mapping all labeled non-in-nuclei pixels to respective nucleus
  //    contour points
  //
  // In this way we have will be able to capture both the voronoi cell areas
  // and their correspondence to specific nuclei.

  std::map<int, std::pair<int, int>> label_to_contour_point;

  for (int y = 0; y < labels.rows; y++) {
    auto const labels_row    = reinterpret_cast<const int*>(labels.ptr(y));
    auto const distances_row = reinterpret_cast<const float*>(distances.ptr(y));

    for (int x = 0; x < labels.cols; x++) {
      auto const label = labels_row[x];
      auto const dist = distances_row[x];
      auto const xy = std::make_pair(x, y);
      bool const in_nucleus = not (dist > 0.f);
      bool const in_contour = in_nucleus and contour_point_to_contour_index.count(xy);

      if (in_contour) {
        if (not label_to_contour_point.count(label)) {
          label_to_contour_point.emplace(label, xy);
        } else {
          assert(label_to_contour_point.at(label) == xy);
      } }
  } } // for y, for x


#ifdef DEBUG
  // -- displaying a diagnostic image

   cv::Mat img_debug = cv::Mat::zeros(labels.size(), CV_8UC3);

  for (int y = 0; y < labels.rows; y++) {
    auto const labels_row    = reinterpret_cast<const int*>(labels.ptr(y));
    auto const distances_row = reinterpret_cast<const float*>(distances.ptr(y));
    auto const blurred_row   = reinterpret_cast<const unsigned char*>(masks_shrunk_blurred_filtered.ptr(y));

    auto img_debug_row = reinterpret_cast<unsigned char*>(img_debug.ptr(y));
    for (int x = 0; x < labels.cols; x++) {
      auto const label = labels_row[x];
      auto const dist = distances_row[x];
      auto const xy = std::make_pair(x, y);
      auto const avoidance_ = static_cast<int>(blurred_row[x])/255.f;
      auto const avoidance  = avoidance_*avoidance_*avoidance_;

          if (dist > 0.) {
            if (label_to_contour_point.count(label)) {
              float const scale = avoidance * scaling_fun(dist, diameter);  // crucial!
            //auto const contour_xy = label_to_contour_point.at(label);
            //auto const contour_i = contour_point_to_contour_index.at(contour_xy);
              img_debug_row[x*3    ] = scale*255;
              img_debug_row[x*3 + 1] = scale*255;
              img_debug_row[x*3 + 2] = scale*255;
            } else {
              // very rare, no need to be handled
            }
          } else {
            if (contour_point_to_contour_index.count(xy)) {
              img_debug_row[x*3    ] = 255;
              img_debug_row[x*3 + 1] = 128;
              img_debug_row[x*3 + 2] = 255;
            } else { // then nucleus interior
              img_debug_row[x*3    ] = 255;
              img_debug_row[x*3 + 1] =   0;
              img_debug_row[x*3 + 2] =   0;
            }
          }
  } } // for y, for x

  cv::imshow("avoidance and distance-related weights in voronoi cells", img_debug);
  cv::waitKey(0);
#endif

  corresp_voronois_points_distances_t
      corresp_voronois_points_distances(n_nuclei, voronoi_weighted_points_t{});

  for (auto& vwps: corresp_voronois_points_distances) {
    vwps.reserve(img_size.height * img_size.width / n_nuclei * 2);
  }

  for (int y = 0; y < labels.rows; y++) {
    auto const labels_row    = reinterpret_cast<const int*>(labels.ptr(y));
    auto const distances_row = reinterpret_cast<const float*>(distances.ptr(y));
    auto const blurred_row   = reinterpret_cast<const unsigned char*>(masks_shrunk_blurred_filtered.ptr(y));

    for (int x = 0; x < labels.cols; x++) {
      auto const label = labels_row[x];

      if (label_to_contour_point.count(label)) {
        auto const contour_xy = label_to_contour_point.at(label);
        auto const contour_i = contour_point_to_contour_index.at(contour_xy);
        auto const xy = cv::Point{x, y};
        float const distance = distances_row[x];
        float const avoidance_ = static_cast<int>(blurred_row[x])/255.f;
        float const avoidance  = avoidance_*avoidance_*avoidance_;

        assert(contour_i < n_nuclei);
        corresp_voronois_points_distances[contour_i].emplace_back(xy, distance, avoidance);
      }
  } } // for y, for x

  for (auto& vwps: corresp_voronois_points_distances) {
    vwps.shrink_to_fit();
  }

  return {corresp_voronois_points_distances, diameter};
}





// -----------------------------------------------------------------------------
// Code below is an apatation of:
//  * opencv-4.2.0/modules/imgproc/src/grabcut.cpp,
//  * opencv-4.2.0/modules/imgproc/include/opencv2/imgproc/detail/gcgraph.hpp
//    (see header for licensing info).
// -----------------------------------------------------------------------------

namespace nuccut {

int constexpr kGC_NONE = cv::GC_BGD + cv::GC_PR_BGD + cv::GC_FGD + cv::GC_PR_FGD;

GMM::GMM()
: model{1, kModelSize*kComponents, CV_64FC1, cv::Scalar{0}},
  component_weight{model.ptr<double>(0)},
  mean{component_weight + kComponents},
  cov{mean + 3*kComponents}
{
  for (int ci = 0; ci < kComponents; ci++) {
    if (component_weight[ci] > 0.) {
      calcInverseCovAndDet_(ci, 0.);
  } }
}


void GMM::beginLearning()
{
  for (int ci = 0; ci < kComponents; ci++) {
    sums[ci][0] = sums[ci][1] = sums[ci][2] = 0.;
    prod_sums[ci][0][0] = prod_sums[ci][0][1] = prod_sums[ci][0][2] = 0.;
    prod_sums[ci][1][0] = prod_sums[ci][1][1] = prod_sums[ci][1][2] = 0.;
    prod_sums[ci][2][0] = prod_sums[ci][2][1] = prod_sums[ci][2][2] = 0.;
    nSamples[ci] = 0;
  }
  nSamplesTotal = 0;
}


void GMM::addSample(int ci, cv::Vec3d const& color)
{
  sums[ci][0] += color[0];
  sums[ci][1] += color[1];
  sums[ci][2] += color[2];
  prod_sums[ci][0][0] += color[0]*color[0];
  prod_sums[ci][0][1] += color[0]*color[1];
  prod_sums[ci][0][2] += color[0]*color[2];
  prod_sums[ci][1][0] += color[1]*color[0];
  prod_sums[ci][1][1] += color[1]*color[1];
  prod_sums[ci][1][2] += color[1]*color[2];
  prod_sums[ci][2][0] += color[2]*color[0];
  prod_sums[ci][2][1] += color[2]*color[1];
  prod_sums[ci][2][2] += color[2]*color[2];
  nSamples[ci]++;
  nSamplesTotal++;
}


void GMM::endLearning(cv::Vec3d const& channel_weights)
{
  // Covariance matrix is stretched in dimensions of weights that are < 1.0.
  constexpr double weight_eps = 1e-3;
  const double w0 = 1./(weight_eps + channel_weights[0]);
  const double w1 = 1./(weight_eps + channel_weights[1]);
  const double w2 = 1./(weight_eps + channel_weights[2]);

  for (int ci = 0; ci < kComponents; ci++) {
    int n = nSamples[ci];
    if (n == 0) {
      component_weight[ci] = 0.;
    } else {
      assert(nSamplesTotal > 0);
      double const inv_n = 1./n;

      component_weight[ci] = AS_DOUBLE(n)/nSamplesTotal;

      double* const m = mean + 3*ci;
      m[0] = sums[ci][0]*inv_n;
      m[1] = sums[ci][1]*inv_n;
      m[2] = sums[ci][2]*inv_n;

      double* const c = cov + 9*ci;
      c[0] = w0*(prod_sums[ci][0][0]*inv_n - m[0]*m[0]);
      c[1] = w1*(prod_sums[ci][0][1]*inv_n - m[0]*m[1]);
      c[2] = w2*(prod_sums[ci][0][2]*inv_n - m[0]*m[2]);
      c[3] = w0*(prod_sums[ci][1][0]*inv_n - m[1]*m[0]);
      c[4] = w1*(prod_sums[ci][1][1]*inv_n - m[1]*m[1]);
      c[5] = w2*(prod_sums[ci][1][2]*inv_n - m[1]*m[2]);
      c[6] = w0*(prod_sums[ci][2][0]*inv_n - m[2]*m[0]);
      c[7] = w1*(prod_sums[ci][2][1]*inv_n - m[2]*m[1]);
      c[8] = w2*(prod_sums[ci][2][2]*inv_n - m[2]*m[2]);
      calcInverseCovAndDet_(ci, 0.01);
  } }
}


void GMM::calcInverseCovAndDet_(int ci, double singular_fix)
{
  constexpr double kMinDet = 1e-6;
  if (component_weight[ci] > 0.) {
    double* const c = cov + 9*ci;
    double det =  c[0]*(c[4]*c[8] - c[5]*c[7])
                - c[1]*(c[3]*c[8] - c[5]*c[6])
                + c[2]*(c[3]*c[7] - c[4]*c[6]);
    if ((det <= kMinDet) and (singular_fix > 0.)) {
        c[0] += singular_fix;  // Add some noise
        c[4] += singular_fix;  // to avoid singular
        c[8] += singular_fix;  // covariance matrix.
        det =  c[0]*(c[4]*c[8] - c[5]*c[7])
             - c[1]*(c[3]*c[8] - c[5]*c[6])
             + c[2]*(c[3]*c[7] - c[4]*c[6]);
    }
    covDets[ci] = det;

    assert(det > std::numeric_limits<double>::epsilon());
    double const inv_det = 1./det;
    inverseCovs[ci][0][0] =  (c[4]*c[8] - c[5]*c[7])*inv_det;
    inverseCovs[ci][1][0] = -(c[3]*c[8] - c[5]*c[6])*inv_det;
    inverseCovs[ci][2][0] =  (c[3]*c[7] - c[4]*c[6])*inv_det;
    inverseCovs[ci][0][1] = -(c[1]*c[8] - c[2]*c[7])*inv_det;
    inverseCovs[ci][1][1] =  (c[0]*c[8] - c[2]*c[6])*inv_det;
    inverseCovs[ci][2][1] = -(c[0]*c[7] - c[1]*c[6])*inv_det;
    inverseCovs[ci][0][2] =  (c[1]*c[5] - c[2]*c[4])*inv_det;
    inverseCovs[ci][1][2] = -(c[0]*c[5] - c[2]*c[3])*inv_det;
    inverseCovs[ci][2][2] =  (c[0]*c[4] - c[1]*c[3])*inv_det;
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void MaxFlowGraph::addEdges(int i, int j, MaxFlowGraph::Weight w_forward,
                                          MaxFlowGraph::Weight w_reverse)
{
  assert(0 <= i and i < SSIZE(vertices));
  assert(0 <= j and j < SSIZE(vertices));
  assert(0 <= w_forward and 0 <= w_reverse);
  assert(i != j);

  if (edges.empty()) { edges.resize(2); }

  Edge const e_from{j, vertices[i].first, w_forward};
  vertices[i].first = SSIZE(edges);
  edges.push_back(e_from);

  Edge const e_to{i, vertices[j].first, w_reverse};
  vertices[j].first = SSIZE(edges);
  edges.push_back(e_to);
}


void MaxFlowGraph::addTermWeights(int i, MaxFlowGraph::Weight w_source,
                                         MaxFlowGraph::Weight w_sink)
{
  assert(0 <= 0 and i < SSIZE(vertices));

  MaxFlowGraph::Weight dw = vertices[i].weight;
  if (dw > 0.) { w_source += dw; }
  else         { w_sink   -= dw; }
  flow += (w_source < w_sink) ? w_source : w_sink;
  vertices[i].weight = w_source - w_sink;
}


MaxFlowGraph::Weight MaxFlowGraph::maxFlow()
{
  constexpr int kTerminal = -1,
                kOrphan = -2;
  Vertex stub, *nilNode = &stub,
               *first = nilNode,
               *last = nilNode;
  int curr_ts = 0;
  stub.next = nilNode;
  Vertex *vtxPtr = &vertices[0];
  Edge *edgePtr = &edges[0];

  std::vector<Vertex*> orphans;

  // initialize the active queue and the graph vertices
  for (int i = 0; i < SSIZE(vertices); i++) {
    Vertex* v = vtxPtr + i;
    v->ts = 0;
    if (not gtest::almost_zero(v->weight)) {
      last = last->next = v;
      v->dist = 1;
      v->parent = kTerminal;
      v->t = v->weight < 0;
    } else {
      v->parent = 0;
  } }
  first = first->next;
  last->next = nilNode;
  nilNode->next = 0;

  // run the search-path -> augment-graph -> restore-trees loop
  while (true) {
    Vertex* v, *u;
    int e0 = -1,  ei = 0,  ej = 0;
    MaxFlowGraph::Weight minWeight, weight;
    uchar vt;

    // grow S & T search trees, find an edge connecting them
    while (first != nilNode) {
      v = first;
      if (v->parent) {
        vt = v->t;
        for (ei = v->first; ei != 0; ei = edgePtr[ei].next) {
          if (gtest::almost_zero(edgePtr[ei^vt].weight)) { continue; }
          u = vtxPtr + edgePtr[ei].dst;
          if (not u->parent) {
            u->t = vt;
            u->parent = ei ^ 1;
            u->ts = v->ts;
            u->dist = v->dist + 1;
            if (!u->next) {
              u->next = nilNode;
              last = last->next = u;
            }
            continue;
          } // if not u->parent

          if (u->t != vt) { e0 = ei ^ vt;  break; }

          if (u->dist > v->dist+1 and u->ts <= v->ts) {
            // re-assign the parent
            u->parent = ei ^ 1;
            u->ts = v->ts;
            u->dist = v->dist + 1;
          }
        } // for ei
        if (e0 > 0) { break; }
      }
      // exclude the vertex from the active list
      first = first->next;
      v->next = 0;
    }

    if (e0 <= 0) { break; }

    // find the minimum edge weight along the path
    minWeight = edgePtr[e0].weight;
    assert(minWeight > 0);
    // k = 1: source tree, k = 0: destination tree
    for (int k = 1; k >= 0; k--) {
      for (v = vtxPtr + edgePtr[e0^k].dst;; v = vtxPtr + edgePtr[ei].dst) {
        if ((ei = v->parent) < 0) { break; }
        weight = edgePtr[ei^k].weight;
        minWeight = MIN(minWeight, weight);
        assert(minWeight > 0);
      }
      weight = fabs(v->weight);
      minWeight = MIN(minWeight, weight);
      assert(minWeight > 0);
    }

    // modify weights of the edges along the path and collect orphans
    edgePtr[e0].weight -= minWeight;
    edgePtr[e0^1].weight += minWeight;
    flow += minWeight;

    // k = 1: source tree, k = 0: destination tree
    for (int k = 1; k >= 0; k--) {
      for (v = vtxPtr + edgePtr[e0^k].dst;; v = vtxPtr + edgePtr[ei].dst) {
        if ((ei = v->parent) < 0) { break; }
        edgePtr[ei^(k^1)].weight += minWeight;
        if (gtest::almost_zero(edgePtr[ei^k].weight -= minWeight)) {
          orphans.push_back(v);
          v->parent = kOrphan;
      } }

      v->weight = v->weight + minWeight*(1 - k*2);
      if (gtest::almost_zero(v->weight)) {
        orphans.push_back(v);
        v->parent = kOrphan;
      }
    } // for k

    // restore the search trees by finding new parents for the orphans
    curr_ts++;
    while (not orphans.empty()) {
      Vertex* v2 = orphans.back();
      orphans.pop_back();

      int d,  minDist = INT_MAX;
      e0 = 0;
      vt = v2->t;

      for (ei = v2->first; ei != 0; ei = edgePtr[ei].next) {
        if (gtest::almost_zero(edgePtr[ei^(vt^1)].weight)) { continue; }
        u = vtxPtr + edgePtr[ei].dst;
        if (u->t != vt or u->parent == 0) { continue; }

        // compute the distance to the tree root
        for (d = 0;;) {
          if (u->ts == curr_ts) { d += u->dist; break; }
          ej = u->parent;
          d++;
          if (ej < 0) {
            if (ej == kOrphan) { d = INT_MAX-1; }
            else { u->ts = curr_ts;  u->dist = 1;}
            break;
          }
          u = vtxPtr+edgePtr[ej].dst;
        }

        // update the distance
        if (++d < INT_MAX) {
          if (d < minDist) { minDist = d;  e0 = ei; }
          for (u = vtxPtr + edgePtr[ei].dst; u->ts != curr_ts;
               u = vtxPtr + edgePtr[u->parent].dst) {
            u->ts = curr_ts;
            u->dist = --d;
      } } } // for ei

      if ((v2->parent = e0) > 0) {
        v2->ts = curr_ts;
        v2->dist = minDist;
        continue;
      }

      // no parent found
      v2->ts = 0;
      for (ei = v2->first; ei != 0; ei = edgePtr[ei].next) {
        u = vtxPtr + edgePtr[ei].dst;
        ej = u->parent;
        if (u->t != vt or !ej) { continue; }
        if (not gtest::almost_zero(edgePtr[ei^(vt^1)].weight) and not u->next) {
          u->next = nilNode;
          last = last->next = u;
        }
        if (ej > 0 and vtxPtr+edgePtr[ej].dst == v2) {
          orphans.push_back(u);
          u->parent = kOrphan;
      } }
    } // while orphans
  } // main loop
  return flow;
} // maxFlow()

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void initMask(cv::Mat& mask, cv::Rect& bbox,
              st::contour_t const& contour,
              float contour_area,
              cv::Size image_size,
              float median_area,
              float fg_area_fraction,
              float prbg_radius_extension,
              float bg_radius_further_extension)
{
  assert(not contour.empty());
  assert(contour_area > 0.f);
  assert((0.f < fg_area_fraction) and (fg_area_fraction <= 1.f));
#define PREVIEW 0

  float const relative_area = contour_area/median_area;
  float const fg_area = contour_area
      *((relative_area < 1.f) ? 1.f - (1.f - fg_area_fraction)*relative_area
                              : fg_area_fraction);
  float const approx_fg_radius = std::sqrt(fg_area/3.141592f);
  float const approx_radius = std::sqrt(contour_area/3.141592f);

  int shrink_to_fg = 2.f*(approx_radius - approx_fg_radius) + 0.5f;
  int enlarge_to_prbg =  approx_radius*2*prbg_radius_extension
                        /((relative_area < 1.f) ? relative_area : 1.f);
  int enlarge_further_to_bg = approx_radius*2*bg_radius_further_extension;
//assert(shrink_to_sure > 0);  // TODO: handle the case of ==0 separately?
  assert(enlarge_to_prbg > 0);
  assert(enlarge_further_to_bg > 0);

  shrink_to_fg          += static_cast<int>((shrink_to_fg          %2) == 0);
  enlarge_to_prbg       += static_cast<int>((enlarge_to_prbg       %2) == 0);
  enlarge_further_to_bg += static_cast<int>((enlarge_further_to_bg %2) == 0);
  assert(shrink_to_fg % 2 == 1);
  assert(enlarge_to_prbg % 2 == 1);
  assert(enlarge_further_to_bg% 2 == 1);

  int const enlarge_total = enlarge_to_prbg/2 + enlarge_further_to_bg/2 + 1;
  bbox = boundingBox(contour, enlarge_total, image_size);
  cv::Point const bbox_tl = bbox.tl();

  mask.create(bbox.size(), CV_8UC1);  mask.setTo(kGC_NONE);

  cv::Mat m_nuc = cv::Mat::zeros(bbox.size(), CV_8UC1);  // nuclear contour
  drawMask(m_nuc, contour, -bbox_tl);

  auto mkElem = [](int sz) {
    return cv::getStructuringElement(cv::MORPH_ELLIPSE, cv::Size{sz, sz});
  };

  cv::Mat m_fg;
  cv::erode(m_nuc, m_fg, mkElem(shrink_to_fg));

  cv::Mat const m_prfg = m_nuc - m_fg;

  cv::Mat m_enlarged;
  cv::dilate(m_nuc, m_enlarged, mkElem(enlarge_to_prbg));

  cv::Mat m_enlarged2 = m_enlarged.clone();
  cv::dilate(m_enlarged, m_enlarged2, mkElem(enlarge_further_to_bg));

  cv::Mat const m_prbg = m_enlarged  - m_nuc;
  cv::Mat const m_bg   = m_enlarged2 - m_enlarged;

#if PREVIEW
  std::vector<cv::Point> fg_ps, bg_ps, prfg_ps, prbg_ps;
#endif
  cv::Size const bbox_size = bbox.size();
  for (cv::Point p{0, 0}; p.y < bbox_size.height; p.y++) {
    for (p.x = 0; p.x < bbox_size.width; p.x++) {
      if (m_bg.at<uchar>(p) == 255) {
        mask.at<uchar>(p) = cv::GC_BGD;
#if PREVIEW
        bg_ps.push_back(p);
#endif
      } else if (m_prbg.at<uchar>(p) == 255) {
#if PREVIEW
        prbg_ps.push_back(p);
#endif
        mask.at<uchar>(p) = cv::GC_PR_BGD;
      } else if (m_fg.at<uchar>(p) == 255) {
        mask.at<uchar>(p) = cv::GC_FGD;
#if PREVIEW
        fg_ps.push_back(p);
#endif
      } else if (m_prfg.at<uchar>(p) == 255) {
        mask.at<uchar>(p) = cv::GC_PR_FGD;
#if PREVIEW
        prfg_ps.push_back(p);
#endif
  } } }

#if PREVIEW
  auto const RED  = cv::Vec3b{  0,   0, 255};
  auto const PINK = cv::Vec3b{230, 130, 255};
  auto const BLUE = cv::Vec3b{255,   0,   0};
  auto const LBLU = cv::Vec3b{255, 255, 160};
  cv::Mat preview(mask.size(), CV_8UC3, cv::Scalar{0});
  for (auto const& p: bg_ps)   { preview.at<cv::Vec3b>(p) = BLUE; }
  for (auto const& p: fg_ps)   { preview.at<cv::Vec3b>(p) = RED;  }
  for (auto const& p: prbg_ps) { preview.at<cv::Vec3b>(p) = LBLU; }
  for (auto const& p: prfg_ps) { preview.at<cv::Vec3b>(p) = PINK; }
  cv::drawContours(preview, contours_vec_t{contour}, 0, cv::Scalar{0}, 1,
                   cv::LINE_4, cv::noArray(), std::numeric_limits<int>::max(),
                   -bbox_tl);
  cv::imshow("NucCut: labels_in_mask", preview);  // cv::waitKey();
#endif
#undef PREVIEW
} // initMask()

#define AS_VEC3D_YX(IMG, Y, X)  static_cast<cv::Vec3d>(IMG.at<cv::Vec3b>(Y, X))
#define AS_VEC3F_PT(IMG, P)     static_cast<cv::Vec3f>(IMG.at<cv::Vec3b>(P))

// recommended_beta = 1/( 2*avg(sqr(||color[i] - color[j]||)) )
double calcRecommendedBeta(cv::Mat const& img)
{
  double beta = 0.;
  for (int y = 0; y < img.rows; y++) {
    for (int x = 0; x < img.cols; x++) {
      cv::Vec3d const color = img.at<cv::Vec3b>(y, x);
      if (x > 0) {  // left
        cv::Vec3d const diff = color - AS_VEC3D_YX(img, y, x - 1);
        beta += diff.dot(diff);
        if (y > 0) {  // up-left
          cv::Vec3d const diff2 = color - AS_VEC3D_YX(img, y - 1, x - 1);
          beta += diff2.dot(diff2);
      } }
      if (y > 0) {  // up
        cv::Vec3d const diff = color - AS_VEC3D_YX(img, y - 1, x);
        beta += diff.dot(diff);
        if (x < img.cols - 1) {  // up-right
          cv::Vec3d const diff2 = color - AS_VEC3D_YX(img, y - 1, x + 1);
          beta += diff2.dot(diff2);
  } } } }

  if (beta <= std::numeric_limits<double>::epsilon()) {
    beta = 0.;
  } else {
    beta = 1./(2 * beta/(4*img.cols*img.rows - 3*img.cols - 3*img.rows + 2));
  }
  return beta;
}


void calcNontermWeights(cv::Mat& wL, cv::Mat& wUL, cv::Mat& wU, cv::Mat& wUR,
                        cv::Mat const& img, cv::Mat const& img_edges,
                        double beta, double gamma, double delta)
{
  assert((img.type() == CV_8UC3) and (img_edges.type() == CV_32FC1));
  assert(img.size() == img_edges.size());
#define PREVIEW 0 //1

  double constexpr sqrt2 = 1.4142135623730951;
  double constexpr two_sqrt2 = 2*sqrt2;
  double const gammaDivSqrt2 = gamma/sqrt2;
  wL .create(img.rows, img.cols, CV_64FC1);
  wUL.create(img.rows, img.cols, CV_64FC1);
  wU .create(img.rows, img.cols, CV_64FC1);
  wUR.create(img.rows, img.cols, CV_64FC1);
  for (int y = 0; y < img.rows; y++) {
    for (int x = 0; x < img.cols; x++) {
      cv::Vec3d const color = img.at<cv::Vec3b>(y, x);
      float const edginess_yx = (1.f - img_edges.at<float>(y, x))/2.;
    //float const delta_edge = delta*(1.f - edginess);
    //float const deltaDivSqrt2_edge = deltaDivSqrt2*(1.f - edginess);
      if (x-1 >= 0) {  // left
        cv::Vec3d const d = color - AS_VEC3D_YX(img, y, x-1);
        double const e = edginess_yx + (1. - img_edges.at<float>(y, x-1))/2.;
        wL.at<double>(y, x) = gamma*exp(-beta*d.dot(d)) + delta*e;
      } else {
        wL.at<double>(y, x) = 0.;
      }
      if (x-1 >= 0 and y-1 >= 0) {  // up-left
        cv::Vec3d const d = color - AS_VEC3D_YX(img, y-1, x-1);
        double const e = edginess_yx + (1. - img_edges.at<float>(y-1, x-1))/two_sqrt2;
        wUL.at<double>(y, x) = gammaDivSqrt2*exp(-beta*d.dot(d)) + delta*e;
      } else {
        wUL.at<double>(y, x) = 0.;
      }
      if (y-1 >= 0) {  // up
        cv::Vec3d const d = color - AS_VEC3D_YX(img, y-1, x);
        double const e = edginess_yx + (1. - img_edges.at<float>(y-1, x))/2.;
        wU.at<double>(y, x) = gamma*exp(-beta*d.dot(d)) + delta*e;
      } else {
        wU.at<double>(y, x) = 0.;
      }
      if (x+1 < img.cols and y-1 >= 0) {  // up-right
        cv::Vec3d const d = color - AS_VEC3D_YX(img, y-1, x+1);
        double const e = edginess_yx + (1. - img_edges.at<float>(y-1, x+1))/two_sqrt2;
        wUR.at<double>(y, x) = gammaDivSqrt2*exp(-beta*d.dot(d)) + delta*e;
      } else {
        wUR.at<double>(y, x) = 0.;
  } } }

#if PREVIEW
  cv::Mat preview = cv::Mat::zeros(img.rows, img.cols, CV_64FC1);
  preview += wL + wUL + wU + wUR;
  double mi, ma;
  cv::minMaxIdx(preview, &mi, &ma);
//printf("NucCut: wNontermSum range: %f...%f\n", mi, ma);
  cv::normalize(preview, preview, 0.0, 1.0, cv::NORM_MINMAX, CV_32FC1);
  cv::imshow("NucCut: wNontermSum", preview);
#endif
#undef PREVIEW
}


// Initialize background and foreground models using k-means.
void initModels(Model& bg, Model& fg, const cv::Mat& mask, const cv::Mat& img,
                cv::Vec3d const& channel_weights,
                int kmeans_n_iter, int kmeans_init_type)
{
  assert((mask.type() == CV_8UC1) and (img.type() == CV_8UC3));
  assert(mask.size() == img.size());

  cv::Mat bg_labels, fg_labels;
  std::vector<cv::Vec3f> bg_samples, fg_samples;
  for (cv::Point p{0, 0}; p.y < img.rows; p.y++) {
    for (p.x = 0; p.x < img.cols; p.x++) {
      uchar const m = mask.at<uchar>(p);
      if (m == cv::GC_BGD or m == cv::GC_PR_BGD) {
        bg_samples.push_back(AS_VEC3F_PT(img, p));
      } else if (m == cv::GC_FGD or m == cv::GC_PR_FGD) { // prev.excl.GC_PR_FGD
        fg_samples.push_back(AS_VEC3F_PT(img, p));
  } } }
  assert(not bg_samples.empty() and not fg_samples.empty());

  // note: outcome of kmeans() depends on the state of the RNG
  auto term = cv::TermCriteria{cv::TermCriteria::Type::COUNT, kmeans_n_iter, 0.};
  cv::Mat _bg_samples(bg_samples.size(), 3, CV_32FC1, &bg_samples[0][0]);
  int const bg_n_clus = std::min(bg.nComponents(), AS_INT(bg_samples.size()));
  kmeans(_bg_samples, bg_n_clus, bg_labels, term, 0, kmeans_init_type);
  cv::Mat _fg_samples(fg_samples.size(), 3, CV_32FC1, &fg_samples[0][0]);
  int const fg_n_clust = std::min(fg.nComponents(), AS_INT(fg_samples.size()));
  kmeans(_fg_samples, fg_n_clust, fg_labels, term, 0, kmeans_init_type);

  bg.beginLearning();
  for (int i = 0; i < SSIZE(bg_samples); i++) {
    bg.addSample(bg_labels.at<int>(i, 0), bg_samples[i]);
  }
  bg.endLearning(channel_weights);

  fg.beginLearning();
  for (int i = 0; i < SSIZE(fg_samples); i++) {
    fg.addSample(fg_labels.at<int>(i, 0), fg_samples[i]);
  }
  fg.endLearning(channel_weights);
}

#undef AS_VEC3D_YX
#undef AS_VEC3F_PT

void learn(Model& bg, Model& fg, cv::Mat const& mask, cv::Mat const& img,
           cv::Vec3d const& channel_weights)
{
  assert((mask.type() == CV_8UC1) and (img.type() == CV_8UC3));
  assert(mask.size() == img.size());

  // (1/2) assign components
  cv::Mat compIdxs(img.size(), CV_32SC1);
  for (cv::Point p{0, 0}; p.y < img.rows; p.y++) {
    for (p.x = 0; p.x < img.cols; p.x++) {
      uchar const m = mask.at<uchar>(p);
      cv::Vec3d const color = img.at<cv::Vec3b>(p);
      compIdxs.at<int>(p) = (m == cv::GC_BGD or m == cv::GC_PR_BGD)
                            ? bg.whichComponent(color)
                              : (m == cv::GC_FGD or m == cv::GC_PR_FGD)
                              ? fg.whichComponent(color)
                              : kGC_NONE;
  } }

  // (2/2) learn GMMs
  static_assert(std::is_same<Model, GMM>::value, "Only GMM can be learned.");
  bg.beginLearning();
  fg.beginLearning();
  for (int ci = 0; ci < GMM::kComponents; ci++) {
    for (cv::Point p{0, 0}; p.y < img.rows; p.y++) {
      for (p.x = 0; p.x < img.cols; p.x++) {
        if (compIdxs.at<int>(p) == ci) {
          uchar const m = mask.at<uchar>(p);
          if (m == cv::GC_BGD or m == cv::GC_PR_BGD) {
            bg.addSample(ci, img.at<cv::Vec3b>(p));
          } else if (m == cv::GC_FGD or m == cv::GC_PR_FGD) {
            fg.addSample(ci, img.at<cv::Vec3b>(p));
  } } } } }
  bg.endLearning(channel_weights);
  fg.endLearning(channel_weights);
}


void segment(cv::Mat& mask, cv::Mat const& img, const Model& bg, const Model& fg,
             cv::Mat const& wL, cv::Mat const& wUL, cv::Mat const& wU, cv::Mat const& wUR,
             double lambda)
{
  assert((mask.type() == CV_8UC1) and (img.type() == CV_8UC3));
  assert(mask.size() == img.size());
#define PREVIEW 0 //1
#if PREVIEW
  cv::Mat preview_fg = cv::Mat::zeros(mask.size(), CV_8UC1);
  cv::Mat preview_bg = cv::Mat::zeros(mask.size(), CV_8UC1);
  cv::Mat preview_rr = cv::Mat::zeros(mask.size(), CV_8UC1);
#endif
  int const n_vertices = img.cols*img.rows,
            n_edges = 2*(4*img.cols*img.rows - 3*(img.cols + img.rows) + 2);

  MaxFlowGraph graph{n_vertices, n_edges};
  for (cv::Point p{0, 0}; p.y < img.rows; p.y++) {
    for (p.x = 0; p.x < img.cols; p.x++) {
      int const vtxIdx = graph.addVertex();
      auto const m = mask.at<uchar>(p);
      auto const color = img.at<cv::Vec3b>(p);

      double fromSource, toSink;
      if (m == cv::GC_PR_BGD or m == cv::GC_PR_FGD) {
        fromSource = -std::log(bg.p(color));  // source terminal -- background
        toSink     = -std::log(fg.p(color));  // sink -- foreground (object)
#if PREVIEW
//        preview_fg.at<uchar>(p) = 255 * fg.p(color);
//        preview_bg.at<uchar>(p) = 255 * bg.p(color);
//        preview_rr.at<uchar>(p) = 255 * fg.p(color)/(fg.p(color) + bg.p(color));
#endif
      } else if (m == cv::GC_FGD) {
        fromSource = lambda;
        toSink = 0.;
      } else if (m == cv::GC_BGD) {
        fromSource = 0.;
        toSink = lambda;
      } else {
        assert(m == kGC_NONE);
        continue;
      }
      graph.addTermWeights(vtxIdx, fromSource, toSink);

      // set weights of NON-TERMINAL edges
      if (m != kGC_NONE) {
        if (p.x > 0) {
          auto const w = wL.at<double>(p);
          graph.addEdges(vtxIdx, vtxIdx-1, w, w);
          if (p.y > 0) {
            auto const w2 = wUL.at<double>(p);
            graph.addEdges(vtxIdx, vtxIdx-img.cols-1, w2, w2);
        } }
        if (p.y > 0) {
          auto const w = wU.at<double>(p);
          graph.addEdges(vtxIdx, vtxIdx-img.cols, w, w);
          if (p.x < img.cols-1) {
            auto const w2 = wUR.at<double>(p);
            graph.addEdges(vtxIdx, vtxIdx-img.cols+1, w2, w2);
  } } } } }

  graph.maxFlow();
  for (cv::Point p{0, 0}; p.y < mask.rows; p.y++) {
    for (p.x = 0; p.x < mask.cols; p.x++) {
      uchar const m = mask.at<uchar>(p);
      if (m == cv::GC_PR_BGD or m == cv::GC_PR_FGD) {
        int const vertex_index = p.y*mask.cols + p.x;
        mask.at<uchar>(p) = graph.inSourceSegment(vertex_index)
                            ? cv::GC_PR_FGD : cv::GC_PR_BGD;
  } } }
#if PREVIEW
//  cv::imshow("NucCut maxflow p_fg", preview_fg);
//  cv::imshow("NucCut maxflow p_bg", preview_bg);
//  cv::imshow("NucCut maxflow p_rr", preview_rr);
#endif
}


st::contour_t extractContour(cv::Mat const& mask)
{
  cv::Mat z = cv::Mat::zeros(mask.size(), CV_8UC1);
  for (cv::Point p{0, 0}; p.y < mask.rows; p.y++) {
    for (p.x = 0; p.x < mask.cols; p.x++) { // TODO -> .ptr()
      if (   mask.at<uchar>(p) == cv::GC_PR_FGD
          or mask.at<uchar>(p) == cv::GC_FGD) {
        z.at<uchar>(p) = 255;
  } } }

  st::contours_vec_t cs;
  cv::findContours(z, cs, cv::RETR_EXTERNAL, cv::CHAIN_APPROX_NONE);

  if (cs.size() == 1) {
    assert(not cs.front().empty());
    return cs.front();
  } else if (cs.size() > 1) {
    cv::Point const center{mask.size()/2};
    int nearest_i = -1;  // "centermost"
    float nearest_distance_sq = std::numeric_limits<float>::max();
    for (int i=0; i<SSIZE(cs); ++i) {
      cv::Point2f const com = st::centerOfMass(cs[i]);
      float const dx = com.x - center.x,
                  dy = com.y - center.y;
      float const distance_sq = dx*dx + dy*dy;
      if (distance_sq < nearest_distance_sq) {
        nearest_distance_sq = distance_sq;
        nearest_i = i;
    } }
    assert(not cs[nearest_i].empty());
    return cs[nearest_i];
  } else {
    assert(cs.empty());
    throw std::runtime_error("Could not extract a NucCut contour!");
  }
}


st::contour_t nucCut(cv::Mat const& img_edges, cv::Mat const& img_orig,
                     cv::Vec3d const& channel_weights,
                     st::contour_t const& nuc_contour, float nuc_area,
                     float median_nuc_area, int max_iters,
                     double delta, double gamma, double beta, double lambda)
{
  assert(not img_orig.empty() and not img_edges.empty());
  assert(img_orig.type() == CV_8UC3 and img_edges.type() == CV_32FC1);
  assert(0. <= channel_weights[0] and channel_weights[0] <= 1.);
  assert(0. <= channel_weights[1] and channel_weights[1] <= 1.);
  assert(0. <= channel_weights[2] and channel_weights[2] <= 1.);
  assert(max_iters >= 1);

  cv::Mat mask;  cv::Rect bbox;
  initMask(mask, bbox, nuc_contour, nuc_area, img_orig.size(), median_nuc_area);
  cv::Mat const img = img_orig(bbox);
  cv::Mat const imgE = img_edges(bbox);

  Model bg, fg;
  initModels(bg, fg, mask, img);

  cv::Mat wL, wUL, wU, wUR; // may change below beta -> calcRecommendedBeta(img)
  calcNontermWeights(wL, wUL, wU, wUR, img, imgE, beta, gamma, delta);

  contour_t refined_contour;
  for (int i = 0; i < max_iters; i++) {

    learn(bg, fg, mask, img, channel_weights);              // |_  do!
    segment(mask, img, bg, fg, wL, wUL, wU, wUR, lambda);   // |

    if (refined_contour.empty()) {                          // |_ done?
      assert(i == 0);                                       // .
      refined_contour = extractContour(mask);               // .
    } else {                                                // .
      assert(i > 0);
      contour_t c = extractContour(mask);
      if ((c.size() == refined_contour.size()) and (c == refined_contour)) {
        break;
      } else {
        refined_contour.swap(c);
    } }
  } // main refinement loop

  assert(not refined_contour.empty());
  auto const bbox_tl = bbox.tl();
  for (auto& p: refined_contour) { p += bbox_tl; }

  return refined_contour;
} // nucCut()


st::contours_vec_t refine(std::vector<annotated_contour_t>&& nuclei_data,
                          cv::Mat const& image_for_edge_detection,
                          int canny_threshold_low, float canny_thresholds_ratio,
                          cv::Mat const& image_full, cv::Vec3d const& ch_weights,
                          double delta, double gamma, bool preview)
{
  assert(image_full.type() == CV_8UC3);
  assert(image_for_edge_detection.type() == CV_8UC1);
  assert(ch_weights.dot(ch_weights) > 0.);

  if (nuclei_data.empty()) {
    fprintf(stderr, "Warning: no nuclei to refine.\n");
    return {};
  }

  // -- (1/3) create an image of enhanced edges
  std::vector<float> nuc_areas;  nuc_areas.reserve(nuclei_data.size());
  for (auto const& nd: nuclei_data) { nuc_areas.push_back(nd.area); } // shouldn't exclde by provenience?
  float const median_nuc_area = nuc_areas[SSIZE(nuc_areas)/2];
  float constexpr two_pi_over_sqrt_pi = 2.f*3.141592f/1.772454f;
  float const median_nuc_len = two_pi_over_sqrt_pi*std::sqrt(median_nuc_area);
  cv::Mat const imageE = st::edgeImage(image_for_edge_detection, preview,
                   median_nuc_len, canny_threshold_low, canny_thresholds_ratio);
//return {};//TMP
  // -- (2/3) sort contours by their current match with edges
  for (auto& nd: nuclei_data) {nd.edge_score = st::edgeScore(imageE, *nd.contour);}
  std::sort(nuclei_data.begin(), nuclei_data.end(),  // better score to the front
            [](auto& ndA, auto& ndB) { return ndA.edge_score > ndB.edge_score; });

  // -- (3/3) refine individual nuclear contours -- TODO: parallelize
  contours_vec_t refined_contours;
  refined_contours.reserve(nuclei_data.size());
  for (auto const& nuc: nuclei_data) {
    refined_contours.push_back( st::nuccut::nucCut(
        imageE, image_full, ch_weights, *nuc.contour, nuc.area,
        median_nuc_area, /*max_iters:*/10, delta, gamma) );  // PARAM
    // TODO: "mask" on/"remove" from the 3-channel image
  }

  assert(refined_contours.size() == nuclei_data.size());
  return refined_contours;
}

} // namespace nuccut
} // namespace st
