// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#ifndef SHUTTLETRACKER_IMAGE_IMAGEDECOMPOSITION_HPP
#define SHUTTLETRACKER_IMAGE_IMAGEDECOMPOSITION_HPP

#include "Types.hpp"
#include "ImageCommon.hpp"
#include "ImageChannel.hpp"
#include "SharedUtilities.hpp"
#include <map>

namespace st {

// "true" voronoi
void partitionIntoVoronoiCellsCoarse(contours_vec_t& voronoi_contours,
   std::vector<std::pair<int, int>>& voronoi_neighbors,
   std::vector<cv::Point2f> const& points,
   cv::Size sz, float homothetic_scaling, int erosion_width);



void suggestBackgroundTiles(std::map<ImageChannel, cv::Mat> const& images,
                            cv::Mat const&xmask_img, contours_vec_t*bg_contours,
                            int n_subdivs, int n_contours, int n_subshifts = 0);


struct weighted_point_t {
  weighted_point_t(cv::Point const& p, float d, float s)
  : point{p}, distance{d}, scaling{s} {}

  weighted_point_t() = delete;

  cv::Point point;  float distance;  float scaling;
};
using voronoi_weighted_points_t = std::vector<weighted_point_t>;
using corresp_voronois_points_distances_t = std::vector<voronoi_weighted_points_t>;

std::pair<corresp_voronois_points_distances_t, float>
getCorrespVoronoisPointsDistances_(cv::Size const& img_size,
                                  pcontours_vec_t const& contours_ptrs
#ifdef DEBUG
                                  , float (*scaling_fun)(float, float)
#endif
                                  );

inline float linear_(float distance, float lambda,
                     float multiplicative_expansion)
{
  auto const s = 1.f - distance/(lambda * multiplicative_expansion);
  return (s > 0.f) ? s : 0.f;
}

inline float linear_025(float d, float l) { return linear_(d, l, 0.25);}
inline float linear_050(float d, float l) { return linear_(d, l, 0.5); }
inline float linear_075(float d, float l) { return linear_(d, l, 0.75);}
inline float linear_100(float d, float l) { return linear_(d, l, 1.0); }
inline float linear_150(float d, float l) { return linear_(d, l, 1.5); }
inline float linear_200(float d, float l) { return linear_(d, l, 2.0); }
inline float linear_300(float d, float l) { return linear_(d, l, 3.0); }
inline float linear_400(float d, float l) { return linear_(d, l, 4.0); }
inline float linear_500(float d, float l) { return linear_(d, l, 5.0); }


inline float sigmiodal_(float distance, float lambda,
                        float multiplicative_expansion,
                        float steepness)
{
  auto const s = 1.f/(1.f + std::pow(distance, steepness)
                           /(lambda * multiplicative_expansion));
  return (s > 0.f) ? s : 0.f;
}

inline float sigmoidal_10_100(float d, float l) { return sigmiodal_(d, l, 1.0, 1.00); }
inline float sigmoidal_10_125(float d, float l) { return sigmiodal_(d, l, 1.0, 1.25); }
inline float sigmoidal_10_150(float d, float l) { return sigmiodal_(d, l, 1.0, 1.50); }
inline float sigmoidal_10_175(float d, float l) { return sigmiodal_(d, l, 1.0, 1.75); }
inline float sigmoidal_10_200(float d, float l) { return sigmiodal_(d, l, 1.0, 2.00); }

inline float sigmoidal_15_100(float d, float l) { return sigmiodal_(d, l, 1.5, 1.00); }
inline float sigmoidal_15_125(float d, float l) { return sigmiodal_(d, l, 1.5, 1.25); }
inline float sigmoidal_15_150(float d, float l) { return sigmiodal_(d, l, 1.5, 1.50); }
inline float sigmoidal_15_175(float d, float l) { return sigmiodal_(d, l, 1.5, 1.75); }
inline float sigmoidal_15_200(float d, float l) { return sigmiodal_(d, l, 1.5, 2.00); }

inline float sigmoidal_20_100(float d, float l) { return sigmiodal_(d, l, 2.0, 1.00); }
inline float sigmoidal_20_125(float d, float l) { return sigmiodal_(d, l, 2.0, 1.25); }
inline float sigmoidal_20_150(float d, float l) { return sigmiodal_(d, l, 2.0, 1.50); }
inline float sigmoidal_20_175(float d, float l) { return sigmiodal_(d, l, 2.0, 1.75); }
inline float sigmoidal_20_200(float d, float l) { return sigmiodal_(d, l, 2.0, 2.00); }




// -----------------------------------------------------------------------------
// Code below is an apatation of:
//
//  * opencv-4.2.0/modules/imgproc/src/grabcut.cpp,
//  * opencv-4.2.0/modules/imgproc/include/opencv2/imgproc/detail/gcgraph.hpp
//
// licensed as follows:
//
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
// -----------------------------------------------------------------------------

namespace nuccut {

struct GMM
{
  constexpr static int kComponents = 20;  // directly impacts performance
  constexpr static int kChannels = 3;
  constexpr static int kModelSize = /*weight:*/1 + /*mean:*/3 + /*covariance:*/9;

  inline static int nComponents() { return kComponents; }

  GMM();

  double p(int ci, cv::Vec3d const& color) const NOEXCEPT;   // |
  double p(cv::Vec3d const& color) const NOEXCEPT;           // +- inlined
  int whichComponent(cv::Vec3d const& color) const NOEXCEPT; // |

  void beginLearning();
  void addSample(int ci, cv::Vec3d const& color);
  void endLearning(cv::Vec3d const& channel_weights = {1., 1., 1.});

private:
  void calcInverseCovAndDet_(int ci, double singular_fix);

  cv::Mat model;
  double* const component_weight;
  double* const mean;
  double* const cov;

  double inverseCovs[kComponents][kChannels][kChannels];
  double covDets[kComponents];

  double sums[kComponents][kChannels];
  double prod_sums[kComponents][kChannels][kChannels];
  int nSamples[kComponents];
  int nSamplesTotal;
}; // struct GMM


inline double GMM::p(int ci, cv::Vec3d const& color) const NOEXCEPT
{
  if (component_weight[ci] > 0.) {
    assert(covDets[ci] > std::numeric_limits<double>::epsilon());
    cv::Vec3d diff = color;
    double const* const m = mean + 3*ci;
    diff[0] -= m[0];
    diff[1] -= m[1];
    diff[2] -= m[2];
    double const mult = diff[0]*(  diff[0]*inverseCovs[ci][0][0]
                                 + diff[1]*inverseCovs[ci][1][0]
                                 + diff[2]*inverseCovs[ci][2][0])
                      + diff[1]*(  diff[0]*inverseCovs[ci][0][1]
                                 + diff[1]*inverseCovs[ci][1][1]
                                 + diff[2]*inverseCovs[ci][2][1])
                      + diff[2]*(  diff[0]*inverseCovs[ci][0][2]
                                 + diff[1]*inverseCovs[ci][1][2]
                                 + diff[2]*inverseCovs[ci][2][2]);
    return 1./std::sqrt(covDets[ci])*std::exp(-0.5*mult);
  } else {
    return 0.;
  }
}


inline double GMM::p(cv::Vec3d const& color) const NOEXCEPT
{
  double res = 0.;
  for (int ci = 0; ci < kComponents; ci++) {
    res += component_weight[ci] * p(ci, color);
  }
  return res;
}


inline int GMM::whichComponent(cv::Vec3d const& color) const NOEXCEPT
{
  int ci_p_max = 0;
  double p_max = 0.;
  for (int ci = 0; ci < kComponents; ci++) {
    double const p_ = p(ci, color);
    if (p_ > p_max) {
      ci_p_max = ci;
      p_max = p_;
  } }
  return ci_p_max;
}


using Model = GMM;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

namespace accuracy {
using MaxFlowGraphWeightType = float;
}

struct MaxFlowGraph
{
  using Weight = accuracy::MaxFlowGraphWeightType;

  MaxFlowGraph(int n_vertices, int n_edges)
  : flow{static_cast<Weight>(0)}
  { vertices.reserve(n_vertices);  edges.reserve(n_edges + 2); }

  int addVertex()
  {
    Vertex v;
    memset(&v, 0, sizeof(Vertex));
    vertices.push_back(v);
    return SSIZE(vertices) - 1;
  }

  void addEdges(int i, int j, Weight w, Weight w_rev);
  void addTermWeights(int i, Weight w_source, Weight w_sink);
  Weight maxFlow();

  bool inSourceSegment(int i)
  { assert(0 <= i and i < SSIZE(vertices));  return vertices[i].t == 0; }

private:
  struct Vertex {
    Vertex *next;  // initialized and used in maxFlow() only
    int parent;
    int first;
    int ts;
    int dist;
    Weight weight;
    uchar t;
  };

  struct Edge {
    int dst;
    int next;
    Weight weight;
  };

  std::vector<Vertex> vertices;
  std::vector<Edge> edges;
  Weight flow;
}; // MaxFlowGraph

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void initMask(cv::Mat& mask, cv::Rect& bbox,
              st::contour_t const& contour,
              float contour_area,
              cv::Size image_size,
              float nuc_area_median,
              float sure_area_fraction = 0.56f,  // GC_FGD
              float unlike_radius_extension = 0.5f, // GC_PR_BGD
              float surely_never_radius_further_extension = 0.55f); // GC_BGD

double calcRecommendedBeta(cv::Mat const& img);

void calcNontermWeights(cv::Mat& wL, cv::Mat& wUL, cv::Mat& wU, cv::Mat& wUR,
                        cv::Mat const& img, cv::Mat const& edges_img,
                        double beta, double gamma, double delta);

void initModels(Model& bg, Model& fg, cv::Mat const& mask, cv::Mat const& img,
                cv::Vec3d const& channel_weights = {1., 1., 1.},
                int kmeans_n_iter = 16,
                int kmeans_init_type = cv::KmeansFlags::KMEANS_PP_CENTERS);

void learn(Model& bg, Model& fg, cv::Mat const& mask, cv::Mat const& img,
           cv::Vec3d const& channel_weights = {1., 1., 1.});

void segment(cv::Mat& mask, cv::Mat const& img, Model const& bg, Model const& fg,
             cv::Mat const& wL,cv::Mat const& wUL,cv::Mat const& upW,cv::Mat const& wUR,
             double lambda);

st::contour_t nucCut(cv::Mat const& image_edges, cv::Mat const& image_full,
                     cv::Vec3d const& image_full_channel_weights,
                     st::contour_t const& nuc_contour, float nuc_area,
                     float median_nuc_area,
                     int max_iters  = 10,
                     double delta  = 15.,
                     double gamma  = 5.,
                     double beta   = 2e-3,
                     double lambda = 50.);  // in GrabCut, recommended ~10*gamma

struct annotated_contour_t {
  annotated_contour_t(std::size_t i, contour_t const& a_contour, float a_area)
  : index{i}, contour{&a_contour}, area{a_area}, bbox{st::boundingBox(a_contour)},
    edge_score{-1.f}
  { }

  //annotated_contour_t& operator=(annotated_contour_t const& other) = default;

  std::size_t index;
  st::contour_t const* contour;
  float area;
  cv::Rect bbox;
  float edge_score;
};

st::contours_vec_t refine(std::vector<annotated_contour_t>&& nuclei_data,
                          cv::Mat const& image_for_edge_detection,
                          int canny_threshold_low, float canny_thresholds_ratio,
                          cv::Mat const& image_full, cv::Vec3d const& ch_weights,
                          double delta, double gamma, bool preview = false);
} // namespace nuccut
} // namespace st

#endif // !defined(SHUTTLETRACKER_IMAGE_IMAGEDECOMPOSITION_HPP)
