// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "Types.hpp"
#include "ImageOperations.hpp"
#include <contrib/gtest/FloatingPoint.hpp>
#include <vector>
#include <cstdio>

#if SHUTTLETRACKER_IMAGE_PROCESSING_IS_MULTITHREADED
#include <algorithm>  // std::min
#include <thread>
#include <mutex>
#include <boost/asio/thread_pool.hpp>
#include <boost/asio/post.hpp>
#endif

#if CV_MAJOR_VERSION >= 3
#  include <opencv2/imgproc.hpp>
#  include <opencv2/photo.hpp>
#  include <opencv2/highgui.hpp>
#elif CV_MAJOR_VERSION == 2
#  include <opencv2/imgproc/imgproc.hpp>
#  include <opencv2/photo/photo.hpp>
#  include <opencv2/highgui/highgui.hpp>
#endif

namespace st {

void identity(cv::Mat&) {}


void normalize(cv::Mat& image)
{ cv::normalize(image, image, 0, 255, cv::NORM_MINMAX); }


void equalize(cv::Mat& image)
{
  // In OpenCV the order of channels is BGR.
  // Sum being equal to area is fulfilled by empty indexed images.
  cv::Scalar_<long int> const s = sum(image);
  bool has_b = (s[0] > 0),  has_g = (s[1] > 0),  has_r = (s[2] > 0);
  if (s[0]==s[1] and s[1]==s[2] and s[2] == image.size().area()) {
    has_b = has_g = has_r = false;
  }

  cv::Mat bgr[] = { cv::Mat(image.size(), CV_8UC1),
                    cv::Mat(image.size(), CV_8UC1),
                    cv::Mat(image.size(), CV_8UC1) };
  cv::split(image, bgr);

  if (has_b) { cv::equalizeHist(bgr[0], bgr[0]); }
  if (has_g) { cv::equalizeHist(bgr[1], bgr[1]); }
  if (has_r) { cv::equalizeHist(bgr[2], bgr[2]); }

  std::vector<cv::Mat> image_iii;
  image_iii.push_back(bgr[0]);
  image_iii.push_back(bgr[1]);
  image_iii.push_back(bgr[2]);
  cv::merge(image_iii, image);

  // The usual procedure:
  //
  // cv::cvtColor(im, im, cv::COLOR_BGR2GRAY);
  // cv::equalizeHist(im, im);
  // cv::cvtColor(im, image, cv::COLOR_GRAY2BGR);
  //
  // is inappropriate due to unequally weighteded linear combination in cvtColor.
}


void adjustGamma(cv::Mat& image, double gamma)
{
  double const inv_gamma = 1./gamma;
  cv::Mat lut_matrix(1,256,CV_8UC1);
  unsigned char* ptr = lut_matrix.ptr();
  for (int i=0; i<256; i++) {
    ptr[i] = static_cast<unsigned char>(static_cast<int>(
        pow(static_cast<double>(i)/255.0, inv_gamma) * 255.0));
  }
  cv::LUT(image, lut_matrix, image);
}


void denoise(cv::Mat& image, double strength)
{
  assert(strength > 0.);

  // weak coupling of otherwise independent parameters
  // (note: at strength == 1, templateWindowSize and searchWindowSize have
  // their recommended values of 7 and 21 pixels, respectively)
  int search_window_sz = 21 + static_cast<int>(5.0*log(strength));
  search_window_sz += static_cast<int>(not(search_window_sz % 2));
  int template_window_sz = 7 + static_cast<int>(2.5*log(strength));
  template_window_sz += static_cast<int>(not(template_window_sz % 2));

  fastNlMeansDenoising(image,image,static_cast<float>(strength),
                       template_window_sz,search_window_sz);
}


void denoiseBilateral(cv::Mat& image, double strength)
{
  assert(strength > 0);

  cv::Mat image2 = image.clone();  // filter below doesn't work in-place
  bilateralFilter(image, image2, static_cast<int>(2*strength+1),
                  7, 7); // no in-place
  cv::swap(image, image2);
}


// TODO: with OpenCV 2.7.1 comes adaptiveBilinearFilter -- use this
void smoothen(cv::Mat& image, int size_base)
{
  assert(size_base > 0);

  cv::Mat const img_tmp = image.clone();//bilateral filter doesn't work in-place
  auto const diam     = size_base;    // diameter of neighborhood
  auto const sigma_c = static_cast<double>(size_base)*2.75;// stdev in color space
  auto const sigma_s = static_cast<double>(size_base)*1;   // stdev in coords space
  //cv::Mat img_tmp2;
  cv::bilateralFilter(img_tmp, /*img_tmp2*/image, diam, sigma_c, sigma_s);
  //cv::medianBlur(img_tmp2, img_out, 3);
}


void smoothenFixedSigma(cv::Mat& image, int size_base)
{
  assert(size_base > 0);

  cv::Mat const img_tmp = image.clone();//bilateral filter doesn't work in-place
  auto const diam     = 2*size_base+1;    // diameter of neighborhood
  auto const sigma_c = 5;
  auto const sigma_s = 5;
  //cv::Mat img_tmp2;
  cv::bilateralFilter(img_tmp, /*img_tmp2*/image, diam, sigma_c, sigma_s);
  //cv::medianBlur(img_tmp2, img_out, 3);
}


void autolevel(cv::Mat& image, float low, float high)
{
  // in gimp's auto-level, the value of 0.006 is hardcoded (for both thresholds)
  auto const threshold_low = static_cast<long int>(low * image.cols*image.rows),
             threshold_high= static_cast<long int>(high* image.cols*image.rows);

  int const hsize = 256;
  float const range[] = {0, 256}; // upper is exclusive
  float const* hrange = { range };
  int const channels[] = {0}; // TODO: check

  cv::Mat hist;
  cv::calcHist(&image, 1, channels, cv::Mat(), hist, 1, &hsize, &hrange, true, false);
//int bin_i = 0;
//for (cv::MatIterator_<float> it = hist.begin<float>(),
//                             en = hist.end<float>(); it != en; ++it)
//{ std::cout << "DEVEL HIST BIN_" << bin_i++ << ": " << *it << std::endl; }

  long int threshold_bin_low = 0, threshold_bin_high = 255;
  auto sum = 0.f;
  for (cv::MatIterator_<float> it = hist.begin<float>(),
                               en = hist.end<float>(); it != en; ++it) {
    sum += *it;
    if (sum < threshold_low) threshold_bin_low++; else break;
  }

  sum = 0.f;
  for (cv::MatIterator_<float> be = hist.begin<float>(),
                               it = --(hist.end<float>()); it != be; --it) {
    sum += *it;
    if (sum < threshold_high) threshold_bin_high--; else break;
  }

  //std::cout << "Thrs: " << threshold_bin_low  << " "
  //                      << threshold_bin_high << std::endl;

  if (not (threshold_bin_low < threshold_bin_high)) {
//  fprintf(stderr,"Warning: Cannot adjust image intensity levels.\n");
    return;
  }

  cv::Mat lut_matrix(1,256,CV_8UC1);
  unsigned char* plut = lut_matrix.ptr();
  for (int i=0; i<256; i++) {
    auto const value = static_cast<int unsigned>(
        (i <= threshold_bin_low)  ?
            0u :
            (i >= threshold_bin_high) ?
                255u :
                255u*static_cast<float>(i - threshold_bin_low)
                    /static_cast<float>(threshold_bin_high - threshold_bin_low));
    plut[i] = static_cast<unsigned char>(value);
  }
  cv::LUT(image, lut_matrix, image);
} // autolevel()


// TODO: improve performance using in-row iteration.
// alpha - contrast, beta - brightness
void adjustBrightnessContrast(cv::Mat& image, double alpha, int beta)
{
  for (int y = 0; y < image.rows; y++) {
    for (int x = 0; x < image.cols; x++) {
      image.at<uchar>(y, x) = cv::saturate_cast<uchar>(alpha*image.at<uchar>(y, x) + beta);
  } }
}

void brighten(cv::Mat& image) { adjustBrightnessContrast(image, 1.5, 0); }


void autolevelWeak  (cv::Mat& image) { autolevel(image, 0.00001f, 0.003f); }

void autolevelMedium(cv::Mat& image) { autolevel(image, 0.00002f, 0.02f); }

void autolevelStrong(cv::Mat& image) { autolevel(image, 0.00004f, 0.08f); }


void threshold(cv::Mat& image, int block_size, int baseline, cv::Mat const& xmask)
{
  int const bs = block_size + static_cast<int>(not (block_size % 2));  // oddify
  if (xmask.empty()) {
    // Method cv::ADAPTIVE_THRESH_MEAN_C is slightly faster, however appears
    // to be also less accurate.
    cv::adaptiveThreshold(image, image, 255,
        cv::ADAPTIVE_THRESH_GAUSSIAN_C, cv::THRESH_BINARY,
        bs, baseline);
  } else {
    st::internal::adaptiveThresholdWithMask_(image, image, xmask, 255,
        cv::ADAPTIVE_THRESH_GAUSSIAN_C, cv::THRESH_BINARY,
        bs, baseline);
  }
}


void morphologyOpen(cv::Mat& image, int ntimes, int erosion_size,
                                                int dilation_size)
{
  for (int opening_i=0; opening_i<ntimes; ++opening_i) {
    int size_base = 2*(erosion_size + opening_i);
    auto const elem_erode = cv::getStructuringElement(cv::MORPH_ELLIPSE,
                                    cv::Size{ size_base+1, size_base+1},
                                    cv::Point{size_base/2, size_base/2} );
    size_base = 2*(dilation_size + opening_i);
    auto const elem_dilate = cv::getStructuringElement(cv::MORPH_ELLIPSE,
                                     cv::Size{ size_base+1, size_base+1},
                                     cv::Point{size_base/2, size_base/2} );
    cv::erode (image, image, elem_erode);
    cv::dilate(image, image, elem_dilate);
  }
}


void morphologyClose(cv::Mat& image, int ntimes, int dilation_size,
                                                 int erosion_size)
{
  for (int closing_i=0; closing_i<ntimes; ++closing_i) {
    int size_base = 2*(dilation_size + closing_i);
    auto const elem_dilate = cv::getStructuringElement(cv::MORPH_ELLIPSE,
                                     cv::Size{ size_base+1, size_base+1},
                                     cv::Point{size_base/2, size_base/2} );
    size_base = 2*(erosion_size + closing_i);
    auto const elem_erode = cv::getStructuringElement(cv::MORPH_ELLIPSE,
                                    cv::Size {size_base+1, size_base+1},
                                    cv::Point{size_base/2, size_base/2} );
    cv::dilate(image, image, elem_dilate);
    cv::erode (image, image, elem_erode);
  }
}

// TODO: faster iteration!
void blackout(cv::Mat& img, cv::Mat const& mask_img)
{
  if (img.channels() == 1) {        // for B,G,R,BF images
    for (int yi=0; yi<mask_img.rows; yi++) {
      for (int xi=0; xi<mask_img.cols; xi++) {
        if (static_cast<int>(mask_img.at<uchar>(yi, xi)) != 0) {
          img.at<uchar>(yi, xi) = 0;
    } } }
  } else if (img.channels() == 3) { // for image overlays
    for (int yi=0; yi<mask_img.rows; yi++) {
      for (int xi=0; xi<mask_img.cols; xi++) {
        if (static_cast<int>(mask_img.at<uchar>(yi, xi)) != 0) {
          img.at<cv::Vec3b>(yi, xi) = 0;
    } } }
  } else {
    throw std::runtime_error("Cannot mask an image of neither 1 nor 3 channels.");
  }
}


void dim(cv::Mat& image, cv::Mat const& xmask, double reduction, int block_size,
         const std::vector<int>& sigmas)
{
  assert(image.type() == CV_8UC1);
  assert((0. < reduction) and (reduction < 1.));
  assert(((block_size % 2) == 1) and (block_size > 1));

  // common definitions
  auto const ksize = cv::Size{block_size, block_size};
  auto const elem = cv::getStructuringElement(cv::MORPH_ELLIPSE, cv::Size{3,3});

  // sligthly enlarge the exclusion mask and get image parts within it
  cv::Mat xxmask;  cv::dilate(xmask, xxmask, elem);  // 'xx': eXpanded eXclusion
  cv::Mat const xparts = image & xxmask;

  // blur the image with masked parts
  cv::Mat xpartsf;  xparts.convertTo(xpartsf, CV_32F);
  xpartsf = 1. - xpartsf/255.;  // inversion => parts black on white background
  cv::Mat z = cv::Mat::ones(image.size(), CV_32FC1);
  if (not sigmas.empty()) {
    std::vector<cv::Mat> xpartsblurfs;
#if SHUTTLETRACKER_IMAGE_PROCESSING_IS_MULTITHREADED
    boost::asio::thread_pool thread_pool(
        std::min(std::thread::hardware_concurrency(), AS_UINT(sigmas.size())));
    std::mutex _;
#endif
    for (auto sigma: sigmas) {
#if SHUTTLETRACKER_IMAGE_PROCESSING_IS_MULTITHREADED
      boost::asio::post(thread_pool, [&_, &xpartsf, &xpartsblurfs, sigma, ksize]()
      {
#endif
        cv::Mat xpartsblurf;
        cv::GaussianBlur(xpartsf, xpartsblurf, ksize, sigma, sigma,
                         cv::BORDER_REPLICATE | cv::BORDER_ISOLATED);
#if SHUTTLETRACKER_IMAGE_PROCESSING_IS_MULTITHREADED
        std::scoped_lock __(_);
#endif
        xpartsblurfs.push_back(xpartsblurf);
#if SHUTTLETRACKER_IMAGE_PROCESSING_IS_MULTITHREADED
      });
#endif
    }
#if SHUTTLETRACKER_IMAGE_PROCESSING_IS_MULTITHREADED
    thread_pool.join();
    assert(xpartsblurfs.size() == sigmas.size());
#endif
    for (auto const& xpartsblurf: xpartsblurfs) { z = z.mul(xpartsblurf); }
  } // if sigmas non-empty

  // make the source image darker in the masked parts
  cv::Mat srcf;  image.convertTo(srcf, CV_32F);
  srcf = srcf.mul(1. - (1. - z)*reduction);
  srcf.convertTo(image, image.type());  // replace
} // dim()


cv::Mat mergeWeighted(cv::Mat const& imageB, double weightB,
                      cv::Mat const& imageG, double weightG,
                      cv::Mat const& imageR, double weightR)
{
  auto scaledImage_ = [&imageB](cv::Mat const& image, double weight) -> cv::Mat
  {
    if (0. < weight and weight < 1.) {
      assert(image.type() == CV_8UC1);
      double min_, max_;
      cv::minMaxLoc(image, &min_, &max_);
      double const factor = (1. - weight)*(max_ - min_)/2.;
      double const alpha = min_ + factor,
                   beta  = max_ - factor;
      assert(alpha <= beta);
      cv::Mat scaled_image;
      cv::normalize(image, scaled_image, alpha, beta, cv::NORM_MINMAX);
      return scaled_image;
    } else if (gtest::almost_zero(std::fabs(weight - 1.))) {
      return image;
    } else {
      return static_cast<cv::Mat>(cv::Mat::zeros(imageB.size(), CV_8UC1));
    }
  };
  cv::Mat image3;
  cv::merge(std::vector<cv::Mat>{scaledImage_(imageB, weightB),
                                 scaledImage_(imageG, weightG),
                                 scaledImage_(imageR, weightR)}, image3);
  return image3;
}


cv::Mat edgeImage(cv::Mat const& img,
                  bool preview,
                  float median_contour_length,
                  int canny_threshold_low,
                  float canny_thresholds_ratio,
                  int canny_sobel_aperture,
                  std::vector<std::pair<cv::Size, double>> const& blur_kernels,
                  int blur_kernel_range,
                  double edges_weighting_by_gradient,
                  double /*edges_prominence*/
                )
{
  assert(img.type() == CV_8UC1);
  assert(median_contour_length > 0.f);
  assert(canny_sobel_aperture > 0);
  assert(canny_thresholds_ratio > 1.f);
  assert(blur_kernels.size() >= 2);
  assert(blur_kernels.at(0).first.width == blur_kernels.at(0).first.height);
  assert(blur_kernels.at(1).first.width == blur_kernels.at(1).first.height);
  assert(blur_kernels.at(0).first.width  < blur_kernels.at(1).first.width);
  assert(blur_kernels.at(0).first.height < blur_kernels.at(1).first.height);
  assert(blur_kernel_range > 0);
  assert(0.<= edges_weighting_by_gradient and edges_weighting_by_gradient <= 1.);
#define PREVIEW 0

  cv::Size const image_size = img.size();
  double const canny_threshold_high = canny_thresholds_ratio*canny_threshold_low;

  // (1/3) -- generate a preliminary image with contours (grayish on black)
  cv::Mat img_canny_accum = cv::Mat::zeros(image_size, CV_8UC1);
  std::vector<cv::Mat> weighted_edges_images;
#if SHUTTLETRACKER_IMAGE_PROCESSING_IS_MULTITHREADED
  boost::asio::thread_pool thread_pool(std::min(std::thread::hardware_concurrency(),
                                                AS_UINT(blur_kernels.size())));
  std::mutex _;
#endif
  for (auto const& kernel: blur_kernels) {
#if SHUTTLETRACKER_IMAGE_PROCESSING_IS_MULTITHREADED
    boost::asio::post(thread_pool,
        [&_, &weighted_edges_images, &kernel, &img, image_size,
         blur_kernel_range, canny_threshold_low, canny_threshold_high,
         median_contour_length, canny_sobel_aperture]()
    {
#endif
      cv::Mat img_blurred;
      cv::GaussianBlur(img, img_blurred, /*size:*/kernel.first,
                       blur_kernel_range, blur_kernel_range, cv::BORDER_DEFAULT);

      cv::Mat img_edges;
      cv::Canny(img_blurred, img_edges, canny_threshold_low, canny_threshold_high,
                canny_sobel_aperture, true);
      contours_vec_t edges;  edges.reserve(4096);
      findContours(img_edges, edges, cv::RETR_EXTERNAL, cv::CHAIN_APPROX_NONE);

      img_edges = cv::Mat::zeros(image_size, CV_8UC1);
      for (int i=0; i<SSIZE(edges); i++) {
        auto const szf = AS_FLOAT(edges[i].size());
        auto const szf3 = szf*szf*szf;
        auto const med3 = 0.125f*median_contour_length
                                *median_contour_length
                                *median_contour_length;
        auto const weight_color = cv::saturate_cast<uchar>(255*szf3/(szf3 + med3));
        drawContours(img_edges, edges, i, weight_color, 2, cv::LINE_AA);
      }
#if SHUTTLETRACKER_IMAGE_PROCESSING_IS_MULTITHREADED
      std::scoped_lock __(_);
#endif
      weighted_edges_images.push_back(std::move(/*weight*/kernel.second*img_edges));
#if SHUTTLETRACKER_IMAGE_PROCESSING_IS_MULTITHREADED
    });
#endif
  } // for each gaussian blur kernel
#if SHUTTLETRACKER_IMAGE_PROCESSING_IS_MULTITHREADED
  thread_pool.join();
#endif

  for (auto const& weighted_edges_image: weighted_edges_images) {
      img_canny_accum += weighted_edges_image;
  } // for each gaussian blur kernel

  // (2/3) -- independently, calculate gradient of the original image
  cv::Mat img_for_grad;
  cv::GaussianBlur(img, img_for_grad,  blur_kernels[0].first,
                   blur_kernels[0].first.width, blur_kernels[0].first.height);
  //
  cv::Mat grad_x, grad_y;
  cv::Sobel(img_for_grad, grad_x, CV_16S, 1, 0);
  cv::Sobel(img_for_grad, grad_y, CV_16S, 0, 1);
  //
  cv::Mat abs_grad_x, abs_grad_y;
  cv::convertScaleAbs(grad_x, abs_grad_x);  // also -> CV_8UC1
  cv::convertScaleAbs(grad_y, abs_grad_y);  // also -> CV_8UC1
  //
  cv::Mat grad;
  cv::addWeighted(abs_grad_x, 0.5, abs_grad_y, 0.5, 0, grad);
  cv::Mat grad_f;
  grad.convertTo(grad_f, CV_32FC1);
  if (preview) { cv::imshow("Refinement: Gradient", grad); }
#if PREVIEW
  cv::imshow("Refinement: Gradient", grad);
#endif

  // (3/3) -- weight (parts of) edges image according to the gradient image
  cv::Mat img_canny_accum_f;
  img_canny_accum.convertTo(img_canny_accum_f, CV_32FC1);
  cv::Mat img_consensus = (img_canny_accum_f/255.f)
                      .mul(  (1. - edges_weighting_by_gradient)
                           + (     edges_weighting_by_gradient)*grad_f/255.f);
//if (preview) { cv::imshow("Refinement: Edges-pre", img_consensus); }  // TMP
  cv::GaussianBlur(img_consensus, img_consensus,  blur_kernels[1].first,
                   blur_kernels[1].first.width  + 2,
                   blur_kernels[1].first.height + 2);
  if (preview) { cv::imshow("Refinement: Edges", img_consensus); }
#if PREVIEW
  cv::imshow("Refinement: Edges", img_consensus);
#endif

/* -------------- *
  cv::Mat img_consensus2 = 255 * img_consensus;
  img_consensus2.convertTo(img_consensus2, CV_8UC1);
  //
  cv::Mat grad_x2, grad_y2;
  cv::Sobel(img_consensus2, grad_x2, CV_16S, 1, 0);
  cv::Sobel(img_consensus2, grad_y2, CV_16S, 0, 1);
  //
  cv::Mat abs_grad_x2, abs_grad_y2;
  cv::convertScaleAbs(grad_x2, abs_grad_x2);  // also -> CV_8UC1
  cv::convertScaleAbs(grad_y2, abs_grad_y2);  // also -> CV_8UC1
  //
  cv::Mat grad2;
  cv::addWeighted(abs_grad_x2, 0.5, abs_grad_y2, 0.5, 0, grad2);
  grad2.convertTo(grad2, CV_8UC1);

  cv::Mat img_consensus3 = (  img_consensus2*(edges_prominence     )
                            + 255           *(1. - edges_prominence)) - grad2;
  img_consensus3.convertTo(img_consensus3, CV_32FC1, 1./255.);
//  if (preview) { cv::imshow("Refinement: Edges, enhanced-pre", img_consensus3); }  // TMP
//  cv::GaussianBlur(img_consensus3, img_consensus3,  blur_kernels[0].first,
//                   blur_kernels[0].first.width  + 2,
//                   blur_kernels[0].first.height + 2);
  if (preview) { cv::imshow("Refinement: Edges, enhanced", img_consensus3); }
#if PREVIEW
  cv::imshow("Refinement: Edges enhanced", 255*img_consensus3);
#endif
  return img_consensus3;
#undef PREVIEW
 * --------------- */

  return img_consensus;
} // edgeImage()


// based on opencv-4.1.2/modules/imgproc/src/thresh.cpp
void internal::adaptiveThresholdWithMask_(
    cv::Mat const& src, cv::Mat& dst, cv::Mat const& xmask /* exclusion mask */,
    double max_value, int method, int type, int block_size, double delta)
{
  assert(src.type() == CV_8UC1);
  assert(xmask.type() == CV_8UC1);
  assert(block_size % 2 == 1 && block_size > 1);
  if (method != cv::ADAPTIVE_THRESH_GAUSSIAN_C) {
    throw std::runtime_error("Error: non-gaussian thresholding not available.");
  }
  if (type != cv::THRESH_BINARY and type != cv::THRESH_BINARY_INV) {
    throw std::runtime_error("Error: unsupported threshold type.");
  }
  if (max_value < 0.) { dst = cv::Scalar{0};  return; }

  cv::Size size = src.size();
  dst.create(size, src.type());

  auto const ksize = cv::Size{block_size, block_size};
  auto const btype = cv::BORDER_REPLICATE | cv::BORDER_ISOLATED;
  auto const selem = cv::getStructuringElement(cv::MORPH_ELLIPSE, cv::Size{3, 3});
  auto constexpr triple_256 = 3*256;

  cv::Mat mean = (src.data != dst.data) ? dst : cv::Mat{};
  cv::Mat xxmask = xmask.clone();
  cv::dilate(xmask, xxmask, selem);
  cv::Mat const xxparts = src & xxmask;  // xx stands for eXpanded and eXcluded

  cv::Mat srcfloat, meanfloat;
  src.convertTo(srcfloat, CV_32F);
  meanfloat = srcfloat;
  cv::GaussianBlur(srcfloat, meanfloat, ksize, 0, 0, btype);

  cv::Mat xxpartsfloat;  xxparts.convertTo(xxpartsfloat, CV_32F);
  cv::Mat xxpartsfloatblur;
  cv::GaussianBlur(xxpartsfloat, xxpartsfloatblur, ksize, 0, 0, btype);
  meanfloat -= xxpartsfloatblur;
  meanfloat.convertTo(mean, src.type());

  uchar const imaxval = cv::saturate_cast<uchar>(max_value);
  int idelta = (type == cv::THRESH_BINARY) ? cvCeil(delta) : cvFloor(delta);
  uchar tab[triple_256];

  if (type == cv::THRESH_BINARY) {
    for (int i = 0; i < triple_256; i++)
    { tab[i] = static_cast<uchar>(i - 255 >  -idelta ? imaxval : 0); }
  } else if (type == cv::THRESH_BINARY_INV) {
    for (int i = 0; i < triple_256; i++)
    { tab[i] = static_cast<uchar>(i - 255 <= -idelta ? imaxval : 0); }
  }

  if (src.isContinuous() and mean.isContinuous() and dst.isContinuous()) {
    size.width *= size.height;
    size.height = 1;
  }
  for (int i = 0; i < size.height; i++) {
    const uchar* sdata = src.ptr(i);
    const uchar* mdata = mean.ptr(i);
    uchar* ddata = dst.ptr(i);
    for (int j = 0; j < size.width; j++) {
      ddata[j] = tab[sdata[j] - mdata[j] + 255];
  } }
} // adaptiveThresholdWithMask_()


// poorly tested, works on 3-channel images AFAIR, obsolete
//
// This function will not desaturate images with more than one active channel
// (desaturating multi-channel images, think overlays, results in images of
// hardly recognizable details) -- its returns quietly in such cases without
// effect.
void obsolete::desaturate3(cv::Mat& image)
{
  if (image.channels() == 1) {

    std::vector<cv::Mat> merged(3, image.clone());
    cv::Mat overlay; // = cv::Mat::zeros(image.size(), CV_8UC3);
    cv::merge(merged, overlay);
    image = overlay;
    return;

 } else {

  // In OpenCV the order of channels is BGR.
  // Sum being equal to area is sually fulfilled by empty indexed images.
  cv::Scalar_<long int> const s = sum(image);
  bool has_b = (s[0] > 0),  has_g = (s[1] > 0),  has_r = (s[2] > 0);
  if (s[0]==s[1] and s[1]==s[2] and s[2]==image.size().area()) {
    // has_b = has_g = has_r = false; // has_b would be never read
    fprintf(stderr,"Warning: desaturation: could not determine the "
                   "data-containing channel.\n");
    return;
  }

  if (  static_cast<int>(has_b)
      + static_cast<int>(has_g)
      + static_cast<int>(has_r) > 1) {
    //return;
    //throw runtime_error("Error: desaturation: "
    //                    "the image has data in more than 1 channel.");
  }

  if (not has_b and not has_g and not has_r) {
    //throw runtime_error("Error: desaturation: "
    //                    "the image has no data in any channel.");
    //return;
  }

  int const chi = (has_b ? 0 : (has_g ? 1 : (has_r ? 2 : -1)));
  if (chi < 0) {
    //throw runtime_error("Error: desaturation: "
    //                    "Cannot determine the data-containing channel.");
    fprintf(stderr,"Warning: while desaturating, could not determine the "
                   "data-containing channel.\n");
    return;
  }

  cv::Mat image_bgr[] = { cv::Mat(image.size(), CV_8UC1),
                          cv::Mat(image.size(), CV_8UC1),
                          cv::Mat(image.size(), CV_8UC1) };
  cv::split(image, image_bgr);
  std::vector<cv::Mat> image_iii;
  image_iii.push_back(image_bgr[chi]);
  image_iii.push_back(image_bgr[chi]);
  image_iii.push_back(image_bgr[chi]);
  cv::merge(image_iii, image);

  // commented out, because 3-bit imgs are expected externally,
  //cv::cvtColor(image, image, cv::COLOR_BGR2GRAY);
 }
} // desaturate3()

} // namespace st
