﻿// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "ImageSequence.hpp"
#include "ImageSequenceChannelInfo.hpp"
#include "ImageSequenceFilenamePattern.hpp"
#include "ImageDecomposition.hpp"
#include <opencv2/imgproc.hpp>
#include <boost/regex.hpp>
#include <boost/system/error_code.hpp>
#include <boost/filesystem.hpp>
#include <boost/algorithm/string.hpp>  // boost::{split,is_any_of}
#include <tinyxml2/tinyxml2.h>
#include <vector>
#include <tuple>
#include <algorithm>  // sort
#include <cmath>
#include <string>
#include <fstream>
#include <cstdio>
#include <new>  // std::bad_alloc
#include <clocale>  // std::locale::{global,classic}
#include <cassert>

namespace st {

const std::string ImageSequence::sCustomMetadataFilename
    = "shuttletracker_metadata.txt";

using vec_channels_t = std::vector<ImageSequence::Channel>;


int ImageSequence::loadDirectory(std::string dir_s,
                                 vec_channels_t const& exclude,
                                 vec_channels_t const& exclude_from_overlays)
EXCEPTION_TRY
{
  assert(progress_);
  if (not storage_) {
    THROW_ERROR("Cannot load images without specified storage strategy.");
  }
  if (    settings_.computing_overlays_upon_image_loading_
      and storage_->mode() != StorageStrategy::RAM) {
    THROW_ERROR("Cannot generate image overlays when original images "
                " are not stored in memory.");
  }

  auto with_exclusions = [&exclude, &exclude_from_overlays]
                         (ImageSequenceChannelInfo&& lut) {
    for (auto const& xch: exclude)  {
      if (lut.channel_to_external_index_.count(xch)) { lut.exclude(xch); }
    }
    for (auto const& xch: exclude_from_overlays) {
      if (lut.channel_to_external_index_.count(xch)) { lut.excludeFromOverlays(xch); }
    }
    return std::move(lut);
  };

  progress()->yieldBusy();

  std::vector<path_t> const files = listDirectoryContents_(dir_s);
  std::vector<path_t> const leica_metadata_files = findLeicaMetadataFiles_(files);
  std::vector<path_t> const custom_metadata_files = findCustomMetadataFiles_(files);

  if (not leica_metadata_files.empty()) {
    path_t const xml_file_path = leica_metadata_files.front();
    lut_ = with_exclusions(channelInfoFromLeicaMetadata_(xml_file_path));
    findImageFiles_(files);
    storage_->loadImages();
    readTimestampsFromLeicaMetadata_(xml_file_path);
  } else if (not custom_metadata_files.empty()) {
    path_t const custom_file_path = custom_metadata_files.front();
    lut_ = with_exclusions(channelInfoFromCustomFile_(custom_file_path));
    findImageFiles_(files);
    storage_->loadImages();
    setTimestampsBasedOnCustomFile(custom_file_path);
  } else {
    THROW_ERROR("Cannot arrange images due to missing metadata. Hint: create "
                "a custom metadata file '" + sCustomMetadataFilename + "'.");
  }

  progress()->yieldIdle();
  return EXIT_SUCCESS;
}
EXCEPTION_RETHROW_MASTER


ImageSequence::paths_t
ImageSequence::listDirectoryContents_(std::string dir_s) const
EXCEPTION_TRY
{
  using filesystem_error_t = boost::filesystem::filesystem_error;
  using error_code_t = boost::system::error_code;

  // check if directory name is not empty
  if (dir_s.size() == 0) {
    THROW_ERROR("Empty directory name given.");
  }

  path_t const dir{dir_s};

  // check if path is valid
  if (not boost::filesystem::exists(dir)) {
    throw filesystem_error_t(__FUNCTION__, dir,
        error_code_t(boost::system::errc::no_such_file_or_directory,
                     boost::system::generic_category()));
  }

  // check if path points to a directory
  if (not boost::filesystem::is_directory(dir)) {
    throw filesystem_error_t(__FUNCTION__, dir,
        error_code_t(boost::system::errc::not_a_directory,
                     boost::system::generic_category()));
  }

  paths_t files;
  std::copy(dir_iter_t(dir), dir_iter_t(), std::back_inserter(files));
  if (files.empty()) { THROW_ERROR("Directory is empty."); }
  std::sort(files.begin(), files.end());

  return files;
}
EXCEPTION_RETHROW_SLAVE


ImageSequence::paths_t
ImageSequence::findLeicaMetadataFiles_(ImageSequence::paths_t const& files) const
EXCEPTION_TRY
{
  using boost::regex;
  std::vector<regex> const xml_filename_patterns{
      regex{".*_Properties.xml"},  // latest Leica LAS (as of 2019)
      regex{"^(?:MetaData/)?.*Pos([0-9]+)_S([0-9]+)_Properties.xml"}, // old LAS
      regex{".*Pos([0-9]+)_S([0-9]+)_Properties.xml"} };              // old LAS

  // search for XML files in the current image directory
  std::set<path_t> properties_files;
  for (auto const& f: files) {
    boost::smatch matches;
    for (auto const& pattern: xml_filename_patterns) {
      if (boost::regex_match(f.filename().string(), matches, pattern)) {
        if (matches.size() > 0) {
          properties_files.insert(f);
  } } } }

  // recent Leica LAS exports XML files to a separate "MetaData" subdirectory
  if (properties_files.empty()) {
    path_t const* p_metadata = nullptr;
    for (auto const& p: files) {
      if (p.filename() == "MetaData") { p_metadata = &p;  break; }
    }
    if (p_metadata and boost::filesystem::exists(*p_metadata)) {
      paths_t metadata_dir_files;
      std::copy(dir_iter_t(*p_metadata), dir_iter_t(),
                std::back_inserter(metadata_dir_files));
      for (auto const& f: metadata_dir_files) {
        boost::smatch matches;
        for (auto const& pattern: xml_filename_patterns) {
          if (regex_match(f.filename().string(), matches, pattern)) {
            if (matches.size() > 0) {
              properties_files.insert(f);
  } } } } } }

  if (properties_files.size() > 1) {
    int fi = 0;
    for (auto const& f: properties_files) {
      fprintf(stderr, "Leica metadata file (%d of %zd): \"%s\"\n",
                      ++fi, properties_files.size(), f.string().c_str());
    }
    THROW_ERROR("More than one Leica metadata XML file found.");
  }

  assert(properties_files.size() == 0 or properties_files.size() == 1);
  paths_t returned_properties_files;
  std::copy(properties_files.begin(), properties_files.end(),
            std::back_inserter(returned_properties_files));
  return returned_properties_files;
}
EXCEPTION_RETHROW_SLAVE

ImageSequence::paths_t
ImageSequence::findCustomMetadataFiles_(ImageSequence::paths_t const& files) const
EXCEPTION_TRY
{
  paths_t returned_paths;
  for (auto const& path: files) {
    if (path.filename().string() == sCustomMetadataFilename) {
      returned_paths.push_back(path);
  } }
  return returned_paths;
}
EXCEPTION_RETHROW_SLAVE


void ImageSequence::readTimestampsFromLeicaMetadata_(
    ImageSequence::path_t const& xml_file_path)
EXCEPTION_TRY
{
  assert(timestamps_.empty() and not timepoints_.empty());

  // "config"
  boost::regex const timestamp_pattern{
      "^      <TimeStamp HighInteger=\"[0-9]+\" LowInteger=\"[0-9]+\" "
      "RelativeTime=\"([\\.,0-9]+)\" "};
  unsigned int CONSTEXPR n_timestamp_line_expected_matches{2u}; // matches in regexp above
  auto const n_channels = AS_SIZE(lut_.nonexcludedChannelsCount());

  // processing the XML file
  std::ifstream ifs(xml_file_path.string().c_str());
  if (ifs) {
    std::string l;
    std::vector<std::string> tupple;
    while (getline(ifs, l)) {
      boost::cmatch matches;
      if (regex_search(l.c_str(), matches, timestamp_pattern)) {
        if (matches.size() == n_timestamp_line_expected_matches) {
          tupple.push_back(matches[1]);
          if (tupple.size() == n_channels) {
            timestamps_.push_back(tupple);
            tupple.clear();
  } } } } }

  if (timestamps_.empty()) {
    // processing the XML file, another attempt for another Leica LAS version
    boost::regex const timestamp_hex_pattern{"^    <TimeStampList "
        "NumberOfTimeStamps=\\\"[0-9]+\\\">([0-9a-f ]+)</TimeStampList>$"};
    std::ifstream ifs2(xml_file_path.string().c_str());
    if (ifs2) {
      bool t0_unset = true;
      double t0 = -1;
      std::string l;
      std::vector<std::string> hex_timepts, tupple;
      while (getline(ifs2, l)) {
        boost::smatch matches;
        if (regex_search(l, matches, timestamp_hex_pattern)) {
          auto const capture = matches[1].str();
          boost::split(hex_timepts,capture, boost::is_any_of(" "));
          for (auto const& htp: hex_timepts) {
            std::stringstream ss;  ss << std::hex << htp;
            long _t;               ss >> _t;
            double t = _t;
            t /= 1.0e7;  // PARAM:LEICA
            if (t0_unset) { t0 = t; t0_unset = false; }
            t -= t0;  // also t0
            tupple.push_back(std::to_string(t));
            if (tupple.size() == n_channels) {
              timestamps_.push_back(tupple);
              tupple.clear();
          } }
          break;
  } } } }

  // fall back solution of a mismatch detected
  if (timestamps_.size() != timepoints_.size()) {
    fprintf(stderr,"Warning: the number of time points in metadata (%zu) "
                   "and the number of image files per channel (%zu) are "
                   "inconsistent. Didn't you add/remove any images by hand?\n",
                   timestamps_.size(), timepoints_.size());
  //timepoints_relative_times_ = {timepoints_.size(), {n_channels, "?"}};
  }
}
EXCEPTION_RETHROW_SLAVE

void ImageSequence::setTimestampsBasedOnCustomFile(
    ImageSequence::path_t const& file_path)
EXCEPTION_TRY
{
  assert(timestamps_.empty() and not timepoints_.empty());

  std::ifstream ifs(file_path.string().c_str());
  if (not ifs) { return; }

  std::string line;
  while (getline(ifs, line)) {
    std::stringstream lss(line);
    std::string token;
    while (lss >> token) {
      if (token == "time_interval") {
        std::string interval_s;   lss >> interval_s;
        double const interval = std::atof(interval_s.c_str());
        assert(interval > 0.);
        double acquisition_time = 0.;
        for (auto i=0l; i<timepointsCount(); ++i) {
          auto const acquisition_time_s = std::to_string(acquisition_time);
          using _t = std::vector<std::string>;
          timestamps_.emplace_back(_t(AS_SIZE(channelsCount()), acquisition_time_s));
          acquisition_time += interval;
        }
        return;
  } } }
}
EXCEPTION_RETHROW_SLAVE


ImageSequenceChannelInfo
ImageSequence::channelInfoFromLeicaMetadata_(
    ImageSequence::path_t const& xml_file_path) const
EXCEPTION_TRY
{
  using namespace tinyxml2;

  XMLDocument props_xml;
  if (props_xml.LoadFile(xml_file_path.string().c_str()) != XML_SUCCESS) {
    THROW_ERROR("Cannot read the Leica metadata XML file!");
  }

  ImageSequenceChannelInfo channel_info;
  XMLElement* channels_xml = props_xml.FirstChildElement("Data")
                                     ->FirstChildElement("Image")
                                     ->FirstChildElement("ImageDescription")
                                     ->FirstChildElement("Channels");
  int channel_ext_index = 0;
  for (auto ch_xml = channels_xml->FirstChildElement(); ch_xml != nullptr;
       ch_xml = ch_xml->NextSiblingElement()) {
    if (not strcmp(ch_xml->Value(), "ChannelDescription")) {
      auto const bit_depth = ch_xml->IntAttribute("Resolution");
      std::string const pseudocolor = ch_xml->Attribute("LUTName");
      Channel const channel = channelFrom(pseudocolor);
      if (channel == Channel::UNDETERMINED) {
        THROW_ERROR("Unknown pseudocolor (\"LUT\") '"+ pseudocolor + "' "
                    "in Leica metadata.");
      }
      channel_info.assign(channel, channel_ext_index, bit_depth);
    } //if
    ++channel_ext_index;
  } // for

  return channel_info;
}
EXCEPTION_RETHROW_SLAVE


ImageSequenceChannelInfo ImageSequence::channelInfoFromCustomFile_(
    ImageSequence::path_t const& file_path) const
EXCEPTION_TRY
{
  ImageSequenceChannelInfo channel_info;
  std::ifstream ifs(file_path.c_str());
  std::string line;
  while (getline(ifs, line)) {
    // make sure that full stop is treated as decimal separator
    std::locale::global(std::locale::classic());  // effective on Linux
    std::stringstream lss(line);
    lss.imbue(std::locale(std::locale::classic()));  // ineffective on Linux
    std::string token;
    while (lss >> token) {
      if (token == "channel") {
        std::string channel_i_s;  lss >> channel_i_s;
        std::string description;  lss >> description;
        std::string pseudocolor;  lss >> pseudocolor;
        std::string bit_depth_s;  lss >> bit_depth_s;
        int const channel_ext_index = stoi(channel_i_s);
        double const bit_depth = stod(bit_depth_s);
        std::string min_value_s = "0";        // last column, that may contain
        double min_value = stod(min_value_s); // bground intensity, is optional
        if (lss >> min_value_s) { min_value = stod(min_value_s); }
        Channel const channel = channelFrom(pseudocolor);
        if (channel == Channel::UNDETERMINED) {
          THROW_ERROR("Unknown pseudocolor '"+ pseudocolor + "' in metadata.");
        }
        channel_info.assign(channel, channel_ext_index, bit_depth, min_value,
                            description);
  } } }
  return channel_info;
}
EXCEPTION_RETHROW_SLAVE


void
ImageSequence::findImageFiles_(ImageSequence::paths_t const& all_files)
EXCEPTION_TRY
{
  // -- analyze contents of the directory -------

  filename_pattern_ = std::make_unique<FilenamePattern>(all_files);
  std::set<long> tt;          // timepoints
  std::set<int> cc;           // (external) channel indices
  std::set<std::string> ee_s; // image file name extensions
  long t_max = 0l;
  using image_file_info_t = std::tuple<path_t, long, int>;
  auto filePath        = [](image_file_info_t const& I) {return std::get<0>(I);};
  auto timepoint       = [](image_file_info_t const& I) {return std::get<1>(I);};
  auto channelExtIndex = [](image_file_info_t const& I) {return std::get<2>(I);};
  using image_files_info_t = std::vector<image_file_info_t>;
  image_files_info_t image_files_info;
  for (auto const& path: all_files) {
    std::string const fn_s = path.filename().string();
//printf("\nfn = %s\n", fn_s.c_str()); fflush(stdout);
    boost::smatch matches;
    if (boost::regex_match(fn_s, matches, (*filename_pattern_)())) {
      auto const t_s = filename_pattern_->timepoint(matches);
      auto const t = std::stol(t_s);
      auto const c_s = filename_pattern_->channel(matches);
      auto const c = std::stoi(c_s);
      auto const e_s = filename_pattern_->extension(matches);
//printf("   :: t=%ld c=%d\n", t, c);
//for (int i=0; i<matches.size(); ++i) {
//printf("  match %d: %s :: t=%ld c=%d\n", i, matches[i].str().c_str(), t, c);
//fflush(stdout); }
      tt.insert(t);  cc.insert(c);  ee_s.insert(e_s);
      t_max = ((t > t_max) ? t : t_max);
      image_files_info.emplace_back(path, t, c);
  } }

  // -- perform consistency checks --------------

  if (image_files_info.empty()) {
    THROW_ERROR("Cannot find any suitable image files.");
  }

  if ((image_files_info.size() % cc.size()) != 0) {
    THROW_ERROR("Image files count is not divisible by the number of channels.");
  }

  if (ee_s.empty()) {
    THROW_ERROR("Cannot extract the graphics image file extension.");
  } else if (SSIZE(ee_s) > 1) {
    for (auto& e: ee_s) {fprintf(stderr, "Image extension: \"%s\"\n", e.c_str());}
    THROW_ERROR("Image files with different graphics file extensions detected.");
  }

  for (auto const& channel: AllChannels) {
    if (has(channel)) {
      auto const ch_ext_i = lut_.externalIndex(channel);
      if (not cc.count(ch_ext_i)) {
        auto const ch_s = st::toString(channel);
        THROW_ERROR("Missing image files for channel '" + ch_s + "' that is "
                    "present in metadata.");
  } } }

  for (auto const& ext_ch_i: cc) {
    if (not lut_.external_index_to_channel_.count(ext_ch_i)) {
      auto const ext_ch_i_s = std::to_string(ext_ch_i);
      THROW_ERROR("No metadata for image files of channel index "+ext_ch_i_s+".");
  } }

  if (filename_pattern_->is_series() and (SSIZE(tt) != t_max + 1)) {
    auto const n_missing_tfs = (t_max + 1) - SSIZE(tt);
    fprintf(stderr, "Warning: missing %ld time frame(s).\n", n_missing_tfs);
  }

  if (lut_.channels().size() != cc.size()) {
    auto const n_missing_chs = SSIZE(lut_.channels()) - SSIZE(cc);
    if (n_missing_chs > 0) {
      fprintf(stderr, "Warning: missing %ld channel%s.\n", n_missing_chs,
                      (n_missing_chs > 1) ? "s" : "");
    } else {
      fprintf(stderr, "Warning: extra %ld channel%s.\n", -n_missing_chs,
                      (-n_missing_chs > 1) ? "s" : "");
  } }

  // assign a graphics file extension
  image_file_extension_ = *ee_s.begin();
  assert(   "." + image_file_extension_
         == boost::filesystem::extension(
                filePath(image_files_info.front()).filename().string()));

  // remember timepoints
  std::copy(tt.begin(), tt.end(), std::back_inserter(timepoints_));

  // pre-allocate the sequence of masking contours
  masking_.contours_ = {timepoints_.size(), st::contours_vec_t{}};
  masking_.mask_images_ = {timepoints_.size(), cv::Mat(cv::Size{0, 0}, CV_8UC1)};

  // -- fill structured files' paths ------------

  // preallocate
  image_files_paths_ = {timepoints_.size(),
                        {AS_SIZE(lut_.nonexcludedChannelsCount()), ""}};

  // sort image filenames by timepoint index and then by channel index
  std::sort(image_files_info.begin(), image_files_info.end(),
            [timepoint, channelExtIndex]
            (image_file_info_t const& a, image_file_info_t const& b)
            { if      (timepoint(a) < timepoint(b)) { return true;  }
              else if (timepoint(a) > timepoint(b)) { return false; }
              else { assert(channelExtIndex(a) != channelExtIndex(b));
                     return channelExtIndex(a) <  channelExtIndex(b); }});
  // assign
  long ti = 0l; // last time index (initial ti has to be 0; it's an arrary index)
  long tp_last = -1l; // last timepoint (given in file name; initial may be > 0)
  for (auto const& finfo: image_files_info) {
    int const ch_ext_index = channelExtIndex(finfo);
    if (lut_.isChannelExcluded(ch_ext_index)) { continue; }
    auto const ci = lut_.channelIndexExternalToInternal(ch_ext_index);
    if (timepoint(finfo) != tp_last) {
      if (tp_last > -1l) { ++ti; }
      tp_last = timepoint(finfo); }
    assert(ti < SSIZE(timepoints_));
    image_files_paths_[AS_SIZE(ti)][AS_SIZE(ci)] = filePath(finfo);
  }
}
EXCEPTION_RETHROW_SLAVE


cv::Mat ImageSequence::convertedTo8bit_(cv::Mat const& image,
                                        ImageSequence::Channel channel) const
EXCEPTION_TRY
{
  bool const normalize = settings_.normalizing_to_bit_depth_;
  double const max_bit       = normalize? lut_.bitDepth(channel)    : 16.;
  double const min_intensity = normalize? lut_.minIntensity(channel):  0.;
  double const intensity_scaling = 1.0/pow(2., max_bit - 8.);

  cv::Mat const image_after_subtraction = image - min_intensity;
  cv::Mat converted_image;
  image_after_subtraction.convertTo(converted_image, CV_8U, intensity_scaling);
  return converted_image;
}
EXCEPTION_RETHROW_SLAVE


cv::Mat ImageSequence::loadImage_(ImageSequence::path_t const& path,
                                  ImageSequence::Channel channel)
EXCEPTION_TRY
{
  assert(boost::filesystem::exists(path));
  cv::Mat const image = cv::imread(path.string(), kImageReadingMode);

  // get image width and height
  if (image.empty() or image.size().area() == 0) {
    THROW_ERROR("Cannot read image file '" + path.string() + "'. "
                "(This may indiciate that you are running out of memory. "
                "If you believe this is not the case, make sure that "
                "it is a genuine single-channel grayscale image.)");
  } else {
    if ((image_size_ != cv::Size{0, 0}) and (image.size() != image_size_)) {
      fprintf(stderr, "Warning: Images in the sequence appear to have different"
                      " dimensions (width, height).\n");
    }
    image_size_ = image.size();
  }

  switch (image.type()) {
    case CV_8UC1: {
      auto const declared_bit_depth_int = std::lrint(lut_.bitDepth(channel));
      if (declared_bit_depth_int != 8l) {
        fprintf(stderr, "Warning: Metadata suggests that the image is %ld-bit, "
                        "whereas it appears to be a genuine 8-bit image.\n",
                        declared_bit_depth_int);
      }
      return image;
    }

    case CV_16UC1: {
      return convertedTo8bit_(image, channel);
    }

    case CV_8UC3:
    case CV_16UC3: {
      THROW_ERROR("Reading (pseudo)colored images is not supported. "
                  "Hint: export raw images (in grayscale, without LUT).");
    }

    default: {
      THROW_ERROR("Cannot read images of non-standard bit depths or more than "
                  "one channel. Hint: convert your images to 8-bit or 16-bit "
                  "single-channel images.");
    }
  } // switch image type
}
EXCEPTION_RETHROW_SLAVE


// wrapper used by DiskStrategy
cv::Mat
ImageSequence::generateOverlay_(long timepoint_index, ImageFilter filter)
{
  assert((0l <= timepoint_index) and (timepoint_index < timepointsCount()));
  assert(static_cast<int>(SSIZE(image_files_paths_.at(AS_SIZE(timepoint_index))))
         == channelsCount());
  assert(storage_->mode()== StorageStrategy::DISK);

  timepoint_image_buffers_t images;
  for (int ch_index=0; ch_index < channelsCount(); ++ch_index) {
    auto const channel = lut_.channelFromInternalIndex(ch_index);
    images.emplace_back(image(channel, timepoint_index));
  }
  return generateOverlay_(images, filter);
}


cv::Mat ImageSequence::generateOverlay_(
    ImageSequence::timepoint_image_buffers_t const& images,
    ImageFilter filter)
{
  // channels: R, G, B
  cv::Mat const empty = cv::Mat::zeros(imageSize(), CV_8UC1);
  timepoint_image_buffers_t merger;
  for (auto const& channel: {Channel::B, Channel::G, Channel::R}) {
    if (hasInOverlay_(channel)) {
      auto image = images[AS_SIZE(lut_.internalIndex(channel))].clone();
      if (filter) { filter(image); }
      merger.push_back(image);
    } else {
      merger.push_back(empty.clone());
  } }
  assert(merger.size() == 3);

  // channels: C, M, Y
  using _t = std::vector<std::pair<Channel, std::pair<Channel, Channel>>>;
  for (auto const& ch_comps: _t{{Channel::C, {Channel::B, Channel::G}},
                                {Channel::M, {Channel::B, Channel::R}},
                                {Channel::Y, {Channel::G, Channel::R}}}) {
    auto const ch = ch_comps.first;
    if (hasInOverlay_(ch)) {
      auto image = images[AS_SIZE(lut_.internalIndex(ch))].clone();
      if (filter) { filter(image); }
      auto const component1 = ch_comps.second.first;
      auto const component2 = ch_comps.second.second;
      merger[TrichannelCvIndices.at(component1)] += image;
      merger[TrichannelCvIndices.at(component2)] += image;
  } }

  // 3x single-channel images --> single 3-channel image
  cv::Mat overlay = cv::Mat::zeros(imageSize(), CV_8UC3);
  cv::merge(merger, overlay);

  // bright-field channel
  if (hasInOverlay_(Channel::BF)) {
    auto const& bf = images[AS_SIZE(lut_.internalIndex(Channel::BF))];
    timepoint_image_buffers_t merger_bf{3, bf};
    cv::Mat image_bf;  cv::merge(merger_bf, image_bf);
    cv::add(image_bf, overlay, overlay);
  }

  return overlay;
}


void ImageSequence::RAMStrategy::loadImages()
EXCEPTION_TRY
{
  imseq_.progress()->yieldMaxProgress(
                            static_cast<int>(imseq_.image_files_paths_.size()));
  try {
    for (auto const& timepoint_channels_paths: imseq_.image_files_paths_) {
      channels_images_.emplace_back(timepoint_image_buffers_t{});
      for (auto const& path: timepoint_channels_paths) {
        auto const ch_int_index = static_cast<int>(channels_images_.back().size());
        auto const channel = imseq_.lut_.channelFromInternalIndex(ch_int_index);
        channels_images_.back().emplace_back(imseq_.loadImage_(path, channel));
      }
      if (imseq_.settings_.computing_overlays_upon_image_loading_) {
        overlay_images_.emplace_back(
            imseq_.generateOverlay_(channels_images_.back()));
      }
      imseq_.progress()->yieldNextProgress();
    }
  } catch (const std::bad_alloc&) {
    THROW_ERROR("Could not allocate memory required to store all images.");
  } catch (...) { throw; }
}
EXCEPTION_RETHROW_SLAVE

void ImageSequence::DiskStrategy::loadImages()
EXCEPTION_TRY
{
  assert(not imseq_.settings_.computing_overlays_upon_image_loading_);
  assert(not imseq_.image_files_paths_.empty());

  // just probe channel images in the first time point
  imseq_.progress()->yieldMaxProgress(
                    static_cast<int>(imseq_.image_files_paths_.front().size()));
  int ch_int_index=0;
  for (auto const& path: imseq_.image_files_paths_.front()) {
    auto const channel = imseq_.lut_.channelFromInternalIndex(ch_int_index);
    imseq_.loadImage_(path, channel);
    ++ch_int_index;
    imseq_.progress()->yieldNextProgress();
  }
}
EXCEPTION_RETHROW_SLAVE


cv::Mat ImageSequence::RAMStrategy::image(Channel channel, long timepoint_index)
const {
  assert(imseq_.lut_.channel_to_internal_index_.count(channel));
  return channels_images_[AS_SIZE(timepoint_index)]
                         [AS_SIZE(imseq_.lut_.internalIndex(channel))].clone();
}


cv::Mat& ImageSequence::RAMStrategy::image_(Channel channel, long timepoint_index)
{
  assert(imseq_.lut_.channel_to_internal_index_.count(channel));
  return channels_images_[AS_SIZE(timepoint_index)]
                         [AS_SIZE(imseq_.lut_.internalIndex(channel))];
}


cv::Mat ImageSequence::DiskStrategy::image(Channel channel, long timepoint_index)
const {
  auto const path = imseq_.imageFilePath(channel, timepoint_index);
  return imseq_.loadImage_(path, channel);
}

cv::Mat ImageSequence::DiskStrategy::overlayImage(long timepoint_index,
                                                  ImageFilter filter)
{ return imseq_.generateOverlay_(timepoint_index, filter); }


cv::Mat ImageSequence::RAMStrategy::overlayImage(long timepoint_index,
                                                 ImageFilter filter)
{
  if (storesOverlays()) {
    assert((0l <= timepoint_index) and (timepoint_index < SSIZE(overlay_images_)));
    return overlay_images_[AS_SIZE(timepoint_index)].clone();
  } else {
    return imseq_.generateOverlay_(channels_images_.at(AS_SIZE(timepoint_index)),
                                   filter);
  }
}


cv::Mat& ImageSequence::RAMStrategy::overlayImage_(long timepoint_index)
{
  assert(storesOverlays());
  return overlay_images_[AS_SIZE(timepoint_index)];
}


bool ImageSequence::RAMStrategy::storesOverlays() const
{
  if (not overlay_images_.empty()) {
    assert(SSIZE(overlay_images_) == imseq_.timepointsCount());
  }
  return (not overlay_images_.empty());
}

bool ImageSequence::DiskStrategy::storesOverlays() const
{ return false; }


void ImageSequence::RAMStrategy::regenerateOverlays(ImageFilter filter)
EXCEPTION_TRY
{
  assert(SSIZE(channels_images_) == imseq_.timepointsCount());
  imseq_.progress()->yieldMaxProgress(static_cast<int>(imseq_.timepointsCount()));
  overlay_images_.clear();
  for (auto const& timepoint_images_: channels_images_) {
    overlay_images_.emplace_back(
        imseq_.generateOverlay_(timepoint_images_, filter));
    imseq_.progress()->yieldNextProgress();
  }
}
EXCEPTION_RETHROW_SLAVE

void ImageSequence::DiskStrategy::regenerateOverlays(ImageSequence::ImageFilter){}


ImageSequence::StorageStrategy::StorageStrategy(ImageSequence& imseq, Mode choice)
  : imseq_{imseq}, mode_{choice} {}

ImageSequence::DiskStrategy::DiskStrategy(ImageSequence& imseq)
  : ImageSequence::StorageStrategy
                 ::StorageStrategy(imseq, ImageSequence::StorageStrategy::DISK) {}

ImageSequence::RAMStrategy::RAMStrategy(ImageSequence& imseq)
  : ImageSequence::StorageStrategy
                 ::StorageStrategy(imseq, ImageSequence::StorageStrategy::RAM) {}

ImageSequence::StorageStrategy::~StorageStrategy() {}
ImageSequence::RAMStrategy::~RAMStrategy() {}
ImageSequence::DiskStrategy::~DiskStrategy() {}


cv::Mat ImageSequence::image(Channel channel, long timepoint_index) const
{
  assert(has(channel));
  assert((0l <= timepoint_index) and (timepoint_index < timepointsCount()));

  cv::Mat image = storage_->image(channel, timepoint_index);
  masking_.maybeMaskRequestedImage(storage_.get(), image, timepoint_index);
  return image;
}


cv::Mat ImageSequence::image(std::string channel_s, long timepoint_index) const
{
  assert(not channel_s.empty());
  return image(st::channelFrom(channel_s), timepoint_index);
}


cv::Mat ImageSequence::image3(Channel channel, long timepoint_index,
                              bool grayscale, ImageFilter filter) const
{
  // Note: Filters work in-place, so images should be cloned.
  // (There's no need to clone only when filter==st::identity, but it's
  // better to have things expressed uniformly with clear intent rather
  // than optimized on a per-case basis).

  assert(has(channel));
  assert(filter);
  assert((0l <= timepoint_index) and (timepoint_index < timepointsCount()));
  assert(imageSize().area());

  auto img = image(channel, timepoint_index);
  timepoint_image_buffers_t to_be_merged(3,
      grayscale ? img
                : (channel == Channel::BF)
                    ? img
                    : cv::Mat::zeros(img.size(), CV_8UC1));
  if (not grayscale and channel != Channel::BF) {
    switch (channel) {
      case Channel::B:  to_be_merged[0] = img;                    break;
      case Channel::G:  to_be_merged[1] = img;                    break;
      case Channel::R:  to_be_merged[2] = img;                    break;
      case Channel::C:  to_be_merged[0] = to_be_merged[1] = img;  break;
      case Channel::M:  to_be_merged[0] = to_be_merged[2] = img;  break;
      case Channel::Y:  to_be_merged[2] = to_be_merged[1] = img;  break;
      case Channel::BF: // fall-through
      case Channel::UNDETERMINED:
        THROW_ERROR("Cannot generate a 3-channel image based on a single-channel"
                    " image of channel '" + st::toString(channel) + "'.");
  } }
  for (auto& img1: to_be_merged) { filter(img1); }
  cv::Mat merged(image_size_, CV_8UC3);
  cv::merge(to_be_merged, merged);
  return merged;
}


cv::Mat ImageSequence::overlayImage(long timepoint_index, ImageFilter filter)const
{
  assert((0l <= timepoint_index) and (timepoint_index < timepointsCount()));
  cv::Mat image = storage_->overlayImage(timepoint_index, filter);
  masking_.maybeMaskRequestedImage(storage_.get(), image, timepoint_index);
  return image;
}


void ImageSequence::regenerateOverlays(ImageFilter filter)
EXCEPTION_TRY
{
  assert(hasOverlays());
  progress()->yieldBusy();
  storage_->regenerateOverlays(filter);
  progress()->yieldIdle();
}
EXCEPTION_RETHROW_MASTER


std::string ImageSequence::imageFilePath(Channel channel, 
                                         long timepoint_index) const
{
  assert(has(channel));
  assert((0l <= timepoint_index) and (timepoint_index < timepointsCount()));
  assert(timepoint_index < SSIZE(image_files_paths_));
  assert(image_files_paths_[AS_SIZE(timepoint_index)]
                           [AS_SIZE(lut_.internalIndex(channel))].string()!="");
  return image_files_paths_[AS_SIZE(timepoint_index)]
                           [AS_SIZE(lut_.internalIndex(channel))].string();
}


std::string ImageSequence::imageFilePathNoChannel(long timepoint_index)const
{
  assert((0l <= timepoint_index) and (timepoint_index < timepointsCount()));
  assert(timepoint_index < SSIZE(image_files_paths_));
  assert(not image_files_paths_[AS_SIZE(timepoint_index)].empty());
  assert(image_files_paths_[AS_SIZE(timepoint_index)].front().string() != "");
  assert(filename_pattern_ != nullptr);

  auto const fpath = image_files_paths_[AS_SIZE(timepoint_index)].front();
  auto const fdir = fpath.parent_path();
  auto const fname = fpath.filename();
  auto const fname_no_channel = boost::regex_replace(
      fname.string(), filename_pattern_->channel_chunk_re(), std::string{""});

  return (fdir / fname_no_channel).string();
}


std::vector<std::string> ImageSequence::time(unsigned int timepoint_index) const
{
  if (hasRelativeTimes_()) {
    assert((0l <= timepoint_index) and (timepoint_index < SSIZE(timestamps_)));
    return timestamps_[timepoint_index];
  } else {
    return std::vector<std::string>(AS_SIZE(channelsCount()), "?");
  }
}


void ImageSequence::setDescription(Channel channel, std::string const& description)
{ lut_.setDescription(channel, description); }


void ImageSequence::setMask(long timepoint_index,
                            st::contours_vec_t const& contours)
{
  assert((0l <= timepoint_index) and (timepoint_index < timepointsCount()));
  assert(timepoint_index < SSIZE(masking_.contours_));

  if (contours.empty()) { return; }

  // store masking contours and mask image
  masking_.contours_[AS_SIZE(timepoint_index)].clear();
  for (auto const& c: contours) {
    masking_.contours_[AS_SIZE(timepoint_index)].push_back(c);
  }

  masking_.maybeStoreMaskImage(storage_.get(), timepoint_index, image_size_);
  masking_.maybeMaskStoredImages(storage_.get(), lut_.channels(), timepoint_index);
}


bool ImageSequence::hasMask(long timepoint_index) const
{
  assert((0l <= timepoint_index) and (timepoint_index < timepointsCount()));
  return masking_.hasMask(timepoint_index);
}


long ImageSequence::calcMaskedArea(long timepoint_index) const
{ return cv::countNonZero(masking_.drawMask_(timepoint_index, image_size_)); }


st::contours_vec_t const&
ImageSequence::maskingContours(long timepoint_index) const
{
  assert((0l <= timepoint_index) and (timepoint_index < timepointsCount()));
  return masking_.contours(timepoint_index);
}


cv::Mat ImageSequence::Masking::drawMask_(long timepoint_index,
                                          cv::Size image_size) const
{
  assert((0l <= timepoint_index) and (timepoint_index < SSIZE(contours_)));
  assert(image_size.area() > 0);

  cv::Mat mask_img = cv::Mat::zeros(image_size, CV_8UC1);
  cv::drawContours(mask_img, contours(timepoint_index), -1, cv::Scalar{255},
                   cv::FILLED, 4);

  // correct for bottom-right "shadow"
  // TODO: faster iteration in OpenCV is possible
  std::vector<cv::Point> to_remove;
  for (int yi=0; yi<mask_img.rows; yi++) {
    for (int xi=0; xi<mask_img.cols; xi++) {
      if (static_cast<int>(mask_img.at<uchar>(yi,xi)) == 0) { continue; }
      if ((    (    (yi > 0)
                and (static_cast<int>(mask_img.at<uchar>(yi-1, xi  )) != 0))
           and (    (yi<mask_img.rows-1)
                and (static_cast<int>(mask_img.at<uchar>(yi+1, xi  )) == 0))
          ) or (
               (    (xi > 0)
                and (static_cast<int>(mask_img.at<uchar>(yi  , xi-1)) != 0))
           and (    (xi < mask_img.cols-1)
                and (static_cast<int>(mask_img.at<uchar>(yi  , xi+1)) == 0))
         )) {
        to_remove.emplace_back(xi, yi);
      }
      if ((    (    (yi > 0)
                and (static_cast<int>(mask_img.at<uchar>(yi-1, xi  )) != 0))
           and (    (yi < mask_img.rows-1)
                and (static_cast<int>(mask_img.at<uchar>(yi+1, xi  )) == 0))
          ) and (
               (    (xi > 0)
                and (static_cast<int>(mask_img.at<uchar>(yi  , xi-1)) != 0))
           and (    (xi < mask_img.cols-1)
                and (static_cast<int>(mask_img.at<uchar>(yi  , xi+1)) == 0))
         )) {
        to_remove.emplace_back(xi-1, yi);
  } } }
  for (auto &p: to_remove) { mask_img.at<uchar>(p) = 0; }

  return mask_img;
}


void ImageSequence::Masking::maybeStoreMaskImage(
    StorageStrategy const*const storage, long timepoint_i, cv::Size image_size)
{
  if (storage->mode() == StorageStrategy::RAM) {
    assert(timepoint_i < SSIZE(mask_images_));
    mask_images_[AS_SIZE(timepoint_i)] = drawMask_(timepoint_i, image_size);
  }
}


void ImageSequence::Masking::maybeMaskStoredImages(
    StorageStrategy* storage,  std::vector<Channel> const& channels,
    long timepoint_index) const
{
  assert(storage->mode() != StorageStrategy::UNDETERMINED);
  if (    (mode_ == Masking::STORE_AND_BLACK_OUT)
      and (storage->mode() == StorageStrategy::RAM)) {
    assert(timepoint_index < SSIZE(mask_images_));
    auto ram_storage = dynamic_cast<RAMStrategy*>(storage);
    assert(ram_storage);
    cv::Mat const& mask = mask_images_[AS_SIZE(timepoint_index)];
    for (Channel channel: channels) {
      assert(storage->imseq_.has(channel));
      cv::Mat& image = ram_storage->image_(channel, timepoint_index);
      blackout(image, mask);
    }
    if (storage->storesOverlays()) {
      cv::Mat& image = ram_storage->overlayImage_(timepoint_index);
      blackout(image, mask);
  } }
}


void ImageSequence::Masking::maybeMaskRequestedImage(
    StorageStrategy* storage, cv::Mat& image, long timepoint_index) const
{
  assert(storage->mode() != StorageStrategy::UNDETERMINED);
  if (hasMask(timepoint_index)) {
    if (storage->mode() == StorageStrategy::DISK) {
       cv::Mat const mask = drawMask_(timepoint_index, image.size());
       st::blackout(image, mask);
    } else if (storage->mode() == StorageStrategy::RAM) {
      assert(hasMaskImage(storage->mode(), timepoint_index));
  } }
}


st::contours_vec_t
ImageSequence::suggestBackgroundRegions(long timepoint_index, int n_subdivs,
                                        int n_contours, bool merge_adjacent) const
{
  // (1/2) Collect images in which background will be probed
  //       based on their availablity and inclusion in overlays.
  std::map<st::ImageChannel, cv::Mat> images;
  for (Channel ch: AllChannels) {
    if (not hasInOverlay_(ch)) { continue; }
    cv::Mat const img = image(ch, timepoint_index);  // can refer to a temporary
    cv::Scalar_<long> const s = sum(img);
    if (s[0] + s[1] + s[2] + s[3] > 0) {
      assert(not images.count(ch));
      images.emplace(ch, img.clone());
    } else {
      fprintf(stderr,"Warning: when proposing background contours, channel %s "
                     "that contains an empty image has been omitted.\n",
                     lut_.description(ch).c_str());
  } }

  // (2/2) Tile the image using a rectangular grid and select contours of
  //       minimal average intensity in all channels.
  cv::Mat const xmask_img =
      hasMask(timepoint_index)
       ?  masking_.hasMaskImage(storage_->mode(), timepoint_index)
           ?  masking_.maskImage_(timepoint_index)
           :  masking_.drawMask_(timepoint_index, image_size_)
       :  cv::Mat::zeros(image_size_, CV_8UC1);

  st::contours_vec_t bg_tiles{};
  st::suggestBackgroundTiles(images, xmask_img, &bg_tiles, n_subdivs, n_contours);

  if (not merge_adjacent) {
    return bg_tiles;
  } else {
    cv::Mat tiles_img(image_size_, CV_8UC1, cv::Scalar{0});
    cv::drawContours(tiles_img, bg_tiles, -1, cv::Scalar{255});
    st::contours_vec_t merged_bg_tiles;
    cv::findContours(tiles_img, merged_bg_tiles, cv::RETR_EXTERNAL,
                                                 cv::CHAIN_APPROX_SIMPLE);
    return merged_bg_tiles;
  }
} // suggestBackgroundRegions()

} // namespace st
