// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#ifndef SHUTTLETRACKER_IMAGE_IMAGESEQUENCE_HPP
#define SHUTTLETRACKER_IMAGE_IMAGESEQUENCE_HPP

#include "ImageCommon.hpp"
#include "ImageChannel.hpp"
#include "ImageSequenceChannelInfo.hpp"
#include "ImageSequenceFilenamePattern.hpp"
#include "ErrorHandling.hpp"
#include "ProgressReporter.hpp"
#include "ImageOperations.hpp"
#include <string>
#include <vector>
#include <deque>
#include <map>
#include <set>
#include <memory>
#include <stdexcept>
#include <limits>
#include <string>
#include <boost/filesystem.hpp>
#include <boost/regex.hpp>
#include <opencv2/imgproc.hpp>
#include <opencv2/highgui.hpp>  // cv::IMREAD_* modes

namespace st {

class ImageSequence {

  // The most memory-efficient indexed data structure to store a sequence of
  // images and/or its associated data appests to be deque.
  template<typename T> using sequence_t = std::deque<T>;

  using path_t = boost::filesystem::path;
  using paths_t = std::vector<path_t>;
  using image_files_paths_t = sequence_t<paths_t>;
  using timepoint_image_buffers_t = std::vector<cv::Mat>;
  using inplace_image_filter_t = void(*)(cv::Mat&);

public:
  using Channel = st::ImageChannel;
  using ChannelInfo = st::ImageSequenceChannelInfo;
  using FilenamePattern = st::ImageSequenceFilenamePattern;
  using ImageFilter = inplace_image_filter_t;

  struct Settings {
    friend class ImageSequence;

    void setNormalizeToBitDepth(bool tf)
    { normalizing_to_bit_depth_ = tf; }

    void setComputeOverlays(bool tf)
    { computing_overlays_upon_image_loading_ = tf; }

  protected:
    bool normalizing_to_bit_depth_ = true;
    bool computing_overlays_upon_image_loading_ = true;
  };

  struct Masking;

  struct StorageStrategy {
#ifndef NDEBUG              //
    friend struct Masking;  // needed for an assert
#endif                      //
    enum Mode { RAM, DISK, UNDETERMINED };
    StorageStrategy(ImageSequence& imseq, Mode);
    virtual ~StorageStrategy();
    Mode mode() const { return mode_; }
    virtual void loadImages() = 0;
    virtual bool storesOverlays() const = 0;
    virtual cv::Mat image(Channel, long timepoint_i) const = 0;
    virtual cv::Mat overlayImage(long timepoint_i, ImageFilter=st::identity) = 0;
    virtual void regenerateOverlays(ImageFilter = st::identity) = 0;
  protected:
    ImageSequence& imseq_;
    cv::Mat last_image_;
    Mode mode_ = UNDETERMINED;
  };


  struct RAMStrategy: public StorageStrategy {
    friend struct Masking;
    RAMStrategy(ImageSequence& imseq);
    ~RAMStrategy() OVERRIDE;
    void loadImages() OVERRIDE;
    bool storesOverlays() const OVERRIDE;
    cv::Mat image(Channel, long timepoint_i) const OVERRIDE;
    cv::Mat overlayImage(long timepoint_i, ImageFilter = st::identity) OVERRIDE;
    void regenerateOverlays(ImageFilter) OVERRIDE;
  protected:
    cv::Mat& image_(Channel, long timepoint_i);   // |_ for friends
    cv::Mat& overlayImage_(long timepoint_i);     // |
    ImageSequence::sequence_t<timepoint_image_buffers_t> channels_images_;
    ImageSequence::sequence_t<cv::Mat> overlay_images_;
  };


  struct DiskStrategy: public StorageStrategy {
    DiskStrategy(ImageSequence& imseq);
    ~DiskStrategy() OVERRIDE;
    void loadImages() OVERRIDE;
    bool storesOverlays() const OVERRIDE;
    cv::Mat image(Channel, long timepoint_i) const OVERRIDE;
    cv::Mat overlayImage(long timepoint_i, ImageFilter = st::identity) OVERRIDE;
    void regenerateOverlays(ImageFilter = st::identity) OVERRIDE;
  };


  struct Masking {
    friend class ImageSequence;
    using contours_seq_vec_t =

    enum Mode { ONLY_STORE, STORE_AND_BLACK_OUT };
    Mode mode() const { return mode_; }

    contours_vec_t const& contours(long timepoint_index) const
    {
      assert(timepoint_index < SSIZE(contours_));
      return contours_[AS_SIZE(timepoint_index)];
    }

    void maybeStoreMaskImage(StorageStrategy const*const storage,
                             long timepoint_i, cv::Size image_size);

    void maybeMaskStoredImages(StorageStrategy* storage,
                               std::vector<Channel> const& channels,
                               long timepoint_index) const;

    void maybeMaskRequestedImage(StorageStrategy* storage,
                                 cv::Mat& image,
                                 long timepoint_index) const;

    bool hasMask(long timepoint_index) const
    {
      //assert((0l < timepoint_index) and (timepoint_index < SSIZE(contours_)));
      return not contours_.empty() and not contours(timepoint_index).empty();
    }

    bool hasMaskImage(StorageStrategy::Mode storage_mode, long timepoint_index)
    const {
      if (mode() == ONLY_STORE or mode() == STORE_AND_BLACK_OUT) {
        if (storage_mode == StorageStrategy::RAM) {
          return mask_images_[AS_SIZE(timepoint_index)].size().area() > 0;
        } else { return false; }
      } else { return false; }
    }

  protected:
    cv::Mat maskImage_(long timepoint_index) const
    {
      assert(not mask_images_.empty());
      assert((0l < timepoint_index) and (timepoint_index < SSIZE(mask_images_)));
      return mask_images_[AS_SIZE(timepoint_index)];
    }

    cv::Mat drawMask_(long timepoint_index, cv::Size image_size) const;

    // Nuclei detection is not aware of masking contours, so this is default:
    Mode const mode_ = STORE_AND_BLACK_OUT;

    // both allocated in findImageFiles_()
    ImageSequence::sequence_t<contours_vec_t> contours_;
    ImageSequence::sequence_t<cv::Mat> mask_images_;
    cv::Mat last_masked_image_;
  }; // struct Masking


  ImageSequence(StorageStrategy* strategy = nullptr,
                st::ProgressReporter* reporter = new SilentReporter{})
  : progress_{reporter},
    storage_{strategy}
  { assert(progress_); }

  ImageSequence(ImageSequence const& other) = delete;
  ImageSequence(ImageSequence&& other) = delete;
  ImageSequence& operator=(ImageSequence const& other) = delete;
  ImageSequence& operator=(ImageSequence&&) = delete;

  void set(StorageStrategy::Mode storage_choice)
  {
    if (not empty()) {
      THROW_ERROR("Non-empty ImageSequence cannot change its storage strategy.");
    }
    switch (storage_choice) {
      case StorageStrategy::RAM:  storage_.reset(new RAMStrategy(*this));  break;
      case StorageStrategy::DISK: storage_.reset(new DiskStrategy(*this)); break;
      case StorageStrategy::UNDETERMINED:
        THROW_ERROR("Inappropriate storage strategy.");
    }
    if (storage_->mode() != StorageStrategy::RAM) { // then enforce:
      settings_.computing_overlays_upon_image_loading_ = false;
    }
  }

  void set(Settings&& settings) { settings_ = settings; }

  // modifies: properties_.normalizing_to_camera_bit_depth_
  int loadDirectory(std::string dir_path,
                    std::vector<Channel>const& exclude = {},
                    std::vector<Channel>const& exclude_from_overlays = {});
private:
  using dir_iter_t = boost::filesystem::directory_iterator;

  paths_t listDirectoryContents_(std::string dir_path) const;

  paths_t findLeicaMetadataFiles_(paths_t const& all_files) const;

  paths_t findCustomMetadataFiles_(paths_t const& all_files) const;

  ChannelInfo channelInfoFromLeicaMetadata_(path_t const& xml_file_path) const;

  ChannelInfo channelInfoFromCustomFile_(path_t const& file_path) const;

  // .--> timepoints_relative_times_
  void readTimestampsFromLeicaMetadata_(path_t const& xml_file_path);
  void setTimestampsBasedOnCustomFile(path_t const& file_path);

  // modifies: image_filepaths_, image_file_extension_, timepoints_
  void findImageFiles_(paths_t const& all_files);

  // Reads in an image from file and coverts to the 8-bit type if necessary.
  // modifies:image_size
  cv::Mat loadImage_(path_t const& path, Channel=Channel::UNDETERMINED);

  cv::Mat convertedTo8bit_(cv::Mat const& image, Channel) const;

  cv::Mat generateOverlay_(timepoint_image_buffers_t const& timepoint_images,
                           ImageFilter=st::identity);

  cv::Mat generateOverlay_(long timepoint_i, ImageFilter=st::identity);


public:
  bool has(Channel channel) const
  {
    bool const has_channel = lut_.channel_to_external_index_.count(channel);
    bool const channel_excluded = lut_.channels_excluded_.count(channel);
    return has_channel and not channel_excluded;
  }

  // only non-excluded channels (that's how has() works).
  std::vector<Channel> channels() const
  {
    std::vector<Channel> available_channels;
    std::copy_if(AllChannels.cbegin(), AllChannels.cend(),
                 std::back_inserter(available_channels),
                 [this](Channel ch) { return has(ch); });
    assert(available_channels.size() == AS_SIZE(channelsCount()));
    return available_channels;
  }

private:
  bool hasInOverlay_(Channel ch) const
  { return has(ch) and not lut_.channels_excluded_from_overlays_.count(ch); }

  bool hasRelativeTimes_() const
  { return (not timestamps_.empty()) and (timestamps_.size()==timepoints_.size()); }

public:
  std::set<Channel>& channelsExcludedFromOverlays()
  { return lut_.channels_excluded_from_overlays_; }

  bool hasOverlays() const { return storage_->storesOverlays(); }

  std::string description(Channel kind) { return lut_.description(kind); }

  void setDescription(Channel kind, std::string const& new_description);

  cv::Size imageSize() const { return image_size_; }

  void regenerateOverlays(ImageFilter);

  std::string imageFilePath(Channel, long timepoint_index) const;

  std::string imageFilePathNoChannel(long timepoint_index) const;

  std::string const& imageFileExtension() const { return image_file_extension_; }

  long timepointsCount() const { return SSIZE(timepoints_); }

  bool empty() const { return timepointsCount() == 0l; }

  int channelsCount() const { return lut_.nonexcludedChannelsCount(); }

  cv::Mat image(Channel, long timepoint_index) const;

  cv::Mat image(std::string channel_s, long timepoint_index) const;

  cv::Mat image3(Channel, long timepoint_index, bool grayscale = false,
                 ImageFilter = st::identity) const;

  cv::Mat overlayImage(long timepoint_index, ImageFilter = st::identity) const;

  std::vector<std::string> time(unsigned int timepoint_index) const;

  st::ProgressReporter* progress() { assert(progress_); return progress_.get(); }

  Masking::Mode maskingMode() const { return  masking_.mode(); }

  void setMask(long timepoint_index, st::contours_vec_t const&);

  bool hasMask(long timepoint_index) const;

  StorageStrategy::Mode storageStrategy() const { return storage_->mode(); }

  st::contours_vec_t const& maskingContours(long timepoint_index) const;

  long calcMaskedArea(long timepoint_index) const;

  st::contours_vec_t suggestBackgroundRegions(long timepoint_index, int n_subdivs,
                                              int n_contours, bool merge_adjacent)
                                                                          const;
protected:
  static const std::string sCustomMetadataFilename;

  static CONSTEXPR int kImageReadingMode =  cv::IMREAD_ANYDEPTH
                                          | cv::IMREAD_ANYCOLOR;

  std::unique_ptr<st::ProgressReporter> mutable progress_;
  std::unique_ptr<StorageStrategy> storage_;

  ImageSequenceChannelInfo lut_;
  Settings settings_;
  Masking masking_;

  image_files_paths_t image_files_paths_;   // <- findImageFiles_()
  std::string image_file_extension_;        // <- findImageFiles_()
  std::vector<long> timepoints_;            // <- findImageFiles_()
  std::unique_ptr<FilenamePattern> filename_pattern_ = nullptr; // <- findImageFiles_()
  cv::Size image_size_ = cv::Size{0, 0};    // <- loadImage_()

  // v-- readTimepointsFromLeicaMetadata_()
  using channels_timepoints_t_ = std::vector<std::string>;
  sequence_t<channels_timepoints_t_> timestamps_;
}; // class ImageSequence

} // namespace st

#endif // !defined(SHUTTLETRACKER_IMAGE_IMAGESEQUENCE_HPP)
