// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "ImageSequenceChannelInfo.hpp"
#include <cctype>

namespace st {

void ImageSequenceChannelInfo::assign(
    channel_t channel, channel_index_external_t ext_index, double bit_depth,
    double min_value, std::string description)
{
  if (external_index_to_channel_.count(ext_index)) {
    auto const i = std::to_string(ext_index);
    THROW_ERROR("Channel of index "+ i +" already registered.");
  }
  if (   channel_to_external_index_.count(channel)
      or channel_bitdepth_.count(channel)
      or channel_minvalue_.count(channel)) {
    auto const ch_s = toString(channel);
    THROW_ERROR("Assignment of pseudocolor '" + ch_s + "' is non-unique.");
  }

  channel_to_external_index_.emplace(channel, ext_index);
  external_index_to_channel_.emplace(ext_index, channel);

  auto const int_index = static_cast<int>(SSIZE(channel_to_internal_index_));
  channel_to_internal_index_.emplace(channel, int_index);
  internal_index_to_channel_.emplace(int_index, channel);

  channel_bitdepth_.emplace(channel, bit_depth);
  channel_minvalue_.emplace(channel, min_value);

  if (not description.empty()) {
    channel_descriptions_.emplace(channel, description);
  } else {
    channel_descriptions_.emplace(channel, customDescription_(channel));
  }

  assert(channel_to_external_index_.size()==external_index_to_channel_.size());
} // assign()


void ImageSequenceChannelInfo::exclude(channel_t channel)
{
  if (channels_excluded_.count(channel)) {
    THROW_ERROR("Channel '" + toString(channel) + "' already excluded.");
  }
  if (not channel_to_internal_index_.count(channel)) {
    THROW_ERROR("Channel '" + toString(channel) + "' is absent and thus "
                "cannot be excluded.");
  }

  channels_excluded_.insert(channel);

  auto excluded_channel_internal_index = channel_to_internal_index_.at(channel);
  channel_to_internal_index_.erase(channel);
  for (auto& ch_idx: channel_to_internal_index_) {
    if (ch_idx.second > excluded_channel_internal_index) { --ch_idx.second; }
  }

  internal_index_to_channel_.clear();
  for (auto const& ch_idx: channel_to_internal_index_) {
    internal_index_to_channel_.emplace(ch_idx.second, ch_idx.first);
  }
} // exclude()


std::string ImageSequenceChannelInfo::customDescription_(channel_t channel)const
{
  switch (channel) {
    case channel_t::G:  case channel_t::R:  case channel_t::B:
    case channel_t::C:  case channel_t::M:  case channel_t::Y:
    case channel_t::BF: {
      assert(not toString(channel).empty());
      auto initial = toString(channel).at(0);
      return std::string(1, static_cast<char>(std::toupper(initial))); }
    case channel_t::UNDETERMINED:
      THROW_ERROR("No mini-description for the undetermined channel!");
    default: {
      assert(channel_to_external_index_.count(channel));
      int const ext_ch_i = channel_to_external_index_.find(channel)->second;
      return "Ch" + std::to_string(ext_ch_i); }
  } //switch
} // customDescription_()


void ImageSequenceChannelInfo::clear()
{
  external_index_to_channel_.clear();
  channel_to_external_index_.clear();
  channel_to_internal_index_.clear();
  internal_index_to_channel_.clear();
  channel_bitdepth_.clear();
  channel_minvalue_.clear();
  channels_excluded_.clear();
  channels_excluded_from_overlays_.clear();
  channel_descriptions_.clear();
} // clear()

} // namespace st
