// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#ifndef SHUTTLETRACKER_IMAGE_IMAGESEQUENCECHANNELINFO_HPP
#define SHUTTLETRACKER_IMAGE_IMAGESEQUENCECHANNELINFO_HPP

#include "ImageCommon.hpp"
#include "ImageChannel.hpp"
#include "ErrorHandling.hpp"
#include <set>
#include <algorithm>
#include <map>
#include <string>
#include <cassert>

namespace st {

// Some other people would name this class ExtendedLUT.
struct ImageSequenceChannelInfo {

  using channel_t = st::ImageChannel;
  using channel_index_internal_t = int;
  using channel_index_external_t = int;

  void assign(channel_t channel, channel_index_external_t ext_index,
              double bit_depth, double min_value=0., std::string description="");

  void exclude(channel_t channel);

  void excludeFromOverlays(channel_t channel)
  {
    if (channels_excluded_from_overlays_.count(channel)) {
      THROW_ERROR("Channel '" + toString(channel) + "' already excluded "
                  "from overlays.");
    }
    channels_excluded_from_overlays_.insert(channel);
  }


  int nonexcludedChannelsCount() const
  {
    auto const n_nonex =  SSIZE(channel_to_external_index_)
                        - SSIZE(channels_excluded_);
    assert(n_nonex == SSIZE(channel_to_internal_index_));
    assert(n_nonex >= 0);
    return static_cast<int>(n_nonex);
  }


  channel_index_external_t externalIndex(channel_t channel) const
  {
    if (not channel_to_external_index_.count(channel)) {
      THROW_ERROR("No such channel: '" + toString(channel) + "'.");
    }
    return channel_to_external_index_.find(channel)->second;
  }


  channel_index_internal_t
  channelIndexExternalToInternal(channel_index_external_t ext_index)
  {
    if (not external_index_to_channel_.count(ext_index)) {
      THROW_ERROR("No such channel index: " + std::to_string(ext_index) + "'.");
    }
    auto const ch = external_index_to_channel_.find(ext_index)->second;
    if (not channel_to_internal_index_.count(ch)) {
      THROW_ERROR("No such channel: '" + toString(ch) + "'.");
    }
    return channel_to_internal_index_.find(ch)->second;
  }


  channel_t channelFromInternalIndex(channel_index_internal_t int_index) const
  {
    if (  (int_index >= SSIZE(internal_index_to_channel_))
        or not internal_index_to_channel_.count(int_index)) {
      THROW_ERROR("No such channel internal index: "+std::to_string(int_index)+".");
    }
    return internal_index_to_channel_.find(int_index)->second;
  }


  channel_index_internal_t internalIndex(channel_t channel) const
  {
    if (not channel_to_internal_index_.count(channel)) {
      THROW_ERROR("No such channel: '" + toString(channel) + "'.");
    }
    return channel_to_internal_index_.find(channel)->second;
  }


  double bitDepth(channel_t channel) const
  {
    if (not channel_bitdepth_.count(channel)) {
      THROW_ERROR("No such channel: '" + toString(channel) + "'.");
    }
    return channel_bitdepth_.find(channel)->second;
  }


  double minIntensity(channel_t channel) const
  {
    return channel_minvalue_.count(channel)
           ? channel_minvalue_.find(channel)->second
           : 0.;
  }


  bool isChannelExcluded(channel_index_external_t ext_index) const
  {
    if (not external_index_to_channel_.count(ext_index)) {
      THROW_ERROR("No such channel index: " + std::to_string(ext_index));
    }
    auto const tested_ch = external_index_to_channel_.find(ext_index)->second;
    return channels_excluded_.count(tested_ch);
  }


  std::string description(channel_t channel) const
  {
    if (not channel_descriptions_.count(channel)) {
      auto const ch_s = toString(channel);
      THROW_ERROR("Missing description for channel '" + ch_s + "'.");
    }
    return channel_descriptions_.at(channel);
  }


  void setDescription(channel_t channel, std::string const& description)
  {
    if (channel_descriptions_.count(channel)) {
      channel_descriptions_.at(channel) = description;
    } else {
      channel_descriptions_.emplace(channel, description);
    }
  }


  std::vector<channel_t> channels() const
  {
    std::vector<channel_t> channels;
    std::transform(std::begin(channel_to_internal_index_),
                   std::end(channel_to_internal_index_),
                   std::back_inserter(channels),
                   [](decltype(channel_to_internal_index_)::value_type const& p)
                   { return p.first;});
    return channels;
  }

  std::string customDescription_(channel_t channel) const;

  void clear();

protected:
  friend class ImageSequence;

  // bi-map: channel indices in filenames from/to//to/from-channel kinds (LUT)
  // filled by "readMetadata()" methods
  // *not* impacted by exclude()
  std::map<int, channel_t> external_index_to_channel_;
  std::map<channel_t, int> channel_to_external_index_;

  // impacted by exclude()
  // filled by "readImages()" methods
  std::map<channel_t, int> channel_to_internal_index_;
  std::map<int, channel_t> internal_index_to_channel_;

  // These two numbers are used together to apply an affine transform of image
  // pixel intensities.
  std::map<channel_t, double> channel_bitdepth_;
  std::map<channel_t, double> channel_minvalue_;

  // these channels are not keys in above maps
  std::set<channel_t> channels_excluded_;

  // these channels have no relation to the keys in above maps
  std::set<channel_t> channels_excluded_from_overlays_;

  // meaningful but short (identifier-like) descriptions
  std::map<channel_t, std::string> channel_descriptions_;
}; // class ImageSequenceChannelInfo

} // namespace st

#endif // defined(SHUTTLETRACKER_IMAGE_IMAGESEQUENCECHANNELINFO_HPP)
