// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#ifndef SHUTTLETRACKER_IMAGE_IMAGESEQUENCEFILENAMEPATTERN_HPP
#define SHUTTLETRACKER_IMAGE_IMAGESEQUENCEFILENAMEPATTERN_HPP

#include "ImageCommon.hpp"
#include "ErrorHandling.hpp"
#include <vector>
#include <string>
#include <boost/regex.hpp>

namespace st {

struct ImageSequenceFilenamePattern
{
  static std::string const re_leica_prefix_chunk_tc;
  static std::string const re_leica_prefix_chunk_ct;
  static std::string const re_leica_timepoint_chunk;
  static std::string const re_leica_channel_chunk;
  static std::string const re_image_extension_chunk;

  static std::string const re_operetta_prefix_chunk;
  static std::string const re_operetta_channel_chunk;
  static std::string const re_operetta_timepoint_chunk;
  static std::string const re_operetta_suffix_chunk;

  static boost::regex const filename_pattern_operetta_series;
  static boost::regex const filename_pattern_leica_series_tc;
  static boost::regex const filename_pattern_leica_series_ct;
  static boost::regex const filename_pattern_leica_single;

  template <class PathsT>
  ImageSequenceFilenamePattern(PathsT const& files)
  {
    assert(not files.empty());  // should be checked rather on the caller's side

    using boost::regex;
    using boost::regex_search;
    using boost::regex_match;
    using boost::smatch;

    auto patterns_votes = std::map<boost::regex const*, int>{
      {&filename_pattern_leica_series_tc, 0},
      {&filename_pattern_leica_series_ct, 0},
      {&filename_pattern_leica_single,    0},
      {&filename_pattern_operetta_series, 0},
    };

    auto by_count = [](auto& a, auto& b) { return a.second < b.second; };

    auto const image_file_ext_pattern = regex{re_image_extension_chunk + "$"};

    for (auto const& f: files) {
      auto const fn_s = f.filename().string();

      if (not regex_search(fn_s, image_file_ext_pattern)) { continue; }

      auto patterns_matches = std::vector<std::pair<regex const*, int>>{
        {&filename_pattern_leica_series_tc, 0},
        {&filename_pattern_leica_series_ct, 0},
        {&filename_pattern_leica_single,    0},
        {&filename_pattern_operetta_series, 0},
      };

      for (auto& [pattern, score]: patterns_matches) {
        smatch match;
        if (regex_match(fn_s, match, *pattern)) {
          for (auto const& s: match) { if (not s.str().empty()) { ++score; } }
      } }

      std::sort(patterns_matches.begin(), patterns_matches.end(), by_count);
      regex const* best_scoring_pattern = patterns_matches.back().first;
      patterns_votes.at(best_scoring_pattern)++;
    }

    regex const* most_voted_pattern = nullptr;
    int most_votes = 0;
    for (auto& [pattern, votes]: patterns_votes) {
      if (votes > most_votes) {
        most_voted_pattern = pattern;
        most_votes = votes;
    } }

    if (most_voted_pattern == nullptr) {
      THROW_ERROR("Cannot recognize the image file name pattern.");
    }

    image_filename_pattern_ = *most_voted_pattern;

    if (image_filename_pattern_ == filename_pattern_leica_series_tc) {
      //printf("Info: ImageSequence image file naming: 'series-t-c'\n");
      timepoint_chunk_re_ = boost::regex{re_leica_timepoint_chunk};
      channel_chunk_re_ = boost::regex{re_leica_channel_chunk};
      is_series_pattern_ = true;
      timepoint_group_index_ = 2;
      channel_group_index_   = 3;
      extension_group_index_ = 4;

    } else if (image_filename_pattern_ == filename_pattern_leica_series_ct) {
      //printf("Info: ImageSequence image file naming: 'series-c-t'\n");
      timepoint_chunk_re_ = boost::regex{re_leica_timepoint_chunk};
      channel_chunk_re_ = boost::regex{re_leica_channel_chunk};
      is_series_pattern_ = true;
      timepoint_group_index_ = 3;
      channel_group_index_   = 2;
      extension_group_index_ = 4;

    } else if (image_filename_pattern_ == filename_pattern_leica_single) {
      //printf("Info: ImageSequence image file naming: 'single'\n");
      channel_chunk_re_ = boost::regex{re_leica_channel_chunk};
      is_series_pattern_ = false;
      channel_group_index_   = 2;
      extension_group_index_ = 3;

    } else if (image_filename_pattern_ == filename_pattern_operetta_series) {
      //printf("Info: ImageSequence image file naming: 'series-operetta'\n");
      timepoint_chunk_re_ = boost::regex{re_operetta_timepoint_chunk};
      channel_chunk_re_ = boost::regex{re_operetta_channel_chunk};
      is_series_pattern_ = true;
      timepoint_group_index_ = 3;
      channel_group_index_   = 2;
      extension_group_index_ = 5;

    } else {
      THROW_ERROR("[internal] Cannot recognize the image file name pattern.");
    }
  } // c'tor


  boost::regex const& operator()() const { return image_filename_pattern_; }

  std::string timepoint(boost::smatch const& m) const
  { assert((timepoint_group_index()>=0)? timepoint_group_index() < SSIZE(m): true);
    return (timepoint_group_index()>=0)? m[timepoint_group_index()].str(): "0";
  }

  std::string channel(boost::smatch const& m) const
  { assert(channel_group_index() < SSIZE(m));
    return m[channel_group_index()].str();
  }

  std::string extension(boost::smatch const& m) const
  { assert(extension_group_index() < SSIZE(m));
    return m[extension_group_index()].str();
  }

  boost::regex const& channel_chunk_re() const { return channel_chunk_re_; }
  boost::regex const& timepoint_chunk_re() const { return timepoint_chunk_re_; }

  bool is_series() const { return is_series_pattern_; }
  int timepoint_group_index() const { return  timepoint_group_index_; }
  int channel_group_index() const { return channel_group_index_; }
  int extension_group_index() const { return extension_group_index_; }

protected:
  boost::regex image_filename_pattern_;
  boost::regex channel_chunk_re_;
  boost::regex timepoint_chunk_re_;
  bool is_series_pattern_ = false;
  int timepoint_group_index_ = -1;
  int channel_group_index_   = -1;
  int extension_group_index_ = -1;
};

} // namespace st

#endif // defined(SHUTTLETRACKER_IMAGE_IMAGESEQUENCEFILENAMEPATTERN_HPP)
