// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "ImageOperations.hpp"
#include "NucleiDetection.hpp"
#include "NucleiDetectionStained.hpp"
#include "NucleiDetectionNonstained.hpp"
#include "Quantification.hpp"
#include "SharedUtilities.hpp"
#include <set>
#include <tuple>
#include <map>
#include <cmath>
#include <algorithm>
#include <cstdio>
#include <future>
#include <thread>
#if CV_MAJOR_VERSION >= 3
#  include <opencv2/imgproc.hpp>
#  include <opencv2/highgui.hpp>
#elif CV_MAJOR_VERSION == 2
#  include <opencv2/imgproc/imgproc.hpp>
#  include <opencv2/highgui/highgui.hpp>
#endif

namespace st {

using internal::ContourProperties;

#if 0
using cut_points_vec_t = std::vector<std::pair<cv::Point,cv::Point>>;
/* called with params:
 *
              0.10* 2*radius_of_median_circ_contour,
              0.35* 2*radius_of_median_circ_contour,
              0.90* 2*radius_of_median_circ_contour,
              static_cast<int>(0.33*lenght_of_median_circ_contour));
 */
int find_cut_points_using_convexity_defects_UNFINISHED(
    contour_t const& c,
    contours_vec_t& dividees,
    cut_points_vec_t& cut_points,
    float min_depth,
    float min_sum_depths,
    float max_dist,
    int min_defects_proximity_on_contour)
{
  float const max_dist_sq = max_dist*max_dist;
  int const contour_length = c.size();

  // -- (1/?) Detect convexity defects and sort them by their depths in the
  //          descending order.
  //
  contouri_t ihull; // TODO define type here
  cv::convexHull(c, ihull, false);
  std::vector<cv::Vec4i> defects;
  cv::convexityDefects(c, ihull, defects);
  std::sort(defects.begin(), defects.end(),
      [](cv::Vec4i const& d1, cv::Vec4i const& d2)
          { return d1[3] > d2[3]; }); // depth idx is 3

  // DEVEL printing
  /*
  std::cout << "Devel: #defects0 = " << defects.size() << std::endl;
  for (auto const& d: defects) {
    float const depth = static_cast<float>(d[3])/256.f;
    std::cout << "        d= " << depth  << std::endl;
  } */

  // -- (2/?) Remove shallow defects using min_depth.
  //
  // In the OpenCV code (see contours.cpp therein), in order to encode depth
  // using an integer with a reasonable resolution, there is a factor of 256.
  // To decode the real depth (in pixels), one has to get rid of the factor.
  int const n_deep = std::count_if(defects.begin(),defects.end(),
      [&min_depth](cv::Vec4i const& d) {
          return static_cast<float>(d[3])/256.f > min_depth; });
  defects.erase(defects.begin()+n_deep,defects.end());

  // DEVEL printing
  /*
  std::cout << "Devel: #defectsE = " << defects.size() << std::endl;
  for (auto const& d: defects) {
    std::cout <<"        d= "<< static_cast<float>(d[3])/256.f <<std::endl;
  }
  std::cout <<"-----------------------------------------\n"; */

  auto calc_dist_sq = [](cv::Point const& pt1, cv::Point const& pt2)
  {  cv::Point const d = pt1 - pt2;  return d.x*d.x + d.y*d.y;  };

  cut_points_vec_t trial_cut_points;

  for (unsigned int i = 0; i<defects.size(); ++i) {
    auto const& di = defects[i];
    for (unsigned int j = i+1; j<defects.size(); ++j) {
      auto const& dj = defects[j];

      // Is the sum of defects' depths sufficiently large?
      //
      float const sum_depths = static_cast<float>(di[3]+dj[3])/256.f;
      bool const deep_enough = (sum_depths > min_sum_depths);
      if (not deep_enough) { continue; }

      // Are the convexity defects of the selected points sufficiently far away
      // on the contour?
      //
      auto calc_dist_on_loop = [](int i, int j, int L)
      { auto const adij = std::abs(i-j); return std::min(adij, L - adij); };
      //
      int const len_ib_je = calc_dist_on_loop(di[0],dj[1],contour_length),
                len_ib_jb = calc_dist_on_loop(di[0],dj[0],contour_length),
                len_ie_je = calc_dist_on_loop(di[1],dj[1],contour_length),
                len_ie_jb = calc_dist_on_loop(di[1],dj[0],contour_length);
      int defects_proximity = std::min( std::min(len_ib_je,len_ib_jb),
                                        std::min(len_ie_je,len_ie_jb));
      if (defects_proximity < min_defects_proximity_on_contour) { continue; }

      // Now, try to improve cut points by searching for the pair of points
      // which are nearest points of two convexity defects to which 2 points
      // found as far belong.

      int min_link_len_sq = INT_MAX;
      int closest_ci = -1, closest_cj = -1;
      for (int ci=di[0]+10; ci<=di[1]-10; ++ci) {
        for (int cj=dj[0]+10; cj<=dj[1]-10; ++cj) {
          int const link_len_sq = calc_dist_sq(c[ci],c[cj]);
          if (link_len_sq < min_link_len_sq) {
            min_link_len_sq = link_len_sq;
            closest_ci = ci;  closest_cj = cj;
      } } }

      int cut_pi, cut_pj;
      if (closest_ci + closest_cj < 0) {
        cut_pi = di[2];      cut_pj = dj[2];
      } else {
        cut_pi = closest_ci; cut_pj = closest_cj;
      }

      // Are the points close enough in space?
      //
      int const dist_sq = calc_dist_sq(c[cut_pi], c[cut_pj]);
      bool const close_enough = (dist_sq < max_dist_sq);
      if (not close_enough) { continue; }

      //
      // All conditions met at this stage!
      //
//std::cout << "Devel: cut possible!\n";
      trial_cut_points.emplace_back(c[cut_pi],c[cut_pj]);

  } }

  // Select (greedely) only the narrowset passage to cut.
  //
  if (trial_cut_points.size() > 1) {
    using pair_of_points_t = std::pair<cv::Point,cv::Point>;
    std::sort(trial_cut_points.begin(),trial_cut_points.end(),
        [calc_dist_sq](pair_of_points_t const& ppt1, pair_of_points_t const& ppt2)
        { return  calc_dist_sq(ppt1.first, ppt1.second)
                > calc_dist_sq(ppt2.first, ppt2.second); });
    trial_cut_points.erase(trial_cut_points.begin()+1,trial_cut_points.end());
  }


  // sort cut points by their distances (shortes first)
  // for every pair of proposed cut points:
  //   perform trial cut
  //   for both contours resulting from the cut:
  //     if it seems to be a reasonable nucleus (solidity,min-max area wrt. median area -- prarams needed!):
  //       return {1,0 unreslv,new contour}
  //     else:
  //       if it is large enough to believe that it can be divided:
  //         run/return {convexity_defects_based_segmentation_()}
  //       else:
  //         return {0,1 unreslv,empty}
  //

  // TMP
  for (auto const& cpts: trial_cut_points) { cut_points.push_back(cpts); }

  return 0; // TMP
}
#endif


// Finds cut points using nearest pockets between contour and its
// convex hull.
//
//  @return values:
//            0 unable to split into two contours
//           >0 number of new contours resulting from splitting
//
//  Note: even when 0 is returned, dividees may be modified
//       (this is for performance reasons).
//
int findCutPoints_(contour_t const& contour,
    ContourProperties const& props,
    contours_vec_t& dividees,
    std::pair<cv::Point,cv::Point>& cut_points,
    unsigned int skip_n_nearest,
    unsigned int min_pocket_area = 8)// [px]
{
  using cv::Size; using cv::Point; using cv::Scalar; using cv::Mat;
  using std::get;

  int const roi_margin = 1;
  Size const roi_size = props.size<Size>() + Size(2*roi_margin,2*roi_margin);
  Point const roi_offset = props.offset<Point>() - Point(roi_margin,roi_margin);
  Scalar const white{255, 255, 255},  black{0, 0, 0};

  Mat img_c = Mat::zeros(roi_size, CV_8UC1);
  cv::drawContours(img_c, contours_vec_t{contour}, 0, white, cv::FILLED, 4,
                   cv::noArray(), INT_MAX, -roi_offset);

  Mat img_h = Mat::zeros(roi_size, CV_8UC1);
  cv::drawContours(img_h, contours_vec_t{props.hull()}, 0, white, cv::FILLED, 4,
                   cv::noArray(), INT_MAX, -roi_offset);

  Mat img_d = img_c ^ img_h;

  // for PREVIEW (further below)
  //Mat img_d_back = img_d.clone();

  // -- filter out inessential pockets
  //
  contours_vec_t all_pockets;
  std::vector<std::pair<contour_t const*,float>> pockets_and_areas;
  cv::findContours(img_d,all_pockets,cv::RETR_EXTERNAL,cv::CHAIN_APPROX_NONE);
  for (auto const& pocket: all_pockets) {
    auto const area = static_cast<float>(contourArea(pocket));
    if (area >= min_pocket_area) { pockets_and_areas.emplace_back(&pocket,area); }
  }

  if (pockets_and_areas.size() <= skip_n_nearest) {
//std::cout << "Devel: good cut not possible: had to skip " << skip_n_nearest << std::endl;
    return 0u;
  }

  // -- find distances between each pair of pockets
  //
  using two_ctrs_with_dist = std::tuple<contour_t const*,Point,
                                        contour_t const*,Point,unsigned long>;
  std::vector<two_ctrs_with_dist> contours_and_distances;

  for (std::size_t i=0ul, n=AS_SIZE(pockets_and_areas.size()); i<n; ++i) {
    auto const* const pocket_i = pockets_and_areas[i].first;
    for (auto j=i+1; j<n; ++j) {
      auto const* const pocket_j = pockets_and_areas[j].first;

      auto min_dist_sq = INT_MAX;
      Point nearest_i{0, 0},  nearest_j{0,0};
      for (auto const& pt_i: *pocket_i) {
        for (auto const& pt_j: *pocket_j) {
          auto const dx = pt_i.x - pt_j.x,
                     dy = pt_i.y - pt_j.y;
          auto const dist_sq = dx*dx+dy*dy;
          if (dist_sq < min_dist_sq) {
            nearest_i = pt_i;  nearest_j = pt_j;
            min_dist_sq = dist_sq;
          }
      } }
      contours_and_distances.emplace_back(pocket_i, nearest_i,
                                          pocket_j, nearest_j, min_dist_sq);
  } } // for i, for j

  // -- sort pairs of pockets: closest at the beginning
  //
  std::sort(contours_and_distances.begin(),contours_and_distances.end(),
      [](two_ctrs_with_dist const& A,two_ctrs_with_dist const& B)
          { return get<4>(A) < get<4>(B); });

  for (int i=skip_n_nearest, n=static_cast<int>(contours_and_distances.size());
       i<n; ++i) {
    auto const& c_d = contours_and_distances[i];
    // line-type "8" needed to assure contour shape invariance
    Mat img_c_cut = Mat::zeros(roi_size, CV_8UC1);
    cv::drawContours(img_c_cut, contours_vec_t{contour}, 0, white,cv::FILLED,
                     cv::LINE_8, cv::noArray(), INT_MAX, -roi_offset);
    cv::line(img_c_cut, get<1>(c_d),get<3>(c_d), black, 1, cv::LINE_4);
    //
    // Line of width=1 may be not sufficient to have a contour cut, because
    // "Polish Air Force checkerboards" at line ends (and in their proximity)
    // would hinder proper contour splitting. More work on the image is needed.
    //
    // Note: Due to this "cleaning" of the original contour, the dividee may
    // differ slightly (the cleaned contour is usually slightly more round).
    //
    for (int yi=0; yi<img_c_cut.rows-1; yi++) {
      for (int xi=0; xi<img_c_cut.cols-1; xi++) {
        if (    ((int)img_c_cut.at<uchar>(yi  ,xi  ) == 255)
            and ((int)img_c_cut.at<uchar>(yi+1,xi+1) == 255)
            and ((int)img_c_cut.at<uchar>(yi+1,xi  ) ==   0)
            and ((int)img_c_cut.at<uchar>(yi  ,xi+1) ==   0)) {
          img_c_cut.at<uchar>(yi,xi  ) = 0; // sacrifice one white pixel
        } else
        if (    ((int)img_c_cut.at<uchar>(yi  ,xi  ) ==   0)
            and ((int)img_c_cut.at<uchar>(yi+1,xi+1) ==   0)
            and ((int)img_c_cut.at<uchar>(yi+1,xi  ) == 255)
            and ((int)img_c_cut.at<uchar>(yi  ,xi+1) == 255)) {
          img_c_cut.at<uchar>(yi,xi+1) = 0; // sacrifice one white pixel
        }
    } }

    // for PREVIEW (below)
    //Mat img_c_cut_back = img_c_cut.clone();

    cv::findContours(img_c_cut, dividees, cv::RETR_EXTERNAL,
                     cv::CHAIN_APPROX_NONE);

    // ----------------------------------------
    // PREVIEW
    //
    /*
    int const preview_countours_n = 6;
    static int ctr_i = 0;
    contours_vec_t contours_to_draw;
    for (auto const p_a: pockets_and_areas) {
      contours_to_draw.push_back(*p_a.first);
    }
    Mat img_dc = Mat::zeros(props.height(), props.width(), CV_8UC3);
    for (int i=0, n=contours_to_draw.size(); i<n; ++i) {
      Scalar rnd_color(rand()*255.f/RAND_MAX,128,rand()*255.f/RAND_MAX);
      cv::drawContours(img_dc,contours_to_draw,i,rnd_color, CV_FILLED,8);
    }

      std::stringstream ss;
      ss << ctr_i++;
      // ss << static_cast<int>(
      //           (static_cast<float>(rand())/RAND_MAX)*preview_countours_n);
      std::string s;  ss >> s;
      std::cerr << "******************************************* " << ctr_i-1 << std::endl;

    {
      Mat img_o;  auto mag = 5;
      cv::resize(img_d_back,img_o,Size(0,0),mag,mag,CV_INTER_NN);
      cv::imshow((s + "d").c_str(),img_o);
    }

    {
      Mat img_o;  auto mag = 5;
      cv::resize(img_c,img_o,Size(0,0),mag,mag,CV_INTER_NN);
      cv::imshow((s + "c").c_str(),img_o);
    }

    {
      Mat img_o;  auto mag = 5;
      cv::resize(img_dc,img_o,Size(0,0),mag,mag,CV_INTER_NN);
      cv::imshow((s + "dc").c_str(),img_o);
    }

    {
      cv::cvtColor(img_c_cut_back,img_c_cut_back,CV_GRAY2BGR);
      cv::line(img_c_cut_back, get<1>(c_d),get<3>(c_d), cv::Scalar(255,255,0),1,4);
      Mat img_o;  auto mag = 5;
      cv::resize(img_c_cut_back,img_o,Size(0,0),mag,mag,CV_INTER_NN);
      s += "-cut";
      //TMP OUT cv::imshow(s.c_str(),img_o);
    }
    */
    // ----------------------------------------

//std::cout << "Devel: #dividees="<< dividees.size() << std::endl;

    for (auto& d: dividees) {
      for (auto& p: d) { p += roi_offset; }
    }

    if (dividees.size() > 1) {
      cut_points = { get<1>(c_d) + roi_offset,
                     get<3>(c_d) + roi_offset };
//std::cout << "Devel: cut FOUND." << std::endl;
      return static_cast<int>(dividees.size());
    }
  } // for each pair, begining with a skip

//std::cout << "Devel: cut NOT found." << std::endl;
  return 0;
}


double detectNuclei(bool stained,
    cv::Mat const& img_orig,
    cv::Mat const& img_preprocd_main,
    cv::Mat const& img_preprocd_aux,
    contours_vec_t* a_contours_of_nuclei,
    contours_vec_t* a_contours_of_debris,
    contours_vec_t* a_contours_nonsegmentable,
    contours_vec_t* a_contours_of_nuclei_from_segmentation,
    contours_vec_t* a_contours_of_orphans_from_segmentation,
    ImagePreprocessingParameters const& preprocessing_p,
    stained::NucleiDetectionParameters const& stained_p,
    nonstained::NucleiDetectionParameters const& nonstained_p,
    NucleiPreselectionParameters const& presel_p,
    NucleiPreselectionAssessmentParameters const& assess_p,
    bool declump_adjacent,
    bool preview,
    bool dry_run,
    std::atomic<bool>const& stop_enforced)
{
  auto const detection_func = (stained and stained_p.uniformization)
                             ? detectNucleiWithUniformization_
                             : detectNuclei_;
  return detection_func(stained,
                        img_orig, img_preprocd_main, img_preprocd_aux,
                        a_contours_of_nuclei,
                        a_contours_of_debris,
                        a_contours_nonsegmentable,
                        a_contours_of_nuclei_from_segmentation,
                        a_contours_of_orphans_from_segmentation,
                        preprocessing_p,
                        stained_p, nonstained_p, presel_p, assess_p,
                        declump_adjacent, preview, dry_run,
                        stop_enforced);
}


double detectNuclei_(bool stained,
    cv::Mat const& img_orig,
    cv::Mat const& img_preprocd_main,
    cv::Mat const& img_preprocd_aux,
    contours_vec_t* a_contours_of_nuclei,
    contours_vec_t* a_contours_of_debris,
    contours_vec_t* a_contours_nonsegmentable,
    contours_vec_t* a_contours_of_nuclei_from_segmentation,
    contours_vec_t* a_contours_of_orphans_from_segmentation,
    ImagePreprocessingParameters const& preprocessing_p,
    stained::NucleiDetectionParameters const& /*stained_p*/,
    nonstained::NucleiDetectionParameters const& nonstained_p,
    NucleiPreselectionParameters const& presel_p,
    NucleiPreselectionAssessmentParameters const& assess_p,
    bool declump_adjacent,
    bool preview,
    bool dry_run,
    std::atomic<bool>const& stop_enforced)
{
  using cv::Size; using cv::Point; using cv::Scalar;
  using cv::Mat; using cv::Rect;

  auto kHelplesslySmall = 6u;  // PARAM, last change: 9->6
  double const kInvalidScore = -1.e12;  // PARAM, CONVENTION

  contours_vec_t dry_contours_of_nuclei,
                 dry_contours_of_debris,
                 dry_contours_nonsegmentable,
                 dry_contours_of_nuclei_from_segmentation,
                 dry_contours_of_orphans_from_segmentation;

  contours_vec_t* contours_of_nuclei = nullptr;
  contours_vec_t* contours_of_debris = nullptr;
  contours_vec_t* contours_nonsegmentable = nullptr;
  contours_vec_t* contours_of_nuclei_from_segmentation = nullptr;
  contours_vec_t* contours_of_orphans_from_segmentation = nullptr;

  if (dry_run) { // then rebind references to local data structures, which will
                 // have a lifetime that is proper for a multi-threaded use
    contours_of_nuclei                    = &dry_contours_of_nuclei;
    contours_of_debris                    = &dry_contours_of_debris;
    contours_nonsegmentable               = &dry_contours_nonsegmentable;
    contours_of_nuclei_from_segmentation  = &dry_contours_of_nuclei_from_segmentation;
    contours_of_orphans_from_segmentation = &dry_contours_of_orphans_from_segmentation;
  } else {
    contours_of_nuclei                    = a_contours_of_nuclei;
    contours_of_debris                    = a_contours_of_debris;
    contours_nonsegmentable               = a_contours_nonsegmentable;
    contours_of_nuclei_from_segmentation  = a_contours_of_nuclei_from_segmentation;
    contours_of_orphans_from_segmentation = a_contours_of_orphans_from_segmentation;
  }


  // -- (1) Find contours

  contours_vec_t found_contours_;

  if (stained) {
    stained::findNuclearContours(img_preprocd_main, found_contours_, preview);
  } else {
    nonstained::findNuclearContours(img_orig, img_preprocd_main,
        img_preprocd_aux, found_contours_, nonstained_p, preview);
  }

  if (found_contours_.empty() or (dry_run and stop_enforced)) {
    return kInvalidScore;
  }

  // -- (between-1-and-2) weed out spurious contours if necessary
  if (preprocessing_p.thresholding_baseline > 0) {
    stained::removeSpuriousNuclearContours(img_orig, found_contours_);
  }


  // just in case, enforce constness for the rest of the function scope
  contours_vec_t const& found_contours = found_contours_;


  // -- (2) Measure contours' properties

  std::map<contour_t const*const, ContourProperties> contours_props;
  for (auto const& c: found_contours) {
    contours_props.emplace(&c, ContourProperties(c));
  }


  // -- (3) Iteratively classify contours based on their geometric properties

  std::set<contour_t const*> contours_sel_for_median; // not a vector to enable by-value erasure
  for (auto const& c: found_contours) {
    assert(contours_props.find(&c) != contours_props.end());
    // ommit contours which are undisputably too small
    if (contours_props.find(&c)->second.area() > kHelplesslySmall) {
      contours_sel_for_median.insert(&c);
  } }

  if (contours_sel_for_median.empty()) { return kInvalidScore; }

  std::vector<std::pair<Point,Point>> cut_points;
  contours_vec_t contours_groups; // not returned, collected only for preview

  int const n_iterations = 3; // PARAM
  float median_contour_area = -1;
  for (int iteration_i = 0; iteration_i<n_iterations; ++iteration_i) {
//std::cerr << "ITERATION " << iteration_i  <<
//      " =============================================================\n";

    if (dry_run and stop_enforced) { return kInvalidScore; }

    cut_points.clear(); // the contents should come from the last iteration
    contours_of_nuclei->clear();
    //assert(contours_of_nuclei_from_segmentation.empty());  |
    //assert(contours_of_orphans_from_segmentation.empty());  >- turned off for
    //assert(contours_nonsegmentable.empty());               |   dry run
                                                        // (dangerously shared)
    // -- (3a) Calculate median contour area
    //
    if (iteration_i + 1 < n_iterations) {
      std::vector<float> contours_areas;
      for (auto pc: contours_sel_for_median) {
        assert(contours_props.find(pc) != contours_props.end());
        float const contour_area = contours_props.find(pc)->second.area();
        contours_areas.push_back(contour_area);
      }

      std::sort(contours_areas.begin(),contours_areas.end());
      median_contour_area = (iteration_i == 0)
          ? contours_areas[ static_cast<int>(0.6*contours_areas.size()) ] // PARAM
          : contours_areas[ static_cast<int>(0.5*contours_areas.size()) ];
      //                   ^
      // For the 1st iteration, frac>0.5 works usually better
      // than the actual median because there are usually much more "debris"
      // (which might result from insufficient denoising) than groups of
      // "touching" or overlapping nuclei.
    }

    // -- (3b) Classify according to contour area
    //

    auto too_small = [&presel_p, &median_contour_area](float contour_area) -> bool
    { return contour_area < presel_p.min_median_contour_area_fraction*median_contour_area; };

    auto too_large = [&presel_p, &median_contour_area](float contour_area) -> bool
    { return contour_area > presel_p.max_median_contour_area_fraction*median_contour_area; };

//std::cout << "Info: (iter="<< iteration_i <<") "
//          << "contours_sel="       << contours_sel_for_median.size() <<", "
//          << "median_contour_area="<< median_contour_area            <<std::endl;

    int n_too_small = 0,  n_too_large = 0,  n_solid = 0,  n_nonsolid = 0;
    std::vector<contour_t const*> erase_them; // ...for the next iteration
    for (auto pc: contours_sel_for_median) {
      if (dry_run and stop_enforced) { return kInvalidScore; }
      assert(contours_props.find(pc) != contours_props.end());
      float const contour_area = contours_props.find(pc)->second.area();
      if (too_small(contour_area)) {
        n_too_small++;
        if (iteration_i + 1 == n_iterations) {
          contours_of_debris->push_back(*pc); // in 99.99% cases 1 elem
        } else {
          erase_them.push_back(pc); // would skew the median
        }
      } else if (too_large(contour_area)) {
        n_too_large++;
        assert(contours_props.find(pc) != contours_props.end());
        double const contour_solidity = contours_props.find(pc)->second.solidity();
        if (contour_solidity > presel_p.min_solidity) { // then assuming: unable/unnecessary to split group
          n_solid++;
          if (iteration_i + 1 == n_iterations) {
            contours_of_nuclei->push_back(*pc);
          } else {
            erase_them.push_back(pc); // large and solid, but would skew the median
          }
        } else { // large and not solid => attempt segmentation
          n_nonsolid++;
          contours_groups.push_back(*pc);

          //||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
          if (declump_adjacent and (contour_area < 250*250)) { // PARAM
            auto n_hopeless = 0u;
            contours_vec_t to_split{*pc}, from_splitting_nuclei, from_splitting_orphans;
            std::vector<ContourProperties> to_split_props{contours_props.find(pc)->second};
            while ((not to_split.empty()) and (to_split.size() != n_hopeless)) {

//std::cout << "Devel: splitting ------------------- " << &to_split.back()
//          << " sz=" <<  to_split.back().size() << std::endl;

              int success = 1;
              int skip = 0;
              while (success > 0) {
                if (dry_run and stop_enforced) { return kInvalidScore; }
                assert(not to_split.empty());
                contours_vec_t dividees;
                std::pair<Point,Point> contour_cut_points;
                success = findCutPoints_(to_split.back(), to_split_props.back(),
                                         dividees, contour_cut_points, skip,
                                         presel_p.min_pocket_area);

                // Prefix 'local' denotes 'after an attempt to make a cut'.
                contours_vec_t local_to_split,
                               local_from_splitting_nuclei,
                               local_from_splitting_orphans;
                std::vector<ContourProperties> local_to_split_props;

//std::cout << "Devel: dividees = " << dividees.size() <<std::endl;

                //|//////////////////
                if (success > 0) { // there was a split, cool

                  int n_found_nuclei = 0, n_found_groups = 0, n_found_debris = 0;

                  for (auto const& c: dividees) {
                    auto const area = static_cast<float>(contourArea(c));
                    //
                    // TODO: MAYBE solidity can be used somehow here?
                    //
                    if (too_large(area)) {
                      n_found_groups++;
                      local_to_split.emplace_back(c);
                      local_to_split_props.emplace_back(ContourProperties(c));
                    } else { // not too large, bingo! -- we have a nucleus
                      if (not too_small(area)) {
                        n_found_nuclei++;
                        local_from_splitting_nuclei.emplace_back(c);
                      } else {
                        // the shape is too small to be either a group or a nucleus
                        // this is a segmentation orphan, let's collect it
                        //
                        n_found_debris++;
                        local_from_splitting_orphans.emplace_back(c);
                      }
                    }
                  } // for each dividee

//std::cout << "Devel: Split results:" << "("<< n_found_nuclei <<","
//                                           << n_found_groups <<","
//                                           << n_found_debris <<")"<< std::endl;

                  if (n_found_nuclei > 0) {

                    // remove the contour under analysis from the queue
                    to_split.pop_back();
                    to_split_props.pop_back();

                    // after split record groups, nuclei, and orphans (==debris); and cut points
                    //
//std::cerr << "Devel: {to_split = " << to_split.size() << std::endl;
                    for (auto const& c_ts: local_to_split) { to_split.emplace_back(c_ts); }
//std::cerr << "Devel: }to_split = " << to_split.size() << std::endl;
                    for (auto const& c_tsp:local_to_split_props) { to_split_props.emplace_back(c_tsp); }

                    for (auto const& c_fs: local_from_splitting_nuclei) { from_splitting_nuclei.emplace_back(c_fs); }
                    for (auto const& c_fs: local_from_splitting_orphans){ from_splitting_orphans.emplace_back(c_fs); }
                    cut_points.emplace_back(contour_cut_points);
//std::cerr << "Devel: nucleiIN  = " <<  from_splitting_nuclei.size() << "  [@ iter=" << iteration_i << "]\n";
                    if (iteration_i + 1 == n_iterations) {
                      for (auto const& c: from_splitting_nuclei)  {
                        contours_of_nuclei_from_segmentation->emplace_back(c); }
                      from_splitting_nuclei.clear();

                      for (auto const& c: from_splitting_orphans) {
                        contours_of_orphans_from_segmentation->emplace_back(c); }
                      from_splitting_orphans.clear();
                    }
//std::cerr << "Devel: nucleiOUT = " << contours_of_nuclei_from_segmentation.size() << std::endl;

//skip = 0;

                    break; // cut ok, stop while (success > 0)

                  } else if ((n_found_nuclei == 0) and (n_found_groups == 1) /* debris irrelevant */) {

//std::cerr << "Devel: cut ineffective." <<std::endl;
                    skip++;  continue; // wrong cut, no progress, take another chance

                  } else if ((n_found_nuclei == 0) and (n_found_groups >= 2)) {

                    // remove the contour under analysis from the queue
                    to_split.pop_back();
                    to_split_props.pop_back();

                    for (auto const& c_ts: local_to_split) { to_split.emplace_back(c_ts); }
//std::cerr << "Devel: to_split = " << to_split.size() << "::   ";
//for (auto const& c_ts: to_split) { std::cerr <<"  "<< c_ts.size() <<""<< c_ts[0] <<""; }
//std::cerr << std::endl;
                    for (auto const& c_tsp: local_to_split_props) { to_split_props.emplace_back(c_tsp); }

                    if (n_found_debris > 0) {
                      for (auto const& c_fs: local_from_splitting_orphans){ from_splitting_orphans.emplace_back(c_fs);  }
                    }
                    cut_points.emplace_back(contour_cut_points);

                    if (iteration_i + 1 == n_iterations) {
                      for (auto const& c_fs: from_splitting_orphans) {
                        contours_of_orphans_from_segmentation->emplace_back(c_fs);
                      }
                      from_splitting_orphans.clear();
                    }

                    break;

                  } else if ((n_found_nuclei == 0) and (n_found_groups == 0) and (n_found_debris > 0)) {

                    to_split.pop_back();
                    to_split_props.pop_back();
//std::cerr << "Devel: {}to_split = " << to_split.size() << std::endl;

                    for (auto const& c_fs: local_from_splitting_orphans){ from_splitting_orphans.emplace_back(c_fs);  }
                    cut_points.emplace_back(contour_cut_points);

                    if (iteration_i + 1 == n_iterations) {
                      for (auto const& c_fs: from_splitting_orphans) {
                        contours_of_orphans_from_segmentation->emplace_back(c_fs);
                      }
                      from_splitting_orphans.clear();
                    }

                    break;

                  } else {

                    fprintf(stderr, \
                        "Warning: " // <<" ["<<__FUNCTION__ <<":"<< __LINE__<<" ]"
                        "Nuclei split resolver is leaky!\n");

//std::cout << "Devel: Split results @ \"leak\": " << "("<< n_found_nuclei <<","
//                                                 << n_found_groups <<","
//                                                 << n_found_debris <<")"<< std::endl;

                    to_split.pop_back();       // this contour/these contours will be LOST!
                    to_split_props.pop_back();
//std::cerr << "Devel: #to_split = " << to_split.size() << std::endl;
                    break;

                  }

                } else /* no success*/ { // there was no split found

                  if (iteration_i + 1 == n_iterations) {  // in 99.99% cases will add 1 elem
                    assert(not to_split.empty());
                    assert(not to_split_props.empty());

                    //contours_nonsegmentable.push_back(to_split.back());
//std::cerr << " ----- Ns: #to_split()=" << to_split.size() << " :   ";
//                    for (auto const& c_ts: to_split) {
//std::cerr << "  " << c_ts.size() <<""<< c_ts[0] <<"";
//                    //contours_nonsegmentable.push_back(c_ts);
//                    }
//std::cerr << std::endl;

                    contours_nonsegmentable->push_back(to_split.back());
//std::cerr << " >>>>>                       -> " << to_split.back().size() << to_split.back()[0] << std::endl;

                    to_split.pop_back();
                    to_split_props.pop_back();
//std::cerr << "Devel: #to_split = " << to_split.size() << std::endl;

                  } // if last iter
                    else {
                    n_hopeless++;
                  }

                  break;  // no cut, stop "while (success > 0)"
                          // This break is rather redundant -- stop condition
                          // in the while loop would give the same effect.
                } // if split found
                //|//////////////////

              } // while finding splits but non-satisfactory

            } // while still contours to split


//std::cerr << " #####                       -> " << to_split.size() <<std::endl;
            if (iteration_i + 1 == n_iterations) {
              for (auto const& c_ts: to_split) {
//std::cerr << "  " << c_ts.size() <<""<< c_ts[0] <<"";
                contours_nonsegmentable->push_back(c_ts);
            } }
            to_split.clear();
            to_split_props.clear();

          } // if split_adjacent
          //||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
            else {

            if (iteration_i + 1 == n_iterations) {
//std::cerr << " ----- Ns:     ++     " << std::endl;;
              contours_nonsegmentable->push_back(*pc);
            }

          }  // segmentation needed but turned off

        } // if segmentation needed (large & non-solid)
      } else { // size OK, i.e., min__ < median_contour_area < max__

        if (iteration_i + 1 == n_iterations) {
          contours_of_nuclei->push_back(*pc); // in 99.99% cases 1 elem
        } else {
          // leave the contour among those selected for calculating median
        }

      }
    } // for each contour sel for median calc (except for the last iteration)

//std::cout <<"               "
//          <<"too_small="<< n_too_small <<", "
//          <<"too_large="<< n_too_large <<", "
//          <<"solid="    << n_solid     <<", "
//          <<"nonsolid=" << n_nonsolid  << std::endl;

    if (iteration_i + 1 < n_iterations) { // not the last iteration
      for (auto const& pc: erase_them) { contours_sel_for_median.erase(pc); }
    }

    // This is dirty quick-fix, but we do not need to have the same pass
    // as in these loop for the "production" segmentation. It is done just
    // in the last iteration of this weed-out & median calculations.
    if (iteration_i + 2 == n_iterations) { // at the last but one, prepare for the last
      contours_sel_for_median.clear();
      for (auto const& c: found_contours) {
        assert(contours_props.find(&c) != contours_props.end());
        // ommit contours which are undisputably too small
        if (contours_props.find(&c)->second.area() > kHelplesslySmall) {
          contours_sel_for_median.insert(&c);
        }
      } // for
    } // if

  } // for iteration


  // (4) Score ---

#if 0

  std::sort(contours_areas.begin(),contours_areas.end());

  double const median_contour_area = contours_areas[ contours_areas.size()/2 ];
  //std::cout << "Contour area:  median = " << median_contour_area << "\n";
  contours_areas.clear(); // was needed only for statistics

  double score = 0;

  //for (auto const& c : found_contours) {
    //auto const& cp = contours_props[&c];
  double areas_acceptable = 0.0, areas_problematic = 0.0;
  for (auto const& c_cp: contours_props) {
    contour_t const*const    c  = c_cp.first;
    internal::ContourProperties const& cp = c_cp.second;
    // Small contours (which are usually "solid") are unwanted debris.
    if (cp.contour_area_ < presel_p.min_median_contour_area_fraction*median_contour_area){
      contours_debris.push_back(*c);
      areas_problematic += cp.contour_area_;
      score += too_small_penalty;
    } else {
      if (cp.solidity_ > presel_p.min_solidity) {
        contours_nuclei.push_back(*c);
        areas_acceptable += cp.contour_area_;
        if (cp.contour_area_ < presel_p.max_median_contour_area_fraction*\
                               median_contour_area) {
          score += solidity_ok_size_ok_penalty;
        } else {
          score += solidity_ok_size_too_large_penalty;
        }
      } else {
        // size OK, but maybe rough border (or holes?)
        if (cp.contour_area_ < max_median_contour_area_fraction*\
                               median_contour_area) {
          contours_nuclei.push_back(*c);
          areas_acceptable += cp.contour_area_;
          score += size_ok_solidity_not_penalty;
        } else {
          contours_groups.push_back(*c);
          areas_problematic += cp.contour_area_;
          score += group_to_divide_penalty
                   *(static_cast<double>(cp.contour_area_)/median_contour_area);
  } } } }

#endif

  if (dry_run and stop_enforced) { return kInvalidScore; }

  double score_area = 1.;
  if (assess_p.area_cov_weight > 0.0) {

    float areas_sum = 0.f;
    std::vector<float> areas;
    auto collect_areas = [&areas,&areas_sum](contours_vec_t const* contours)
    {
      for (auto const& c: *contours) {
        auto area = static_cast<float>(contourArea(c));
        areas_sum += area;
        areas.push_back(area);
      }
    };

    collect_areas(contours_of_nuclei);
    collect_areas(&contours_groups);
  //collect_areas(contours_of_debris);
    collect_areas(contours_of_nuclei_from_segmentation);
    collect_areas(contours_of_orphans_from_segmentation);
    collect_areas(contours_nonsegmentable);

    if (areas_sum <  (0.9069f * 0.75)
                    *img_preprocd_main.cols*img_preprocd_main.rows) {
      float mean_area = 0.f;  // PARAM: 75% of optimal circle packing density ;]
      for (auto const& a: areas) { mean_area += a; }
      mean_area /= areas.size();

      float area_stdv = 0.f;
      for(auto const&a: areas) { auto d = mean_area - a;  area_stdv += d*d; }
      area_stdv /= areas.size();
      area_stdv = std::sqrt(area_stdv);

      score_area = 1.0/(area_stdv/mean_area); // (Coefficient of Variation)^-1
    } else { // only one element found, probably it's the whole image
      score_area = 1e-6;
    }
  }

  double score_solidity = 1.f;
  if (assess_p.solidity_cov_weight > 0.0) {

    std::vector<float> solidities;
    auto collect_solidities = [&solidities](contours_vec_t const* contours)
    { for (auto const& c: *contours) {
        contour_t h;  cv::convexHull(c,{h},false);
        auto const sol = static_cast<float>(contourArea(c)/contourArea(h));
        if ((not std::isnan(sol)) and (sol > 0)) { solidities.push_back(sol); }
      }
    };
    collect_solidities(contours_of_nuclei);
  //collect_solidities(contours_groups);
  //collect_solidities(contours_of_debris);
    collect_solidities(contours_of_nuclei_from_segmentation);
  //collect_solidities(contours_of_orphans_from_segmentation);
  //collect_solidities(contours_nonsegmentable);

    if (solidities.size() > 1) {
#if 0
      double mean_solidity = 0.f;
      for (auto const& a: solidities) { mean_solidity += a; }
      mean_solidity /= solidities.size();

      //score_solidity = mean_solidity;

      // These deviations are calculated w.r.t 1.0, not w.r.t the mean.
      //
      float solidity_stdv = 0.f;
      for (auto const&a: solidities) {
        auto d = /*mean_solidity*/ 1.0f- a;  solidity_stdv += d*d;
      }
      solidity_stdv /= solidities.size();
      solidity_stdv = std::sqrt(solidity_stdv);

      score_solidity = 1.0f/(solidity_stdv / /*mean_solidity*/1.0f); // (Coefficient of Variation)^-1
#endif
      float solidity_dvtn = 0.f;
      for (auto const&a: solidities) { auto d = 1.0f- a; solidity_dvtn += d*d; }
      solidity_dvtn /= solidities.size();
      solidity_dvtn = std::sqrt(solidity_dvtn);

      score_solidity = 1.0f/(solidity_dvtn);

    } else { // only one element found, probably it's the whole image
      score_solidity = static_cast<float>(1e-6);
    }
  }

  double const score = score_area * score_solidity;
  assert(score >= 0);


  // (5) Filter out nuclei of area 0 ---

  for (auto& pcontours: {contours_of_nuclei,
                         contours_of_debris,
                         contours_nonsegmentable,
                         contours_of_nuclei_from_segmentation,
                         contours_of_orphans_from_segmentation})
  {
    pcontours->erase(std::remove_if(pcontours->begin(), pcontours->end(),
                                    [](contour_t const& c)
                                      { return cv::contourArea(c) < 0.5f; }),
                     pcontours->end());
  }


  // (6) Preview ---

  if (preview) {

    cv::RNG rng(123);
    Mat drawing = Mat::zeros(img_preprocd_main.size(), CV_8UC3);

    for (auto i=0ul, s=static_cast<unsigned long>(found_contours.size());
         i<s; i++) {
      Scalar fill_color(rng.uniform(50,110),
                            rng.uniform( 0, 80),
                            rng.uniform(20, 80));
      cv::drawContours(drawing, found_contours,i, fill_color, cv::FILLED,cv::LINE_AA);
    }

    // draw convex hulls of nuclei
    //
    std::vector<std::vector<Point>> hulls(contours_of_nuclei->size());
    for (auto i=0ul, s=static_cast<unsigned long>(contours_of_nuclei->size());
         i<s; i++) {
      std::vector<Point> hull;
      cv::convexHull((*contours_of_nuclei)[i], hulls[i], false);
      hulls.push_back(hull);
      Scalar color_hull(224,224,224);
      cv::drawContours(drawing, hulls,i, color_hull,1,cv::LINE_AA);
    }

    // draw convex hulls of groups of adjacent/overlapping nuclei
    //
    for (auto i=0ul, s=static_cast<unsigned long>(contours_groups.size());
             i<s; i++) {
      std::vector<std::vector<Point>> hulls2(contours_groups.size());
      cv::convexHull(contours_groups[i], hulls2[i], false);
      Scalar const color_group(90,50,255);
      cv::drawContours(drawing, hulls2, i, color_group, 1, cv::LINE_AA);
    }

    // mark debris with X
    //
    for (auto i=0ul, s=static_cast<unsigned long>(contours_of_debris->size());
         i<s; i++) {
      Point center(0,0);
      for (auto const& p: (*contours_of_debris)[i]) { center += p; }
      center.x /= static_cast<int>((*contours_of_debris)[i].size());
      center.y /= static_cast<int>((*contours_of_debris)[i].size());
      Scalar const color(255,160,200);
      int const r = 5;
      Point const stretch11(-r,+r), stretch22 = -stretch11;
      Point const stretch12(+r,+r), stretch21 = -stretch12;
      cv::line(drawing, center+stretch11,center+stretch22, color,1,cv::LINE_AA);
      cv::line(drawing, center+stretch21,center+stretch12, color,1,cv::LINE_AA);
    }

#if 0
    // add ellipses
    //
    for (size_t i=0, n=contours.size(); i<n; i++) {

      size_t contour_len = contours[i].size();
      if (contour_len < 5) { continue; }

      try {
        Mat pointsf;
        Mat(contours[i]).convertTo(pointsf, CV_32F);
        cv::RotatedRect box = cv::fitEllipse(pointsf);

      //cv::ellipse(drawing, box, Scalar(0,0,255), 1, CV_AA);
        cv::Size2f sz(box.size);
        sz = sz*0.5f + cv::Size2f(5,5);
        cv::ellipse(drawing, box.center, sz, box.angle, 0, 360,
                    Scalar(0,255,255), 1, CV_AA);
      } catch (...) {
      }
    }
#endif

    // Draw results of segmentation
    //
    Scalar const greenish(0,255,96);
    for (auto i=0ul, s=static_cast<unsigned long>(
         contours_of_nuclei_from_segmentation->size()); i<s; i++) {
      cv::drawContours(drawing, *contours_of_nuclei_from_segmentation,i,
                       greenish, 2, cv::LINE_AA);
    }
    //
    Scalar const orangish(0,192,255);
    for (auto i=0ul, s=static_cast<unsigned long>(
         contours_of_orphans_from_segmentation->size()); i<s; i++) {
      cv::drawContours(drawing, *contours_of_orphans_from_segmentation,i,
                       orangish, 2, cv::LINE_AA);
    }
    //
    Scalar const reddish(0,32,255);
    for (auto i=0ul, s=static_cast<unsigned long>(
         contours_nonsegmentable->size()); i<s; i++) {
      cv::drawContours(drawing, *contours_nonsegmentable,i, reddish,2,cv::LINE_AA);
    }
    //
    //
    for (auto const& cps: cut_points) {
      cv::line(drawing, cps.first, cps.second, CV_RGB(0,0,0), 1, cv::LINE_AA);
    }

    cv::imshow("Nuclei detection: declumped contours", drawing);
  } // if preview

  return score;
} // detectNuclei()


// TODO: maybe move to NucleiDetectionStained.cpp?
double detectNucleiWithUniformization_(bool stained,
    cv::Mat const& img_orig,
    cv::Mat const& img_preprocd_main,
    cv::Mat const& img_preprocd_aux,
    contours_vec_t* a_contours_of_nuclei,
    contours_vec_t* a_contours_of_debris,
    contours_vec_t* a_contours_nonsegmentable,
    contours_vec_t* a_contours_of_nuclei_from_segmentation,
    contours_vec_t* a_contours_of_orphans_from_segmentation,
    ImagePreprocessingParameters const& preprocessing_p,
    stained::NucleiDetectionParameters const& stained_p,
    nonstained::NucleiDetectionParameters const& nonstained_p,
    NucleiPreselectionParameters const& presel_p,
    NucleiPreselectionAssessmentParameters const& assess_p,
    bool declump_adjacent,
    bool preview,
    bool dry_run,
    std::atomic<bool>const& stop_enforced)
{
  // below, handle a special case: multileveled stained nuclei
  assert(stained and stained_p.uniformization);
  assert(not img_preprocd_aux.empty());

  // -- (1) detect nuclei, with assessment and splitting -----------------------
  contours_vec_t contours_of_nuclei,
                 contours_of_debris,
                 contours_nonsegmentable,
                 contours_of_nuclei_from_segmentation,
                 contours_of_orphans_from_segmentation;
  std::vector<contours_vec_t*> const all_containers {
      &contours_of_nuclei,
      &contours_of_debris,
      &contours_nonsegmentable,
      &contours_of_nuclei_from_segmentation,
      &contours_of_orphans_from_segmentation  };

  detectNuclei_(stained /*asserted true*/,
                img_orig, img_preprocd_main, cv::Mat{},
                &contours_of_nuclei,
                &contours_of_debris,
                &contours_nonsegmentable,
                &contours_of_nuclei_from_segmentation,
                &contours_of_orphans_from_segmentation,
                preprocessing_p,
                stained_p, nonstained_p, presel_p, assess_p,
                declump_adjacent, preview,
                /*dry_run*/false /*otherwise no contours returned!*/,
                stop_enforced);

  // -- (2) quantify nuclei intensities ----------------------------------------
  using intensity_t = decltype(st::QuantResults::mean_intensity);
  using addr_mintens_t = std::tuple<int, int, intensity_t>;
  auto container_index=[](addr_mintens_t const& ami){ return std::get<0>(ami); };
  auto contour_index  =[](addr_mintens_t const& ami){ return std::get<1>(ami); };
  auto mean_intensity =[](addr_mintens_t const& ami){ return std::get<2>(ami); };

  std::vector<addr_mintens_t> addr_repr_intenss; // "addressed representative intensities"
  auto const n_contours =   contours_of_nuclei.size()
                          + contours_of_nuclei_from_segmentation.size()
                          + contours_nonsegmentable.size()
                          + contours_of_debris.size()
                          + contours_of_orphans_from_segmentation.size();
  if (n_contours == 0) {
    fprintf(stderr, "Warning: No nuclei for uniformization.\n");
    return -1.;
  }
  addr_repr_intenss.reserve(n_contours);
  int container_idx = 0;
  for (auto const& container: all_containers) {
    int contour_idx = 0;
    for (auto const& contour: *container) {
      auto const mean_inten = st::quantifyInContoursMeanIntensity(
                                      img_orig, contours_vec_t{contour});
      addr_repr_intenss.emplace_back(container_idx, contour_idx, mean_inten);
      ++contour_idx;
    }
    ++container_idx;
  }
  assert(not addr_repr_intenss.empty());

  // -- (3) sort nuclei by intensities, in descding order, determine cutoff levels
  std::sort(addr_repr_intenss.begin(), addr_repr_intenss.end(),
      [mean_intensity](addr_mintens_t const& a, addr_mintens_t const& b)
                      { return mean_intensity(a) > mean_intensity(b); });
  float const cutoff_highest = mean_intensity(addr_repr_intenss.front()),
              cutoff_lowest = mean_intensity(addr_repr_intenss.back()),
              cutoff_range = cutoff_highest - cutoff_lowest;

  // -- (4) create a mask of brightest nuclei ----------------------------------
  cv::Mat brightest_mask = cv::Mat::zeros(img_orig.size(), CV_8UC1);
  auto const min_brightest_intensity = cutoff_lowest
      + cutoff_range*static_cast<float>(100 - stained_p.uniformization_threshold)/100.f;
  for (addr_mintens_t const& ami: addr_repr_intenss) {
    auto const& contours = *all_containers[AS_SIZE(container_index(ami))];
    drawMask(brightest_mask, contours, contour_index(ami));
    if (mean_intensity(ami) < min_brightest_intensity) { break; }
  }
  cv::Mat const brightest_mask_neg = cv::Scalar{255} - brightest_mask;
  // TODO: cannot use not operator?

  // -- (5) apply the mask to dim the brightest nuclei -------------------------
  assert(preprocessing_p.smoothing_size > 0);
  auto const block_size = (preprocessing_p.smoothing_size + 2)*2 + 1;  // PARAM
  float const reduction = 1.125f  // PARAM
      *cutoff_range*(stained_p.uniformization_threshold/100.f)/cutoff_highest;
  cv::Mat img_preprocd_main_eq = dry_run ? img_preprocd_aux.clone()  // parallel
                                         : img_preprocd_aux;         // serial
  dim(img_preprocd_main_eq, brightest_mask, AS_DOUBLE(reduction), block_size);

  if (preview) {
    cv::imshow("Nuclei detection (blob-based): after uniformization",
               img_preprocd_main_eq);
  }

  // -- (6) apply further preproc. stages and then run usual detection ---------
  ImagePreprocessingParameters p{preprocessing_p};
  p.normalizing = false; // |_ normalization and denoising already done on aux
  p.denoising = false;   // |
  stained::preprocessImage(img_preprocd_main_eq, p,
      stained::ImagePreprocessingStage::FINAL, img_preprocd_main_eq);

  return detectNuclei_(stained,
                       img_orig, img_preprocd_main_eq, cv::Mat{},
                       a_contours_of_nuclei,
                       a_contours_of_debris,
                       a_contours_nonsegmentable,
                       a_contours_of_nuclei_from_segmentation,
                       a_contours_of_orphans_from_segmentation,
                       preprocessing_p,
                       stained_p, nonstained_p, presel_p, assess_p,
                       declump_adjacent, preview, dry_run,
                       stop_enforced);
} // detectNucleiWithUniformization_()


double detectNuclei_DryWithTimeout(bool stained_mode,
    cv::Mat const& img_orig,
    cv::Mat const& img_preprocd_main,
    cv::Mat const& img_preprocd_aux,
    ImagePreprocessingParameters const& img_preproc_params,
    stained::NucleiDetectionParameters const& stained_params,
    nonstained::NucleiDetectionParameters const& nonstained_params,
    NucleiPreselectionParameters const& params,
    NucleiPreselectionAssessmentParameters const& assess_params,
    bool split_adjacent,
    bool preview,
    double timeout_secs)
{
  using namespace std;
//cout << "+ timed_detect() " << this_thread::get_id() << endl;

  // these are for formal parameters, will not be used (as dry_run := true)
  // API -- TODO
  st::contours_vec_t contours_nuclei, contours_debris,
      contours_nonsegmentable, contours_of_nuclei_from_segmentation,
      contours_of_orphans_from_segmentation;

  atomic<bool> stop_enforced{false};
  packaged_task<decltype(detectNuclei)> task{detectNuclei};
  future<double> fut{task.get_future()};
  thread thr{move(task), stained_mode,
      cref(img_orig), cref(img_preprocd_main), cref(img_preprocd_aux),
      &contours_nuclei,
      &contours_debris,
      &contours_nonsegmentable,
      &contours_of_nuclei_from_segmentation,
      &contours_of_orphans_from_segmentation,
      img_preproc_params, stained_params, nonstained_params, params, assess_params,
      split_adjacent,
      preview, true, cref(stop_enforced)} ;

  chrono::microseconds const dt(static_cast<unsigned long>(1e6*timeout_secs));
  future_status after_dt;
  do {
    after_dt = fut.wait_for(dt);
    if (after_dt == future_status::ready) {
      thr.join();
      return fut.get();
    } else {
      assert(after_dt == future_status::timeout);
      stop_enforced = true;
      thr.join();
      return -1e12; // PARAM
    }
    // below commented out because never executed:
    //this_thread::sleep_for(chrono::milliseconds(100));
  } while (after_dt == future_status::deferred) ; // TODO: probably a dead code

  thr.join();
  return -1e12; // PARAM
}


st::contours_vec_t splitNucleus(st::contour_t const& nuclear_contour,
                                st::contour_t& demarcation_line,
                                cv::Size const& original_image_size)
{
  auto bbox = st::boundingBox(nuclear_contour, 1, original_image_size);
  cv::Point const bbox_tl = bbox.tl();

  cv::Mat m_nuc = cv::Mat::zeros(bbox.size(), CV_8UC1);
  drawMask(m_nuc, nuclear_contour, -bbox_tl);

  for (auto& p: demarcation_line) { p -= bbox_tl; }
  for (int pi=0, ps=demarcation_line.size(); pi < ps-1; ++pi) {
    auto const& pA = demarcation_line[pi];
    auto const& pB = demarcation_line[pi + 1];
    cv::line(m_nuc, pA, pB, cv::Scalar{0}, 1, cv::LINE_4);
  }

  st::contours_vec_t contours;
  cv::findContours(m_nuc, contours, cv::RETR_EXTERNAL, cv::CHAIN_APPROX_NONE);
  contours.erase(std::remove_if(contours.begin(), contours.end(),
                                [](contour_t const& c) { return c.size() < 3; }),
                 contours.end());
  for (auto& contour: contours) { for (auto &p: contour) { p += bbox_tl; } }

  return contours;
}


} // namespace st
