// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#ifndef SHUTTLETRACKER_IMAGE_NUCLEIDETECTIONCOMMON_HPP
#define SHUTTLETRACKER_IMAGE_NUCLEIDETECTIONCOMMON_HPP

#include "ImageCommon.hpp"
#include "Types.hpp"
#if CV_MAJOR_VERSION >= 3
#  include <opencv2/imgproc.hpp>          // cv::contourArea
#elif CV_MAJOR_VERSION == 2
#  include <opencv2/imgproc/imgproc.hpp>  // cv::contourArea
#endif
#include <atomic>

namespace cv { class Mat; }

namespace st {

// Structs below have no defaults so that they can be used as PODs.

struct ImagePreprocessingParameters {
  bool   normalizing;
  bool   denoising;
  double denoising_strength;
  bool   smoothing;
  int    smoothing_size;
  bool   thresholding;
  int    thresholding_block_size;
  int    thresholding_baseline;
  bool   opening;
  int    opening_ntimes;
  int    opening_erosion_size;
  int    opening_dilation_size;
  bool   closing;
  int    closing_ntimes;
  int    closing_dilation_size;
  int    closing_erosion_size;
};


// TODO: rename
struct NucleiPreselectionAssessmentParameters {
  double area_cov_weight;
  double solidity_cov_weight;
};

namespace stained {
struct NucleiDetectionParameters {
  bool uniformization;
  int  uniformization_threshold;
};
}

namespace nonstained {
struct NucleiDetectionParameters {
  int   threshold_low;
  float thresholds_ratio;
  float tolerance;
  bool  reconstruct;
};
}

// TODO:rename
struct NucleiPreselectionParameters {
  double min_solidity;
  double min_median_contour_area_fraction;
  double max_median_contour_area_fraction;
  int min_pocket_area;
};


// @return score
// contours vectors are filled
double detectNuclei_(bool stained_mode,
    cv::Mat const& img_orig,
    cv::Mat const& img_preprocd_main, cv::Mat const& img_preprocd_aux,
    contours_vec_t *contours_nuclei,
    contours_vec_t *contours_debris,
    contours_vec_t *contours_nonsegmentable_ret,
    contours_vec_t *contours_of_nuclei_from_segmentation_ret,
    contours_vec_t *contours_of_orphans_from_segmentation_ret,
    ImagePreprocessingParameters const& preprocessing_p,
    stained::NucleiDetectionParameters const& stained_p,
    nonstained::NucleiDetectionParameters const& nonstained_params,
    NucleiPreselectionParameters const& params,
    NucleiPreselectionAssessmentParameters const& assess_params,
    bool declump_adjacent,
    bool preview = false,
    bool dry_run = false,
    const std::atomic<bool>& stop_enforced = std::atomic<bool>{false});


// @return score
// contours vectors are filled
double detectNucleiWithUniformization_(bool stained_mode,
    cv::Mat const& img_orig,
    cv::Mat const& img_preprocd_main, cv::Mat const& img_preprocd_aux,
    contours_vec_t *contours_nuclei,
    contours_vec_t *contours_debris,
    contours_vec_t *contours_nonsegmentable_ret,
    contours_vec_t *contours_of_nuclei_from_segmentation_ret,
    contours_vec_t *contours_of_orphans_from_segmentation_ret,
    ImagePreprocessingParameters const& preprocessing_p,
    stained::NucleiDetectionParameters const& stained_p,
    nonstained::NucleiDetectionParameters const& nonstained_params,
    NucleiPreselectionParameters const& params,
    NucleiPreselectionAssessmentParameters const& assess_params,
    bool declump_adjacent,
    bool preview = false,
    bool dry_run = false,
    const std::atomic<bool>& stop_enforced = std::atomic<bool>{false});


// @return score
// contours vectors are filled
double detectNuclei(bool stained_mode,
    cv::Mat const& img_orig,
    cv::Mat const& img_preprocd_main, cv::Mat const& img_preprocd_aux,
    contours_vec_t *contours_nuclei,
    contours_vec_t *contours_debris,
    contours_vec_t *contours_nonsegmentable_ret,
    contours_vec_t *contours_of_nuclei_from_segmentation_ret,
    contours_vec_t *contours_of_orphans_from_segmentation_ret,
    ImagePreprocessingParameters const& preprocessing_p,
    stained::NucleiDetectionParameters const& stained_p,
    nonstained::NucleiDetectionParameters const& nonstained_params,
    NucleiPreselectionParameters const& params,
    NucleiPreselectionAssessmentParameters const& assess_params,
    bool declump_adjacent,
    bool preview = false,
    bool dry_run = false,
    std::atomic<bool>const& stop_enforced = std::atomic<bool>{false});

// @return score
// dry run, because contours vectors are not filled
// (we are interested only in score)
double detectNuclei_DryWithTimeout(bool stained_mode,
    cv::Mat const& img_orig,
    cv::Mat const& img_preprocd,
    cv::Mat const& img_preprocd2,
    ImagePreprocessingParameters const& img_preproc_params,
    stained::NucleiDetectionParameters const& stained_params,
    nonstained::NucleiDetectionParameters const& nonstained_params,
    NucleiPreselectionParameters const& params,
    NucleiPreselectionAssessmentParameters const& assess_params,
    bool declump_adjacent,
    bool preview = false,
    double timeout_secs = 10);

#if 0
double OBSOLETE_detect_(bool stained_mode,
    cv::Mat const& img_orig,
    cv::Mat const& img_preprocd,
    contours_vec_t* contours_nuclei,
    contours_vec_t* contours_debris,
    contours_vec_t* contours_nonsegmentable_ret,
    contours_vec_t* contours_of_nuclei_from_segmentation_ret,
    contours_vec_t* contours_of_orphans_from_segmentation_ret,
    int nonstained_threshold_low,
    float nonstained_thresholds_ratio,
    float nonstained_tolerance,
    bool nonstained_reconstruct,
    double min_solidity = 0.93,
    double min_median_contour_area_fraction = 0.10,
    double max_median_contour_area_fraction = 1.15,
    unsigned int min_pocket_area = 10,
    double area_cov_weight = 0.0,
    double solidity_cov_weight = 1.0,
    bool declump_adjacent = true,
    bool preview = false,
    bool dry_run = false,
    const std::atomic<bool>& stop_enforced = std::atomic<bool>{false}) ;
#endif


st::contours_vec_t splitNucleus(st::contour_t const& nuclear_contour,
                                contour_t& demarcation_line,
                                cv::Size const& original_image_size);

namespace internal {

class ContourProperties {
public:
  ContourProperties(contour_t const& c)
  {
    contour_area_ = static_cast<float>(contourArea(c));

    cv::convexHull(c, hull_, false);
    hull_area_ = static_cast<float>(contourArea(hull_));

    for (auto const& p: c) {
      if (p.x < min_x_) { min_x_ = p.x; }
      if (p.x > max_x_) { max_x_ = p.x; }
      if (p.y < min_y_) { min_y_ = p.y; }
      if (p.y > max_y_) { max_y_ = p.y; }
    }
    width_  = max_x_ - min_x_ + 1;
    height_ = max_y_ - min_y_ + 1;
  }

  float area() const { return contour_area_; }
  float solidity() const { assert(hull_area_>0.f); return contour_area_/hull_area_; }
  contour_t const& hull() const { return hull_; }
  template <class S> S size() const { return {width_,height_}; }
  template <class P> P offset() const { return {min_x_,min_y_}; }
  int width() const { return width_; }
  int height() const { return height_; }

//inline void set_offset(int x,int y)
//{ min_x_ = x;               min_y_ = y;
//  max_x_ = x + width_ - 1;  max_y_ = y + height_ - 1; }

protected:
  contour_t hull_;
  float contour_area_ = -1.f,  hull_area_ = -1.f;
  int min_x_ = INT_MAX, max_x_ = INT_MIN,
      min_y_ = INT_MAX, max_y_ = INT_MIN;
  int width_ = -1, height_ = -1;
};

} // namespace st::internal
} // namespace st

#endif // !defined(SHUTTLETRACKER_IMAGE_NUCLEIDETECTIONCOMMON_HPP)
