// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "NucleiDetectionNonstained.hpp"
#include "NucleiDetection.hpp"
#include "ImageOperations.hpp"
#if CV_MAJOR_VERSION >= 3
#  include <opencv2/imgproc.hpp>
#  include <opencv2/highgui.hpp>
#  include <opencv2/ml.hpp>
#elif CV_MAJOR_VERSION == 2
#  include <opencv2/imgproc/imgproc.hpp>
#  include <opencv2/highgui/highgui.hpp>
#  include <opencv2/ml/ml.hpp>
#endif
#include <boost/geometry.hpp>
#include <boost/geometry/geometries/point.hpp>
#include <boost/geometry/geometries/box.hpp>
#include <boost/geometry/index/rtree.hpp>
#include <cstdlib>
#include <cstdio>
#include <cmath>
#include <algorithm>
#include <numeric> // accumulate
#include <tuple>
#include <future>
//#include <iostream>

namespace st {
namespace nonstained {

void preprocessImage(cv::Mat const& img_in,
                     cv::Mat& img_out_complete,
                     cv::Mat& img_out_partial,
                     ImagePreprocessingParameters const& p,
                     bool preview)
{
  assert(img_in.channels() == 1);
  img_out_complete = img_in.clone();

  if (p.normalizing) {
    st::normalize(img_out_complete);
    if (preview) { cv::imshow("Nuclei detection (edge-based): after normalization",
                              img_out_complete); }
  }

  // noise can be bright or dark
  if (p.denoising) {
    st::denoiseBilateral(img_out_complete, p.denoising_strength);
    if (preview) { cv::imshow("Nuclei detection (edge-based): after denoising",
                              img_out_complete); }
  }

  // noise can be bright or dark
  if (p.smoothing) {
    st::smoothenFixedSigma(img_out_complete, p.smoothing_size);
    if (preview) { cv::imshow("Nuclei detection (edge-based): after smoothing",
                              img_out_complete); }
  }

  img_out_complete.copyTo(img_out_partial);

  // noise is dark; contours are greatly enhanced
  if (p.closing) {
    st::morphologyClose(img_out_complete, p.closing_ntimes,
                        p.closing_dilation_size, p.closing_erosion_size);
    if (preview) { cv::imshow("Nuclei detection (edge-based): after closing",
                              img_out_complete); }
  }
} // preprocessImage()



// ------ CLEAN-UP necessary below ---------------------------------------------

namespace {

using namespace cv;
using namespace std;
namespace bg = boost::geometry;
namespace bgi = boost::geometry::index;

using contour_t = std::vector<cv::Point>;
using contours_vec_t = std::vector<contour_t>;
using contour_roi_t = std::pair<cv::Mat, cv::Point>;

cv::Scalar const black{  0};
cv::Scalar const white{255};

} // namespace (anonymous)


contour_roi_t contour_roi_(contour_t const& c, int margin=1)
{
  // use boundingRect() -- (one attempt failed) ?
  int x_mn = INT_MAX,       x_mx = INT_MIN,
      y_mn = INT_MAX,       y_mx = INT_MIN;
  for (auto const& p: c) {
    x_mn = min(x_mn, p.x);  x_mx = max(x_mx, p.x);
    y_mn = min(y_mn, p.y);  y_mx = max(y_mx, p.y);
  }

  Size const sz{x_mx - x_mn + 1 +2*margin,
                y_mx - y_mn + 1 +2*margin};
  return {Mat(sz, CV_8UC1, Scalar::all(0)),
          Point{-x_mn+margin, -y_mn+margin}};
}

// Note: this function may modify ROI image.
bool is_contour_fillable_(contours_vec_t const& contours,
                          int contour_i, contour_roi_t* roi)
{
  assert(roi and countNonZero(roi->first) == 0);

  // draw the contour filled (if possible), and then remove its border
  drawContours(roi->first, contours, contour_i, white, cv::FILLED,
               cv::LINE_8, cv::noArray(), INT_MAX, roi->second);
  drawContours(roi->first, contours, contour_i, black, 1,
               cv::LINE_8, cv::noArray(), INT_MAX, roi->second);

  // if empty image, then non-fillable (only border that was removed);
  // if non-empty image, then interior still present
  return countNonZero(roi->first) > 0;
}


float nonellipticity_(contour_t const& c, int extra_margin = 1)
{
  if (c.size() < 5) { return 1.f; }

  std::pair<Mat,Point> img_off = contour_roi_(c, 1 + extra_margin);
  auto& img = img_off.first;
  auto& off = img_off.second;
  assert(countNonZero(img) == 0);
  auto  img_fit = img.clone();

  RotatedRect box = fitEllipse(c);
  drawContours(img, std::vector<contour_t>{c}, 0, white,
               cv::FILLED, cv::LINE_8, noArray(), INT_MAX, off);
  box.center += Point2f(off);
  ellipse(img_fit, box, white, cv::FILLED, cv::LINE_8);
#if CV_MAJOR_VERSION >= 4
  double const d = matchShapes(img_fit, img, cv::CONTOURS_MATCH_I2, 0);
#elif CV_MAJOR_VERSION < 4
  double const d = matchShapes(img_fit, img, CV_CONTOURS_MATCH_I2, 0);
#endif
  return static_cast<float>(d);
}


inline
float ellipse_mismatch_(contours_vec_t const& contours, vector<Rect> const& rois,
                        int contour_i, int extra_margin=2) // PARAM:16f19be2
{
  assert(contour_i < static_cast<int>(contours.size()));
  assert(contour_i < static_cast<int>(rois.size()));

  auto roi = rois[static_cast<std::size_t>(contour_i)];
  auto off = Point{-roi.x+extra_margin, -roi.y+extra_margin};
  roi.x -= extra_margin;  roi.width  += 2*extra_margin;
  roi.y -= extra_margin;  roi.height += 2*extra_margin;

  if (contours[static_cast<std::size_t>(contour_i)].size() < 5) { return 1.f; }

  RotatedRect ell_box = fitEllipse(contours[static_cast<std::size_t>(contour_i)]);
  ell_box.center += Point2f{off};

  Mat img_cou{roi.size(), CV_8UC1, Scalar{0}},
      img_ell{roi.size(), CV_8UC1, Scalar{0}};
  drawContours(img_cou, contours, contour_i, white, cv::FILLED, cv::LINE_8,
               noArray(), INT_MAX, off);
  ellipse(img_ell, ell_box, white, cv::FILLED, cv::LINE_8);

  return static_cast<float>(countNonZero(img_cou ^ img_ell))
        /static_cast<float>(countNonZero(img_cou));
}


// @return minor to major axes ratio in range (0..1]
inline
float ellipse_axes_ratio_(contours_vec_t const& contours,
                          std::vector<Rect> const& rois,
                          int contour_i, int extra_margin=2) // PARAM:16f19be2
{
  auto roi = rois[static_cast<std::size_t>(contour_i)];
//auto off = Point{-roi.x+extra_margin, -roi.y+extra_margin};
  roi.x -= extra_margin;  roi.width  += 2*extra_margin;
  roi.y -= extra_margin;  roi.height += 2*extra_margin;

  if (contours[static_cast<std::size_t>(contour_i)].size() < 5) { return 0.f; }

  RotatedRect ell_box = fitEllipse(contours[static_cast<std::size_t>(contour_i)]);
  assert(ell_box.size.height > 0);
  float const r = ell_box.size.width/ell_box.size.height;

  return (r < 1.f) ? r : 1.f/r;
}


inline
float mean_intensity_adjacent_(Mat const& img, contours_vec_t const& contours,
                               vector<Rect> const& rois, int contour_i,
                               int extra_margin=5)
{
  auto roi = rois[static_cast<std::size_t>(contour_i)];
//auto off = Point{-roi.x+extra_margin, -roi.y+extra_margin};
  roi.x -= extra_margin;  roi.width  += 2*extra_margin;
  roi.y -= extra_margin;  roi.height += 2*extra_margin;
  roi.x = max(roi.x, 0);
  roi.y = max(roi.y, 0);
  if (roi.x + roi.width > img.size().width-1)  { roi.width  = img.size().width -1 - roi.x; }
  if (roi.y + roi.height> img.size().height-1) { roi.height = img.size().height-1 - roi.y; }
  assert(roi.x >=0 and roi.x < img.size().width);
  assert(roi.y >=0 and roi.y < img.size().height);

  Mat mask(roi.size(), CV_8UC1, Scalar{0});
  drawContours(mask, contours, contour_i, white, cv::FILLED, cv::LINE_8,
               noArray(), INT_MAX, {-roi.x, -roi.y});
  for (int i=0; i<extra_margin; ++i) {
    dilate(mask, mask, getStructuringElement(cv::MORPH_ELLIPSE,
                                             Size{3,3}, Point{1,1}));
  }
  drawContours(mask, contours, contour_i, Scalar{0}, cv::FILLED, cv::LINE_8,
               noArray(), INT_MAX, {-roi.x, -roi.y});
  auto const area_px = countNonZero(mask);
  assert(area_px > 0);

  img(roi).copyTo(mask, mask);
  return static_cast<float>(sum(mask)[0])/area_px;
}


inline
float mean_intensity_(Mat const& img, contours_vec_t const& contours,
                      vector<Rect> const& rois, int contour_i)
{
  auto const& roi = rois[static_cast<std::size_t>(contour_i)];
  Mat mask = Mat::zeros(roi.size(), CV_8UC1);
  drawContours(mask, contours, contour_i, white, cv::FILLED, cv::LINE_8,
               noArray(), INT_MAX, {-roi.x, -roi.y});
  int const area_px = countNonZero(mask);
  img(roi).copyTo(mask, mask);
  return static_cast<float>(sum(mask)[0])/area_px;
}


inline
int shared_edge_px_(contours_vec_t const& contours, vector<Rect> const& rois,
                    int contour_i, int contour_j)
{
  auto const q = rois[static_cast<std::size_t>(contour_i)],
             r = rois[static_cast<std::size_t>(contour_j)];
  Rect const roi{ Point{min(q.x,         r.x),
                        min(q.y,         r.y)},
                  Point{max(q.x+q.width, r.x+r.width),
                        max(q.y+q.height, r.y+r.height)} };
  //cout << q << endl << r << endl << roi << endl << endl;

  Mat img_i{roi.size(), CV_8UC1, Scalar{0}},
      img_j{roi.size(), CV_8UC1, Scalar{0}};
  drawContours(img_i, contours, contour_i, white, cv::FILLED, cv::LINE_8,
               noArray(), INT_MAX, {-roi.x, -roi.y});
  drawContours(img_j, contours, contour_j, white, cv::FILLED, cv::LINE_8,
               noArray(), INT_MAX, {-roi.x, -roi.y});
  Mat a{roi.size(), CV_8UC1, Scalar{0}};
  return countNonZero(img_i & img_j);
}


double intensity_zscore_(Mat const& img_gray, contour_t const& c)
{
  std::pair<Mat, Point> img_off = contour_roi_(c);
  auto& mask = img_off.first;
  auto& off = img_off.second;

  drawContours(mask, std::vector<contour_t>{c}, 0, white,
               cv::FILLED, cv::LINE_8, noArray(), INT_MAX, off);

  long int s = 0, n = 0;
  for(int j=0; j<mask.rows; j++) {
    for (int i=0; i<mask.cols; i++) {
      if (mask.at<unsigned char>(j, i) == 255) {
        s += img_gray.at<unsigned char>(j - off.x, i - off.y);
        ++n;
  } } }

  double mean = static_cast<double>(s)/n;
  double ss = 0.;
  for(int j=0;j<mask.rows;j++) {
    for (int i=0;i<mask.cols;i++) {
      if (mask.at<uchar>(j,i) == 255) {
        auto const v = img_gray.at<unsigned char>(j-off.x, i-off.y);
        ss += (v - mean)*(v - mean);
  } } }

  assert(n > 1);
  return sqrt(ss/(n - 1))/mean;
}


struct CorrectionSettings {
  float max_chord_length_sq = std::numeric_limits<float>::max(); // for editing
  float min_contour_length = 0.f;                     // for discriminating
  float min_area = 0.f;                               // for discriminating
  float max_area = std::numeric_limits<float>::max(); // for discriminating
  float avg_area = -1.f;                              // for scoring
  float max_ellipse_mismatch = 0.001f;                // PARAM
  int min_contour_n_points = 12;                      // PARAM (also a8ae7ef4)
};


//bool intersects_with_existing(Mat const&


std::tuple<contour_t, Point, int>
correct_nonfillable_(contours_vec_t const& contours,
                     std::vector<contour_roi_t>& contours_rois,
                     int contour_i,
                     CorrectionSettings const& sett/*,Mat const&img_orig*/,
                     std::vector<Mat_<int>> const& endpoint_kernels)
{
  assert(contours.size() == contours_rois.size());
  assert(contour_i < static_cast<int>(contours.size()));
  assert(sett.avg_area > 0.f);

  contour_t const& c = contours[contour_i];
  if (arcLength(c, false) < sett.min_contour_length) { return{{},{},contour_i}; }

  contour_roi_t& img_off = contours_rois[contour_i];
  auto& img = img_off.first;
  auto& off = img_off.second;
  assert(static_cast<int>( sum(img)[0] + sum(img)[1]
                          +sum(img)[2] + sum(img)[3]) == 0);

  Mat img_endpoints           {img.size(), CV_8UC1, black},
      img_endpoints_one_kernel{img.size(), CV_8UC1, black},
      img_with_chord          {img.size(), CV_8UC1, black};

  drawContours(img, std::vector<contour_t>{c}, 0, white,
               1, cv::LINE_8, noArray(), INT_MAX, off);

  // find endpoints
  //
  for (auto const& ek: endpoint_kernels) {
    img_endpoints_one_kernel = Scalar::all(0);
    morphologyEx(img, img_endpoints_one_kernel, cv::MORPH_HITMISS, ek);
    img_endpoints |= img_endpoints_one_kernel;
  }
  //imshow("EP", img_endpoints);
  //
  std::vector<Point> endpoints;
  endpoints.reserve(6); // practically, max is 6
  for(int j = 0; j < img_endpoints.rows; ++j) {
    uchar* row = img_endpoints.ptr<uchar>(j);
    for (int i = 0; i < img_endpoints.cols; ++i) {
      //if(img_endpoints.at<uchar>(j,i)==255){endpoints.push_back(Point(i,j));}
      if (row[i] == 255) { endpoints.emplace_back(i, j); }
  } }

  contour_t contour_off = c;  for (auto& p: contour_off) { p += off; }

  vector<contour_t> cs;
  using ranking_value_t = std::pair<contour_t, float>;
  vector<ranking_value_t> ranking;
  ranking.reserve(endpoints.size()*contour_off.size());
  for (auto const& ep: endpoints) {
    for (auto const& p_off: contour_off) {
      auto const dx = ep.x - p_off.x,  dy = ep.y - p_off.y;
      if (dx*dx + dy*dy > sett.max_chord_length_sq) { continue; }

      img_endpoints = Scalar::all(0); // reuse
      line(img_endpoints, ep, p_off, white); // draw chord
      img_endpoints |= img; // original contour with a chord
      findContours(img_endpoints, cs, cv::RETR_EXTERNAL, cv::CHAIN_APPROX_SIMPLE);

      img_with_chord = Scalar::all(0); // reuse
      drawContours(img_with_chord, cs, -1, white, cv::FILLED);
      img_with_chord -= img;  // remove original contour, leave the new if closed
      findContours(img_with_chord, cs, cv::RETR_EXTERNAL, cv::CHAIN_APPROX_NONE);

      for (auto const& c2: cs) {
        if (static_cast<int>(c2.size()) < sett.min_contour_n_points) { continue; }
        if (arcLength(c2, true) < sett.min_contour_length) { continue; }
        float const area = static_cast<float>(contourArea(c2));
        if (area < sett.min_area or area > sett.max_area) { continue; }
        //if (not is_elliptic(c)) { continue; }
        contour_t h;  cv::convexHull(c2, {h}, false);
        float const solidity = area/static_cast<float>(contourArea(h));
     // float izsc  = intensity_zscore(img, c); // typical range: [0:3] (never seen >3)
        float const ellipticity = 1.f - nonellipticity_(c2/*, sett.max_ellipse_mismatch*/);
        if (ellipticity < 1.f - sett.max_ellipse_mismatch) { continue; }
        float const area_scaling = (area/sett.avg_area)/((area/sett.avg_area) + 0.1f);
        float const score = area_scaling*solidity*ellipticity*ellipticity*ellipticity;
        ranking.emplace_back(c2, score);
      }
    } // for contour points
  } // for endpoints
  //cvtColor(img_endpoints, img_endpoints, cv::COLOR_GRAY2BGR);
  //imshow("fx", img_endpoints);

  //img_with_chord = Scalar::all(0);
  sort(ranking.begin(), ranking.end(),
       [](ranking_value_t const& c_s1, ranking_value_t const& c_s2)
         { return c_s1.second > c_s2.second; });
  /*
  if (not ranking.empty()) {
  drawContours(img_with_chord, std::vector<contour_t>{ranking.front().first},
               0, Scalar(255,255,255), cv::FILLED);
  cvtColor(img_with_chord, img_with_chord, cv::COLOR_GRAY2BGR);
  imshow("fx-dev", img_with_chord);
  }
  */
  if (not ranking.empty()) {
    img_with_chord = Scalar::all(0);
    drawContours(img_with_chord, vector<contour_t>{ranking.front().first}, 0,
                 white, cv::FILLED);
    Mat const element = getStructuringElement(cv::MORPH_ELLIPSE, Size{3,3},
                                                                 Point{1,1});
    dilate(img_with_chord, img_with_chord, element);  // 5418c7f5
    findContours(img_with_chord, cs, cv::RETR_EXTERNAL, cv::CHAIN_APPROX_SIMPLE);
    assert(cs.size() == 1);
    return {cs.front(), -off, contour_i};
  } else {
    return {contour_t{}, -off, contour_i};
  }
}


void assess_fillabilities_(contours_vec_t const&  contours,
                           vector<int>*           indices_fillable,
                           vector<int>*           indices_nonfillable,
                           vector<contour_roi_t>* contours_rois,
                           int index_base = 0)
{
  assert(indices_fillable);
  assert(indices_nonfillable);
  assert(contours_rois);
  assert(indices_fillable->size() + indices_nonfillable->size()
         == contours_rois->size());

  for (int i=index_base, s=static_cast<int>(contours.size()); i<s; i++) {
    contours_rois->emplace_back(contour_roi_(contours[i]));
    if (is_contour_fillable_(contours, i, &contours_rois->back())) {
      indices_fillable->push_back(i);
      assert(countNonZero(contours_rois->back().first) > 0); // assert as reminder
    } else {
      indices_nonfillable->push_back(i);
      assert(countNonZero(contours_rois->back().first) == 0);// assert as reminder
    }
  } // for each contour
} // assess_fillabilities_()


void extract_fillable_contour_leftovers_(contours_vec_t*        edges,
                                         vector<int>*           indices_fillable,
                                         vector<int>*           indices_nonfillable,
                                         vector<contour_roi_t>* contours_rois)
{
  int last_indices_fillable_size = 0;
  while (true) {

    auto const from = static_cast<int>(  indices_fillable->size()
                                       + indices_nonfillable->size());
    assess_fillabilities_(*edges, indices_fillable, indices_nonfillable,
                          contours_rois, from);

    // remove filled part in fillable contours, add remaining contour to analysis
    contours_vec_t edges_leftovers;
    cv::Mat const element =
        getStructuringElement(cv::MORPH_ELLIPSE, Size{3,3}, Point{1,1});

    for (auto i: *indices_fillable) {
      if (i < from) { continue; }

      // draw filled part and dilate to remove adjacent part of the contour
      cv::Mat interior = (*contours_rois)[i].first.clone();
      assert(countNonZero(interior) > 0);
      cv::dilate(interior, interior, element);

      // draw complete contour
      cv::Mat complete = cv::Mat::zeros(interior.size(), CV_8UC1);
      auto const off = (*contours_rois)[i].second;
      drawContours(complete, *edges, i, white, cv::FILLED, cv::LINE_8,
                   noArray(), INT_MAX, off);

      // subtract to get "leftovers"
      //imshow("complete", complete);  imshow("interior", interior);
      complete -= interior; // complete now contains leftovers
      //imshow("leftovers", complete);

      findContours(complete, edges_leftovers, cv::RETR_EXTERNAL, cv::CHAIN_APPROX_NONE);
      for (auto& cl: edges_leftovers) {
        if (cl.size() >= 12) { // PARAM (see also a8ae7ef4)
          for (auto& p: cl) { p -= off; }
          edges->push_back(cl);
      } }
      edges_leftovers.clear();
    } // for i

    if (static_cast<int>(indices_fillable->size()) == last_indices_fillable_size) {
      break; // nihil novi
    } else {
      last_indices_fillable_size = static_cast<int>(indices_fillable->size());
      continue;
    }
  } // while

  assert(contours_rois->size() ==  indices_fillable->size()
                                 + indices_nonfillable->size());
} // extract_fillable_contour_leftovers_()


void skeletonize_(Mat& img_canny, bool add_border = true)
{
  // The input image should have contours drawn using thickness == 3.
  // Because of this, it is required that n_iters == 2.
  int const n_iters = 2;

  Mat const element = getStructuringElement(cv::MORPH_CROSS, Size{3,3});
  auto const sz = img_canny.size();
  for (int i=0; i<n_iters; i++) {
    Mat eroded(sz, CV_8UC1);  erode(img_canny,eroded, element);
    Mat temp(sz, CV_8UC1);    dilate(eroded,temp, element);
                              subtract(img_canny,temp, temp);
    Mat skel(sz, CV_8UC1);    bitwise_or(skel,temp, skel);
    img_canny = eroded;
  }

  if (add_border) {
    line(img_canny, {0, 0}, {0,  sz.width-1}, white);
    line(img_canny, {0, 0}, {sz.height-1, 0}, white);
    line(img_canny, {sz.width-1, sz.height-1}, {sz.width-1 , 0}, white);
    line(img_canny, {sz.height-1,sz.height-1}, {0, sz.height-1}, white);
  }
}


struct Stats {
  float mean, std;
  Stats(vector<float>const& vs)
  {
    assert(vs.size() > 1);
    mean = accumulate(vs.begin(), vs.end(), 0.f)/vs.size();
    std = 0.f;
    for (float v: vs) { float d = v - mean;  std += d*d; }
    std = sqrt(std/(vs.size() - 1));
  }
};

struct ContourFeatures  {
  float const area;
  float const avg_ints;
  float const dif_ints;
  float const ell_mmch;
  float const ell_axrt;
};

struct ContourFeaturesStats  {
  Stats const area_stat;
  Stats const avg_ints_stat;
  Stats const dif_ints_stat;
  Stats const ell_mmch_stat;
  Stats const ell_axrt_stat;
};

enum class CellLine {
  MCF10A_AlbeckLab_c1
};


bool is_contour_acceptable_(ContourFeatures const& F,
                            ContourFeaturesStats const& S,
                            float TOL,
                            CellLine cell_line = CellLine::MCF10A_AlbeckLab_c1)
{
  switch (cell_line) {

  case CellLine::MCF10A_AlbeckLab_c1:
  {
    // Tolerance-scaling factors below were tuned by hand.

    // large intensity and adjacent region not much brighter
    if (    (F.avg_ints > S.avg_ints_stat.mean + TOL*S.avg_ints_stat.std*0.75f)
        and (F.dif_ints < S.dif_ints_stat.mean - TOL*S.dif_ints_stat.std*0.75f))
    { return false; }

    // large area and low average intensity => intercellular space (background)
    // (elliptic mismatch is OK for large contours)
    if (    (F.area > S.area_stat.mean + TOL*S.area_stat.std*1.0f)
        and (   (F.avg_ints < S.avg_ints_stat.mean - TOL*S.avg_ints_stat.std*1.00f)
             or (F.dif_ints < S.dif_ints_stat.mean - TOL*S.dif_ints_stat.std*0.75f)))
    { return false; }

    // non-large area and {elliptic mismatch or adjacent region brigtness problems}
    if (    (F.area <= S.area_stat.mean + TOL*S.area_stat.std)
        and (   (F.ell_mmch > S.ell_mmch_stat.mean + TOL*S.ell_mmch_stat.std*1.25f)
             or (F.dif_ints < S.dif_ints_stat.mean - TOL*S.dif_ints_stat.std*1.40f)  ) )
    { return false; }

    // rather small and elliptic mismatch
    // (small non-elliptic nuclei tend to have underappreciated elliptic mismatch)
    if (    (F.area < S.area_stat.mean + (0.5f - TOL)*S.area_stat.std)
        and (   (F.ell_mmch > S.ell_mmch_stat.mean + TOL*S.ell_mmch_stat.std*0.5f)
             or (F.ell_axrt < 0.45f))) // PARAM
    { return false; }

    // average size (so likely non-splittable), but long
    if (    (F.area < S.area_stat.mean + S.area_stat.std*1.5f)
        and (F.ell_axrt < 0.40f))
    { return false; }

    // VERY small and small intensity difference
    if (    (F.area < S.area_stat.mean - TOL*S.area_stat.std*0.8f)
        and (F.dif_ints < S.dif_ints_stat.mean - TOL*S.dif_ints_stat.std*0.2f))
    { return false;  }

    // JUST small and small intensity difference
    if (    (F.area < S.area_stat.mean - TOL*S.area_stat.std*0.25f)
        and (F.dif_ints < S.dif_ints_stat.mean - TOL*S.dif_ints_stat.std*1.25f))
    { return false;  }

    // average size but likely background
    if (    (F.area > S.area_stat.mean - TOL*S.area_stat.std*0.5f)
        and (F.area < S.area_stat.mean + TOL*S.area_stat.std*0.5f)
        and (    (F.avg_ints < S.avg_ints_stat.mean - TOL*S.avg_ints_stat.std*1.0f)
             and (F.dif_ints < S.dif_ints_stat.mean - TOL*S.dif_ints_stat.std*1.0f)))
    { return false;  }

    return true;
  } // case MCF10A_AlbeckLab_c1

  //default:                                   // commented out because switch
  //  return true; // accept unconditionally   //   covers all enum values

  } // switch cell line

  return true; // out-of-switch default
} // is_contour_acceptable_()


using point_t = bg::model::point<int, 2, bg::cs::cartesian>;
using box_t = bg::model::box<point_t>;
using geom_value_t = std::pair<box_t, int>;
using rtree_t = bgi::rtree<geom_value_t, bgi::quadratic<16>>;


inline box_t roi_to_box_(Rect const& roi)
{
  return { point_t{roi.x,           roi.y           },
           point_t{roi.x+roi.width, roi.y+roi.height} };
}


void reconstruct_contours_(contours_vec_t & corrected,
                           contours_vec_t & recycled_edges,
                           contours_vec_t const& canny_edges,
                           vector<Rect>& rois,
                           Stats const& area_stat,
                           Size const& img_orig_size,
                           float TOL, bool preview, Mat& img_preview)
{
  //t0 = cv::getTickCount();
  assert(canny_edges.size() == rois.size());

  static const Mat kernel_E = (Mat_<int>(3, 3) <<
       0, -1, -1,
      +1, +1, -1,
       0, -1, -1);
  static const Mat kernel_W = (Mat_<int>(3, 3) <<
      -1, -1,  0,
      -1, +1, +1,
      -1, -1,  0);
  static const Mat kernel_N = (Mat_<int>(3, 3) <<
      -1, -1, -1,
      -1, +1, -1,
       0, +1,  0);
  static const Mat kernel_S = (Mat_<int>(3, 3) <<
       0, +1,  0,
      -1, +1, -1,
      -1, -1, -1);
  static const Mat kernel_SE = (Mat_<int>(3, 3) <<
      +1,  0, -1,
       0, +1, -1,
      -1, -1, -1);
  static const Mat kernel_SW = (Mat_<int>(3, 3) <<
      -1,  0, +1,
      -1, +1,  0,
      -1, -1, -1);
  static const Mat kernel_NW = (Mat_<int>(3, 3) <<
      -1, -1, -1,
      -1, +1,  0,
      -1,  0, +1);
  static const Mat kernel_NE = (Mat_<int>(3, 3) <<
      -1, -1, -1,
       0, +1, -1,
      +1,  0, -1);

  static const std::vector<Mat_<int>> endpoint_kernels{
      kernel_N,  kernel_E,  kernel_S,  kernel_W,
      kernel_NE, kernel_NW, kernel_SE, kernel_SW
  };


  float const CORR_TOL = 1.667f*TOL;  // influences performace, not acurracy

  rtree_t rtree;
  vector<float> lengths;  lengths.reserve(canny_edges.size());
  int const roi_trim_px = 5;  // PARAM  (let square have soft edges, contours are round)
  for (int i=0, s=static_cast<int>(canny_edges.size()); i<s; ++i) {
    auto roi = rois[i];
    roi.x += roi_trim_px;  roi.width  -= 2*roi_trim_px;
    roi.y += roi_trim_px;  roi.height -= 2*roi_trim_px;
    if (roi.width <= 0 or roi.height <= 0) { continue; }
    rtree.insert(std::make_pair(roi_to_box_(roi), i));
    lengths.emplace_back(arcLength(canny_edges[i], /*true*/ false));
  }

  vector<int> indices_fillable, indices_nonfillable;
  vector<contour_roi_t> contours_rois;
  extract_fillable_contour_leftovers_(&corrected,
     &indices_fillable, &indices_nonfillable, &contours_rois);

  if (preview) {
    auto const color_ok_fill   = Scalar{96,255, 64}; // green
  //auto const color_ok_border = Scalar{ 0,  0,  0}; // black
    for (auto i: indices_fillable) {
      drawContours(img_preview, corrected, i, color_ok_fill,   cv::FILLED, cv::LINE_8);
    //drawContours(img_preview, corrected, i, color_ok_border, 1,          cv::LINE_8);
    }
    for (auto i: indices_nonfillable) {
      assert(countNonZero(contours_rois[i].first) == 0);
      auto const rand_color = Scalar(rand()&255, rand()&255, rand()&255);
      drawContours(img_preview, corrected, i, rand_color, 1, cv::LINE_8);
    }
  } // if preview

  Stats len_stat{lengths};
  CorrectionSettings sett;
      sett.max_chord_length_sq = (len_stat.mean/3.141592f) * CORR_TOL;
      sett.min_contour_length = max(len_stat.mean - CORR_TOL*len_stat.std,0.f);
      sett.min_area = max(area_stat.mean - (CORR_TOL*0.67f)*area_stat.std,0.f); // PARAM
      sett.max_area =     area_stat.mean + (CORR_TOL*1.50f)*area_stat.std;      // PARAM
      sett.avg_area = area_stat.mean;
      sett.max_ellipse_mismatch *= CORR_TOL;

  using correction_return_t = vector<std::tuple<contour_t,Point,int>>;
  vector<future<correction_return_t>> future_corrections;
  int const task_chunk = 0x20;
  Rect const img_rect = Rect{Point{0,0}, img_orig_size};
  for (int j=0, jsz=SSIZE(indices_nonfillable); j<jsz; j+=task_chunk) {
    future_corrections.emplace_back(
        std::async(std::launch::async, [&, j, jsz]() {
            vector<std::tuple<contour_t, Point, int>> ret;
            vector<geom_value_t> overlaps;
            for (int jj=j, je=min(j+task_chunk, jsz); jj<je; ++jj) {
              int const i = indices_nonfillable[jj];

              // check 1
              Rect corr_c_bounds = boundingRect(corrected[i]);
              if (   corr_c_bounds.x < 0
                  or corr_c_bounds.y < 0
                  or corr_c_bounds.x + corr_c_bounds.width >= img_rect.width
                  or corr_c_bounds.y + corr_c_bounds.height >= img_rect.height)
              { continue; }

              // check 2
              corr_c_bounds.x += roi_trim_px;
              corr_c_bounds.y += roi_trim_px;
              corr_c_bounds.width  -= 2*roi_trim_px;
              corr_c_bounds.height -= 2*roi_trim_px;
              if (   corr_c_bounds.width  < 0
                  or corr_c_bounds.height < 0) { continue; }
              rtree.query(bgi::intersects(roi_to_box_(corr_c_bounds)),
              back_inserter(overlaps));
              bool const overlapping = not overlaps.empty();
              overlaps.clear();  // will be reused
              if (overlapping) { continue; }

              ret.emplace_back( correct_nonfillable_(
                  cref(corrected), ref(contours_rois), i,
                  cref(sett)/*,img_{src?,orig?}*/,
                  endpoint_kernels));

            } // for each index in a chunk
            return ret; }));
  } // for each chunk of tasks
  auto const img_orig_wm1 = img_orig_size.width - 1,
             img_orig_hm1 = img_orig_size.height - 1;
  for (auto& fc: future_corrections) {
    auto chunk_of_fixed = fc.get();
    for (auto &fixed: chunk_of_fixed) {
      auto& fixed_contour     = get<0>(fixed);
      if (not fixed_contour.empty()) {
        auto& fixed_contour_off = get<1>(fixed);
        for (auto& p: fixed_contour) {
          p += fixed_contour_off;
          // dilation in a roi with a margin can lead to contours beyond
          // image border, which critically need adjustment -- see 5418c7f5
          if (p.x < 0){p.x = 0;} else if(p.x > img_orig_wm1){p.x = img_orig_wm1;}
          if (p.y < 0){p.y = 0;} else if(p.y > img_orig_hm1){p.y = img_orig_hm1;}
        }
        recycled_edges.push_back(fixed_contour);
        // formally, also (commented out):
        //canny_edges[get<2>(fixed)] = fixed_contour; // associated roi is not updated!
  } } }
//  std::cout << "Contours recyclable: "<< recycled_edges.size() << std::endl;

  if (preview) {
    auto const orange_light = Scalar{0, 128, 255},
               orange_dark  = Scalar{0, 64,  255};
    for (int i=0, s=static_cast<int>(recycled_edges.size()); i<s; ++i) {
      drawContours(img_preview, recycled_edges, i, orange_light, FILLED, LINE_AA);
      drawContours(img_preview, recycled_edges, i, orange_dark,  1,      LINE_AA);
    }

    Mat remain{img_orig_size, CV_8UC1, black};
    for (auto i: indices_nonfillable) {
      drawContours(remain, corrected, i, white, 1, cv::LINE_8);
    }
    Mat remain_final{remain.size(), CV_8UC1, black};
    Mat remain_local = remain_final.clone();
    for (auto const& ek: endpoint_kernels) {
      remain_local = Scalar::all(0);
      morphologyEx(remain, remain_local, cv::MORPH_HITMISS, ek);
      remain_final |= remain_local;
    }
    cvtColor(remain_final, remain_final, cv::COLOR_GRAY2BGR);
    img_preview |= remain_final;
  }  // if preview
} // reconstruct_contours_()


void findNuclearContours(cv::Mat const& img_orig,
                         cv::Mat const& img_preprocessed_complete,
                         cv::Mat const& img_preprocessed_partial,
                         contours_vec_t& returned_contours,
                         nonstained::NucleiDetectionParameters const& P,
                         bool const preview)
{
  assert(returned_contours.empty());
//int64 t0 = cv::getTickCount();

  int const canny_kernel_size = 5; // PARAM

  Mat img_preview;
  if (preview) {
    img_preview.create(img_orig.size(), CV_8UC3);
    img_preview.setTo(Scalar{0, 0, 0}); // or img_orig.copyTo(img_preview);
  }

  Mat img_canny;
  assert(P.thresholds_ratio >= 1.f);
  Canny(img_preprocessed_complete, img_canny, P.threshold_low,
        P.threshold_low*P.thresholds_ratio, canny_kernel_size, true);

  contours_vec_t canny_edges;  canny_edges.reserve(1024);
  findContours(img_canny, canny_edges, cv::RETR_EXTERNAL, cv::CHAIN_APPROX_NONE);
  drawContours(img_canny, canny_edges, -1, white, 3, cv::LINE_8);

  contours_vec_t canny_edges_saved_for_reconstruction;
  if (P.reconstruct) {
      copy(canny_edges.begin(), canny_edges.end(),
           back_inserter(canny_edges_saved_for_reconstruction));
  }

  skeletonize_(img_canny);
//imshow("skel", img_canny);
  vector<Vec4i> hierarchy;
  findContours(img_canny, canny_edges, hierarchy, cv::RETR_TREE, cv::CHAIN_APPROX_NONE);

  Mat Z;
  if (preview) {
    Z.create(img_canny.size(), CV_8UC3);  Z.setTo(black);
    auto const blueish = Scalar{64, 84, 96};
    for (int i=0; i < SSIZE(canny_edges); ++i) {
      drawContours(Z, canny_edges, i, blueish, 1, cv::LINE_AA);
    }
//  Mat A(img.size(), CV_8UC3, black);
//  Mat B(img.size(), CV_8UC3, black);
//  Mat C(img.size(), CV_8UC3, black);
//  Scalar color{128,128,128};
//  drawContours(A, canny_edges, -1, color, 1, cv::LINE_8, hierarchy, 1);
//  drawContours(B, canny_edges, -1, color, 1, cv::LINE_8, hierarchy, 2);
//  drawContours(C, canny_edges, -1, color, 1, cv::LINE_8, hierarchy, 3);
  }
//imshow("Z", Z);


  // remove non-children and thin contours ----------------

  assert(canny_edges.size() == hierarchy.size());
  auto it_last  = canny_edges.end();
  auto it_lasth = hierarchy.end();
  int n_rem=0;
  vector<float> canny_areas;  canny_areas.reserve(canny_edges.size());
  vector<int> marked_for_deletion_here;
  for (int i=0, s=static_cast<int>(canny_edges.size()); i < s-n_rem; ) {
    bool remove = false;
    float area;
    if (hierarchy[i][2] >= 0) {  // then has children
      remove = true;
    } else {
      float const length = arcLength(canny_edges[i], true);
      area = static_cast<float>(contourArea(canny_edges[i]));
      if (area < 1.f or (length*length)/(4.f*area*3.141592f) > 2.5f) {// PARAM - TODO:TOL
        remove = true;
    } }
    if (remove) {
      std::iter_swap(canny_edges.begin() + i, --it_last);
      std::iter_swap(hierarchy.begin()   + i, --it_lasth);
      ++n_rem;
    } else {
       canny_areas.push_back(area);
      ++i;
  } }
  canny_edges.erase(it_last, canny_edges.end());
  hierarchy.clear();
  canny_areas.shrink_to_fit();
  float avg_area =  std::accumulate(canny_areas.begin(), canny_areas.end(), 0.f)
                   /canny_areas.size();

  constexpr float four_pi = 4.f*3.141592f;
  for (int i=0; i < SSIZE(canny_edges); ++i) {
    auto const& c = canny_edges[AS_SIZE(i)];
    float const area = AS_FLOAT(canny_areas[AS_SIZE(i)]);
    float const len = AS_FLOAT(arcLength(c, true));
    float z = len*len/(area*four_pi);
    Scalar color = (z > 2.5f) ? Scalar{255, 0, 0} : Scalar{0, 0, 255};
    drawContours(Z, canny_edges, i, color, 1, cv::LINE_AA);
  }
//imshow("preZ", Z);

  // remove small contours --------------------------------

  bool removal_due = true;
  while (removal_due) {

    // setup
    int const n_clusters = 2,
              n_samples = static_cast<int>(canny_areas.size());
    assert(canny_edges.size() == canny_areas.size());
    if (not (n_clusters <= n_samples))  { break; }
    //
    Mat samples(n_samples, 1, CV_64FC1);
    samples = samples.reshape(1, 0);
    for (int i=0; i<n_samples; i++) { samples.at<double>(i) = canny_areas[i]; }

    // learning
    Ptr<ml::EM> em_model = ml::EM::create();
    em_model->setClustersNumber(n_clusters);
    em_model->setCovarianceMatrixType(ml::EM::COV_MAT_SPHERICAL);
    em_model->setTermCriteria(TermCriteria(
        TermCriteria::COUNT + TermCriteria::EPS, 3000, 1e-6));
    //
    Mat means(n_clusters, 1, CV_64FC1);
    means.at<double>(0) = 3.;  means.at<double>(1) = 300.; // PARAM
    //
    Mat probs(n_samples, n_clusters, CV_64FC1);
    Mat labels;
    em_model->trainE(samples, means,
                     /*cov*/noArray(), /*wght*/noArray(), /*outLogL*/noArray(),
                     labels, probs);
    means = em_model->getMeans();
//cout << n_samples << endl;
//for (int i=0; i<15; ++i) { cout << " " << canny_areas[i] << endl; }
//cout << means << endl << endl;
    int const label_smaller = (means.at<double>(0) < means.at<double>(1)) ? 0 : 1;
    int const label_larger  = 1 - label_smaller;
    double const mean_smaller = means.at<double>(label_smaller);
    double const mean_larger  = means.at<double>(label_larger);
    bool const means_ok = (mean_smaller < 0.67f*avg_area and 1.f*avg_area < mean_larger); // PARAM
    int const label_to_remove = means_ok ? label_smaller : label_larger;
//  cout << "\navg_area: " << avg_area  << endl;
//  cout << "means: S="<< mean_smaller <<" L="<< mean_larger << (means_ok?" OK":" BAD") << endl;
//  cout << "counts: "<< n_samples - sum(labels)[0] <<" -&- "<< sum(labels)[0] << endl;
//for (int i=0; i<n_samples; i++) {
//cout << "Z " << i << " " << canny_areas[i] << "  class:" << labels.at<int>(i)
//     << "  p_in_its_class:" << probs.at<double>(i, labels.at<int>(i)) << endl;
//}
    if (preview) {
      for (int i=0; i<n_samples; i++) {
        Scalar const color =
            (labels.at<int>(i) == label_to_remove) ? Scalar{128, 96, 0}
                                                   : Scalar{0, 172, 0};
        drawContours(Z, canny_edges, i, color, 1, cv::LINE_AA);
    } }

    // class-based removal
    n_rem = 0;
    it_last = canny_edges.end();
    auto it_lastl = labels.size().height-1;
    auto it_lasta = canny_areas.end();
    for (int i=0; i<static_cast<int>(canny_edges.size()) - n_rem; /**/) {
      if (labels.at<int>(i) == label_to_remove) {
        std::iter_swap(canny_edges.begin() + i, --it_last);
        std::iter_swap(canny_areas.begin() + i, --it_lasta);
        {
          auto ve = labels.at<int>(it_lastl);           // |
          labels.at<int>(it_lastl) = labels.at<int>(i); // +- swap
          labels.at<int>(i) = ve;                       // |
        }
        ++n_rem;
        --it_lastl;
      } else {
        ++i;
    } }
    canny_edges.erase(it_last, canny_edges.end());
    canny_areas.erase(it_lasta, canny_areas.end());

    // commented out because unused before next deletion:
//  avg_area = accumulate(canny_areas.begin(), canny_areas.end(), 0.f)/canny_areas.size();

    removal_due = not means_ok;
  } // while removal_due
//{
//Mat _(img_orig.size(), CV_8UC3, Scalar{0, 0, 0});
//drawContours(_, canny_edges, -1, Scalar{64,255,96}, 1, cv::LINE_8);
//imshow("Y", _);
//}

  // remove border-touching non-elliptic ------------------

  // determine rois and elliptic mismatches
  vector<Rect> rois;  rois.reserve(canny_edges.size());
  vector<float> ell_mmch;  ell_mmch.reserve(canny_edges.size());
  for (int i=0, s=static_cast<int>(canny_edges.size()); i<s; ++i) {
    rois.emplace_back(boundingRect(canny_edges[i]));
    ell_mmch.emplace_back(ellipse_mismatch_(canny_edges, rois, i));
  }

  Stats ell_mmch_stat{ell_mmch};
  vector<int> on_border;
  auto const img_orig_sz = img_orig.size();
  for (int i=0, s=static_cast<int>(canny_edges.size()); i<s; ++i) {
    auto const& roi = rois[i];
    if (   roi.x == 0  or  roi.y == 0
        or roi.x+roi.width  == img_orig_sz.width -1
        or roi.y+roi.height == img_orig_sz.height-1) {
      if (ell_mmch[i] > ell_mmch_stat.mean + ell_mmch_stat.std) { // PARAM
        on_border.push_back(i);
  } } }

  // quick and dirty removal
  std::vector<int> to_be_deleted_border(on_border.begin(),on_border.end());
  sort(to_be_deleted_border.begin(),to_be_deleted_border.end(),[](int a,int b){return a>b;});
  for (int i: to_be_deleted_border) {
    canny_edges.erase(canny_edges.begin() + i);
    canny_areas.erase(canny_areas.begin() + i);
    rois.erase(rois.begin() + i);
    ell_mmch.erase(ell_mmch.begin() + i);
  }
  to_be_deleted_border.clear();
  on_border.clear();

  // remove adjacent artifactual contours -----------------

  // prepare spatial lookup
  vector<float> avg_ints;  avg_ints.reserve(canny_edges.size());
  rtree_t rtree;
  for (int i=0, s=static_cast<int>(canny_edges.size()); i<s; ++i) {
    avg_ints.emplace_back(mean_intensity_(img_preprocessed_partial, canny_edges,
                                                                      rois, i));
    rtree.insert(std::make_pair(roi_to_box_(rois[i]), i));
  }

  set<int> to_be_deleted;
  for (int i=0, s=static_cast<int>(canny_edges.size()); i<s; ++i) {
    //if (to_be_deleted.count(i)) { continue; }
    auto const& query_box = roi_to_box_(rois[i]);
    vector<geom_value_t> overlaps;  // will include itself
    rtree.query(bgi::intersects(query_box), std::back_inserter(overlaps));
    for (auto const& ov: overlaps) {
      int const j = ov.second;
      if (j == i) { continue; } // no self-overlap
      if (shared_edge_px_(canny_edges, rois, i, j) > 0) { // PARAM
        //to_be_deleted.insert((neigh_badness(j) > neigh_badness(i)) ? (int)j : (int)i);
        float cont_ints_cutoff = 1.33f; // PARAM
        if (       avg_ints[i] > cont_ints_cutoff*avg_ints[j]) {
          to_be_deleted.insert(i);
        } else if (avg_ints[j] > cont_ints_cutoff*avg_ints[i]) {
          to_be_deleted.insert(j);
        }
    } }
  } // for each contour
  // Note: The r-tree can be cleared here.
  //
  if (preview) {
    for (int i=0, s=static_cast<int>(canny_edges.size()); i<s; ++i) {
      if (not to_be_deleted.count(i)) {
        drawContours(Z, canny_edges, i, Scalar{32, 255, 0}, 1, cv::LINE_AA);
  } } }

  // quick and dirty removal
  std::vector<int> to_be_deleted2(to_be_deleted.begin(),to_be_deleted.end());
  sort(to_be_deleted2.begin(),to_be_deleted2.end(),[](int a,int b){return a>b;});
  for (int i: to_be_deleted2) {
    canny_edges.erase(canny_edges.begin() + i);
    canny_areas.erase(canny_areas.begin() + i);
    rois.erase(rois.begin() + i);
    avg_ints.erase(avg_ints.begin() + i);
    ell_mmch.erase(ell_mmch.begin() + i);
  }
  to_be_deleted2.clear();
  to_be_deleted.clear();

  assert(canny_edges.size() == canny_areas.size());
  assert(canny_edges.size() == rois.size());
  assert(canny_edges.size() == avg_ints.size());
  assert(canny_edges.size() == ell_mmch.size());

  // remove remaining debris ------------------------------

  vector<float> dif_ints;  dif_ints.reserve(canny_edges.size());
  vector<float> ell_axrt;  ell_axrt.reserve(canny_edges.size());
  for (int i=0, s=static_cast<int>(canny_edges.size()); i<s; ++i) {
    auto const mia = mean_intensity_adjacent_(img_preprocessed_partial,
                                              canny_edges, rois, i);
    dif_ints.push_back(mia - avg_ints[i]);
    ell_axrt.push_back(ellipse_axes_ratio_(canny_edges, rois, i));
  }
  Stats area_stat{canny_areas};
  Stats avg_ints_stat{avg_ints};
  Stats dif_ints_stat{dif_ints};
  ell_mmch_stat = Stats{ell_mmch};
  Stats ell_axrt_stat{ell_axrt};
  ContourFeaturesStats const feats_stats{area_stat, avg_ints_stat,dif_ints_stat,
                                         ell_mmch_stat, ell_axrt_stat};
  set<int> to_be_deleted_debris;
  float const TOL = P.tolerance; // *1./static_cast<float>(tolerance_nominal);
  for (int i=0, s=static_cast<int>(canny_edges.size()); i<s; ++i) {
    ContourFeatures const feats{canny_areas[i], avg_ints[i], dif_ints[i],
                                ell_mmch[i], ell_axrt[i]};
    if (not is_contour_acceptable_(feats, feats_stats, P.tolerance)) {
      to_be_deleted_debris.insert(i);  continue;
  } }
//{
//Mat _(img.size(), CV_8UC3, black3);
//for (int i=0, s=canny_edges.size(); i<s; ++i) {
//  auto color = Scalar{20, 16, 108};
//  drawContours(_, canny_edges, i, color,
//     (to_be_deleted_debris.count(i)) ? 3 : 1, cv::LINE_AA);
//}
//imshow("debris", _);
//}

  std::vector<int> to_be_deleted3(to_be_deleted_debris.begin(),to_be_deleted_debris.end());
  sort(to_be_deleted3.begin(),to_be_deleted3.end(),[](int a,int b){return a>b;});
  for (int i: to_be_deleted3) {
    canny_edges.erase(canny_edges.begin() + i);
    canny_areas.erase(canny_areas.begin() + i);
    rois.erase(rois.begin() + i);
    avg_ints.erase(avg_ints.begin() + i);
    dif_ints.erase(dif_ints.begin() + i);
    ell_mmch.erase(ell_mmch.begin() + i);
  }
  to_be_deleted3.clear();
  to_be_deleted_debris.clear();

  // apply detailed correction if requested ---------------

  //std::cout << (cv::getTickCount()-t0)/cv::getTickFrequency() << " s" << endl;

  contours_vec_t recycled_edges;
  if (P.reconstruct) {
    reconstruct_contours_(canny_edges_saved_for_reconstruction,
        recycled_edges, canny_edges, rois, area_stat, img_orig.size(),
        P.tolerance, preview, img_preview);
    // TODO: check reconstructed for overlaps in a more detailed manner than
    //       with a rough r-tree

    area_stat     = Stats{canny_areas};
    avg_ints_stat = Stats{avg_ints};
    dif_ints_stat = Stats{dif_ints};
    ell_mmch_stat = Stats{ell_mmch};
  } // if reconstruct

  assert(canny_edges.size() == canny_areas.size());
  assert(canny_edges.size() == rois.size());
  assert(canny_edges.size() == avg_ints.size());
  assert(canny_edges.size() == ell_mmch.size());

  // process contours to be recycled -----------------------
  // (these contours were initially excluded as parental in hierarchy)

  vector<Rect> recycled_rois;
  for (int i=0, s = recycled_edges.size(); i<s; ++i) {
    auto const& c = recycled_edges[i];

    if (c.size() < 12) {  // PARAM (see also a8ae7ef4)
      recycled_rois.emplace_back(Rect{0,0,0,0});  // dummy to have indexing in-sync
      continue;
    }

    float const area = static_cast<float>(contourArea(c));
    if (   (area < area_stat.mean - TOL*area_stat.std)
        or (area > area_stat.mean + TOL*area_stat.std)) {
      recycled_rois.emplace_back(Rect{0,0,0,0});  // dummy to have indexing in-sync
      continue;
    }

    recycled_rois.emplace_back(boundingRect(c));

    float const avg_int = mean_intensity_(img_preprocessed_partial, recycled_edges, recycled_rois, i);
    if (   (avg_int < avg_ints_stat.mean - TOL*avg_ints_stat.std)
        or (avg_int > avg_ints_stat.mean + TOL*avg_ints_stat.std)) {
      continue;
    }

    float const dif_int =
          mean_intensity_adjacent_(img_preprocessed_partial, recycled_edges, recycled_rois, i)
        - avg_int;
    if (   (dif_int < dif_ints_stat.mean - TOL*dif_ints_stat.std)
        or (dif_int > dif_ints_stat.mean + TOL*dif_ints_stat.std)) {
      continue;
    }

    float const ell_mmch2 = ellipse_mismatch_(recycled_edges, recycled_rois, i);
    if (   (ell_mmch2 < ell_mmch_stat.mean - TOL*ell_mmch_stat.std)
        or (ell_mmch2 > ell_mmch_stat.mean + TOL*ell_mmch_stat.std)) {
      continue;
    }

    // all conditions met at this point, so:
    canny_edges.push_back(c);
  } // for i

  if (preview) {
    if (P.reconstruct) {
      imshow("Nuclei detection: edge tracing (reconstruction)", img_preview);
    }
    Mat img_preview2;
    cvtColor(img_orig, img_preview2, cv::COLOR_GRAY2BGR);
    img_preview2 += Z;
    drawContours(img_preview2, canny_edges, -1, Scalar{0,224,255}, 1, cv::LINE_AA);
    imshow("Nuclei detection: edge tracing (general)", img_preview2);
  }

  if (returned_contours.empty()) {
    returned_contours = canny_edges;
  } else {
    fprintf(stderr, "Warning: NonstainedNucleiDetection: "
                    "non-empty destination contaner.\n");
    std::copy(canny_edges.begin(), canny_edges.end(),
              std::back_inserter(returned_contours));
  }
} // findNuclearContours()

} // namespace nonstained
} // namespace st
