// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "NucleiDetectionStained.hpp"
#include "ImageOperations.hpp"
#include "Quantification.hpp"
#include "SharedUtilities.hpp"
#if CV_MAJOR_VERSION >= 3
#  include <opencv2/imgproc.hpp>
#  include <opencv2/highgui.hpp>
#elif CV_MAJOR_VERSION == 2
#  include <opencv2/imgproc/imgproc.hpp>
#  include <opencv2/highgui/highgui.hpp>
#endif

namespace st {
namespace stained {

void preprocessImage(cv::Mat const& orig_image,
                     ImagePreprocessingParameters const& p,
                     ImagePreprocessingStage stageA,  cv::Mat& imgA,
                     ImagePreprocessingStage stageB,  cv::Mat& imgB,
                     bool preview /*= false*/)
{
  using cv::imshow;
  using S = ImagePreprocessingStage;

  assert(orig_image.channels() == 1);    // |_ make sure that image format
  assert(orig_image.type() == CV_8UC1);  // |  is as expected by filters

  bool doneA = (stageA == S::NONE),
       doneB = (stageB == S::NONE);
  if (doneA and doneB) { return; }

  cv::Mat pre_image = orig_image.clone();
  std::string const wt = "Nuclei detection (blob-based): ";  // window title

  auto f = [stageA, &imgA, &doneA, stageB, &imgB, &doneB, &pre_image](S s)->bool
  {
    if (s == stageA) { assert(!doneA); imgA = pre_image.clone(); doneA = true; }
    if (s == stageB) { assert(!doneB); imgB = pre_image.clone(); doneB = true; }
    return (doneA and doneB);  // all expected work is done!
  };

  if (p.normalizing) {
    st::normalize(pre_image);
    if (preview) { imshow(wt + "after normalization", pre_image); }
  }
  if (f(S::AFTER_NORMALIZATION)) { return; }

  if (p.denoising) {
    st::denoise(pre_image, p.denoising_strength);
    if (preview) { imshow(wt + "after denoising", pre_image); }
  }
  if (f(S::AFTER_DENOISING)) { return; }

  if (p.smoothing) {
    st::smoothen(pre_image, p.smoothing_size);
    if (preview) { imshow(wt + "after smoothing", pre_image); }
  }
  if (f(S::AFTER_SMOOTHING)) { return; }

  if (p.thresholding) {
    st::threshold(pre_image, p.thresholding_block_size, p.thresholding_baseline);
    if (preview) { imshow(wt + "after tresholding", pre_image); }
  }
  if (f(S::AFTER_THRESHOLDING)) { return; }

  if (p.opening) {
    st::morphologyOpen(pre_image, p.opening_ntimes, p.opening_erosion_size,
                                                    p.opening_dilation_size);
    if (preview) { imshow(wt + "after opening", pre_image); }
  }
  if (f(S::AFTER_OPENING)) { return; }

} // preprocessImage()


void findNuclearContours(cv::Mat const& img_preprocd,
                        contours_vec_t& found_contours,
                        bool /*preview*/)
{
  using namespace cv;
  const int roi_margin = 1;
  Point const roi_offset{roi_margin, roi_margin};
  Size const twice_roi_margin_xy{2*roi_margin, 2*roi_margin};
  Size const img_preprocd_ext_sz = img_preprocd.size() + twice_roi_margin_xy;
  Mat img_preprocd_ext{img_preprocd_ext_sz, CV_8UC1, Scalar{0}};
  Rect const roi{roi_margin, roi_margin, img_preprocd.cols, img_preprocd.rows};
  Mat img_preprocd_ext_roi{img_preprocd_ext, roi};
  img_preprocd.copyTo(img_preprocd_ext_roi);
  findContours(img_preprocd_ext, found_contours, RETR_EXTERNAL,
               CHAIN_APPROX_NONE, -roi_offset);
}


void removeSpuriousNuclearContours(cv::Mat const& img_orig,
                                   st::contours_vec_t& contours,
                                   int peel_size /*=3*/,
                                   double threshold /*=0.*/)
{
  cv::Size const img_orig_size = img_orig.size();
  int const elem_sz = 2*peel_size + 1;
  cv::Mat const elem = cv::getStructuringElement(cv::MorphShapes::MORPH_ELLIPSE,
                                                 cv::Size{elem_sz, elem_sz});
  for(auto contour_it = std::make_reverse_iterator(contours.end());
      contour_it != std::make_reverse_iterator(contours.begin());
      ++contour_it) {
    cv::Rect const roi = st::boundingBox(*contour_it, peel_size+1, img_orig_size);
    cv::Mat const img_orig_roi = img_orig(roi);
    cv::Mat img_contour = cv::Mat::zeros(img_orig_roi.size(), CV_8UC1);
    st::drawMask(img_contour, *contour_it, -roi.tl());

    cv::Mat img_inner;  cv::erode (img_contour, img_inner, elem);
    cv::Mat img_outer;  cv::dilate(img_contour, img_outer, elem);
    img_inner = img_contour & (~img_inner);  // img_inner := img_contour OK too
    img_outer = img_outer   & (~img_contour);
    if (   cv::countNonZero(img_inner) < 10                // small would be
        or cv::countNonZero(img_outer) < 10) { continue; } // non-representative

    // first quartile is less sensitive to intensity from neighboring nuclei
    double q1_in  = st::quantifyInMaskIntensityQuartile(img_orig_roi, img_inner, 1);
    double q1_out = st::quantifyInMaskIntensityQuartile(img_orig_roi, img_outer, 1);
    if (q1_in - q1_out <= threshold) { // TODO: shouldn't test if not the last elem?
      std::swap(*contour_it, contours.back());
      contours.pop_back();
    }
  } // for each found contour
} // removeSpuriousNuclearContours()

} // namespace stained
} // namespace st
