// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "Perinuclei.hpp"
#include "SharedUtilities.hpp"
#include "opencv2/imgproc.hpp"
#include <vector>
#include <cmath>

constexpr int kExtraMargin = 3;

static st::Bicontour
findBicontour_(st::contour_t const& nuc_contour, cv::Rect const& roi,
               cv::Mat const& elem_in, cv::Mat const& elem_out,
               st::PerinucleiParameters const& params)
{
  // (1/4) draw the contour
  cv::Mat img_nuc = cv::Mat::zeros(roi.size(), CV_8UC1);
  cv::Point const roi_tl = roi.tl();
  st::drawMask(img_nuc, nuc_contour, -roi_tl);

  // (2/4) determine inner contour
  st::contours_vec_t found;
  if (params.dilation_in == 0) {
    auto shifted_nuc_contour = nuc_contour;
    for (auto& p: shifted_nuc_contour) { p -= roi_tl; }
    found.push_back(shifted_nuc_contour);
  } else {
    assert(params.dilation_in > 0);
    dilate(img_nuc, img_nuc, elem_in, {-1, -1}, params.dilation_in);
    findContours(img_nuc, found, cv::RETR_EXTERNAL, cv::CHAIN_APPROX_NONE);
  }
  assert(found.size() == 1);
  auto inner_contour = found.front();
  assert(not inner_contour.empty());
  for (auto& p: inner_contour) { p += roi_tl; }

  // (3/4) dilate further to determine outer contour
  cv::dilate(img_nuc, img_nuc, elem_out);
  found.clear();  // reuse
  findContours(img_nuc, found, cv::RETR_EXTERNAL, cv::CHAIN_APPROX_NONE);
  assert(found.size() == 1);
  auto outer_contour = found.front();
  assert(not outer_contour.empty());
  for (auto& p: outer_contour) { p += roi_tl; }

  // (4/4) assemble a final bi-contour
  return st::Bicontour{inner_contour, outer_contour,
                       st::centerOfMass(inner_contour)};
}


static
st::contour_t removeOverlap_(std::vector<st::Bicontour> const& bicontours,
                             unsigned int i, cv::Size img_size,
                             st::PerinucleiParameters const& params)
{
  // create a mask of the ith outer contour
  auto const area = cv::contourArea(bicontours[i].outer);
  auto const diameter = AS_INT(2*std::sqrt(area/3.141592));
  int const roi_margin = (params.dilation_in+params.dilation_out)+5*(10+diameter);
  cv::Rect const roi = st::boundingBox(bicontours[i].outer, roi_margin, img_size);
  cv::Point const roi_tl = roi.tl();
  cv::Mat img_outer = cv::Mat::zeros(roi.size(), CV_8UC1);
  st::drawMask(img_outer, bicontours[i].outer, -roi_tl);

  // create a mask of outer contours in the neighborhood of the ith outer contour
  cv::Mat img_neighboring_outers = cv::Mat::zeros(roi.size(), CV_8UC1);
  auto const com_i = bicontours[i].com;
  for (unsigned int pi=0; pi<bicontours.size(); ++pi) {
    if (pi == i) { continue; }
    auto shift = com_i - bicontours[pi].com;
    if (shift.x*shift.x + shift.y*shift.y > roi_margin*roi_margin) { continue; }
    st::drawMask(img_neighboring_outers, bicontours[pi].outer, -roi_tl);
  }

  // introduce avoidance of countours in proximity, even without overlap
  if (params.neighbor_evasion > 0) {
    int const size_base = 2*params.neighbor_evasion;
    auto const elem = getStructuringElement(cv::MORPH_ELLIPSE,
                                            cv::Size{size_base+1, size_base+1},
                                            cv::Point{size_base/2, size_base/2});
    dilate(img_neighboring_outers, img_neighboring_outers, elem);
  }

  // intersect masks of current and neighboring outer contours to get overlaps
  cv::Mat img_overlaps = img_outer & img_neighboring_outers;

  // inflate overlaps and subtract them from the current contour
  int const n_dilations = 1;
  int const size_base = 2*params.overlap_evasion;
  auto const elem = getStructuringElement(cv::MORPH_ELLIPSE,
                                          cv::Size{size_base+1, size_base+1},
                                          cv::Point{size_base/2, size_base/2});
  dilate(img_overlaps, img_overlaps, elem, {-1, -1}, n_dilations);
  img_outer -= img_overlaps;
  st::contours_vec_t found;
  findContours(img_outer, found, cv::RETR_EXTERNAL, cv::CHAIN_APPROX_NONE);

  if (found.size() == 0) {
    return st::contour_t{};  // empty outer contour
  } else if (found.size() == 1) {
    for (auto& pt: found.front()) { pt += roi_tl; }
    return found.front();  // the only outer contour
  } else {
    std::sort(found.begin(),found.end(),
              [](st::contour_t const& c1, st::contour_t const& c2)
                { return cv::contourArea(c1) < cv::contourArea(c2); });
    for (auto& pt: found.back()) { pt += roi_tl; }
    return found.back();  // the largest outer contour
  }

  // Of note, the outer contour is no longer guaranteed to envelope the inner
  // contour (it can enter the region of the inner contour). These artifacts
  // have to be fixed in a next step.
}


static
st::Perinucleus makePerinucleus_(st::Bicontour const& bicontour,
                                 cv::Mat const& img_bg_mask,
                                 cv::Size img_size,
                                 st::PerinucleiParameters const& params,
                                 float min_solidity_OvsC = 0.8f)  // PARAM
{
  if (bicontour.outer.empty()) {
    return st::Perinucleus{st::Perinucleus::Kind::kEmpty, {}};
  }

  // set ROI
  cv::Rect const outer_bb = st::boundingBox(bicontour.outer, kExtraMargin, img_size);
  auto const to_roi_shift = -outer_bb.tl();
  auto const img_roi_size =  outer_bb.size();

  // draw "ring"
  cv::Mat img_roi_outer = cv::Mat::zeros(img_roi_size, CV_8UC1);  // \_ outer
  st::drawMask(img_roi_outer, bicontour.outer, to_roi_shift);     // /
  cv::Mat img_roi_inner = cv::Mat::zeros(img_roi_size, CV_8UC1);  // \_ inner
  st::drawMask(img_roi_inner, bicontour.inner, to_roi_shift);     // /
  cv::Mat img_roi_splitter = img_roi_outer - img_roi_inner;       // difference
  if (params.bg_evasion) {
    assert(img_bg_mask.total() > 0);
    img_roi_splitter -= img_bg_mask(outer_bb);
  } else {
    assert(img_bg_mask.total() == 0);
  }

  // find contours of the "ring"
  st::contours_vec_t found;
  findContours(img_roi_splitter, found, cv::RETR_EXTERNAL, cv::CHAIN_APPROX_NONE);

  // remove very small contours
  auto too_small = [&params](st::contour_t const& c) -> bool
      { return cv::contourArea(c) < params.min_part_area; };
  found.erase(std::remove_if(found.begin(), found.end(), too_small), found.end());

  // un-ROI-fy contours of the "ring"
  for (auto& c: found) { for (auto& pt: c) { pt -= to_roi_shift; } }

  if (found.empty()) {
    return st::Perinucleus{st::Perinucleus::Kind::kEmpty, {}};
  } else if (found.size() > 1) {
    return st::Perinucleus{st::Perinucleus::Kind::kMultipartC, found};
  } else {
    assert(found.size() == 1);
    st::contour_t h;  convexHull(found.front(), {h}, false);
    auto const sol = AS_FLOAT(cv::contourArea(found.front())/cv::contourArea(h));
    if (sol <= min_solidity_OvsC) {
      return st::Perinucleus{st::Perinucleus::Kind::kCompleteC, found};
    } else {
      // enlarge inner contour by 1 pixel
      cv::Mat img_roi_inner2 = cv::Mat::zeros(img_roi_size, CV_8UC1);
      st::drawMask(img_roi_inner2, bicontour.inner, to_roi_shift);
      auto const elem = cv::getStructuringElement(cv::MORPH_ELLIPSE, cv::Size{3, 3});
      dilate(img_roi_inner2, img_roi_inner2, elem);
      st::contours_vec_t contour_inner_plus1px;
      findContours(img_roi_inner2, contour_inner_plus1px,
                   cv::RETR_EXTERNAL, cv::CHAIN_APPROX_NONE);

      // handle artifacts
      if (contour_inner_plus1px.empty()) {
        return st::Perinucleus{st::Perinucleus::Kind::kEmpty, {}};
      } else if (contour_inner_plus1px.size() > 1) {          // very rare; take
        int max_length = -1, max_length_index = -1;           // the longest
        for (int i=0; i<SSIZE(contour_inner_plus1px); i++) {  // contour and
          if (SSIZE(contour_inner_plus1px[i]) > max_length) { // hope that it
           max_length = SSIZE(contour_inner_plus1px[i]);      // is a reasonable
           max_length_index = i;                              // choice
        } }
        st::contours_vec_t contour_inner_plus1px_sel{
                                       contour_inner_plus1px[max_length_index]};
        contour_inner_plus1px.swap(contour_inner_plus1px_sel);
      }

      // return a complete either "C"-type or "O"-type perinucleus
      assert(contour_inner_plus1px.size() == 1);
      if (cv::contourArea(found.front()) <= cv::contourArea(bicontour.inner)) {
        return st::Perinucleus{st::Perinucleus::Kind::kCompleteC, found};  // rare
      } else {
        for (auto& pt: contour_inner_plus1px.front()) { pt -= to_roi_shift; }
        return st::Perinucleus{st::Perinucleus::Kind::kCompleteO,
                               {bicontour.inner, // inner for quantification
                                found.front(),   // outer for quantification
                                contour_inner_plus1px.front()}}; // inner for drawing
  } } }
}


static
void findBicontours_(std::vector<st::contour_t const*> const& nuclear_contours,
                     std::vector<st::Bicontour>* bicontours, cv::Size img_size,
                     st::PerinucleiParameters const& params)
{
  assert(bicontours and bicontours->empty());
  int const roi_margin =  params.dilation_in + params.dilation_out + kExtraMargin;
  auto const elem_in = getStructuringElement(cv::MORPH_ELLIPSE, cv::Size{3, 3});
  int const sz = 2*params.dilation_out + 1;
  auto const elem_out = getStructuringElement(cv::MORPH_ELLIPSE, cv::Size{sz, sz});
  bicontours->reserve(nuclear_contours.size());
  for (auto const& nc: nuclear_contours) {
    auto const bbox = st::boundingBox(*nc, roi_margin, img_size);
    bicontours->emplace_back(findBicontour_(*nc, bbox, elem_in, elem_out, params));
  }
}


static
void removeOverlaps_(std::vector<st::Bicontour>* bicontours, cv::Size img_size,
                     st::PerinucleiParameters const& params)
{
  assert(bicontours);
  std::vector<st::Bicontour> const bicontours0 = *bicontours;
  for (unsigned int i=0; i<bicontours0.size(); ++i) {
    (*bicontours)[i].outer = removeOverlap_(bicontours0, i, img_size, params);
  }
}


static
void makePerinuclei_(std::vector<st::Bicontour> const& bicontour,
                      std::vector<st::Perinucleus>* perinuclei,
                      cv::Mat const& img_bg_mask,
                      cv::Size img_size,
                      st::PerinucleiParameters const& params)
{
  assert(perinuclei);
  perinuclei->reserve(bicontour.size());
  for (auto const& bc: bicontour) {
    perinuclei->emplace_back(makePerinucleus_(bc, img_bg_mask, img_size, params));
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


void st::derivePerinuclei(std::vector<contour_t const*> const& nuclear_contours,
                          std::vector<Perinucleus>* perinuclei,
                          cv::Mat const& img_bg_mask,
                          cv::Size img_size,
                          PerinucleiParameters const& params)
{
  // (1/3) nuclear contours --> bi-contours
  std::vector<Bicontour> bicontours;
  findBicontours_(nuclear_contours, &bicontours, img_size, params);

  // (2/3) adjustments of bi-contours
  removeOverlaps_(&bicontours, img_size, params);

  // (3/3) adjusted bi-contours --> perinuclei
  makePerinuclei_(bicontours, perinuclei, img_bg_mask, img_size, params);
}


void st::makeBackgroundMask(cv::Mat const& img_orig,
                            cv::Mat* dest_img_bg_mask,
                            PerinucleiParameters const& params,
                            std::pair<float, float> const&& bg_intens_mu_sigma)
{
  assert(dest_img_bg_mask);
  float const threshold =  bg_intens_mu_sigma.first
                         + params.bg_expansion*bg_intens_mu_sigma.second;
  cv::Mat img_bg;
  cv::threshold(img_orig, img_bg, AS_DOUBLE(threshold), /*max_value*/255,
                cv::THRESH_BINARY_INV);

  int const size_base = 2*params.bg_smoothing;
  auto const elem0 = cv::getStructuringElement(cv::MORPH_ELLIPSE, cv::Size{3,3});
  auto const elem1 = cv::getStructuringElement(cv::MORPH_ELLIPSE,
                                               cv::Size{size_base+1, size_base+1},
                                               cv::Point{size_base/2,size_base/2});
  cv::erode(img_bg, img_bg, elem0);  // remove noise (isolated pixels)
  cv::dilate(img_bg, img_bg, elem1); // smoothen

  img_bg.copyTo(*dest_img_bg_mask);
}
