// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#ifndef SHUTTLETRACKER_IMAGE_PROGRESSREPORTER_HPP
#define SHUTTLETRACKER_IMAGE_PROGRESSREPORTER_HPP

#include "ImageCommon.hpp"
#ifdef USE_QT
#  include <QObject>
#  include <QDebug>
#endif
#include <limits>

namespace st {

struct ProgressReporter {
  virtual void yieldBusy() NOEXCEPT = 0;
  virtual void yieldIdle() NOEXCEPT = 0;
  virtual void yieldNextProgress() NOEXCEPT = 0;
  virtual void yieldCurrentProgress(int) NOEXCEPT = 0;
  virtual void yieldMaxProgress(int) NOEXCEPT = 0;
  virtual ~ProgressReporter();
};


struct SilentReporter: public ProgressReporter {
  void yieldBusy() NOEXCEPT OVERRIDE { }
  void yieldIdle() NOEXCEPT OVERRIDE { }
  void yieldNextProgress() NOEXCEPT OVERRIDE { }
  void yieldCurrentProgress(int) NOEXCEPT OVERRIDE {}
  void yieldMaxProgress(int) NOEXCEPT OVERRIDE {}
  ~SilentReporter() OVERRIDE;
};


#ifdef USE_QT
// Note tat the order of base classes matters for Qt MOC.
class QtReporter: public QObject, public ProgressReporter {
Q_OBJECT

signals:
  void busy();
  void idle();
  void progressValue(int);
  void progressMax(int);

public:
  void yieldBusy() NOEXCEPT OVERRIDE
  { //qDebug() << "QtReportingPolicy::busy()";
    busy_ = true;
    progress_ = 0;
    emit busy();
  }

  void yieldIdle() NOEXCEPT OVERRIDE
  { //qDebug() << "QtReportingPolicy::idle()";
    busy_ = false;
    progress_ = (std::numeric_limits<int>::max)();
    emit idle();
  }

  void yieldNextProgress() NOEXCEPT OVERRIDE
  { //qDebug() << "QtReportingPolicy::next_progress()";
    ++progress_;
    emit progressValue(progress_);
  }

  void yieldCurrentProgress(int value) NOEXCEPT OVERRIDE
  { //qDebug() << "QtReportingPolicy::current_progress()";
    emit progressValue(value);
  }

  void yieldMaxProgress(int max_value) NOEXCEPT OVERRIDE
  { //qDebug() << "QtReportingPolicy::max_progress() " << max_value;
    emit progressMax(max_value);
  }

  ~QtReporter() OVERRIDE;

private:
  bool busy_ = false;
  int progress_ = (std::numeric_limits<int>::max)();
}; // class  QtReporter
#endif // if USE_QT

} // namespace st

#endif // !defined(SHUTTLETRACKER_IMAGE_PROGRESSREPORTER_HPP)
