// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "Quantification.hpp"
#include "SharedUtilities.hpp"
#include <opencv2/core.hpp>
#include <opencv2/imgproc.hpp>  // cv::convexHull, ...
#include <vector>
#include <deque>
#include <algorithm>  // std::for_each
#include <execution>
#include <numeric>  // std::accumulate
#include <limits>
#include <cmath>  // std::fabsf, std::sqrt
#include <cstdio>
#include <cassert>

namespace st {

QuantResults quantifyInContours(cv::Mat const& orig_img,
                                contours_vec_t const& contours,
                                contours_vec_t const& exclude_in_contours,
                                bool calculate_inertia)
{
  assert(orig_img.channels() == 1);
  assert(not contours.empty());
  cv::Size const orig_img_size = orig_img.size();

  QuantResults results;  // returned

  std::vector<int> intensities;
  int contour_index = 0;
  for (auto const& contour: contours) {
#ifndef NDEBUG
    auto const r = cv::Rect{0, 0, orig_img.size().width, orig_img.size().height};
    for (auto const& p: contour) { assert(r.contains(p)); }
#endif
    cv::Rect const roi = boundingBox(contour, 1, orig_img_size);

    // will be used to include pixels for quantification
    cv::Mat mask_img_roi = cv::Mat::zeros(roi.size(), CV_8UC1);
    drawMask(mask_img_roi, contour, -roi.tl());

    // will be used for quantification
    cv::Mat const orig_img_roi0(orig_img, roi);
    assert(orig_img_roi0.size() == mask_img_roi.size());
    cv::Mat const orig_img_roi = orig_img_roi0 & mask_img_roi;

   //std::cerr <<"SUM "<<  cv::sum(mask_roi_img) << std::endl;
    int const orig_img_roi_size = orig_img_roi.rows*orig_img_roi.cols;
    intensities.reserve(intensities.size() + AS_SIZE(orig_img_roi_size*0.9));

    if (not exclude_in_contours.empty()) {
      cv::Mat xmask_img = cv::Mat::zeros(orig_img.size(), CV_8UC1);
      drawMask(xmask_img, exclude_in_contours);
      cv::Mat const xmask_img_roi(xmask_img, roi);
      assert(orig_img_roi.size() == xmask_img_roi.size());
      mask_img_roi = mask_img_roi & (~xmask_img_roi);
    }
    for (int row = 0; row < orig_img_roi.rows; ++row) {
      uchar const *prow_orig_img_roi = orig_img_roi.ptr(row);
      uchar const *prow_mask_img_roi = mask_img_roi.ptr(row);
      for (int col = 0; col < orig_img_roi.cols; ++col) {
        uchar const* oimg_px = prow_orig_img_roi;
        uchar const* mask_px = prow_mask_img_roi;
        if (*mask_px == 255) {
          int const v = oimg_px[0];
          results.min_intensity = std::min(results.min_intensity, v);
          results.max_intensity = std::max(results.max_intensity, v);
          intensities.push_back(v);  // for mean, median, sum
        } // if
        ++prow_orig_img_roi;
        ++prow_mask_img_roi;
    } } // for xy

    if (calculate_inertia and contours.size() == 1 and contour_index == 0) {
      results.inertia = st::internal::calcIntensityRelativeInertia__(
          orig_img_roi0, mask_img_roi, results.sum_of_intensities);
    }

    ++contour_index;
  } // for each contour

  if (intensities.empty()) { return results; }

  results.sum_of_intensities =
      std::accumulate(intensities.cbegin(), intensities.cend(), 0l);

  results.mean_intensity = AS_FLOAT(results.sum_of_intensities)
                          /intensities.size();
  std::sort(intensities.begin(), intensities.end());
  results.median_intensity    = intensities[intensities.size()/2];
  results.quartile1_intensity = intensities[intensities.size()*1/4];
  results.quartile3_intensity = intensities[intensities.size()*3/4];

  // means of each of four quartiles
#define  MEAN_BETWEEN(from, to)  AS_FLOAT(std::accumulate(from, to, 0l))/std::distance(from, to);
  //
  auto const ibeg = intensities.cbegin();
  auto const iend = intensities.cend();
  std::size_t const isz = intensities.size();
  //
  results.q1_mean_intensity = \
      MEAN_BETWEEN(ibeg,                     ibeg + AS_SIZE(0.25*isz));
  results.q2_mean_intensity = \
      MEAN_BETWEEN(ibeg + AS_SIZE(0.25*isz), ibeg + AS_SIZE(0.50*isz));
  results.q3_mean_intensity = \
      MEAN_BETWEEN(ibeg + AS_SIZE(0.50*isz), ibeg + AS_SIZE(0.75*isz));
  results.q4_mean_intensity = \
      MEAN_BETWEEN(ibeg + AS_SIZE(0.75*isz), iend);
  //
  results.t1_mean_intensity = \
      MEAN_BETWEEN(ibeg,                         ibeg + AS_SIZE(0.333333*isz));
  results.t2_mean_intensity = \
      MEAN_BETWEEN(ibeg + AS_SIZE(0.333333*isz), ibeg + AS_SIZE(0.666667*isz));
  results.t3_mean_intensity = \
      MEAN_BETWEEN(ibeg + AS_SIZE(0.666667*isz), iend);
  //
  results.h1_mean_intensity = \
      MEAN_BETWEEN(ibeg,                    ibeg + AS_SIZE(0.5*isz));
  results.h2_mean_intensity = \
      MEAN_BETWEEN(ibeg + AS_SIZE(0.5*isz), iend);
  //
  assert(results.q1_mean_intensity <= results.q2_mean_intensity);
  assert(results.q2_mean_intensity <= results.q3_mean_intensity);
  assert(results.q3_mean_intensity <= results.q4_mean_intensity);
  assert(results.t1_mean_intensity <= results.t2_mean_intensity);
  assert(results.t2_mean_intensity <= results.t3_mean_intensity);
  assert(results.h1_mean_intensity <= results.h2_mean_intensity);

  // std dev
  assert(results.mean_intensity >= 0.f);
  float sum_sq = 0.f;
  for (auto v: intensities) {
    float const d = results.mean_intensity - v;
    sum_sq += d*d;
  }
  if (intensities.size() >= 2) {
    results.stddev_intensity = std::sqrt(sum_sq/(intensities.size() - 1));
  } else {
    results.stddev_intensity = 0.001f;  // PARAMETER:CONVENTION
  }

  results.mean80_intensity =
      AS_FLOAT(std::accumulate(intensities.begin() + 0.1*intensities.size(),
                               intensities.end()   - 0.1*intensities.size(), 0l))
     /(0.8*intensities.size());
  sum_sq = 0.f;
  for (auto it = intensities.begin() + 0.1*intensities.size(),
            en = intensities.end()   - 0.1*intensities.size(); it != en; ++it) {
    float const d = results.mean80_intensity - *it;
    sum_sq += d*d;
  }
  if (intensities.size()*0.8 >= 2) {
    results.stddev80_intensity = std::sqrt(sum_sq/((intensities.size()*0.8) - 1));
  } else {
    results.stddev80_intensity = 0.001f;  // PARAMETER:CONVENTION
  }

  assert(results.max_intensity >= 0);
  assert(results.min_intensity < std::numeric_limits<int>::max());
  assert(results.mean_intensity >= 0.f);
  assert(results.stddev_intensity >= 0.f);
  assert(results.median_intensity >= 0);
  assert(results.quartile1_intensity >= 0);
  assert(results.quartile3_intensity >= 0);
  assert(results.sum_of_intensities >= 0);
#ifndef NDEBUG
  if (calculate_inertia and contours.size() == 1 and contour_index == 0) {
    assert(results.inertia >= 0.f);
  }
#endif

  return results;
}


double quantifyInMaskIntensityQuartile(cv::Mat const& orig_img,
                                       cv::Mat const& mask_img,
                                       int quartile)
{
  // intial checks
  assert((orig_img.type() == CV_8UC1) and (mask_img.type() == CV_8UC1));
  assert(orig_img.size() == mask_img.size());
  assert(orig_img.data != mask_img.data);
  assert(cv::countNonZero(mask_img) > 0);

  // collect values of pixels in mask
  std::vector<double> intensities;
  intensities.reserve(AS_SIZE(0.9*mask_img.total()));
  for (int row = 0; row < orig_img.rows; ++row) {
    uchar const *prow_orig_img = orig_img.ptr(row);
    uchar const *prow_mask_img = mask_img.ptr(row);
    for (int col = 0; col < orig_img.cols; ++col) {
      uchar const* oimg_px = prow_orig_img;
      uchar const* mask_px = prow_mask_img;
      if (*mask_px == 255) {
        int const v = oimg_px[0];
        intensities.push_back(v);
      } // if in mask
      ++prow_orig_img;
      ++prow_mask_img;
  } } // for x, for y
  std::sort(intensities.begin(), intensities.end());

  // compute and return requested statistics
  assert(not intensities.empty());
  std::size_t const index = intensities.size()*0.25*AS_DOUBLE(quartile);
  double const q = IS_EVEN(intensities.size())
                   ? AS_DOUBLE(intensities[index])
                   : AS_DOUBLE(  intensities[(index > 0) ? index - 1 : index]
                               + intensities[index])/2.;
  return q;
}


double quantifyInContoursMeanIntensity(cv::Mat const& orig_img,
                                       contours_vec_t const& contours,
                                       contours_vec_t const& exclude_in_contours)
{
  assert(orig_img.channels() == 1);
  assert(not contours.empty());

  double sum_intensities = 0.;
  long int n_pixels = 0l;

  for (auto const& contour: contours) {
#ifndef NDEBUG
    auto const r = cv::Rect{0, 0, orig_img.size().width, orig_img.size().height};
    for (auto const& p: contour) { assert(r.contains(p)); }
#endif
    cv::Rect const roi = boundingBox(contour);

    // will be used to include pixels for quantification
    cv::Mat mask_img_roi = cv::Mat::zeros(roi.size(), CV_8UC1);
    drawMask(mask_img_roi, contour, -roi.tl());

    // will be used for quantification
    cv::Mat const orig_img_roi0(orig_img, roi);
    assert(orig_img_roi0.size() == mask_img_roi.size());
    cv::Mat const orig_img_roi = orig_img_roi0 & mask_img_roi;

    if (not exclude_in_contours.empty()) {
      cv::Mat xmask_img = cv::Mat::zeros(orig_img.size(), CV_8UC1);
      drawMask(xmask_img, exclude_in_contours);
      cv::Mat const xmask_img_roi(xmask_img, roi);
      assert(orig_img_roi.size() == xmask_img_roi.size());
      mask_img_roi = mask_img_roi & (~xmask_img_roi);
    }

    for (int row = 0; row < orig_img_roi.rows; ++row) {
      uchar const *prow_orig_img_roi = orig_img_roi.ptr(row);
      uchar const *prow_mask_img_roi = mask_img_roi.ptr(row);
      for (int col = 0; col < orig_img_roi.cols; ++col) {
        uchar const* oimg_px = prow_orig_img_roi;
        uchar const* mask_px = prow_mask_img_roi;
        if (*mask_px == 255) {
          int const v = oimg_px[0];
          sum_intensities += v;
          ++n_pixels;
        } // if in mask
        ++prow_orig_img_roi;
        ++prow_mask_img_roi;
    } } // for x, for y
  } // for each contour

  if (n_pixels > 0l) { return sum_intensities/AS_DOUBLE(n_pixels); }
  else { return -1.; }  // PARAM CONVENTION ERROR
}


//using vquants_t = std::map<ImageChannel,
//                           std::vector<fnamed_vquant_t>>; // for each distance scaling function
vquants_t processWeigthedVoronoiPoints_(voronoi_weighted_points_t const& nuc_corresp_vwpts,
                                        std::map<ImageChannel, cv::Mat> const& images,
                                        std::map<std::string, std::function<float(float)>> const& d_scalings)
{
  vquants_t vquants;

  for (auto const& [channel, image]: images) {

    vquants.emplace(channel, std::vector<fnamed_vquant_t>{});
    vquants[channel].reserve(d_scalings.size());

    for (auto const& [func_name, scaling_func]: d_scalings) {

      float scaled_intensities_sum = 0.f;
      for (auto const& [pt, distance, scaling]: nuc_corresp_vwpts) {
        auto const value = image.at<unsigned char>(pt);
        scaled_intensities_sum += value * scaling * scaling_func(distance);
      } // for each point

      vquants[channel].emplace_back(func_name, scaled_intensities_sum);

    } // for each named distance scaling function
  } // for each channel/image

  return vquants;
}

std::vector<std::string> distanceScalingsNames()
{
  return {
    "linear_025",
    "linear_050",
    "linear_075",
    "linear_100",
    "linear_200",
    "linear_300",
    "linear_400",
    "linear_500",

    "sigmoidal_10_100",
    "sigmoidal_10_125",
    "sigmoidal_10_150",
    "sigmoidal_10_175",
    "sigmoidal_10_200",

    "sigmoidal_15_100",
    "sigmoidal_15_125",
    "sigmoidal_15_150",
    "sigmoidal_15_175",
    "sigmoidal_15_200",

    "sigmoidal_20_100",
    "sigmoidal_20_125",
    "sigmoidal_20_150",
    "sigmoidal_20_175",
    "sigmoidal_20_200",
  };
}


std::vector<vquants_t>
quantifyInCorrespWeightedVoronois(std::map<ImageChannel, cv::Mat> const& images,
                                  pcontours_vec_t const& pcontours)
{
  assert(images.size());
  cv::Size const image_size = images.begin()->second.size();
#ifndef NDEBUG
  for (auto const& [_, image]: images) { assert(image.size() == image_size); }
#endif

  auto const [nucs_vwpts, lambda] = \
      st::getCorrespVoronoisPointsDistances_(image_size, pcontours);

  // as the characteristic length is known, partially specialize
  auto const d_scalings = std::map<std::string, std::function<float(float)>>{
    { "linear_025", [lambda=lambda](float d){ return st::linear_025(d, lambda); } },
    { "linear_050", [lambda=lambda](float d){ return st::linear_050(d, lambda); } },
    { "linear_075", [lambda=lambda](float d){ return st::linear_075(d, lambda); } },
    { "linear_100", [lambda=lambda](float d){ return st::linear_100(d, lambda); } },
    { "linear_200", [lambda=lambda](float d){ return st::linear_200(d, lambda); } },
    { "linear_300", [lambda=lambda](float d){ return st::linear_300(d, lambda); } },
    { "linear_400", [lambda=lambda](float d){ return st::linear_400(d, lambda); } },
    { "linear_500", [lambda=lambda](float d){ return st::linear_500(d, lambda); } },

    { "sigmoidal_10_100", [lambda=lambda](float d){ return st::sigmoidal_10_100(d, lambda); } },
    { "sigmoidal_10_125", [lambda=lambda](float d){ return st::sigmoidal_10_125(d, lambda); } },
    { "sigmoidal_10_150", [lambda=lambda](float d){ return st::sigmoidal_10_150(d, lambda); } },
    { "sigmoidal_10_175", [lambda=lambda](float d){ return st::sigmoidal_10_175(d, lambda); } },
    { "sigmoidal_10_200", [lambda=lambda](float d){ return st::sigmoidal_10_200(d, lambda); } },

    { "sigmoidal_15_100", [lambda=lambda](float d){ return st::sigmoidal_15_100(d, lambda); } },
    { "sigmoidal_15_125", [lambda=lambda](float d){ return st::sigmoidal_15_125(d, lambda); } },
    { "sigmoidal_15_150", [lambda=lambda](float d){ return st::sigmoidal_15_150(d, lambda); } },
    { "sigmoidal_15_175", [lambda=lambda](float d){ return st::sigmoidal_15_175(d, lambda); } },
    { "sigmoidal_15_200", [lambda=lambda](float d){ return st::sigmoidal_15_200(d, lambda); } },

    { "sigmoidal_20_100", [lambda=lambda](float d){ return st::sigmoidal_20_100(d, lambda); } },
    { "sigmoidal_20_125", [lambda=lambda](float d){ return st::sigmoidal_20_125(d, lambda); } },
    { "sigmoidal_20_150", [lambda=lambda](float d){ return st::sigmoidal_20_150(d, lambda); } },
    { "sigmoidal_20_175", [lambda=lambda](float d){ return st::sigmoidal_20_175(d, lambda); } },
    { "sigmoidal_20_200", [lambda=lambda](float d){ return st::sigmoidal_20_200(d, lambda); } }
  };

  std::vector<vquants_t> vquants;  vquants.reserve(nucs_vwpts.size());
  std::for_each(
    std::execution::seq, // execution::par would mess the order which is crucial
    nucs_vwpts.begin(), nucs_vwpts.end(),
    [&images, &d_scalings, &vquants](voronoi_weighted_points_t const& vwpts) {
      vquants.push_back(processWeigthedVoronoiPoints_(vwpts, images, d_scalings));
    });

  return vquants;
}


QuantResults quantifyImage(cv::Mat const& img,
                           contours_vec_t const& exclude_in_contours)
{
  int const t = 0,  l = 0,  b = img.rows-1,  r = img.cols-1;
  using cv::Point;
  contour_t const whole_img{Point{l,t}, Point{r,t}, Point{r,b}, Point{l,b}};
  return quantifyInContours(img, {whole_img}, exclude_in_contours);
}

// // // // // // // // // // // // // // // // // // // // // // // // // //

// TODO: consider throwing an exception on erroneous input.
//
std::pair<float,float>
internal::calcGeometricFeatures_(contour_t const& contour)
{
  if (contour.size() < 6) { return {-1.f, -1.f}; }  // PARAM:INTERNAL

  cv::Mat pointsf;  cv::Mat(contour).convertTo(pointsf, CV_32F);
  cv::RotatedRect const box = fitEllipse(pointsf);

  float const lambda1 = box.size.width, lambda2 = box.size.height;
  float const eccentricity = fabsf(lambda1 - lambda2)/(lambda1 + lambda2);
  float const orientation = box.angle/360.f;

  return {eccentricity, orientation};
}

//
// TODO: This a more efficient contour extraction code which should be also
// applied in quantify_in_contour().
//
// TODO: It seems that there is some code duplication w.r.to boundingBox()
// (defined in SharedUtilities.cpp).
//
std::tuple<long, float, std::vector<double>>
internal::calcIntensityFeatures_(contour_t const& contour,
                                 cv::Mat const& image,
                                 cv::Point2f inertia_center)
{
  using cv::Mat;  using cv::Scalar; using cv::Size;
  using cv::Rect; using cv::Point;  using cv::Vec3b;


  int min_x = INT_MAX, max_x = INT_MIN,
      min_y = INT_MAX, max_y = INT_MIN;
  for (auto const& pt: contour) {
    min_x = std::min(min_x, pt.x);  max_x = std::max(max_x, pt.x);
    min_y = std::min(min_y, pt.y);  max_y = std::max(max_y, pt.y);
  }

  if ((min_x == max_x) or (min_y == max_y)) {
    fprintf(stderr, "Warning: cannot characterize intensities in an apparently "
                    "empty contour.\n");
    fflush(stderr);
    return {-1l, -1.f, std::vector<double>(kIntensHistBins, -1.)};
  }

  Mat const mask_roi_img = Mat::zeros(Size{max_x - min_x + 0,
                                           max_y - min_y + 0}, CV_8UC1);
  // TODO: code duplication w.r.to drawMask()
  cv::drawContours(mask_roi_img, contours_vec_t{contour}, 0,
                   Scalar{255, 255, 255}, cv::FILLED, 8, cv::noArray(), INT_MAX,
                   cv::Point{-min_x, -min_y});
//std::cerr <<"M "<< mask_roi_img.rows << "x" << mask_roi_img.cols << std::endl;

  Rect const roi{Point{min_x, min_y}, Point{max_x, max_y}};
  Mat const spot_roi_img(image, roi); // will be quantified
//std::cerr <<"S "<< spot_roi_img.rows << "x" << spot_roi_img.cols << std::endl;

//std::cerr <<"SUM "<<  cv::sum(mask_roi_img) << std::endl;

  // -- (1/3) sum of intensities -----------------------------------------------
  std::vector<double> intensities;
  intensities.reserve(AS_SIZE(0.85*mask_roi_img.rows*mask_roi_img.cols));
  // TODO: this iteration can be possibly be faster, use .ptr()
  long sum_of_intensities = 0ul;  // returned
  for (int yi=0; yi<mask_roi_img.rows; yi++) {
    for (int xi=0; xi<mask_roi_img.cols; xi++) {
      if (static_cast<int>(mask_roi_img.at<uchar>(yi, xi)) == 255) {
        auto const v = spot_roi_img.at<unsigned char>(yi, xi);
        intensities.push_back(static_cast<double>(v));
        sum_of_intensities += static_cast<unsigned int long>(v);
  } } } // for, for, if
  intensities.shrink_to_fit();

  // -- (2/3) inertia of intensities -------------------------------------------
  float const intensities_inertia = calcIntensityRelativeInertia__(
      spot_roi_img, mask_roi_img, sum_of_intensities,
      inertia_center - cv::Point2f{AS_FLOAT(min_x), AS_FLOAT(min_y)});

  // -- (3/3) distribution of intensities --------------------------------------
  std::vector<double> intensities_histogram(kIntensHistBins, 0.);  // returned
  if (not intensities.empty()) {
    // X-range normalization (=> luminance independence):
    std::sort(intensities.begin(), intensities.end());
    auto min_v = intensities.front();
    for (auto& v: intensities) { v -= min_v; }
    auto max_v = intensities.back() + 0.001;
    for (auto& v: intensities) { v /= max_v; } // all intens's are now in [0..1]
    //
    // collect histogram and normalize it to obtain PDF:
    for (auto const& v: intensities) {
      intensities_histogram[AS_SIZE(kIntensHistBins*v)]
          += 1./AS_DOUBLE(SSIZE(intensities));
    }
    //
    // turn to CDF:
    for (unsigned int bin_i=1u; bin_i<kIntensHistBins; ++bin_i) {
     intensities_histogram[bin_i] += intensities_histogram[bin_i - 1];
    }
    assert(fabs(1.0 - intensities_histogram[kIntensHistBins - 1]) < 0.000001);
  }

  return {sum_of_intensities, intensities_inertia, intensities_histogram};
}


float internal::calcIntensityRelativeInertia__(cv::Mat const& spot_image,
                                               cv::Mat const& spot_mask,
                                               long         a_sum_intensities,
                                               cv::Point2f  a_inertia_center)
{
  assert(spot_image.size() == spot_mask.size());
  assert(spot_mask.type() == CV_8UC1);
  assert(spot_image.type() == CV_8UC1);

  cv::Mat const masked_spot_image = spot_image & spot_mask;

  // -- (1/2) determine center and sum_intensities (if not given), and area ----
  cv::Point2f inertia_center;
  if (a_inertia_center.x < 0.f or a_inertia_center.y < 0.f) {
    try {  // try-catch is a quick-fix here, more in-depth checks are necessary
    //cv::Moments const M = cv::moments(masked_spot_image);  // old way
      cv::Moments const M = cv::moments(spot_mask);  // v1.5.1+
      if (M.m00 > 0.0) {
         inertia_center = { AS_FLOAT(M.m10/M.m00),    // x
                            AS_FLOAT(M.m01/M.m00) };  // y
      } else {
        fprintf(stderr, "Warning: Cannot compute moments to determine "
                        "the inertia center.\n");
        return -1.f;
      }
    } catch (...) {
      fprintf(stderr, "Warning: Cannot determine the inertia center.\n");
      return -1.f;
    }
  } else {
    inertia_center = a_inertia_center;
  }
  assert(inertia_center.x >= 0.f and inertia_center.y >= 0.f);

  long sum_intensities = 0;
  if (a_sum_intensities >= 0) {
    sum_intensities = a_sum_intensities;
  } else {
    for (int yi = 0; yi < masked_spot_image.rows; yi++) {
      uchar const* msi_row = masked_spot_image.ptr<uchar>(yi);
      for (int xi = 0; xi < masked_spot_image.cols; xi++) {
        assert(spot_mask.at<uchar>(yi, xi)==0 ? msi_row[xi]==0 : true);
        sum_intensities += msi_row[xi];  // it may be cheaper to add 0 than if()
  } } }
  if (sum_intensities == 0l) { return 1.f; }

  auto const surface_area = AS_DOUBLE(cv::countNonZero(spot_mask));
  assert(surface_area > 0.);
  if (surface_area <= 1) { return 1.f; }

  // -- (2/2) compute inertia of the actual image and of the uniform image -----
  double sumsq_image = 0.,
         sumsq_mask  = 0.;
  double const mask_v = 1./surface_area;  // uniform values are assumed as ref
  for (int yi = 0; yi < masked_spot_image.rows; ++yi) {
    uchar const* msi_row_px = masked_spot_image.ptr<uchar>(yi);
    uchar const* msk_row_px = spot_mask.ptr<uchar>(yi);
    for (int xi = 0; xi < masked_spot_image.cols; ++xi) {
      if (*msk_row_px == 255) {
        double const dx = AS_DOUBLE(inertia_center.x - xi),
                     dy = AS_DOUBLE(inertia_center.y - yi);
        double const dsq = (dx*dx + dy*dy) / surface_area; // norm'd sq'd distnc
        double const image_v = AS_DOUBLE(*msi_row_px) / AS_DOUBLE(sum_intensities);
        sumsq_image += dsq * image_v*image_v; // v1.5.1+: curr. intensity^2 (!)
        sumsq_mask  += dsq * mask_v *mask_v;  // v1.5.1+: avg.  intensity^2 (!)
      }
      ++msi_row_px;
      ++msk_row_px;
  } } // for yi, xi
  assert(sumsq_mask > 0.);

  return AS_FLOAT(sumsq_image / sumsq_mask);
}


QuantFeatures calculateFeatures(contour_t const& contour,
                                cv::Mat const& orig_image,
                                float area)
{
  QuantFeatures F; // returned

  F.center_of_mass = centerOfMass(contour);
  F.surface_area = (area > 0.f) ? area  // if a value was passed, use it
                                : AS_FLOAT(cv::contourArea(contour));

  std::tie(F.eccentricity, F.orientation) = internal::calcGeometricFeatures_(contour);

  auto ifs = internal::calcIntensityFeatures_(contour, orig_image, F.center_of_mass);
  F.intensities_sum = std::get<0>(ifs);
  F.intensities_inertia = std::get<1>(ifs);
  F.intensities_distribution.swap(std::get<2>(ifs));

  F.toggled = false; // should be changed if needed by an external code,
                     // which hopefully has access the actual toggled status
  return F;
}


std::vector<std::vector<double>> calculateSimilarities(
    std::map<Outline const*, QuantFeatures> const& features1,
    std::map<Outline const*, QuantFeatures> const& features2,
    SimilarityParameters const& weight,
    double distance_power_exponent)
{
  // -- (1/2) order features ---------------------------------------------------

  std::vector<QuantFeatures const*>
      pfeatures1_sorted_by_surface_area,
      pfeatures1_sorted_by_eccentricity,
      pfeatures1_sorted_by_orientation,
      pfeatures1_sorted_by_intensities_sum,
      pfeatures1_sorted_by_intensities_inertia,
      //
      pfeatures2_sorted_by_surface_area,
      pfeatures2_sorted_by_eccentricity,
      pfeatures2_sorted_by_orientation,
      pfeatures2_sorted_by_intensities_sum,
      pfeatures2_sorted_by_intensities_inertia;

  auto const f1S = features1.size();
  pfeatures1_sorted_by_surface_area.reserve(f1S);
  pfeatures1_sorted_by_eccentricity.reserve(f1S);
  pfeatures1_sorted_by_orientation.reserve(f1S);
  pfeatures1_sorted_by_intensities_sum.reserve(f1S);
  pfeatures1_sorted_by_intensities_inertia.reserve(f1S);
  for (auto const& f1: features1) {
    auto const* pfeat1 = &f1.second;
    pfeatures1_sorted_by_surface_area       .push_back(pfeat1);
    pfeatures1_sorted_by_eccentricity       .push_back(pfeat1);
    pfeatures1_sorted_by_orientation        .push_back(pfeat1);
    pfeatures1_sorted_by_intensities_sum    .push_back(pfeat1);
    pfeatures1_sorted_by_intensities_inertia.push_back(pfeat1);
  }
  //
  auto const f2S = features2.size();
  pfeatures2_sorted_by_surface_area.reserve(f2S);
  pfeatures2_sorted_by_eccentricity.reserve(f2S);
  pfeatures2_sorted_by_orientation.reserve(f2S);
  pfeatures2_sorted_by_intensities_sum.reserve(f2S);
  pfeatures2_sorted_by_intensities_inertia.reserve(f2S);
  for (auto const& f2: features2) {
    auto const* pfeat2 = &f2.second;
    pfeatures2_sorted_by_surface_area       .push_back(pfeat2);
    pfeatures2_sorted_by_eccentricity       .push_back(pfeat2);
    pfeatures2_sorted_by_orientation        .push_back(pfeat2);
    pfeatures2_sorted_by_intensities_sum    .push_back(pfeat2);
    pfeatures2_sorted_by_intensities_inertia.push_back(pfeat2);
  }

  auto cmp_area = [](QuantFeatures const* pf1, QuantFeatures const* pf2)
                    { return pf1->surface_area < pf2->surface_area; };
  std::sort(pfeatures1_sorted_by_surface_area.begin(),
            pfeatures1_sorted_by_surface_area.end(), cmp_area);
  std::sort(pfeatures2_sorted_by_surface_area.begin(),
            pfeatures2_sorted_by_surface_area.end(), cmp_area);

  auto cmp_ecc = [](QuantFeatures const* pf1, QuantFeatures const* pf2)
                   { return pf1->eccentricity < pf2->eccentricity; };
  std::sort(pfeatures1_sorted_by_eccentricity.begin(),
            pfeatures1_sorted_by_eccentricity.end(), cmp_ecc);
  std::sort(pfeatures2_sorted_by_eccentricity.begin(),
            pfeatures2_sorted_by_eccentricity.end(), cmp_ecc);

  auto cmp_ornt = [](QuantFeatures const* pf1, QuantFeatures const* pf2)
                    { return pf1->orientation < pf2->orientation; };
  std::sort(pfeatures1_sorted_by_orientation.begin(),
            pfeatures1_sorted_by_orientation.end(), cmp_ornt);
  std::sort(pfeatures2_sorted_by_orientation.begin(),
            pfeatures2_sorted_by_orientation.end(), cmp_ornt);

  auto cmp_isum = [](QuantFeatures const* pf1, QuantFeatures const* pf2)
                    { return pf1->intensities_sum < pf2->intensities_sum; };
  std::sort(pfeatures1_sorted_by_intensities_sum.begin(),
            pfeatures1_sorted_by_intensities_sum.end(), cmp_isum);
  std::sort(pfeatures2_sorted_by_intensities_sum.begin(),
            pfeatures2_sorted_by_intensities_sum.end(), cmp_isum);

  auto cmp_inrt = [](QuantFeatures const* pf1, QuantFeatures const* pf2)
                    { return pf1->intensities_inertia < pf2->intensities_inertia; };
  std::sort(pfeatures1_sorted_by_intensities_inertia.begin(),
            pfeatures1_sorted_by_intensities_inertia.end(), cmp_inrt);
  std::sort(pfeatures2_sorted_by_intensities_inertia.begin(),
            pfeatures2_sorted_by_intensities_inertia.end(), cmp_inrt);


  // -- (2/2) estimate pairwise-similarity scores ------------------------------
  //
  std::vector<std::vector<double>> similarities;  // returned
  similarities.reserve(features1.size());

  //for (size_t i1=0u, e1=features1.size(); i1<e1; ++i1) {
  for (auto const& feat1_: features1) {

    auto const& feat1 = feat1_.second;

    similarities.emplace_back(std::vector<double>{});
    similarities.back().reserve(features2.size());

    // prevent high scores for small (3-pts or so) contours:
    //
    if (feat1.intensities_sum == 0ul) {
      for (size_t i2=0u, e2=features2.size(); i2<e2; ++i2) {
        similarities.back().push_back(-1e9);// PARAM
      }                                     // CONSISTENCY:absolutely_dissimilar
      continue;
    }

    // COM -- center of the mass, // iKS -- intensts~ Kolmogorov--Smirnov
    std::vector<float>  COM_distances; COM_distances.reserve(features2.size());
    std::vector<double> iKS_distances; iKS_distances.reserve(features2.size());
    //
    auto const eucl_dist_sq = [](cv::Point2f const& p1, cv::Point2f const& p2)
    { auto const dx = p2.x - p1.x,  dy = p2.y - p1.y;  return dx*dx + dy*dy; };
    //
    auto const KS_dist = [](std::vector<double>const& intens1,
                            std::vector<double>const& intens2) -> double
    { double max_ksd = 0.;
      for (unsigned int bin_i=0u; bin_i<kIntensHistBins; ++bin_i) {
        assert(intens1.size() == kIntensHistBins and intens2.size() == kIntensHistBins);
        auto const ksd = fabs(intens1[bin_i] - intens2[bin_i]);
        if (ksd > max_ksd) { max_ksd = ksd; }
      }
      return max_ksd;
    };
    //
    // prepratory pass (for two distribution-based scores):
    //
    auto const& com1 = feat1.center_of_mass;
    auto const& isd1 = feat1.intensities_distribution;
    //for (size_t i2=0u, e2=features2.size(); i2<e2; ++i2) {
    for (auto const& feat2_: features2) {
      auto const& feat2 = feat2_.second;
      COM_distances.push_back(eucl_dist_sq(com1, feat2.center_of_mass));
      iKS_distances.push_back(KS_dist(isd1, feat2.intensities_distribution));
    }
    std::deque<float> sorted_COM_distances(COM_distances.begin(),
                                           COM_distances.end());
    std::deque<double>sorted_iKS_distances(iKS_distances.begin(),
                                           iKS_distances.end());
    std::sort(sorted_COM_distances.begin(),sorted_COM_distances.end());
    std::sort(sorted_iKS_distances.begin(),sorted_iKS_distances.end());

    // pairwise-comparison pass:
    //
    auto i2=0;
    for (auto const& feat2_: features2) {

      auto const& feat2 = feat2_.second;

      // prevent high scores for small (3-pts or so) contours:
      //
      if (feat2.intensities_sum == 0ul) {
        similarities.back().push_back(-1e9);// PARAM
        continue;                           // CONSISTENCY:absolutely_dissimilar
      }

      SimilarityParameters delta;

      // Delta center_of_mass
      auto const& v = sorted_COM_distances;
      auto const d = std::distance(v.begin(),
          std::lower_bound(v.begin(),v.end(), COM_distances[AS_SIZE(i2)]));
      delta.center_of_the_mass = AS_DOUBLE(d)/AS_DOUBLE(f2S);

      // Delta: surface_area, eccentricity, orientation, intensities_{sum,inertia}
      #define RATING(SORTED_CONTAINER, FEATURES, FEATURE_TYPE, FEATURE_NAME) \
        (std::distance(SORTED_CONTAINER.begin(),                           \
             std::lower_bound(SORTED_CONTAINER.begin(),                    \
                              SORTED_CONTAINER.end(),                      \
             FEATURES.FEATURE_NAME,                                        \
             [](QuantFeatures const* pfeats, FEATURE_TYPE cmp_feat_value)  \
               { return pfeats->FEATURE_NAME < cmp_feat_value; })))

      #define RELATIVE_RATING(r1, r2) \
          (fabs(AS_DOUBLE(r1)/AS_DOUBLE(f1S) - AS_DOUBLE(r2)/AS_DOUBLE(f2S)))

      delta.surface_area = RELATIVE_RATING(
          RATING(pfeatures1_sorted_by_surface_area, feat1, float, surface_area),
          RATING(pfeatures2_sorted_by_surface_area, feat2, float, surface_area));

      delta.eccentricity = RELATIVE_RATING(
          RATING(pfeatures1_sorted_by_eccentricity, feat1, float, eccentricity),
          RATING(pfeatures2_sorted_by_eccentricity, feat2, float, eccentricity));

      delta.orientation = RELATIVE_RATING(
          RATING(pfeatures1_sorted_by_orientation, feat1, float, orientation),
          RATING(pfeatures2_sorted_by_orientation, feat2, float, orientation));

      delta.intensities_sum = RELATIVE_RATING(
          RATING(pfeatures1_sorted_by_intensities_sum, feat1, long, intensities_sum),
          RATING(pfeatures2_sorted_by_intensities_sum, feat2, long, intensities_sum));

      delta.intensities_inertia = RELATIVE_RATING(
          RATING(pfeatures1_sorted_by_intensities_inertia,feat1,float,intensities_inertia),
          RATING(pfeatures2_sorted_by_intensities_inertia,feat2,float,intensities_inertia));

      #undef RELATIVE_RATING
      #undef RATING

      // Delta intensities_distribution
      auto const& vv = sorted_iKS_distances;
      auto const dd = std::distance(vv.begin(),
          std::lower_bound(vv.begin(),vv.end(), iKS_distances[AS_SIZE(i2)]));
      delta.intensities_distribution = AS_DOUBLE(dd)/AS_DOUBLE(f2S);

      double const dissimilarity =
          std::pow(eucl_dist_sq(feat1.center_of_mass,
                                feat2.center_of_mass),
                   0.5*distance_power_exponent)
         *SimilarityParameters::dissimilarityScore(delta, weight);
      similarities.back().push_back(-dissimilarity);

    } // for each feat2

    ++i2;

  } // for each feat1

  return similarities;
} // calculateSimilarities()


// obsolete because: works on 3-channel images and has no quants: stddev,q1,q3
QuantResults obsolete::quantifyInContours3(cv::Mat const& orig_img,
                                    contours_vec_t const& contours,
                                    contours_vec_t const& exclude_in_contours)
{
  assert(orig_img.channels() == 3);

  QuantResults results; // returned
  assert(results.max_intensity < 0);
  assert(results.min_intensity == std::numeric_limits<int>::max());
  assert(results.median_intensity < 0);
  assert(results.mean_intensity < 0);
  assert(results.sum_of_intensities < 0);

  cv::Mat xmasks_img = cv::Mat::zeros(orig_img.size(), CV_8UC1);
  drawMask(xmasks_img, exclude_in_contours);

  std::vector<int> intensities;
  for (auto const& contour: contours) {
#ifndef NDEBUG
    auto const& r = cv::Rect{0, 0,
                             orig_img.size().width, orig_img.size().height};
    for (auto const& p: contour) { assert(r.contains(p)); }
#endif
    cv::Mat mask_img = cv::Mat::zeros(orig_img.size(), CV_8UC1);
    drawMask(mask_img, contour);
    cv::Mat mask_img3;
    cv::cvtColor(mask_img, mask_img3, cv::COLOR_GRAY2BGR); // get a 3-channel mask..
    cv::Mat const spot_img = orig_img & mask_img3;    // ..and use it on orig_img

    cv::Rect const roi = boundingBox(contour);
    cv::Mat const   orig_img_roi(spot_img,  roi); // will be quantified
    cv::Mat const   mask_img_roi(mask_img,  roi); // will be used as a 1-ch mask
    cv::Mat const xmasks_img_roi(xmasks_img,roi); // will be used to exclude

   //std::cerr <<"SUM "<<  cv::sum(mask_roi_img) << std::endl;
    auto const orig_img_roi_size = orig_img_roi.rows*orig_img_roi.cols;
    intensities.reserve(intensities.size() + AS_SIZE(orig_img_roi_size/2));
    for (int yi=0; yi<orig_img_roi.rows; yi++) {
      for (int xi=0; xi<orig_img_roi.cols; xi++) {
        if (        (static_cast<int>(  mask_img_roi.at<uchar>(yi,xi)) == 255)
            and not (static_cast<int>(xmasks_img_roi.at<uchar>(yi,xi)) == 255)) {
          auto const vvv = orig_img_roi.at<cv::Vec3b>(yi,xi);
           // below max because C,M,Y imgs have the same values in 2 channels
          int const v = std::max(std::max(vvv[0], vvv[1]), vvv[2]);
          results.min_intensity = std::min(results.min_intensity, v);
          results.max_intensity = std::max(results.max_intensity, v);
          intensities.push_back(v);  // for mean, median, sum
        } // if
    } } // for, for
  } // for each contour

  if (intensities.empty()) { return results; }

  results.sum_of_intensities =
      std::accumulate(intensities.begin(), intensities.end(), 0);

  results.mean_intensity =   static_cast<float>(results.sum_of_intensities)
                           / intensities.size();

  std::sort(intensities.begin(), intensities.end());
  results.median_intensity = intensities[intensities.size()/2];

  return results;
} // quantifyInContours3()

} // namespace st
