// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#ifndef SHUTTLETRACKER_IMAGE_QUANTIFICATION_HPP
#define SHUTTLETRACKER_IMAGE_QUANTIFICATION_HPP

#include "ImageCommon.hpp"
#include "Types.hpp"
#include "ImageDecomposition.hpp"
#include <vector>
#include <map>
#include <limits>

namespace cv { class Mat; }
class Outline;

namespace st {

// -- intensity-based quantifications ------------------------------------------

struct QuantResults {
    int min_intensity       = std::numeric_limits<int>::max();
    int max_intensity       = -1;
  float mean_intensity      = -1.f;
  float stddev_intensity    = -1.f;
  float mean80_intensity    = -1.f;
  float stddev80_intensity  = -1.f;
    int median_intensity    = -1;
    int quartile1_intensity = -1;
    int quartile3_intensity = -1;
  float q1_mean_intensity = -1.f;
  float q2_mean_intensity = -1.f;
  float q3_mean_intensity = -1.f;
  float q4_mean_intensity = -1.f;
  float t1_mean_intensity = -1.f;
  float t2_mean_intensity = -1.f;
  float t3_mean_intensity = -1.f;
  float h1_mean_intensity = -1.f;
  float h2_mean_intensity = -1.f;
   long sum_of_intensities  = -1l;
  float inertia             = -1.f;
};

QuantResults quantifyInContours(cv::Mat const& orig_img,
                                contours_vec_t const& contours,
                                contours_vec_t const& exclude_in_contours = {},
                                bool calculate_inertia = true);

double quantifyInContoursMeanIntensity(cv::Mat const& orig_img,
                                       contours_vec_t const& contours,
                                       contours_vec_t const& exclude_in_contours={});

double quantifyInMaskIntensityQuartile(cv::Mat const& orig_img,
                                       cv::Mat const& mask_img, int quartile);

// function-named in-weighter-voronoi quantification
using fnamed_vquant_t = std::pair<std::string, float>;

using vquants_t = std::map<ImageChannel,
                           std::vector<fnamed_vquant_t>>; // for each distance scaling function

std::vector<vquants_t>
quantifyInCorrespWeightedVoronois(std::map<ImageChannel, cv::Mat> const& images,
                                  pcontours_vec_t const& pcontours);


std::vector<std::string> distanceScalingsNames();

namespace obsolete {
QuantResults quantifyInContours3(cv::Mat const& orig_img,
                                 contours_vec_t const& contours,
                                 contours_vec_t const& exclude_in_contours = {});
}

QuantResults quantifyImage(cv::Mat const& img,
                           contours_vec_t const& exclude_in_contours = {});


// -- quantifiable features, mainly for tracking -------------------------------

CONSTEXPR unsigned int kIntensHistBins = 12;

struct QuantFeatures {
          cv::Point2f center_of_mass;  // center of the mass
                float surface_area;
                float eccentricity;
                float orientation;
                 long intensities_sum;
                float intensities_inertia;
  std::vector<double> intensities_distribution; // range norm'd to [0,kKSHistBins] // TODO: USE small_vector from boost/llvm
                 bool toggled;
};


namespace internal {

std::pair<float, float>  // eccentricity, orientation
calcGeometricFeatures_(contour_t const&);

 // TODO: USE small_vector from boost/llvm
std::tuple<long, float, std::vector<double>>  // intensities sum,inertia,distrib
calcIntensityFeatures_(contour_t const&,
                       cv::Mat const& image,
                       cv::Point2f inertia_center = {-1.f, -1.f});

float calcIntensityRelativeInertia__(cv::Mat const& spot_image,
                                     cv::Mat const& spot_mask,
                                     long a_sum_intensities = -1,
                                     cv::Point2f a_inertia_center={-1.f, -1.f});
} // namespace internal


QuantFeatures calculateFeatures(contour_t const& contour,
                                cv::Mat const& orig_image,
                                float surface_area = -1.f);


// -- quantifiable features-based similarity -----------------------------------

struct SimilarityParameters {
  double center_of_the_mass       = std::numeric_limits<double>::quiet_NaN();
  double surface_area             = std::numeric_limits<double>::quiet_NaN();
  double eccentricity             = std::numeric_limits<double>::quiet_NaN();
  double orientation              = std::numeric_limits<double>::quiet_NaN();
  double intensities_sum          = std::numeric_limits<double>::quiet_NaN();
  double intensities_inertia      = std::numeric_limits<double>::quiet_NaN();
  double intensities_distribution = std::numeric_limits<double>::quiet_NaN();

  inline double sum() const
  {
    return  center_of_the_mass
          + surface_area
          + eccentricity
          + orientation
          + intensities_sum
          + intensities_inertia
          + intensities_distribution;
  }

  static double dissimilarityScore(SimilarityParameters const& delta,
                                   SimilarityParameters const& weight)
  {
    double dissimilarity = 0.;  // returned
    dissimilarity += delta.center_of_the_mass      *weight.center_of_the_mass;
    dissimilarity += delta.surface_area            *weight.surface_area;
    dissimilarity += delta.eccentricity            *weight.eccentricity;
    dissimilarity += delta.orientation             *weight.orientation;
    dissimilarity += delta.intensities_sum         *weight.intensities_sum;
    dissimilarity += delta.intensities_inertia     *weight.intensities_inertia;
    dissimilarity += delta.intensities_distribution*weight.intensities_distribution;
    dissimilarity /= weight.sum();
    return dissimilarity;
  }
}; // struct SimilarityParameters


std::vector<std::vector<double>>
calculateSimilarities(std::map<Outline const*, QuantFeatures> const& features1,
                      std::map<Outline const*, QuantFeatures> const& features2,
                      SimilarityParameters const& beta,
                      double distance_power_exponent);

} // namespace st

#endif // !defined(SHUTTLETRACKER_IMAGE_QUANTIFICATION_HPP)
