// ShuttleTracker, a tracker of nucleocytoplasmic trafficking in cells
//
// Copyright (C)  2015-2023  Marek Kochanczyk <mkochan:at:ippt.pan.pl>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU GPL version 3. See LICENSE.md for details.

#include "SharedUtilities.hpp"
#include "opencv2/core/types.hpp"
#include "opencv2/imgproc.hpp"  // cv::moments
#include <limits>
#include <numeric>  // std::accumulate
#include <cassert>

namespace st {

cv::Rect boundingBox(contour_t const& contour, int add_margin /* = 0 */,
                      cv::Size image_size /* = cv::Size{0, 0} */)
{
  // Cannot add margin if size of the image is not known:
  assert((add_margin == 0) ? (image_size == cv::Size{0, 0}) : true);
  assert((add_margin  > 0) ? (image_size != cv::Size{0, 0}) : true);

  int min_x = std::numeric_limits<int>::max(),
      max_x = std::numeric_limits<int>::min(),
      min_y = std::numeric_limits<int>::max(),
      max_y = std::numeric_limits<int>::min();

  for (auto const& pt: contour) {
    min_x = std::min(min_x, pt.x);
    max_x = std::max(max_x, pt.x);
    min_y = std::min(min_y, pt.y);
    max_y = std::max(max_y, pt.y);
  }

  if (add_margin > 0) {
    min_x = std::max(0,                       min_x - add_margin);
    max_x = std::min(image_size.width - 1*0,  max_x + add_margin);
    min_y = std::max(0,                       min_y - add_margin);
    max_y = std::min(image_size.height - 1*0, max_y + add_margin);
  }

  return cv::Rect{cv::Point{min_x, min_y},
                  cv::Point{max_x, max_y}};
}


// CONVENTION (maybe an exception should be thrown in centerOfMass() instead?)
namespace { const cv::Point2f kCOMFailureResult = cv::Point2f{-1.f, -1.f}; }
//
// Center of mass is computed from moments as described in:
// http://docs.opencv.org/modules/imgproc/doc/structural_analysis_and_shape_descriptors.html
// (which is as convenient as an overkill).
//
// TODO: consider throwing an exception on erroneous input.
//
cv::Point2f centerOfMass(contour_t const& contour)
{
  if (contour.size() < 2) { return kCOMFailureResult; }  // lbound found xptally

  try {  // try-catch is a quick-fix here, more in-depth checks are necessary
    cv::Moments const M = cv::moments(contour);
    if (M.m00 > 0.0) {
      return { static_cast<float>(M.m10/M.m00),
               static_cast<float>(M.m01/M.m00) };
    } else { return kCOMFailureResult; }
  } catch (...) { return kCOMFailureResult; }
}


float edgeScore(cv::Mat const& image, contour_t const& contour)
{
  assert(not contour.empty());
  assert(image.cols*image.rows > 0);
  assert(image.type() == CV_32FC1);
  using value_t = float;

  // compute average and std. deviation
  std::vector<value_t> intens;  intens.reserve(contour.size());
  for (auto const& p: contour) { intens.push_back(image.at<value_t>(p)); }
  double const avg = std::accumulate(intens.begin(), intens.end(), 0.)/intens.size();
  double sd = 0.;
  for (value_t value: intens) { double const d = avg - value;  sd += d*d; }
  sd = std::sqrt(sd / intens.size());

  constexpr double kStddevScaling = 1./3.;  // PARAM
  return AS_FLOAT(avg - sd*kStddevScaling);
}

} // namespace st
