#include "ImageSequence.hpp"
#include "Quantification.hpp"
#include "NucleiDetection.hpp"
#include "NucleiDetectionStained.hpp"
#include "SharedUtilities.hpp"
#include <opencv2/highgui.hpp>
#include <iostream>
#include <chrono>
#include <thread>
#include <cmath>
#include <tuple>

using namespace std;
using namespace st;

using cv::Mat; using cv::Scalar; using cv::Size;
using cv::Rect; using cv::Point; using cv::Vec3b;

using ImSeq = ImageSequence;
using ImKind = ImSeq::Channel;
using ImPreprocParams = st::ImagePreprocessingParameters;

const double pi = std::acos(-1);

struct Annulus {
  enum class Type : unsigned char {kCompleteO, kCompleteC, kMultipartC};
  Type kind;
  contours_vec_t contours;
};

std::tuple<contour_t,contour_t,cv::Point>
find_perinuclear_contours(contour_t const& nuc_contour, cv::Size const img_size)
{
  int dilation_size = 1; // PARAM
  int n_dilations_in = 2; // PARAM
  int n_dilations_out = 5; // PARAM

  Mat img_cur_nuc_mask = Mat::zeros(img_size, CV_8UC1);
  cv::drawContours( img_cur_nuc_mask, contours_vec_t{nuc_contour}, 0,
                    Scalar(255,255,255), cv::FILLED, 4 );

  // -- (1) Derive internal perinuclear contour --------------------------------

  // -- (1a) Establish ROI
  int roi_margin = n_dilations_in + n_dilations_out + 5; // INTERNAL:PARAM
  Rect const roi = boundingBox(nuc_contour, roi_margin, img_size);

  // -- (1b) Use ROI to create an image with and without neighboring nuclei

  Mat const img_cur_nuc_mask_local(  img_cur_nuc_mask, roi);

  // -- (1c) Use the image without neighboring nuclei to dilate it and create
  //         internal contour of the perinuclear region.
  Mat img_cur_nuc_mask_local_dil = img_cur_nuc_mask_local.clone();

  int size_base = 2*dilation_size;
  auto const elem_dilate = cv::getStructuringElement(cv::MORPH_ELLIPSE,
                                   cv::Size(size_base+1, size_base+1),
                                   cv::Point(size_base/2, size_base/2) );
  cv::dilate(img_cur_nuc_mask_local_dil, img_cur_nuc_mask_local_dil, elem_dilate,
             {-1,-1}, n_dilations_in);

  //cv::imshow("PERINUC DEVEL", img_cur_nuc_mask); cv::waitKey(0);
  //cv::imshow("PERINUC DEVEL", img_cur_nuc_mask_local); cv::waitKey(0);

  contours_vec_t found_contours_;
  findContours(img_cur_nuc_mask_local_dil, found_contours_,
     cv::RETR_EXTERNAL, cv::CHAIN_APPROX_NONE);
  if (found_contours_.size() != 1) {
    cerr << "Could not find an inner contour of a perinuclear annulus!\n";
  }
  auto perinuc_inner_contour = found_contours_.front();
  assert(not perinuc_inner_contour.empty());
  auto const roi_corner = roi.tl();
  for (auto& p: perinuc_inner_contour) { p += roi_corner; }

  //cv::imshow("PERINUC DEVEL", img_cur_nuc_mask_local_dil); cv::waitKey(0);


  // -- (1d) Dilate further to create external contour of the perinuclear region.
  for (int dil_i=0; dil_i<n_dilations_out; ++dil_i) {
    cv::dilate(img_cur_nuc_mask_local_dil, img_cur_nuc_mask_local_dil, elem_dilate);
    if (not (dil_i%2)) {
      cv::erode(img_cur_nuc_mask_local_dil, img_cur_nuc_mask_local_dil, elem_dilate);
      ++n_dilations_out;
    }
  }

  found_contours_.clear();
  findContours(img_cur_nuc_mask_local_dil, found_contours_,
     cv::RETR_EXTERNAL, cv::CHAIN_APPROX_NONE);
  if (found_contours_.size() != 1) {
    cerr << "Could not find an outer contour of a perinuclear annulus!\n";
  }
  auto perinuc_outer_contour = found_contours_.front();
  assert(not perinuc_outer_contour.empty());
  for (auto& p: perinuc_outer_contour) { p += roi_corner; }

  //cv::imshow("PERINUC DEVEL", img_cur_nuc_mask_local_dil); cv::waitKey(0);

  return std::make_tuple(perinuc_inner_contour, perinuc_outer_contour,
                         centerOfMass(perinuc_inner_contour));
}


contour_t remove_outer_contour_overlaps(
    std::vector<std::tuple<contour_t,contour_t,Point>> const& perinucs,
    int perinuc_index,
    cv::Size const img_size )
{
  auto const& perinuc_inner_countour = get<0>(perinucs[perinuc_index]);
  auto const& perinuc_outer_countour = get<1>(perinucs[perinuc_index]);
  auto const& com = get<2>(perinucs[perinuc_index]);

  Mat img_cur_perinuc_outer_mask = Mat::zeros(img_size, CV_8UC1);
  cv::drawContours( img_cur_perinuc_outer_mask, contours_vec_t{perinuc_outer_countour}, 0,
                    Scalar(255,255,255), cv::FILLED, 4 );
  Mat img_cur_perinuc_inner_mask = Mat::zeros(img_size, CV_8UC1);
  cv::drawContours( img_cur_perinuc_inner_mask, contours_vec_t{perinuc_inner_countour}, 0,
                    Scalar(255,255,255), cv::FILLED, 4 );
  Mat img_cur_perinuc_mask = Mat::zeros(img_size, CV_8UC1);
  img_cur_perinuc_mask = img_cur_perinuc_inner_mask | img_cur_perinuc_outer_mask;

  //cv::imshow("PERINUC DEVEL", img_cur_perinuc_inner_mask); cv::waitKey(0);
  //cv::imshow("PERINUC DEVEL", img_cur_perinuc_outer_mask); cv::waitKey(0);
  //cv::imshow("PERINUC DEVEL", img_cur_perinuc_mask); cv::waitKey(0);

  int const diameter = 2*std::sqrt(cv::contourArea(perinuc_outer_countour)/pi);
  int const diameter_sq = diameter*diameter;
  int const not_farther_then_diams = 5; // INTERNAL:PARAM

  Mat img_all_other_perinucs_masks = Mat::zeros(img_size, CV_8UC1);
  for (int pi=0; pi<perinucs.size(); ++pi) {
    if (pi == perinuc_index) { continue; }
    auto const& inner_c = get<0>(perinucs[pi]);
    auto const& outer_c = get<1>(perinucs[pi]);
    auto shift = com - get<2>(perinucs[pi]);
    if (shift.x*shift.x + shift.y*shift.y >
        not_farther_then_diams*not_farther_then_diams*diameter_sq) { continue; }
    //cv::drawContours( img_all_other_perinucs_masks, contours_vec_t{inner_c}, 0,
    //                  Scalar(255,255,255), cv::FILLED, 4 );
    cv::drawContours( img_all_other_perinucs_masks, contours_vec_t{outer_c}, 0,
                      Scalar(255,255,255), cv::FILLED, 4 );
  }

  // -- avoidance without overlap, just proximity
  //
  int const size_base0 = 2; // PARAM
  auto const elem_dilate0 = cv::getStructuringElement(cv::MORPH_ELLIPSE,
                                   cv::Size(size_base0+1, size_base0+1),
                                   cv::Point(size_base0/2, size_base0/2) );
  cv::dilate(img_all_other_perinucs_masks, img_all_other_perinucs_masks,
             elem_dilate0, {-1,-1}, 1);
  //

  Mat img_overlaps_masks = Mat::zeros(img_size, CV_8UC1);
  img_overlaps_masks  = img_cur_perinuc_mask & img_all_other_perinucs_masks;

  int const dilation_size = 4;
  int const size_base = 2*dilation_size;
  auto const elem_dilate = cv::getStructuringElement(cv::MORPH_ELLIPSE,
                                   cv::Size(size_base+1, size_base+1),
                                   cv::Point(size_base/2, size_base/2) );
  cv::dilate(img_overlaps_masks, img_overlaps_masks, elem_dilate, {-1,-1}, 3);
  //cv::imshow("Overlaps", img_overlaps_masks); cv::waitKey(0);

  img_cur_perinuc_mask -= img_overlaps_masks;
  //cv::imshow("Overlaps excluded", img_cur_perinuc_mask); cv::waitKey(0);

  contours_vec_t found_contours_;
  findContours(img_cur_perinuc_mask, found_contours_, cv::RETR_EXTERNAL, cv::CHAIN_APPROX_NONE);

  // THIS IS NAIVE:
  //assert(found_contours_.size() == 1);
  return found_contours_.front();
}


void generate_annuli(std::vector<std::tuple<contour_t,contour_t,cv::Point>> const& perinucs,
                     std::vector<Annulus>& annuli, cv::Size img_size)
{
  for (auto const& pn: perinucs) {

    auto const& inner_c = get<0>(pn);
    auto const& outer_c = get<1>(pn);

    int const roi_margin = 5; // INTERNAL:PARAM
    cv::Rect const outer_bb = boundingBox(outer_c, roi_margin, img_size);
    auto const to_roi_shift = -outer_bb.tl();
    auto const img_roi_size = outer_bb.size() + cv::Size{1, 1};

    Mat img_roi_outer = Mat::zeros(img_roi_size, CV_8UC1);
    auto const white = Scalar(255,255,255);
    cv::drawContours( img_roi_outer, contours_vec_t{outer_c}, 0, white,
                      cv::FILLED, 4, cv::noArray(), INT_MAX, to_roi_shift );
    Mat img_roi_inner = Mat::zeros(img_roi_size, CV_8UC1);
    cv::drawContours( img_roi_inner, contours_vec_t{inner_c}, 0, white,
                      cv::FILLED, 4, cv::noArray(), INT_MAX, to_roi_shift );

    Mat img_roi_splitter = Mat::zeros(img_roi_size, CV_8UC1);
    img_roi_splitter = img_roi_outer - img_roi_inner;

    contours_vec_t found_contours_;
    findContours(img_roi_splitter, found_contours_,
        cv::RETR_EXTERNAL, cv::CHAIN_APPROX_NONE);
    for (auto& c: found_contours_) {
      for (auto& pt: c) {
        pt -= to_roi_shift;
    } }

    // TODO below: filter out too little contours

    if (found_contours_.size() == 0) {
      // TODO: error handling
    } else if (found_contours_.size() == 1) {
      contour_t h;  cv::convexHull(found_contours_.front(), {h}, false);
      auto const solidity = static_cast<float>(
            cv::contourArea(found_contours_.front())/cv::contourArea(h));
      if (solidity > 0.8f) { // PARAM
        annuli.push_back(Annulus{Annulus::Type::kCompleteO,
                                 {found_contours_.front(),inner_c}});
      } else {
        annuli.push_back(Annulus{Annulus::Type::kCompleteC, found_contours_});
      }
    } else {
      annuli.push_back(Annulus{Annulus::Type::kMultipartC, found_contours_});
    }

  } // for each perinuc
}



int main()
{  
  ImSeq s;
  s.set(ImSeq::StorageStrategy::RAM);
  auto dir = "/home/marek/tmp/ad_shuttletracker/S1-sub";
  //auto dir = "/home/marek/Syncium/Projekty/MACDOBRY/2018-05-01_MCF10AmutInhib--WORK/A2/S1-sub";
  s.loadDirectory(dir);
  cout << s.timepointsCount() << " frames" << endl;

  auto const& img_orig = s.image(ImKind::B, 4);

  st::contours_vec_t contours_nuclei,
      contours_of_debris,
      contours_nonsegmentable,
      contours_of_nuclei_from_segmentation,
      contours_of_orphans_from_segmentation;

    ImPreprocParams const img_preproc_params = {
        true, // ui_->checkBoxNucDetNormalization->isChecked(),
        false, //ui_->checkBoxNucDetDenoising->isChecked(),
           5, //    static_cast<float>(ui_->dspinBoxNucDetDenoising->value()),
        true, // ui_->checkBoxNucDetSmoothing->isChecked(),
           5, //ui_->spinBoxNucDetSmoothing->value(),
        true, // ui_->checkBoxNucDetThresholding->isChecked(),
            501, // ui_->spinBoxNucDetThresholdingBlocksize->value(),
            0, // ui_->spinBoxNucDetThresholdingBaseline->value(),
        true, // ui_->checkBoxNucDetOpening->isChecked(),
            1, // ui_->spinBoxNucDetOpeningTimes->value(),
            2, // ui_->spinBoxNucDetOpeningErosion->value(),
            2, //ui_->spinBoxNucDetOpeningDilation->value()
    };
    cv::Mat img_preproc;
    cv::Mat img_preproc_UNUSED;
    st::preprocessStainedNucleiImage(
        img_orig, img_preproc_UNUSED, img_preproc, img_preproc_params, /*preview=*/false);

    st::NucleiPreselectionParameters nuc_presel_params = {
            0.93, // ui_->dspinBoxNucDetSolidity->value(),
            0.5, // ui_->dspinBoxNucDetMinAreaFrac->value(),
            1.1, // ui_->dspinBoxNucDetMaxAreaFrac->value(),
            5 //static_cast<unsigned int>(ui_->spinBoxNucDetMinPocketArea->value())
    };

  st::NucleiPreselectionAssessmentParameters nuc_presel_assess_params_ = {
      0.0, 1.0
  };


  double const score = st::detectNuclei(true,
          img_orig, img_preproc, img_preproc,
          &contours_nuclei,
          &contours_of_debris,
          &contours_nonsegmentable,
          &contours_of_nuclei_from_segmentation,
          &contours_of_orphans_from_segmentation,
          NucleiDetectionNonstainedParameters{},
          nuc_presel_params, nuc_presel_assess_params_,
          true,   // ui_->checkBoxNucDetSplitAdjacent->isChecked(),
          false,  // ui_->checkBoxNucDetPreviewContours->isChecked());
          false); // dry-run

  std::cout << contours_nuclei.size() << " nuclei found\n";
  ////////////////////// END OF GUI ACTION EMULATION FOR DEVEL /////////////////


  double t_since, t_until, delta_t;
  t_since = static_cast<double>(cv::getTickCount());
  auto const img_sz = cv::Size{img_orig.cols, img_orig.rows};


  // -- stage 1
t_since = static_cast<double>(cv::getTickCount());
  std::vector<std::tuple<contour_t,contour_t,cv::Point>> perinucs1;
  perinucs1.reserve(contours_nuclei.size());
  for (auto const& nuc_contour: contours_nuclei) {
    auto pn = find_perinuclear_contours(nuc_contour, img_sz);
    perinucs1.emplace_back( pn );
  }
  // TODO: more for other types of nuclei
t_until = static_cast<double>(cv::getTickCount());
delta_t =  static_cast<int>((t_until - t_since) *1000/cv::getTickFrequency());
std::cout << "Time1: " << delta_t << " ms\n";


  // -- stage 2
t_since = static_cast<double>(cv::getTickCount());
  auto perinucs2 = perinucs1;
  for (int i=0; i<perinucs1.size(); ++i){
    get<1>(perinucs2[i]) = remove_outer_contour_overlaps(perinucs1, i, img_sz);
  }
t_until = static_cast<double>(cv::getTickCount());
delta_t =  static_cast<int>((t_until - t_since) *1000/cv::getTickFrequency());
std::cout << "Time2: " << delta_t << " ms\n";

  // -- preview, only for this stage of DEVEL, later delete
  cout << contours_nuclei.size() << " contours" <<endl;
  Mat img_all_nucs_masks = Mat::zeros(img_sz, CV_8UC1);
  for (int i=0; i<contours_nuclei.size(); ++i/* TODO other contours as well*/) {
    cv::drawContours( img_all_nucs_masks, contours_nuclei, i,
                      Scalar(96,96,96), cv::FILLED, 8 );
    cv::drawContours( img_all_nucs_masks, contours_vec_t{get<0>(perinucs2[i])}, 0,
                      Scalar(255,0,255), 1, 8 );
    cv::drawContours( img_all_nucs_masks, contours_vec_t{get<1>(perinucs2[i])}, 0,
                      Scalar(125,125,0), 1, 8 );
  }
  cv::imshow("STAGE 2", img_all_nucs_masks); cv::waitKey(0);


  // -- stage 3
t_since = static_cast<double>(cv::getTickCount());
  std::vector<Annulus> annuli;
  generate_annuli(perinucs2, annuli, img_sz);
t_until = static_cast<double>(cv::getTickCount());
delta_t =  static_cast<int>((t_until - t_since) *1000/cv::getTickFrequency());
std::cout << "Time3: " << delta_t << " ms\n";

  // -- preview, only for this stage of DEVEL, later delete
  Mat img_annuli = Mat::zeros(img_sz, CV_8UC1);
  for (Annulus const& aa: annuli) {
    if (aa.kind == Annulus::Type::kMultipartC) {
      for (contour_t const& a: aa.contours) {
        cv::drawContours( img_annuli, contours_vec_t{a}, 0,
                           Scalar(255,255,255), cv::FILLED, 8 );

      }
    } else if (aa.kind == Annulus::Type::kCompleteC) {
      cv::drawContours( img_annuli, contours_vec_t{aa.contours[0]}, 0,
                        Scalar(150,180,150), cv::FILLED, 8 );
    } else if (aa.kind == Annulus::Type::kCompleteO) {
      cv::drawContours( img_annuli, contours_vec_t{aa.contours[0]}, 0,
                         Scalar(255,255,255), cv::FILLED, 8 );
      cv::drawContours( img_annuli, contours_vec_t{aa.contours[1]}, 0,
                      Scalar(1,1,0), 1, 8 );
    }
  }
  for (int i=0; i<contours_nuclei.size(); ++i/* TODO other contours as well*/) {
    cv::drawContours( img_annuli, contours_nuclei, i,
                      Scalar(46,46,46), cv::FILLED, 8 );
  }
  cv::imshow("STAGE 3", img_annuli); cv::waitKey(0);

  return 0;
}
