//
//            -----==#[  MasterTiffToAtomicTiffs.ijm  ]#==-----
//
// This ImageJ/Fiji macro reads in a stack of images of multiple time-frames
// and multiple channels from a single TIFF file and exports individual images
// to adequately named single-frame single-channel ("atomic") TIFF files.
//
// The macro learns about the file to be processed from the argument passed 
// upon invocation (full absolute path is required). The exported TIFFs have 
// grayscale LUT and zero-padded time-frame index followed by the channel index;
// indexing of both the time frames and channels starts from 0. The output files
// are thus named as follows:
// 
//   exported_T0000_C0.tif  exported_T0000_C1.tif
//   exported_T0001_C0.tif  exported_T0001_C1.tif
//   exported_T0002_C0.tif  exported_T0002_C1.tif
//
// (and so on).
//
// As is the case of many ImageJ/Fiji macros, this macro cannot be launched 
// in the "--headless" mode. If you work on Linux and want to avoid annoying 
// appearance and disappearance of substack windows, you may launch the script 
// using a virtual framebuffer (e.g., Xvfb):
//
//   $ xvfb-run fiji -macro MasterTiffToAtomicTiffs.ijm `pwd`/TO_BE_CONVERTED.tif
//
// (assuming that both the macro file and the big TIFF to split are located
// in the same current directory).
//
// This code is part of ShuttleTracker distribution, for more info see the
// project homepage: http://pmbm.ippt.pan.pl/software/shuttletracker 


// settings for Bio-Formats Exporter
EXPORTED_TIFF_FILE_PREFIX = "exported";  // beginning of the name of each exported TIFF
EXPORTED_TIFF_FILE_COMPRESSION = "LZW";  // "LZW" or "Uncompressed"

// get location of the input TIFF file
open(getArgument());
dir = getDirectory("image");
filename = getTitle();
path = dir + filename;

// count the number of maximum necessary zeros, double underscores mark unused variables
getDimensions(__width, __height, n_channels, __n_slices, n_frames);
n_max_frame_index_zeros = lengthOf("" + (n_frames - 1));

for (frame_index=1; frame_index<=n_frames; frame_index++) {

    print("Exporting time frame " + frame_index + " of " + n_frames + "...");

    // in exported files, time frames are counted from 0
    frame_index_zero_based = frame_index - 1;

    // pad time frame index with appropriate number of "0"s
    len_frame_index_zero_based = lengthOf("" + frame_index_zero_based);
    n_current_frame_index_zeros = n_max_frame_index_zeros - len_frame_index_zero_based;
    zero_padding="";
    for (i=0; i<n_current_frame_index_zeros; i++) { zero_padding += "0";  }

    // build a base filename for duplication and for the Bio-Formats exporter
    out_filename = EXPORTED_TIFF_FILE_PREFIX + "_T" + zero_padding + frame_index_zero_based + ".tif";
    
    // duplicate current time frame with all channels
    channels_chunk = " channels=" + "1-" + n_channels;
    frames_chunk = " frames=" + frame_index;
    duplicate_cmd = "title=" + out_filename + " duplicate " + channels_chunk + frames_chunk;
    run("Duplicate...", duplicate_cmd);

    // assure that exported images will have grayscale LUT
    Stack.setDisplayMode("grayscale");

    // export to individual TIFFs using an OME exporter
    bioformats_options = " export write_each_channel compression=" + EXPORTED_TIFF_FILE_COMPRESSION;
    run("Bio-Formats Exporter", "save=./" + out_filename + bioformats_options);

    // close current image window so that the image sequence on which
    // (partial) duplication will work is again the main sequence 
    close();

    print("Exporting time frame " + frame_index + " of " + n_frames + "... done");
    
} // for each time frame

run("Quit");

