#!/usr/bin/env python3

"""
Example invocation:

./tiff_expander.py --channel-0 WellA_Field1_C1.tif --bit-depth-0 16to8 --dye-0 blue  --label-0 X \
                   --channel-1 WellA_Field1_C2.tif --bit-depth-1 16to8 --dye-1 green --label-1 Y \
                   --channel-2 WellA_Field1_C3.tif --bit-depth-2 16to8 --dye-2 red   --label-2 Z \
                   --basename WellA_Field1
"""

from itertools import permutations
import sys
import os
import re
from optparse import OptionParser, OptionGroup
from PIL import Image, ImageSequence
import numpy as np


def parse_command_line_(args):

    allowed_dyes = 'blue green red cyan yellow magenta bright-field'.split()
    allowed_conversions_base = '16to16 8to8 9to8 10to8 11to8 12to8 13to8 14to8 15to8 16to8'.split()
    allowed_conversions = sorted(allowed_conversions_base
                                 + list(map(lambda c: c+'clip', allowed_conversions_base[2:])),
                                 key=lambda c: 10*int(c.split('to')[0]) + len(c))

    max_n_channels = len(allowed_dyes)

    ## set up parser -----

    parser = OptionParser()

    opt_names = dict()

    for i in range(max_n_channels):
        si = str(i)

        opt_names[i] = {'tiff' + si: '--channel-' + si,
                        'dye' + si: '--dye-' + si,
                        'conversion' + si: '--bit-depth-' + si}

        group = OptionGroup(parser, 'Channel ' + si + ' options')
        group.add_option('--channel-' + si, dest='tiff' + si, metavar='tiff' + si,
                         help='multi-frame TIFF file for this channel ')
        group.add_option('--dye-' + si, dest='dye' + si, metavar='dye' + si,
                         help='one of: ' + ', '.join(allowed_dyes))
        group.add_option('--bit-depth-' + si, dest='conversion' + si,
                         metavar='conversion' + si,
                         help='one of: '+', '.join(allowed_conversions))
        group.add_option('--label-' + si, dest='label' + si, metavar='label' + si)
        parser.add_option_group(group)

    parser.add_option('--basename', dest='basename', metavar='basename',
                      help='prefix common for all output files.')


    ## use parser checking arguments ("options") consistentcy -----

    (opts, args) = parser.parse_args(args)
    opts = opts.__dict__

    # check whether whole a set (that is, a triple) of arguments has been provided
    for i in range(max_n_channels):
        si = str(i)
        for keys in permutations(opt_names[i].keys()):
            if not opts[keys[0]] is None:
                if opts[keys[1]] is None:
                    print('Error: No "' + opt_names[i][keys[1]] + '" option featuring "'
                          + opt_names[i][keys[0]] + '".')
                    return None
                if opts[keys[2]] is None:
                    print('Error: No "' + opt_names[i][keys[2]] + '" option featuring "'
                          + opt_names[i][keys[0]] + '".')
                    return None
        for k in opt_names[i]:
            if 'dye' in k:
                if not opts[k] is None and not opts[k] in set(allowed_dyes):
                    print('Error: Unrecognized dye "' + opts[k] + '" for channel ' + si + '.')
                    print('Error: Allowed dyes are: ' + ', '.join(allowed_dyes) + '.')
                    return None
            elif 'conversion' in k:
                if not opts[k] is None and not opts[k] in set(allowed_conversions):
                    print('Error: Unrecognized bit depth conversion "' + opts[k]
                          + '" for channel ' + si +'.')
                    print('Error: Allowed conversions are: ' + ', '.join(allowed_conversions) + '.')
                    return None

    n_channels = sum([1 if not opts['tiff'+str(i)] is None else 0 for i in range(max_n_channels)])
    if n_channels == 0:
        print('Error: No channels specified.')
        return None

    print('Info: Number of channels:', n_channels, '(output indices: ['
          + ','.join(map(lambda c: '0'+str(c), list(range(n_channels)))) + ']).')

    if opts['tiff0'] is None:
        print('Error: It is expected to have the first channel indexed with 0.')
        return None

    for channel_i in range(1, n_channels):
        if opts['tiff' + str(channel_i)] is None:
            print('Error: It is expected to have continuous channel numbering.')
            return None

    channels_descs = []
    for channel_i in range(n_channels):
        si = str(channel_i)
        label = opts['label' + si] if opts['label' + si] is not None else 'Protein' + si
        desc = (si, opts['tiff' + si], opts['dye' + si], opts['conversion' + si], label)
        channels_descs.append(desc)

    seen_labels = set()
    for _, __, ___, ____, label in channels_descs:
        if not label in seen_labels:
            seen_labels.add(label)
        else:
            print('Error: Duplacation of label "' + label + '".')
            return None

    seen_dyes = set()
    for _, __, dye, ___, ____ in channels_descs:
        if not dye in seen_dyes:
            seen_dyes.add(dye)
        else:
            print('Error: Duplacation of dye "' + dye + '".')
            return None

    if not opts['basename'] is None:
        basename = opts['basename']
    else:
        print('Error: No basename given.')
        return None

    return basename, channels_descs



def calculate_conversion_factor_(conversion):

    pre = int(conversion.split('to')[0])
    return int(2**(16 - pre))



def expand_tiffs_(basename, channels_descriptions):

    for channel_i, tiff_file_name, _, conversion, __ in channels_descriptions:

        im = Image.open(tiff_file_name)

        if im.mode != 'I;16B':
            print('Error: Bit depth of "{}" is not 16 bit.'.format(tiff_file_name))
            return

        if im.n_frames == 1:
            print('Error: File "{}" does not appear to contain multiple frames.'
                  .format(tiff_file_name))
            return

        print('Info: Expansion of "{}" ...'.format(tiff_file_name), end='', flush=True)

        scaling = calculate_conversion_factor_(conversion)/(2**16 - 1)*(2**8 - 1)

        n_overwritten = 0
        assert int(channel_i) < 10
        for i, frame in enumerate(ImageSequence.Iterator(im)):

            frame_array = np.array(frame, dtype=np.uint16)
            frame_array_conv = None
            if 'clip' in conversion:
                conversion_from = int(conversion.split('to')[0])
                frame_array_maybe_clipped = np.clip(frame_array, 0, int(2**conversion_from-1))
                frame_array_conv = (frame_array_maybe_clipped*scaling).astype(np.uint8)
            else:
                if conversion == '16to16':
                    frame_array_conv = frame_array  # effectively, no conversion
                else:
                    assert 'to8' in conversion
                    if np.max(frame_array)*scaling < 2**8:
                        frame_array_conv = (frame_array*scaling).astype(np.uint8)
                    else:
                        print('\nError: Conversion of bit depth "{}" would result in overflow.'
                              .format(conversion))
                        print('Hint: Bit depth (range of values) appears to be wider than assumed.')
                        return

            ti = ('{:0' + str(len(str(im.n_frames - 1))) + 'd}').format(i)
            frame_out_file_name = "{}_t{}_ch0{:1s}.tif".format(basename, ti, channel_i)
            assert frame_out_file_name != tiff_file_name
            if os.path.exists(frame_out_file_name):
                n_overwritten += 1
            Image.fromarray(frame_array_conv).save(frame_out_file_name)

        frames_out_files_names = re.sub(r"_t\d*_ch", "_t[0...{}]_ch"
                                        .format(im.n_frames - 1), frame_out_file_name)
        print('\b\b\b--> "{}".'.format(frames_out_files_names), flush=True)
        if n_overwritten > 0:
            print('Note: Overwritten {} image files.'.format(n_overwritten))



def generate_hints_for_shuttletracker(channels_descriptions):

    hints_out_file_name = 'hints_for_shuttletracker.txt'
    overwritting_hints_file = False
    if os.path.exists(hints_out_file_name):
        overwritting_hints_file = True

    with open(hints_out_file_name, 'w') as hints_of:
        for channel_i, _, dye, __, label in channels_descriptions:
            print('channel {} {} {} {:1d}'
                  .format(channel_i, label, dye, 8), file=hints_of)
        print('time_interval 60', file=hints_of)

    print('Info: File "{}" has been generated.'.format(hints_out_file_name))
    #print('      Frame-to-frame interval of 60 seconds has been provisionally assumed.')

    if overwritting_hints_file:
        print('Note: File "{}" has been overwritten.'.format(hints_out_file_name))



if __name__ == "__main__":

    basename_and_channels_descriptions = parse_command_line_(sys.argv[1:])

    if basename_and_channels_descriptions is None:
        print('Error: Do not know how to proceed further.')
        print('Hint: Use switch -h to learn how to use this script.')
        sys.exit(1)

    bn, chds = basename_and_channels_descriptions

    expand_tiffs_(bn, chds)

    generate_hints_for_shuttletracker(chds)
